# -*- coding:utf-8 -*-
import logging
import os
import datetime as dt
import unittest
import re
import pandas as pd
from rio3.datafeed.common import _get_market
from rio3.quote_core.quote_info import is_future
from rio3.datafeed.dbIO import get_main_future_contract, get_stock_id
from rio3.datafeed.models import OneDayView
import urllib3


def to_sina_code(stockfullcode):
    # type: (str) -> str
    """ 转换代码为新浪的代码，转换失败返回None """
    if is_future(stockfullcode) and stockfullcode.endswith('9001'):
        stockfullcode = get_main_future_contract(stockfullcode=stockfullcode)
    market = _get_market(stockfullcode)
    if market is None:
        return None
    # 新浪股票代码为小写，带市场名
    if market in ('SZ', 'SH'):
        return stockfullcode.lower()
    # 期货为大写，不带市场名
    if market in ('DL', 'SQ', 'SF', 'SE'):
        if stockfullcode.endswith('0001'):
            stockfullcode = stockfullcode.replace('0001', '0')
        return stockfullcode.replace(market, '', 1).upper()
    # 郑商所的代码年份为两位，不带市场名。比如ZZMA709的新浪代码为MA1709
    if market == 'ZZ':
        r = re.search('(^\D+)(\d+)$', stockfullcode)
        if not r:
            return None
        # 获得合约的名字，去掉市场
        contract_code = r.group(1)
        contract_code = contract_code.replace(market, '', 1).upper()
        # 处理合约日期代码
        date_code = r.group(2)
        date_code = int(date_code)
        if stockfullcode.endswith('0001'):
            return contract_code + '0'
        if date_code < 100 or date_code >= 1000:
            return None
        current_date_code = (dt.datetime.now().year % 100) * 100 + dt.datetime.now().month
        while date_code < current_date_code:
            date_code += 1000
        return contract_code + str(date_code)
    return None


def get_latest_kline(stockfullcode):
    # type: (str) -> OneDayView
    """
    获取新浪的最新行情，对于不支持的代码返回None。
    盘中查询就是最新的行情，盘后查就是当天的日线。
    对于股票和商品期货，查询的方式为http://hq.sinajs.cn/list=sh600000。
    对于股指期货，查询的方式为http://hq.sinajs.cn/list=CFF_RE_IF1709。
    股票、商品、股指返回的字符串的解析各不相同。
    商品没有amount信息。
    """
    sina_code = to_sina_code(stockfullcode)
    if sina_code is None:
        return None
    market = _get_market(stockfullcode)
    open = high = low = close = vol = amt = settle = yd_settle = yd_close = oi = yd_oi = max = min = 0
    bp1 = bp2 = bp3 = bp4 = bp5 = sp1 = sp2 = sp3 = sp4 = sp5 = 0
    bv1 = bv2 = bv3 = bv4 = bv5 = sv1 = sv2 = sv3 = sv4 = sv5 = 0
    name = market_name = symbol_name = ''
    date = None
    url_http = urllib3.PoolManager()
    # TODO: Python37下的UrlLib3需要重新调试，以下代码测试不通过 2020/05/28
    if market in ['ZZ', 'DL', 'SQ', 'SE']:
        url = "http://hq.sinajs.cn/list={}".format(sina_code)
        req = url_http.request('GET', url)
        con = url_http.urlopen('GET', url)
        doc = con.read()
        content = doc.split('''"''')[1]
        elems = content.split(',')
        if len(elems) > 1:
            name = elems[0].decode('gb2312')
            open = float(elems[2])
            high = float(elems[3])
            low = float(elems[4])
            yd_close = float(elems[5])
            bp1 = float(elems[6])
            sp1 = float(elems[7])
            close = float(elems[8])
            settle = float(elems[9])
            yd_settle = float(elems[10])
            bv1 = int(elems[11])
            sv1 = int(elems[12])
            oi = int(elems[13])
            vol = int(elems[14])
            market_name = elems[15].decode('gb2312')
            symbol_name = elems[16].decode('gb2312')
            date = pd.to_datetime(elems[17])
    elif market in ['SH', 'SZ']:
        url = "http://hq.sinajs.cn/list={}".format(sina_code)
        req = url_http.request('GET', url)
        con = url_http.urlopen('GET', url)
        doc = con.read()
        content = doc.split('''"''')[1]
        elems = content.split(',')
        if len(elems) > 1:
            name = elems[0].decode('gb2312')
            open = float(elems[1])
            yd_close = float(elems[2])
            close = float(elems[3])
            high = float(elems[4])
            low = float(elems[5])
            bp1 = float(elems[6])
            sp1 = float(elems[7])
            vol = int(elems[8]) / 100
            amt = float(elems[9])
            bv1 = int(elems[10]) / 100
            bp1 = float(elems[11])
            bv2 = int(elems[12]) / 100
            bp2 = float(elems[13])
            bv3 = int(elems[14]) / 100
            bp3 = float(elems[15])
            bv4 = int(elems[16]) / 100
            bp4 = float(elems[17])
            bv5 = int(elems[18]) / 100
            bp5 = float(elems[19])
            sv1 = int(elems[20]) / 100
            sp1 = float(elems[21])
            sv2 = int(elems[22]) / 100
            sp2 = float(elems[23])
            sv3 = int(elems[24]) / 100
            sp3 = float(elems[25])
            sv4 = int(elems[26]) / 100
            sp4 = float(elems[27])
            sv5 = int(elems[28]) / 100
            sp5 = float(elems[29])
            date = pd.to_datetime(elems[30] + ' ' + elems[31])
    elif market == 'SF':
        url = "http://hq.sinajs.cn/list=CFF_RE_{}".format(sina_code)
        req = url_http.request('GET', url)
        con = url_http.urlopen('GET', url)
        doc = con.read()
        content = doc.split('''"''')[1]
        elems = content.split(',')
        if len(elems) > 1:
            open = float(elems[0])
            high = float(elems[1])
            low = float(elems[2])
            close = float(elems[3])
            vol = int(elems[4])
            amt = float(elems[5])
            oi = float(elems[6])
            max = float(elems[9])
            min = float(elems[10])
            yd_close = float(elems[13])
            yd_settle = float(elems[14])
            yd_oi = float(elems[15])
            date = pd.to_datetime(elems[36] + ' ' + elems[37])
    else:
        return None
    stock_id = get_stock_id(stockfullcode)
    if stock_id is None:
        return None
    kline = OneDayView(
        id_stock=stock_id,
        stockfullcode=stockfullcode,
        date_trade=date,
        val_preclose=yd_close,
        val_open=open,
        val_high=high,
        val_low=low,
        val_close=close,
        amt_amount=amt,
        amt_volumn=vol,
        amt_wp=0,
        amt_np=0,
        flg_st=0,
        val_OI=oi,
        val_turnover=0,
        dailyPreclose=yd_close
    )
    return kline


class SinaTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_to_sina_code_test(self):
        sina_code = to_sina_code('SH600000')
        self.assertEqual(sina_code, 'sh600000')
        sina_code = to_sina_code('SZ000001')
        self.assertEqual(sina_code, 'sz000001')
        sina_code = to_sina_code('SFIF1709')
        self.assertEqual(sina_code, 'IF1709')
        sina_code = to_sina_code('SQcu1801')
        self.assertEqual(sina_code, 'CU1801')
        sina_code = to_sina_code('DLa1709')
        self.assertEqual(sina_code, 'A1709')
        sina_code = to_sina_code('SFIF0001')
        self.assertEqual(sina_code, 'IF0')
        sina_code = to_sina_code('ZZMA0001')
        self.assertEqual(sina_code, 'MA0')

    def test_get_sina_kline_test(self):
        kline = get_latest_kline('DLa0001')
        self.assertEqual(kline.stockfullcode, 'DLa0001')
        self.assertEqual(kline.val_close > 0, True)
        self.assertEqual(kline.amt_volumn > 0, True)
        kline = get_latest_kline('SFIF0001')
        self.assertEqual(kline.stockfullcode, 'SFIF0001')
        self.assertEqual(kline.val_close > 0, True)
        self.assertEqual(kline.amt_amount > 0, True)
        kline = get_latest_kline('SH600000')
        self.assertEqual(kline.stockfullcode, 'SH600000')
        self.assertEqual(kline.val_close > 0, True)
        self.assertEqual(kline.amt_amount > 0, True)
        kline = get_latest_kline('SQcu1702')
        self.assertEqual(kline.stockfullcode, 'SQcu1702')


if __name__ == "__main__":
    unittest.main()
