# -*- coding:utf-8 -*-
import logging
import re
import unittest
from rio3.log.setup import setup_file_logger, setup_screen_logger

# 从stockfullcode中提取market和code
STOCK_EXCHANGE_CODE_FORMAT = re.compile(r'(?P<market>(\D{2}))(?P<code>(\S+))')

# 从stock code中提取exchange type/symbol/code 如：SFIF0001 结果为SF，IF, 0001
STOCK_EXCHANGE_SYMBOL_CODE_FORMAT = re.compile(r'(?P<market>(\D{2}))(?P<symbol>(\D*))(?P<code>(\d+))')

# 从contract中提取symbol/code 如：IC1806
FUTURE_SYMBOL_CODE_FORMAT = re.compile(r'(?P<symbol>(\D*))(?P<code>(\d+))')

# Redis connection url
REDIS_CONNECT_URL_FORMAT = re.compile(
    r'redis://(?P<auth_info>(.*))@(?P<address_url>(.*)):(?P<port>(\d+))/(?P<db>(\d+))')

# 期权合约全称格式
OPTIONS_STOCKFULLCODE_GROUP_FORMAT = r'^(?P<market>(DL|SQ|ZZ|SE|SF))(?P<symbol>([a-zA-Z]+))(?P<code>(\d{4}|\d{3}))(?P<con_type>(C|P|-C-|-P-))(?P<exec_price>(\d*))$'
OPTIONS_STOCKFULLCODE_FORMAT = r'^(?:DL|SQ|ZZ|SE|SF)[a-zA-Z]+(?:\d{4}|\d{3})(?:C|P|-C-|-P-)\d*$'

# 期权合约格式
OPTIONS_CONTRACT_GROUP_FORMAT = r'^(?P<symbol>([a-zA-Z]+))(?P<code>(\d{4}|\d{3}))(?P<con_type>(C|P|-C-|-P-))(?P<exec_price>(\d*))$'
OPTIONS_CONTRACT_FORMAT = r'^[a-zA-Z]+(?:\d{4}|\d{3})(?:C|P|-C-|-P-)\d*$'

# 股票代码格式
VALID_STOCK_CODE = []
VALID_STOCK_PAT = [
    re.compile(r'^(?P<market>(SH))(?P<code>(6\d{5}))$'),  # SH6xxxxx, 沪市A股
    re.compile(r'^(?P<market>(SH))(?P<code>(51\d{4}))$'),  # SH51xxxx, 沪市ETF
    re.compile(r'^(?P<market>(SH))(?P<code>(000\d{3}))$'),  # SH000xxx, 沪市指数
    re.compile(r'^(?P<market>(SZ))(?P<code>(0\d{5}))$'),  # SZ0xxxxx, 深市A股
    re.compile(r'^(?P<market>(SZ))(?P<code>(30\d{4}))$'),  # SZ30xxxx, 深市创业板
    re.compile(r'^(?P<market>(SZ))(?P<code>(159\d{3}))$'),  # SZ159xxx, 深市ETF
    re.compile(r'^(?P<market>(SZ))(?P<code>(399\d{3}))$'),  # SZ399xxx, 深市指数
]

INVALID_STOCK_PAT = []

# 商品期货格式
VALID_COMMODITY_CODE = []
VALID_COMMODITY_PAT = [
    re.compile(r'^(?P<market>(DL))(?P<symbol>([a-zA-Z]+))(?P<code>(\d{4}))$'),  # DL*xxxx，例如DLa1709
    re.compile(r'^(?P<market>(SQ))(?P<symbol>([a-zA-Z]+))(?P<code>(\d{4}))$'),  # SQ*xxxx，例如SQcu1709
    re.compile(r'^(?P<market>(ZZ))(?P<symbol>([a-zA-Z]+))(?P<code>(\d{3}))$'),  # ZZ*xxx，例如ZZMA709
    re.compile(r'^(?P<market>(SE))(?P<symbol>([a-zA-Z]+))(?P<code>(\d{4}))$'),  # SE*xxxx，例如SEsc1809
    re.compile(r'^(?P<market>(ZZ))(?P<symbol>([a-zA-Z]+))(?P<code>(9001))$'),  # ZZ*xxx，例如ZZMA709
    re.compile(r'^(?P<market>(ZZ))(?P<symbol>([a-zA-Z]+))(?P<code>(9991))$'),  # ZZ*xxx，例如ZZMA709
    re.compile(r'^(?P<market>(ZZ))(?P<symbol>([a-zA-Z]+))(?P<code>(9992))$'),  # ZZ*xxx，例如ZZMA709
    re.compile(r'^(?P<market>(DL))(?P<symbol>([a-zA-Z]+))(?P<code>(MI|PI|VI))$'),  # DL*MI/PI/VI，例如DLaMI
    re.compile(r'^(?P<market>(SQ))(?P<symbol>([a-zA-Z]+))(?P<code>(MI|PI|VI))$'),  # SQ*MI/PI/VI，例如SQcuMI
    re.compile(r'^(?P<market>(ZZ))(?P<symbol>([a-zA-Z]+))(?P<code>(MI|PI|VI))$'),  # ZZ*MI/PI/VI，例如ZZMAMI
    re.compile(r'^(?P<market>(SE))(?P<symbol>([a-zA-Z]+))(?P<code>(MI|PI|VI))$'),  # SE*MI/PI/VI，例如SEscMI
    # 商品期权代码  ZZZC201P800
    re.compile(
        r'^(?P<market>(DL|SQ|ZZ|SE))(?P<symbol>([a-zA-Z]+))(?P<code>(\d{4}|\d{3}))(?P<con_type>(C|P|-C-|-P-))(?P<exec_price>(\d*))$'),

    # add WuXi ni code validation 2018-03-29
    re.compile(r'^(?P<market>(WX))(?P<symbol>([a-zA-Z]+))(?P<code>(\d{4}))$'),  # WX*xxxx，例如WXni1709
]
INVALID_COMMODITY_PAT = []

# 期权格式
VALID_OPTIONS_CONTRACT_PAT = [
    re.compile(r'^(?P<market>(SQ|DL|ZZ|SE))(?P<symbol>([a-zA-Z]+))(?P<code>(.))$')  # SQal
]

# 股指期货格式
VALID_FUTURE_INDEX_CODE = []
VALID_FUTURE_INDEX_PAT = [
    re.compile(r'^(?P<market>(SF))(?P<symbol>(IF))(?P<code>(\d{4}))$'),  # SFIFxxxx
    re.compile(r'^(?P<market>(SF))(?P<symbol>(IC))(?P<code>(\d{4}))$'),  # SFICxxxx
    re.compile(r'^(?P<market>(SF))(?P<symbol>(IH))(?P<code>(\d{4}))$'),  # SFIHxxxx
    re.compile(r'^(?P<market>(SF))(?P<symbol>(T))(?P<code>(\d{4}))$'),  # SFTxxxx
    re.compile(r'^(?P<market>(SF))(?P<symbol>(TF))(?P<code>(\d{4}))$'),  # SFTFxxxx
    re.compile(r'^(?P<market>(SF))(?P<symbol>(TS))(?P<code>(\d{4}))$'),  # SFTSxxxx
    # 股指期权代码  ZZZC201P800
    re.compile(
        r'^(?P<market>(SF))(?P<symbol>([a-zA-Z]+))(?P<code>(\d{4}|\d{3}))(?P<con_type>(C|P|-C-|-P-))(?P<exec_price>(\d*))$')
]
INVALID_FUTURE_INDEX_PAT = []

# 检查脚本的范围格式代码
VALID_CHECK_STOCK_PAT = [
    re.compile(r'^SH6\d{5}$'),  # SH6xxxxx, 沪市A股
    re.compile(r'^SZ0\d{5}$'),  # SZ0xxxxx, 深市A股
    re.compile(r'^SZ30\d{4}$'),  # SZ30xxxx, 深市创业板
    re.compile(r'^DL([a-zA-Z]+)\d{4}$'),  # DL*xxxx，例如DLa1709
    re.compile(r'^SQ([a-zA-Z]+)\d{4}$'),  # SQ*xxxx，例如SQcu1709
    re.compile(r'^ZZ([a-zA-Z]+)\d{3}$'),  # ZZ*xxx，例如ZZMA709
    re.compile(r'^SE([a-zA-Z]+)\d{4}$'),  # SE*xxxx，例如SQsc1809
    re.compile(r'^SF(IF)\d{4}$'),  # SFIFxxxx
    re.compile(r'^SF(IC)\d{4}$'),  # SFICxxxx
    re.compile(r'^SF(IH)\d{4}$'),  # SFIHxxxx
    re.compile(r'^SF(T)\d{4}$'),  # SFTxxxx
    re.compile(r'^SF(TF)\d{4}$')  # SFTFxxxx
]

# 股票和期货的压缩包格式命名
DICT_STOCK_TYPE_PACKAGE_FORMAT = {
    'stock': 'ws(?P<date>(\d+))fb5.(rar|zip)',
    'commodity_future': 'ws(fu|SE)(?P<date>(\d+))fb.(rar|zip)',
    # CF，SF和SF5切换日期，压缩包时间格式为yyyyMMdd
    'index_future': 'ws(SF5_|CF5_|SF)(?P<date>(\d+)).*.(rar|zip)'
}

# 股票和期货的压缩包内部路径格式
DICT_PACKAGE_INSIDE_PATH_FORMAT = {
    'stock': 'ws{:%Y%m%d}fb',
    'commodity_future': '{:%Y%m%d}',
    'index_future': '{:%Y%m%d}'
}

# 股票和期货的压缩包存放路径格式
DICT_STOCK_TYPE_PACKAGE_PATH = {
    'stock': 'stock\{:%Y}\{:%Y%m}\ws{:%m%d}fb5.rar',
    'commodity_future': 'commodity_future\{:%Y}\{:%Y%m}\wsfu{:%m%d}fb.rar',
    'index_future': 'index_future\SF5\{:%Y}\{:%Y%m}\wsSF5_{:%Y%m%d}.rar'
}

# yyyyMMdd格式的时间
year_month_pattern = re.compile(r'(?P<year>(\d{4}))(?P<month>(\d{2}))')
year_pattern = re.compile(r'\d{4}')
year_date_pattern = re.compile(r'(?P<year>(\d{4}))(?P<date>(\d{4}))')

# Tick/Kline H5的文件格式
RE_H5_FILE_FORMAT = re.compile(r'(?P<stockfullcode>(\S+))_(?P<h5_file_type>(tick|kline)).h5')

# stock_code_ymd  exp. SH000001_20160103.csv
CSV_STOCK_CODE_DATE_FORMAT = re.compile(
    r'(?P<exchange_type>(\D{2}))(?P<rest>((?P<symbol>(\D*))(?P<code>(\d+))|(.+)))_(?P<date>(\d{8})).csv')

# XXXXX_yyyyMMdd exp. EthCny_20160103.csv
CSV_SYMBOL_FORMAT = re.compile(r'(?P<symbol_str>([\s\S]*))_(?P<date>(\d{8})).csv')


# 根据传入的pattern集合来确定给定的字符串的格式
def get_pattern_of_str(input_str, pattern_list):
    pattern_list = [pattern_list, ] if isinstance(pattern_list, re.Pattern) else pattern_list
    str_pattern = None
    for pat in pattern_list:
        if re.match(pat, input_str):
            str_pattern = pat
            break
    return str_pattern


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_get_pattern_of_str(self):
        # 测试商品期货代码
        commodity_stockfullcode_list = ['SQrbMI', 'SQcu9001', 'DLa706', 'ZZMA309']
        commodity_result_list = [
            'MI',
            '9001',
            None,
            '309'
        ]
        for i in range(len(commodity_stockfullcode_list)):
            pat = get_pattern_of_str(commodity_stockfullcode_list[i], VALID_COMMODITY_PAT)
            if pat is not None:
                match_res = re.match(pat, commodity_stockfullcode_list[i])
                self.assertEqual(match_res.group('code'), commodity_result_list[i])
            else:
                self.assertEqual(pat, commodity_result_list[i])

        # 测试股指期货代码
        index_stockfullcode_list = ['SFIF9001', 'SFTF0001']
        index_result_list = [
            'IF',
            'TF'
        ]
        for i in range(len(index_stockfullcode_list)):
            pat = get_pattern_of_str(index_stockfullcode_list[i], VALID_FUTURE_INDEX_PAT)
            if pat is not None:
                match_res = re.match(pat, index_stockfullcode_list[i])
                self.assertEqual(match_res.group('symbol'), index_result_list[i])
            else:
                self.assertEqual(pat, index_result_list[i])

        # 测试股票代码
        stock_stockfullcode_list = ['SH000001', 'SZ399999']
        stock_result_list = [
            'SH',
            'SZ'
        ]
        for i in range(len(stock_stockfullcode_list)):
            pat = get_pattern_of_str(stock_stockfullcode_list[i], VALID_STOCK_PAT)
            if pat is not None:
                match_res = re.match(pat, stock_stockfullcode_list[i])
                self.assertEqual(match_res.group('market'), stock_result_list[i])
            else:
                self.assertEqual(pat, stock_result_list[i])


if __name__ == "__main__":
    logger = logging.getLogger('root')
    setup_file_logger()
    setup_screen_logger()
    unittest.main()
