# -*- coding:utf-8 -*-
import re
import os
import unittest
import logging
import datetime as dt
import numpy as np
import pandas as pd
from functools import wraps
from rio3.base import regex
from rio3.base.regex import STOCK_EXCHANGE_SYMBOL_CODE_FORMAT
from rio3.log.setup import setup_screen_logger
from rio3.base.constant import SECURITY_DIR_DICT, DEFAULT_INDEX_STOCK

# 匹配分钟数字 + m的freq定义
# s: 秒 m: 分钟 d: 日 w: 周 M: 月 Q: 季度
FREQ_REGEX = re.compile(r'^(?P<num>(\d+))(?P<unit>(s|h|m|d|y|w|M|Q))$')

# JQ的期货代码格式
JQ_FUTURE_CODE_FORMAT = re.compile(r'(?P<symbol>(\D*))(?P<code>(\d+)).(?P<market>(\D{4}))')

PERIOD2TEXT_DICT = {
    'tick': 'tick',
    '5s': 'fiveseconds',
    '15s': 'fifteenseconds',
    '30s': 'thirtyseconds',
    '1m': 'oneminute',
    '5m': 'fiveminutes',
    '15m': 'fifteenminutes',
    '30m': 'halfhour',
    '60m': 'onehour',
    '1d': 'oneday'
}

TEXT2Period_DICT = {
    'tick': 'tick',
    'fiveseconds': '5s',
    'fifteenseconds': '15s',
    'thirtyseconds': '30s',
    'oneminute': '1m',
    'fiveminutes': '5m',
    'fifteenminutes': '15m',
    'halfhour': '30m',
    'onehour': '60m',
    'oneday': '1d'
}


def split_market(stockfullcode):
    '''
    拆分stockfullcode，返回市场名和品种合约代码
    :param stockfullcode:
    :return: market, stock
    '''
    market = stockfullcode[0:2]
    stock = stockfullcode[2:]
    return market, stock


def _get_jq_future_symbol(jq_stockfullcode):
    # type: (str) -> str
    jq_reg_format = re.match(JQ_FUTURE_CODE_FORMAT, jq_stockfullcode)
    if jq_reg_format is not None:
        return jq_reg_format.group('symbol')
    else:
        raise TypeError(f'不支持的jq_stockfullcode: {jq_stockfullcode}的转化类型')


def _get_contract_code(stockfullcode):
    """
    从stockfullcode抽取合约代码 SFIF2008 返回2008
    :param stockfullcode:
    :return:
    """
    regex_code_arr = re.match(STOCK_EXCHANGE_SYMBOL_CODE_FORMAT, stockfullcode)
    if regex_code_arr is not None:
        return regex_code_arr.group('code')
    else:
        raise TypeError(f'不支持的stockfullcode: {stockfullcode}的转化类型')


def _get_options_symbol(stockfullcode):
    # type: (str) -> str
    match_pat = regex.get_pattern_of_str(stockfullcode, [re.compile(regex.OPTIONS_STOCKFULLCODE_GROUP_FORMAT)])
    if match_pat is not None:
        match_result = re.match(match_pat, stockfullcode)
        symbol = match_result.group('symbol')
    return symbol


def _get_future_symbol(stockfullcode):
    # type: (str) -> str
    '''
    根据stockfullcode/stockcode截取品种
    :param stockfullcode:
    :return: symbol
    '''
    market, stockcode = split_market(stockfullcode)
    symbol = None
    if market == 'SQ' or market == 'DL' or market == 'ZZ' or market == 'SE':
        patterns = regex.VALID_COMMODITY_PAT
    elif market == 'SF':
        patterns = regex.VALID_FUTURE_INDEX_PAT
    else:
        return symbol
    match_pat = regex.get_pattern_of_str(stockfullcode, patterns)
    if match_pat is not None:
        match_result = re.match(match_pat, stockfullcode)
        symbol = match_result.group('symbol')
    return symbol


def _get_future_symbol_from_stockcode(stock_code):
    """
    从stockcode中提取品种
    :param stock_code:
    :return:
    """
    symbol = stock_code
    stockcode_array = re.match(regex.FUTURE_SYMBOL_CODE_FORMAT, stock_code)
    if stockcode_array is not None:
        symbol = stockcode_array.group('symbol')
        code = stockcode_array.group('code')
    return symbol


def _get_market(stockfullcode):
    # type: (str) -> str
    '''
    根据stockfullcode获取市场名
    :param stockfullcode:
    :return:
    '''
    return split_market(stockfullcode)[0]


def is_valid_stock(stockfullcode):
    if stockfullcode in regex.VALID_STOCK_CODE:
        return True
    for pat in regex.INVALID_STOCK_PAT:
        if re.search(pat, stockfullcode):
            return False
    for pat in regex.VALID_STOCK_PAT:
        if re.search(pat, stockfullcode):
            return True
    return False


def is_valid_commodity(stockfullcode):
    if stockfullcode in regex.VALID_COMMODITY_CODE:
        return True
    for pat in regex.INVALID_COMMODITY_PAT:
        if re.search(pat, stockfullcode):
            return False
    for pat in regex.VALID_COMMODITY_PAT:
        if re.search(pat, stockfullcode):
            return True
    return False


def is_valid_index_future(stockfullcode):
    if stockfullcode in regex.VALID_FUTURE_INDEX_CODE:
        return True
    for pat in regex.INVALID_FUTURE_INDEX_PAT:
        if re.search(pat, stockfullcode):
            return False
    for pat in regex.VALID_FUTURE_INDEX_PAT:
        if re.search(pat, stockfullcode):
            return True
    return False


# 通过市场名字获取股票种类 股票|股指期货|商品期货
def _get_security_type_by_market(market_name):
    for security_type, market_name_list in SECURITY_DIR_DICT.items():
        if market_name in market_name_list:
            return security_type
    return None


# 如果不是本地指定的缓存，就读取NAS的缓存，默认的路径需要根据stockfullcode拼接指定路径
def _get_cache_files_dir(cache_root_path, stockfullcode, is_local, pass_security_type=None):
    '''
    通过stockfullcode，security_type决定cache存放的位置
    :param cache_root_path:
    :param stockfullcode:
    :param is_local:
    :param pass_security_type:
    :return:
    '''
    cache_file_path = cache_root_path
    if is_local:
        return cache_file_path
    else:
        market_name = _get_market(stockfullcode)
        if market_name is not None:
            # 如果外部直接传入security_type，则直接使用，不走stockfullcode判断，这里主要是针对期权数据，因为持有的合约的市场名与stock/future有重合，所以路径需要单独导向
            if pass_security_type is not None:
                cache_file_path = os.path.join(cache_root_path, pass_security_type, market_name)
            else:
                if stockfullcode in DEFAULT_INDEX_STOCK:
                    security_type = 'index'
                else:
                    security_type = _get_security_type_by_market(market_name)
                if security_type is not None:
                    cache_file_path = os.path.join(cache_root_path, security_type, market_name)
    return cache_file_path


def _calculate_zq_kdix(quote_df, freq_str):
    """
    # 单独计算kidx和didx
    didx:   tick的索引 = （kidx - 1）* 每秒更新次数 * 时间间隔，对于使用精确tick数据的，这个是大致等于，可能误差1~2
            (期货：1秒2tick) (股票：5秒1tick)

    kidx:   累积的k线索引 累计数量，类似计数器

    grid:   目前无意义
    :param quote_df:
    :param period:
    :return:
    """
    assert isinstance(quote_df, pd.DataFrame)
    if not quote_df.empty:
        # reset index以保证序列正确
        quote_df.reset_index(inplace=True)
        # 计算didx,kidx,grid
        quote_df['kidx'] = quote_df.index + 1
        quote_df['didx'] = (quote_df['kidx'] - 1) * 2 * int(convert_freq2num(freq_str=freq_str))
        quote_df['grid'] = quote_df['kidx'] + 100
        if 'index' in quote_df.columns:
            quote_df.drop(['index'], axis=1, inplace=True)


def filter_quote_df(quote_df):
    """
    过滤行情数据，增量+条件
    :param quote_df:
    :return:
    """
    # 1d行情的持续更新会造成contract_idx为nan，填充默认值
    quote_df.fillna({'contract_idx': 1}, inplace=True)
    quote_df.dropna(axis=0, how='any', inplace=True)
    # 默认返回所有的K线，不做vol > 0，用户自己根据需求过滤
    # quote_df = quote_df[quote_df['vol'] > 0]
    return quote_df


def _fill_mid_and_midv(quote_df):
    """
    填充mid/midv，目前从JQdata的分钟线数据，由于不是从Tick数据而来，所以缺少可以用来计算mid/midv的条件
    :param quote_df:
    :return:
    """
    assert isinstance(quote_df, pd.DataFrame)
    if quote_df.empty:
        return pd.DataFrame()
    if 'mid' not in quote_df.columns:
        quote_df['mid'] = quote_df['close']
    if 'midv' not in quote_df.columns:
        quote_df['midv'] = quote_df['close']


# 计算Mid和MidV
# 计算公式
# Mid = (SP1 + BP1) / 2
# Midv = (SP1 * BV1 + BP1 * SV1) / (SV1 + BV1)
# 1. if SP1 == 0 then Mid = BP1, Midv = BP1 else
# 2. if BP1 == 0 then Mid = SP1, Midv = SP1
# 3. if SV1 == 0 & BV1 == 0 then 转回1和2的判断 Mid == Midv
def calculate_mid_and_midv(quote_df, close_column='Close'):
    # 先按照原公式进行
    quote_df['mid'] = (quote_df['Ask1Price'] + quote_df['Bid1Price']) / 2
    quote_df['midv'] = (quote_df['Ask1Price'] * quote_df['Bid1Volume'] + quote_df['Bid1Price'] * quote_df[
        'Ask1Volume']) / (quote_df['Ask1Volume'] + quote_df['Bid1Volume'])
    # 当SP1 == 0，需要进行数据纠正
    quote_df['mid'] = np.where(quote_df['Ask1Price'] == 0, quote_df['Bid1Price'], quote_df['mid'])
    quote_df['midv'] = np.where(quote_df['Ask1Price'] == 0, quote_df['Bid1Price'], quote_df['midv'])
    # 当BP1 == 0，需要进行数据纠正
    quote_df['mid'] = np.where(quote_df['Bid1Price'] == 0, quote_df['Ask1Price'], quote_df['mid'])
    quote_df['midv'] = np.where(quote_df['Bid1Price'] == 0, quote_df['Ask1Price'], quote_df['midv'])
    # 当SV1 == 0 & BV1 == 0，即计算Midv的公式出现分母为0，则Midv = Mid
    quote_df['midv'] = np.where(quote_df['midv'] == np.inf, quote_df['mid'], quote_df['midv'])
    quote_df['midv'] = quote_df['midv'].fillna(quote_df['mid'])
    # 如果计算结果Mid和Midv还有0的情况，则等于Close的值
    quote_df['mid'] = np.where(quote_df['mid'] == 0, quote_df[close_column], quote_df['mid'])
    quote_df['midv'] = np.where(quote_df['midv'] == 0, quote_df[close_column], quote_df['midv'])
    # 保留小数点后4位
    quote_df.round({'mid': 4, 'midv': 4})


def _get_stock_code_from_h5_file(h5_file_path):
    """
    # 从H5文件名中分离出stockfullcode
    :param h5_file_path:
    :return:
    """
    file_name = os.path.basename(h5_file_path)
    name_array = file_name.split('_')
    if len(name_array) > 0:
        return name_array[0]
    return None


def _get_h5_file_path(source_path, stockfullcode, h5_type):
    '''
    指定路径和stockfullcode，获取h5_tick或h5_kline文件路径
    h5_type (Tick|Kline)
    :param source_path:
    :param stockfullcode:
    :param h5_type:
    :return:
    '''
    h5_file_path = os.path.join(source_path, stockfullcode + '_{}.h5'.format(h5_type.lower()))
    return h5_file_path


def return_data(return_frame=True):
    if return_frame:
        return pd.DataFrame()
    else:
        return dict()


def convert_freq2num(freq_str):
    '''
    将1分钟以上，1d以下的分钟线做转化
    :param freq_str:
    :return:
    '''
    assert isinstance(freq_str, str)
    if freq_str == 'tick':
        return -1
    if freq_str in TEXT2Period_DICT.keys():
        freq_str = TEXT2Period_DICT.get(freq_str, freq_str)
    # 将freq_str等于单个周期单位的转化为标准周期的转化
    if freq_str in ['m', 'h', 'd', 'w', 'y', 'M', 'Q']:
        freq_str = '1{}'.format(freq_str)
    freq_num_unit = re.match(FREQ_REGEX, freq_str)
    if freq_num_unit is not None:
        freq_num = freq_num_unit.group('num')
        freq_unit = freq_num_unit.group('unit')
        # 如果是传入的是日线单位d，直接返回
        if freq_unit == 's':
            return int(freq_num)
        elif freq_unit == 'd':
            return int(freq_num) * 86400
        elif freq_unit == 'm':
            return int(freq_num) * 60
        elif freq_unit == 'h':
            return int(freq_num) * 3600
        elif freq_unit == 'w':
            return int(freq_num) * 86400 * 7
        elif freq_unit == 'M':
            return int(freq_num) * 86400 * 30
        elif freq_unit == 'Q':
            return int(freq_num) * 86400 * 90
        elif freq_unit == 'y':
            return int(freq_num) * 86400 * 365
    else:
        raise TypeError(f'不支持的频率转化类型: {freq_str}')


def get_resample_interval(freq_str):
    """
    将freq_str转化为resample支持的转化周期
    :param freq_str:
    :return:
    """
    assert isinstance(freq_str, str)
    # 将freq_str等于单个周期单位的转化为标准周期的转化
    if freq_str in ['m', 'h', 'd', 'w', 'M', 'Q', 'y']:
        freq_str = '1{}'.format(freq_str)
    # 根据interval枚举秒数获取h5存储节点名和resample的转化周期
    freq_num_unit = re.match(FREQ_REGEX, freq_str)
    resample_type = '1T'
    if freq_num_unit is not None:
        freq_num = freq_num_unit.group('num')
        freq_unit = freq_num_unit.group('unit')
        if freq_unit == 'd':
            # resample_type = '{}D'.format(freq_num)
            resample_type = '{}H'.format(24 * int(freq_num))
        elif freq_unit == 'h':
            resample_type = '{}H'.format(int(freq_num))
        elif freq_unit == 's':
            resample_type = '{}S'.format(freq_num)
        elif freq_unit == 'm':
            resample_type = '{}T'.format(freq_num)
        elif freq_unit == 'w':
            resample_type = '{}W'.format(freq_num)
        elif freq_unit == 'M':
            resample_type = '{}M'.format(freq_num)
        elif freq_unit == 'Q':
            resample_type = '{}Q'.format(freq_num)
        elif freq_unit == 'y':
            resample_type = '{}Y'.format(freq_num)
        else:
            raise TypeError(f'不支持的频率转化类型: {freq_str}')
    return resample_type


def _to_datetime(tday=None):
    """
    获取tday
    :param tday:
    :return:
    """
    if tday is None:
        tday = pd.to_datetime(dt.datetime.now())
    else:
        tday = pd.to_datetime(tday)
    return tday


def _get_tday(tday=None):
    """
    获取tday，去除时分秒
    :param tday:
    :return:
    """
    if tday is None:
        tday = dt.datetime.now().date()
    else:
        tday = pd.to_datetime(tday).date()
    return tday


def timer(func):
    '''
    计算函数运行时间
    :param func:
    :return:
    '''

    @wraps(func)
    def wrap_func(*args, **kwargs):
        start = dt.datetime.now()
        result = func(*args, **kwargs)
        end = dt.datetime.now()
        usage = end - start
        print('[函数]：', func.__name__)
        print('[用时(s)]：', usage.seconds)
        return result

    return wrap_func


def _split_list_by_divide_unit(list_data, divide_unit=100):
    '''
    将数组数据按照divide_unit切分
    :param list_data:
    :param divide_unit:
    :return:
    '''
    divided_data_list = list()
    start_index = 0
    kline_divided_length, remainder = divmod(len(list_data), divide_unit)
    # 循环归纳出TDAY_DIVIDE_UNIT单位的tday数据
    for i in range(kline_divided_length):
        end_index = start_index + divide_unit
        each_interval_list = list_data[start_index: end_index]
        start_index = end_index
        divided_data_list.append(each_interval_list)
    # 处理剩下的天数
    rest_tdays_list = list_data[-remainder:]
    divided_data_list.append(rest_tdays_list)
    return divided_data_list


class TestCommonFuncs(unittest.TestCase):
    def setUp(self):
        setup_screen_logger()
        self.logger = logging.getLogger('root')

    def tearDown(self):
        pass

    def test_get_future_symbol(self):
        stockfullcode0 = 'DLi2009-C-510'
        symbol0 = _get_future_symbol(stockfullcode=stockfullcode0)
        self.assertEqual(symbol0, 'i')

        stockfullcode01 = 'ZZZC201P800'
        symbol0 = _get_future_symbol(stockfullcode=stockfullcode01)
        self.assertEqual(symbol0, 'ZC')

        stockfullcode02 = 'SFIO2006-C-3550'
        symbol0 = _get_future_symbol(stockfullcode=stockfullcode02)
        self.assertEqual(symbol0, 'IO')

        stockfullcode1 = 'SQrb9001'
        symbol1 = _get_future_symbol(stockfullcode=stockfullcode1)
        self.assertEqual(symbol1, 'rb')

        stockfullcode2 = 'SH600000'
        symbol2 = _get_future_symbol(stockfullcode=stockfullcode2)
        self.assertEqual(symbol2, None)

        stockfullcode3 = 'ZZAP9001'
        symbol3 = _get_future_symbol(stockfullcode=stockfullcode3)
        self.assertEqual(symbol3, 'AP')

    def test_get_market(self):
        stockfullcode1 = 'SH600000'
        market = _get_market(stockfullcode=stockfullcode1)
        self.assertEqual(market, 'SH')
        stockfullcode2 = 'SQcu2005'
        market = _get_market(stockfullcode=stockfullcode2)
        self.assertEqual(market, 'SQ')
        stockfullcode3 = 'datetrade'
        market = _get_market(stockfullcode=stockfullcode3)
        self.assertEqual(market, 'SQ')

    def test_split_market(self):
        stockfullcode1 = 'SH600000'
        market1, stockcode1 = split_market(stockfullcode=stockfullcode1)
        self.assertEqual(market1, 'SH')
        self.assertEqual(stockcode1, '600000')

        stockfullcode2 = 'DLpp9001'
        market2, stockcode2 = split_market(stockfullcode=stockfullcode2)
        self.assertEqual(market2, 'DL')
        self.assertEqual(stockcode2, 'pp9001')

    def test_get_jq_future_symbol(self):
        stockfullcode1 = _get_jq_future_symbol(jq_stockfullcode='AU9999.XSGE')
        self.assertEqual(stockfullcode1, 'AU')
        stockfullcode2 = _get_jq_future_symbol(jq_stockfullcode='AP2005.XSGE')
        self.assertEqual(stockfullcode2, 'AP')

    def test_get_resample_interval(self):
        period1 = '1m'
        period2 = '30s'
        period3 = '30m'
        period4 = '2d'
        resample_type1 = get_resample_interval(period1)
        resample_type2 = get_resample_interval(period2)
        resample_type3 = get_resample_interval(period3)
        resample_type4 = get_resample_interval(period4)
        pass

    def test_get_tday(self):
        tday = _get_tday()
        pass

    def test_get_future_symbol_from_stockcode(self):
        stockcode = 'IF9001'
        symbol = _get_future_symbol_from_stockcode(stockcode)
        self.assertEqual('IF', symbol)
        pass

    def test_get_options_symbol(self):
        stockfullcode = 'DLi2205-C-800'
        symbol = _get_options_symbol(stockfullcode=stockfullcode)
        pass


if __name__ == "__main__":
    unittest.main()
