# -*- coding:utf-8 -*-
import unittest
import traceback
import pandas as pd
import pyodbc
import numpy as np
import logging
import collections
import datetime as dt
import calendar
from collections import defaultdict
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy import create_engine, or_, and_, desc, asc, not_
from sqlalchemy.dialects.mysql import insert
from sqlalchemy.sql import func
from rio3.base import constant
from rio3.base import enumerator
from rio3.datafeed.models import Tday, FutureInfoView, EdInfoView, Stock, Market, OneMinute, FiveMinutes, Strategy, \
    EdFactorView, EdFactorInfoView, EdFactorInfo, EdFactor, TdaySymbolStocksView, EdInfo, EdExecuteInfo, OneDayView, \
    TdaySymbol, TdaySymbolStocks, OneMinuteView, KLineView, KLine, OneDay, FifteenMinutes, Halfhour, OneHour, Holiday
from rio3.quote_core import quote_info
from rio3.log.setup import setup_screen_logger
from rio3.datafeed.common import _get_market, _get_future_symbol, split_market, _split_list_by_divide_unit
from rio3.utils.collect import bidict
from functools import wraps
from sqlalchemy.exc import InterfaceError, TimeoutError
from rio3.datafeed.models import t_VW_oneminute_history, t_VW_fiveminutes_history
from rio3.datafeed.common import _to_datetime, convert_freq2num
from rio3.production.dbIO import get_contract_list_date

STRATEGY_LIST = list()
STOCK_LIST = []
STRATEGY_ID_DICT = bidict()
STOCKFULLCODE_ID_DICT = bidict()
STOCKCODE_ID_DICT = bidict()
MARKET_LIST = dict()
SESSION_DICT = dict()
DB_ENGINE_DICT = dict()
TDAY_LIST = list()
HOLIDAY_LIST = list()
# stockfullcode和future_info的字典
FUTURE_INFO_DICT = dict()
# stockfullcode/symbol和乘数的字典
FUTURE_MULTIPLE_INFO_DICT = dict()

logger = logging.getLogger('root')


def db_retry_on_exception(retry=5):
    def dec(func):
        @wraps(func)
        def func_wrap(*args, **kwargs):
            remain_retry = retry
            if remain_retry == 0:
                return func(*args, **kwargs)
            last_exception = None
            while remain_retry > 0:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    last_exception = e
                    if isinstance(e, InterfaceError) or isinstance(e, TimeoutError):
                        retry_msg = 'Retry database operation because of exception {}'.format(e.message)
                        print(retry_msg)
                        logger.warning(retry_msg)
                    else:
                        raise e
                finally:
                    remain_retry -= 1
            raise last_exception

        return func_wrap

    return dec


def get_db_engine(engine_string):
    global DB_ENGINE_DICT
    if engine_string in DB_ENGINE_DICT:
        return DB_ENGINE_DICT[engine_string]
    db_engine = create_engine(engine_string, echo=False, pool_size=10, max_overflow=100,
                              pool_recycle=3600)
    DB_ENGINE_DICT[engine_string] = db_engine
    return db_engine


def get_db_session(engine_string):
    global SESSION_DICT
    if engine_string in SESSION_DICT:
        session = SESSION_DICT[engine_string]
        try:
            session().execute('SELECT 1')
        except:
            session().close()
            del SESSION_DICT[engine_string]

    if engine_string in SESSION_DICT:
        session = SESSION_DICT[engine_string]
    else:
        engine = get_db_engine(engine_string=engine_string)
        session_factory = sessionmaker(bind=engine)
        session = scoped_session(session_factory)
        SESSION_DICT[engine_string] = session
    return session()


def get_holiday_df(start_dt=None, end_dt=None, engine_string=None):
    """
    转化get_holiday_list的结果为dataframe
    :param start_dt:
    :param end_dt:
    :param engine_string:
    :return:
    """
    holiday_list = get_holiday_list(start_dt=start_dt, end_dt=end_dt, engine_string=engine_string)
    return pd.DataFrame(data=holiday_list, columns=['holiday'])


def get_holiday_list(start_dt=None, end_dt=None, engine_string=None):
    '''
    获取指定时间的法定节假日，数据来源为手工统计
    :param start_dt: 开始节假日
    :param end_dt:  结束节假日
    :param engine_string:   数据库连接字符串
    :return:
    '''
    global HOLIDAY_LIST
    if start_dt is None:
        start_date = constant.DATETIME_MIN
    else:
        start_date = pd.to_datetime(start_dt)
    if end_dt is None:
        end_date = constant.DATETIME_MAX
    else:
        end_date = pd.to_datetime(end_dt)
    if not HOLIDAY_LIST:
        if engine_string is None:
            engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
        session = get_db_session(engine_string)
        holidays = session.query(Holiday).order_by(Holiday.holiday).all()
        HOLIDAY_LIST = [pd.to_datetime(t.holiday) for t in holidays]
        session.close()
    return [holiday for holiday in HOLIDAY_LIST if start_date <= holiday <= end_date]


def get_tday_df(start_dt=None, end_dt=None, engine_string=None):
    """
    转化get_tday_list的结果为dataframe
    :param start_dt:
    :param end_dt:
    :param engine_string:
    :return:
    """
    tday_list = get_tday_list(start_dt=start_dt, end_dt=end_dt, engine_string=engine_string)
    return pd.DataFrame(data=tday_list, columns=['tday'])


def get_tday_list(start_dt=None, end_dt=None, engine_string=None):
    '''
    获取指定时间区间的交易日，默认获取的是上交所的数据库记录
    :param start_dt: 开始交易日
    :param end_dt:  结束交易日
    :param engine_string:   数据库连接字符串
    :return:
    '''
    global TDAY_LIST
    if start_dt is None:
        start_date = constant.DATETIME_MIN
    else:
        start_date = pd.to_datetime(start_dt)
    if end_dt is None:
        end_date = constant.DATETIME_MAX
    else:
        end_date = pd.to_datetime(end_dt)
    if not TDAY_LIST:
        if engine_string is None:
            engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
        session = get_db_session(engine_string)
        tdays = session.query(Tday).order_by(Tday.tday).all()
        TDAY_LIST = [pd.to_datetime(t.tday) for t in tdays]
        session.close()
    return [tday for tday in TDAY_LIST if start_date <= tday <= end_date]


def get_previous_tday(day, count=1):
    if count < 0:
        return None
    query_day = pd.to_datetime(day).replace(hour=0, minute=0, second=0, microsecond=0)
    query_day = query_day - dt.timedelta(days=1)
    if count == 0:
        return query_day
    tdays = get_tday_list(end_dt=query_day)
    if len(tdays) >= count:
        return tdays[0 - count]
    else:
        return None


def get_next_tday(day, count=1):
    if count < 0:
        return None
    query_day = pd.to_datetime(day).replace(hour=0, minute=0, second=0, microsecond=0)
    # query_day = query_day + dt.timedelta(days=1)
    if count == 0:
        if is_valid_tday(query_day):
            return query_day
    query_day = query_day + dt.timedelta(days=1)
    tdays = get_tday_list(start_dt=query_day)
    if len(tdays) >= count:
        return tdays[count - 1]
    else:
        return None


def is_valid_tday(day):
    query_day = pd.to_datetime(day)
    query_day = query_day.replace(hour=0, minute=0, second=0, microsecond=0)
    return query_day in get_tday_list()


def get_trading_contract_list(symbol=None):
    trading_contract_list = list()
    symbol_trading_contract_df = get_future_info(symbol=symbol)
    if not symbol_trading_contract_df.empty:
        symbol_trading_contract_df = symbol_trading_contract_df[
            symbol_trading_contract_df['expire_date'].astype('datetime64[ns]') > dt.datetime.now().date()]
        if not symbol_trading_contract_df.empty:
            trading_contract_list = sorted(symbol_trading_contract_df['stockfullcode'].unique().tolist())
    return trading_contract_list


def get_trading_contract_list_SF(symbol, date=None):
    """
    传入日期，获取大于等于传入date的合约
    :param symbol:
    :param date:
    :return:
    """
    symbol = symbol.upper()
    if date is None:
        today = dt.date.today()
    else:
        today = pd.to_datetime(date).date()

    all_contract_list = [['01', '02', '03', '06'], ['02', '03', '06', '09'], ['03', '04', '06', '09'],
                         ['04', '05', '06', '09'],
                         ['05', '06', '09', '12'], ['06', '07', '09', '12'], ['07', '08', '09', '12'],
                         ['08', '09', '12', '03'],
                         ['09', '10', '12', '03'], ['10', '11', '12', '03'], ['11', '12', '03', '06'],
                         ['12', '01', '03', '06']]
    df_contract = pd.DataFrame(all_contract_list)
    df_contract = df_contract.applymap(lambda x: str(today)[2:4] + x)
    df_contract.iloc[-1, 1] = str(int(df_contract.iloc[-1, 1]) + 100)
    df_contract.iloc[-2:, 2] = df_contract.iloc[-2:, 2].map(lambda x: str(int(x) + 100))
    df_contract.iloc[-5:, 3] = df_contract.iloc[-5:, 3].map(lambda x: str(int(x) + 100))
    contract_now = 'SF' + symbol + str(today)[2:4] + str(today)[5:7]
    if get_contract_list_date(contract_now)[contract_now][1] >= today:
        con_list = df_contract.loc[today.month - 1].map(lambda x: 'SF' + symbol + x).to_list()
    elif today.month == 12:
        con_list = df_contract.loc[0].map(lambda x: str(int(x) + 100)).map(lambda x: 'SF' + symbol + x).to_list()
    else:
        con_list = df_contract.loc[today.month].map(lambda x: 'SF' + symbol + x).to_list()
    return con_list


def get_future_info(stockfullcode=None, symbol=None, open_date=constant.DATETIME_MIN,
                    expire_date=constant.DATETIME_MAX, dataframe=True):
    """
    获取期货信息表
    :param stockfullcode:
    :param symbol:
    :param open_date: 合约开始日期
    :param expire_date: 合约过期日期
    :param dataframe:
    :return:
    """
    future_info_list = list()
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    list_symbol = [symbol, ] if isinstance(symbol, str) else symbol
    open_date = pd.to_datetime(open_date).strftime('%Y-%m-%d')
    expire_date = pd.to_datetime(expire_date).strftime('%Y-%m-%d')
    try:
        db_session = get_db_session(constant.DEFAULT_DB_RD_ENGINE_STRING)
        q = db_session.query(FutureInfoView.stockfullcode, FutureInfoView.symbol,
                             FutureInfoView.open_date, FutureInfoView.expire_date,
                             FutureInfoView.start_deliver_date, FutureInfoView.end_deliver_date,
                             FutureInfoView.volume_multiple, FutureInfoView.long_margin_ratio,
                             FutureInfoView.short_margin_ratio, FutureInfoView.price_tick)
        conditions = None
        if stockfullcode is not None:
            conditions = FutureInfoView.stockfullcode.in_(list_stockfullcode)
        if symbol is not None:
            conditions = FutureInfoView.symbol.in_(list_symbol) if conditions is None else or_(conditions,
                                                                                               FutureInfoView.symbol.in_(
                                                                                                   list_symbol))
        if conditions is None:
            conditions = and_(FutureInfoView.open_date >= open_date, FutureInfoView.expire_date <= expire_date)
        else:
            conditions = and_(conditions, FutureInfoView.open_date >= open_date,
                              FutureInfoView.expire_date <= expire_date)
        q = q.filter(conditions).order_by(asc(FutureInfoView.open_date))
        q = q.order_by(asc(FutureInfoView.symbol))
        future_info_list = q.all()
        future_info_list = [{key: getattr(each_item, key) for key in each_item._fields} for each_item in
                            future_info_list]
    except Exception as ex:
        msg = f'读取期货合约记录失败:\r\n{traceback.format_exc()}'
        logger.error(msg)
    finally:
        db_session.close()
    if dataframe:
        result_df = pd.DataFrame.from_dict(future_info_list)
        if not result_df.empty:
            result_df = result_df.astype(
                {'expire_date': np.str, 'open_date': np.str, 'end_deliver_date': np.str, 'start_deliver_date': np.str},
                copy=True)
        return result_df
    else:
        return future_info_list


@db_retry_on_exception(retry=5)
def get_future_info_dict(engine_string=None):
    global FUTURE_INFO_DICT
    if not FUTURE_INFO_DICT:
        if engine_string is None:
            engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
        session = get_db_session(engine_string)
        future_info_list = session.query(FutureInfoView).all()
        for future_info in future_info_list:
            FUTURE_INFO_DICT[future_info.stockfullcode] = future_info
        session.close()
    return FUTURE_INFO_DICT


def get_volume_multiple_dict(by_symbol=True, engine_string=None):
    future_info_dict = get_future_info_dict(engine_string=engine_string)
    volume_multiple_dict = {}
    for stockfullcode, future_info in future_info_dict.items():
        if by_symbol:
            volume_multiple_dict[future_info.symbol] = future_info.volume_multiple
        else:
            volume_multiple_dict[future_info.stockfullcode] = future_info.volume_multiple
    return volume_multiple_dict


def get_volume_multiple_df():
    """
    构造volume_multiple的数据Dataframe
    :return:
    """
    symbol_volume_multiple_dict = get_volume_multiple_dict()
    data_df = pd.DataFrame(list(symbol_volume_multiple_dict.items()), columns=['symbol', 'volume_multiple'])
    return data_df


@db_retry_on_exception(retry=5)
def _get_future_volume_multiple_dict(engine_string=None):
    '''
    获取期货品种volume_mltiple的字典集合, 返回品种对应乘数, 运行该SQL需要安装pyodbc python包，windows下需要安装mysql-connector-odbc-5.3.7-winx64.msi
    :param engine_string:
    :return:
    '''
    global FUTURE_MULTIPLE_INFO_DICT
    if not FUTURE_MULTIPLE_INFO_DICT:
        if engine_string is None:
            engine_string = constant.DB_HIS_STRING
        conn = pyodbc.connect(engine_string)
        cursor = conn.cursor()
        sql = f"""
        SELECT t.* FROM 
            (SELECT symbol, volume_multiple, open_date FROM vw_lookup_future_info ORDER BY open_date DESC) t 
        GROUP BY symbol
        """
        data_rows = cursor.execute(sql)
        for future_info in data_rows:
            if future_info.symbol is not None:
                FUTURE_MULTIPLE_INFO_DICT[future_info.symbol] = future_info.volume_multiple
    return FUTURE_MULTIPLE_INFO_DICT


def get_volume_multiple(stockfullcode, engine_string=None):
    '''
    传入stockfullcode, 获取期货品种的乘数
    :param stockfullcode:
    :param engine_string:
    :return:
    '''
    market_name = _get_market(stockfullcode)
    # 如果market_name是股票市场，直接返回固定的volume_multiple为100
    if market_name == 'SH' or market_name == 'SZ':
        return 100
    # 如果market_name是比特币市场，直接返回固定的volume_multiple为1
    if market_name == 'BX':
        return 1
    vm_dict = _get_future_volume_multiple_dict(engine_string=engine_string)
    symbol = _get_future_symbol(stockfullcode)
    # Add by Huhaohua 2017/07/12, add symbol volume multiple which is out of date
    if symbol is not None:
        if symbol.lower() == 'tc':
            return 200
        elif symbol.lower() == 'me':
            return 50
        elif symbol.lower() == 'er':
            return 10
        elif symbol.lower() == 'ws':
            return 10
        elif symbol.lower() == 'ro':
            return 5
        elif symbol.lower() == 'wt':
            return 10
        else:
            return vm_dict.get(symbol, None)
    return None


def check_ed_info_df(da, ed_info_df):
    """
    :param da: 行情数据，DataFrame格式，必须含有赋权表索引的字段，对应到相关的ed_info_df的index
    :param ed_info_df: 复权信息数据，DataFrame格式
    :return ed_info_df_sel:去除停牌时多余的复权信息，返回正确的复权信息表
    """
    if da.empty:
        return ed_info_df
    if 'Tday' not in da.columns:
        da['Tday'] = pd.to_datetime(da['today'], format='%Y%m%d')
    wrong_index_list = []
    coef_v = ed_info_df[['fhbl', 'pgbl', 'pgjg', 'sgbl']].values
    time_v = ed_info_df['date_execute'].values
    tday_v = da['Tday'].values
    da_start_time = tday_v[0]
    da_end_time = tday_v[-1]

    for i in range(len(ed_info_df) - 1):
        if (coef_v[i] == coef_v[i + 1]).all():
            start_time = time_v[i]
            end_time = time_v[i + 1]
            if (start_time >= da_start_time) & (end_time <= da_end_time):
                da_sel = da[(da['Tday'] >= start_time) & (da['Tday'] < end_time)]
                if da_sel.empty:
                    wrong_index_list.extend([i])
    if len(wrong_index_list) == 0:
        return ed_info_df
    else:
        ed_info_df_sel = ed_info_df[~ed_info_df.index.isin(wrong_index_list)].copy()
        ed_info_df_sel.reset_index(inplace=True)
        del ed_info_df_sel['index']
        return ed_info_df_sel


def get_ed_info_org(stockfullcode=None, start=constant.DATETIME_MIN, end=constant.DATETIME_MAX,
                    return_dict=False, to_ed_info=False, db_string=None):
    """
    获取ed_info的数据库对象格式
    :param stockfullcode:
    :param start:
    :param end:
    :param return_dict:
    :param to_ed_info:
    :param db_string:
    :return:
    """
    ed_info_dict = defaultdict(list)
    ed_info_list = list()
    if db_string is None:
        db_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    db_session = get_db_session(db_string)
    q = db_session.query(EdInfoView)
    if stockfullcode is not None:
        if isinstance(stockfullcode, list) and len(stockfullcode) > 0:
            q = q.filter(EdInfoView.stockfullcode.in_(stockfullcode))
        elif isinstance(stockfullcode, str):
            q = q.filter(EdInfoView.stockfullcode == stockfullcode)
    q = q.filter(EdInfoView.date_ed >= start)
    q = q.filter(EdInfoView.date_ed <= end)
    q = q.order_by(EdInfoView.date_ed)
    for edinfo in q.all():
        if return_dict:
            ed_info_dict[edinfo.stockfullcode].append(edinfo.to_ed_info())
        else:
            if to_ed_info:
                ed_info_list.append(edinfo.to_ed_info())
            else:
                ed_info_list.append(edinfo)
    db_session.close()
    if return_dict:
        return ed_info_dict
    return ed_info_list


def get_ed_info_df(stockfullcode=None, start=constant.DATETIME_MIN, end=constant.DATETIME_MAX, db_string=None):
    '''
    获取ed_info_df
    :param stockfullcode:
    :param start:
    :param end:
    :param db_string:
    :return:
    '''
    ed_info_list = _get_ed_info_list(stockfullcode=stockfullcode, start=start, end=end, db_string=db_string)
    return pd.DataFrame.from_records(ed_info_list)


def correct_ed_info_df(ed_info_df):
    """
    用来修正ed_info_df中合约来回互换的问题
    :param ed_info_df:
    :return:
    """
    # 按照contract和date_ed进行排序，筛选出重复的合约
    contract_sorted_ed_info_df = ed_info_df.sort_values(by=['stockfullcode_main', 'date_ed'])


# 获取某一个合约的到期时间
def get_expire_date(stockfullcode, engine_string=None):
    future_info_dict = get_future_info_dict(engine_string=engine_string)
    if stockfullcode in future_info_dict:
        future_info = future_info_dict[stockfullcode]
        return future_info.expire_date
    return None


# 获取某一个合约的交割时间
def get_deliver_date(stockfullcode, deliver_type, engine_string=None):
    future_info_dict = get_future_info_dict(engine_string=engine_string)
    if stockfullcode in future_info_dict:
        future_info = future_info_dict[stockfullcode]
        if deliver_type == 'start':
            return future_info.end_deliver_date
        else:
            return future_info.start_deliver_date
    return None


def get_next_contract(stockfullcode=None, engine_string=None):
    future_info_dict = get_future_info_dict(engine_string=engine_string)
    # 获取传入合约的symbol
    symbol = _get_future_symbol(stockfullcode=stockfullcode)
    # 过滤出相关合约
    symbol_contracts = [item for item in sorted(future_info_dict.keys()) if symbol == _get_future_symbol(item)]
    # 找出比当前合约大的最近的合约
    next_contracts = sorted([item for item in symbol_contracts if item > stockfullcode])
    if len(next_contracts) > 0:
        return next_contracts[0]
    return None


@db_retry_on_exception(retry=5)
def sync_ed_info(source_engine_string, dest_engine_string, market=None):
    """
    同步不同数据库的EDInfo信息
    :param source_engine_string:
    :param dest_engine_string:
    :param market:
    :return:
    """
    src_session = get_db_session(source_engine_string)
    dst_session = get_db_session(dest_engine_string)
    market_filter = None
    if market is not None:
        if isinstance(market, enumerator.ENUM_MARKET):
            market_filter = market.name
        else:
            market_filter = str(market)
        src_max_date_ed, = src_session.query(func.max(EdInfoView.date_ed)).filter(
            EdInfoView.stockfullcode.startswith(market_filter)).one()
        dst_max_date_ed, = dst_session.query(func.max(EdInfoView.date_ed)).filter(
            EdInfoView.stockfullcode.startswith(market_filter)).one()
    else:
        src_max_date_ed = src_session.query(func.max(EdInfo.date_ed).label('max_date_ed')).one().max_date_ed
        dst_max_date_ed = dst_session.query(func.max(EdInfo.date_ed).label('max_date_ed')).one().max_date_ed
    if src_max_date_ed is None:
        src_max_date_ed = constant.DATETIME_MIN
    if dst_max_date_ed is None:
        dst_max_date_ed = constant.DATETIME_MIN
    updated_num = 0
    if src_max_date_ed > dst_max_date_ed:
        q = src_session.query(EdInfoView).filter(EdInfoView.date_ed > dst_max_date_ed)
        if market_filter is not None:
            q = q.filter(EdInfoView.stockfullcode.startswith(market_filter))
        src_data_list = q.all()
        for src_data in src_data_list:
            new_edinfo = EdInfo(
                id_stock=src_data.id_stock,
                date_ed=src_data.date_ed,
                fhbl=src_data.fhbl,
                pgbl=src_data.pgbl,
                pgjg=src_data.pgjg,
                sgbl=src_data.sgbl,
                flg_calc_factor=src_data.flg_calc_factor,
                id_main_stock=src_data.id_main_stock
            )
            dst_session.add(new_edinfo)
            new_ed_exe_info = EdExecuteInfo(
                id_stock=src_data.id_stock,
                date_ed=src_data.date_ed,
                date_execute=src_data.date_execute,
                val_preclose=src_data.val_preclose
            )
            dst_session.add(new_ed_exe_info)
            updated_num += 1
        dst_session.commit()
    src_session.close()
    dst_session.close()
    return updated_num


@db_retry_on_exception(retry=5)
def get_daily_preclose(stockfullcode, date_trade, db_engine_string=None):
    """
    获取指定品种指定交易日前一天的日线
    :param stockfullcode:
    :param date_trade:
    :param db_engine_string:
    :return:
    """
    if db_engine_string is None:
        db_engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    preclose = -1
    date_trade = pd.to_datetime(date_trade).to_pydatetime()
    date_trade = date_trade.replace(hour=0, minute=0, second=0, microsecond=0)
    db_session = get_db_session(db_engine_string)
    # 先看有没有当天的日线，有的话直接取当天日线的preclose
    q = db_session.query(OneDayView).filter(OneDayView.stockfullcode == stockfullcode). \
        filter(OneDayView.date_trade == date_trade).one_or_none()
    if q is None:
        # 当天日线不存在的话，取数据库里之前最近的那条日线的close作为preclose
        q = db_session.query(OneDayView).filter(OneDayView.stockfullcode == stockfullcode). \
            filter(OneDayView.date_trade < date_trade).order_by(desc(OneDayView.date_trade)).first()
        if q is not None:
            preclose = float(q.val_close)
    else:
        preclose = float(q.val_preclose)
    db_session.close()
    return preclose


@db_retry_on_exception(retry=5)
def get_close_price(stockfullcode, day, db_string=None):
    if db_string is None:
        db_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    tday = pd.to_datetime(day).to_pydatetime()
    pre_tday = get_previous_tday(tday).to_pydatetime()
    start_time = pre_tday.replace(hour=constant.COMMODITY_DAY_CUTOFF_HOUR, minute=0, second=0)
    end_time = tday.replace(hour=constant.COMMODITY_DAY_CUTOFF_HOUR, minute=0, second=0)
    session = get_db_session(db_string)
    q = session.query(OneDayView)
    q = q.filter(OneDayView.date_trade > start_time)
    q = q.filter(OneDayView.date_trade < end_time)
    q = q.filter(OneDayView.stockfullcode == stockfullcode)
    q = q.order_by(desc(OneDayView.date_trade))
    om = q.first()
    session.close()
    if om is None:
        return None
    return float(om.val_close)


@db_retry_on_exception(retry=5)
def get_open_interest(stockfullcode, day, db_string=None):
    if db_string is None:
        db_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    tday = pd.to_datetime(day).to_pydatetime()
    pre_tday = get_previous_tday(tday).to_pydatetime()
    start_time = pre_tday.replace(hour=constant.COMMODITY_DAY_CUTOFF_HOUR, minute=0, second=0)
    end_time = tday.replace(hour=constant.COMMODITY_DAY_CUTOFF_HOUR, minute=0, second=0)
    session = get_db_session(db_string)
    q = session.query(OneDayView)
    q = q.filter(OneDayView.date_trade > start_time)
    q = q.filter(OneDayView.date_trade < end_time)
    q = q.filter(OneDayView.stockfullcode == stockfullcode)
    q = q.order_by(desc(OneDayView.date_trade))
    om = q.first()
    session.close()
    if om is None:
        return None
    return float(om.val_OI)


def calc_main_contract(tday, stockfullcode=None, market=None, freq='1d', field='vol', db_string=None):
    """
    获取当日主力合约，默认按照成交量来排序，compare_field可以传入oi,则按照持仓量排序
    :param day:
    :param stockfullcode:
    :param market:
    :param period:
    :param field:
    :return:
    """
    if db_string is None:
        db_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    symbol = None
    if stockfullcode is not None:
        symbol = _get_future_symbol(stockfullcode)
    tday = pd.to_datetime(tday).to_pydatetime()
    if not is_valid_tday(tday):
        logger.error('{}不是交易日!'.format(tday))
        return None
    main_symbol_dict = dict()  # 存放当天每个品种主力合约的代码
    main_symbol_vol_dict = collections.defaultdict(int)  # 存放当天每个品种最大成交量
    pre_tday = get_previous_tday(tday).to_pydatetime()
    next_tday = get_next_tday(tday).to_pydatetime()
    calc_day_start = pre_tday.replace(hour=0, minute=0, second=0, microsecond=0) + dt.timedelta(hours=20)
    calc_day_end = tday.replace(hour=0, minute=0, second=0, microsecond=0) + dt.timedelta(hours=16)
    if freq == '1m':
        kline_view = OneMinuteView
    elif freq == '1d':
        kline_view = OneDayView
    else:
        kline_view = OneDayView
    try:
        session = get_db_session(db_string)
        # 判断按照成交量还是持仓量排序
        if field == 'vol':
            q = session.query(kline_view.stockfullcode, func.sum(kline_view.amt_volumn).label('total_volume'))
        elif field == 'oi':
            q = session.query(kline_view.stockfullcode, kline_view.val_OI.label('last_oi'))
        if market is None:
            q = q.filter(or_(kline_view.market_name == 'DL', kline_view.market_name == 'SQ',
                             kline_view.market_name == 'ZZ', kline_view.market_name == 'SF',
                             kline_view.market_name == 'SE'))
        elif isinstance(market, list):
            q = q.filter(kline_view.market_name.in_(market))
        elif isinstance(market, str):
            q = q.filter(kline_view.market_name == market)
        elif isinstance(market, enumerator.ENUM_MARKET):
            q = q.filter(kline_view.market_name == market.name)
        else:
            session.close()
            raise TypeError('不支持的market类型{}'.format(type(market)))
        q = q.filter(and_(kline_view.date_trade >= calc_day_start, kline_view.date_trade <= calc_day_end))
        # 目前只有oneminute和oneday的view表中有stockcode
        if symbol is not None:
            q = q.filter(kline_view.stockcode.like('{}%'.format(symbol)))
        q = q.filter(not_(kline_view.stockcode.like('%0001')))
        q = q.filter(not_(kline_view.stockcode.like('%0002')))
        q = q.filter(not_(kline_view.stockcode.like('%0003')))
        q = q.filter(not_(kline_view.stockcode.like('%0004')))
        q = q.filter(not_(kline_view.stockcode.like('%9001')))
        q = q.filter(not_(kline_view.stockcode.like('%0000')))
        q = q.filter(not_(kline_view.stockcode.like('%MI')))
        q = q.filter(not_(kline_view.stockcode.like('%PI')))
        q = q.filter(not_(kline_view.stockcode.like('%VI')))
        q = q.group_by(kline_view.stockfullcode)
        q = q.order_by(kline_view.stockfullcode)
        for tmp_stockfullcode, tmp_field_val in q.all():
            cf_symbol = _get_future_symbol(tmp_stockfullcode)
            expire_date = get_expire_date(tmp_stockfullcode)
            if expire_date is None or pd.to_datetime(expire_date).to_pydatetime() > next_tday:
                # 找出相同品种各个合约成交量最大的那个作为主力合约
                if main_symbol_vol_dict[cf_symbol] < tmp_field_val:
                    main_symbol_vol_dict[cf_symbol] = tmp_field_val
                    main_symbol_dict[cf_symbol] = tmp_stockfullcode
        if main_symbol_dict:
            logger.debug("{}共有{}个主力合约".format(tday.strftime("%Y%m%d"), len(main_symbol_dict)))
        else:
            logger.warning("在{}没有找到主力合约".format(tday.strftime("%Y%m%d")))
        for tmp_symbol, tmp_stockfullcode in main_symbol_dict.items():
            logger.debug("{}:{}".format(tmp_symbol, tmp_stockfullcode))
    except Exception as ex:
        logger.error("[获取主力合约排名失败]: 交易日:{}, 品种:{}, 市场:{}, 错误:{} ".format(tday, symbol, market, ex))
    finally:
        session.close()
    return main_symbol_dict


@db_retry_on_exception(retry=5)
def add_his_klines(kline_list, freq, max_insert_num=800, db_string=None):
    """
    往历史数据库中插入K线
    :param kline_list:
    :param db_string:
    :return:
    """
    inserted_num = 0
    if db_string is None:
        db_string = constant.DEFAULT_DB_WR_ENGINE_STRING
    session = get_db_session(db_string)
    divided_kline_list = _split_list_by_divide_unit(list_data=kline_list, divide_unit=max_insert_num)
    # 检查freq与quote_model的对应关系，如果quote_model是None, 则跳过
    quote_model = _get_quote_model(freq=freq)
    if quote_model is None:
        return -1
    try:
        for each_kline_list in divided_kline_list:
            insert_kline_list = list()
            # 如果kline中没有id_stock, 从get_stock_id中获取
            for kline in each_kline_list:
                id_stock = kline['id_stock'] if kline.get('id_stock', None) is not None else get_stock_id(
                    kline['Contract'])
                if id_stock is None or id_stock < 0:
                    logger.warning('[找不到对应的]: freq:{}'.format(freq))
                    continue
                # 格式化日期
                insert_date_trade = pd.to_datetime(kline['DateTrade']).to_pydatetime()
                insert_tday = pd.to_datetime(kline['Tday']).to_pydatetime() if kline.get('Tday',
                                                                                         None) is not None else pd.to_datetime(
                    kline['DateTrade']).strftime('%Y%m%d')
                if freq == '1d':
                    new_kline = OneDay(
                        id_stock=id_stock,
                        date_trade=insert_date_trade,
                        tday=insert_tday,
                        val_open=kline['Open'],
                        val_high=kline['High'],
                        val_low=kline['Low'],
                        val_close=kline['Close'],
                        amt_amount=kline['Amount'],
                        amt_volume=kline['Volume'],
                        val_preclose=kline.get('Daily_Preclose', -1),
                        val_OI=kline.get('OpenInterest', -1),
                        flg_st=0,
                        val_sp1=kline['Sp1'] if kline.get('Sp1', None) is not None else 0,
                        val_sv1=kline['Sv1'] if kline.get('Sv1', None) is not None else 0,
                        val_bp1=kline['Bp1'] if kline.get('Bp1', None) is not None else 0,
                        val_bv1=kline['Bv1'] if kline.get('Bv1', None) is not None else 0,
                        val_mid=kline['Mid'] if kline.get('Mid', None) is not None else 0,
                        val_midv=kline['Midv'] if kline.get('Midv', None) is not None else 0
                    )
                else:
                    new_kline = quote_model(
                        id_stock=id_stock,
                        date_trade=insert_date_trade,
                        tday=insert_tday,
                        val_open=kline['Open'],
                        val_high=kline['High'],
                        val_low=kline['Low'],
                        val_close=kline['Close'],
                        amt_amount=kline['Amount'],
                        amt_volume=kline['Volume'],
                        amt_wp=0,
                        amt_np=0,
                        val_preclose=kline.get('Daily_Preclose', -1),
                        val_OI=kline.get('OpenInterest', -1),
                        val_sp1=kline['Sp1'] if kline.get('Sp1', None) is not None else 0,
                        val_sv1=kline['Sv1'] if kline.get('Sv1', None) is not None else 0,
                        val_bp1=kline['Bp1'] if kline.get('Bp1', None) is not None else 0,
                        val_bv1=kline['Bv1'] if kline.get('Bv1', None) is not None else 0,
                        val_mid=kline['Mid'] if kline.get('Mid', None) is not None else 0,
                        val_midv=kline['Midv'] if kline.get('Midv', None) is not None else 0
                    )
                if new_kline is not None:
                    insert_kline_list.append(new_kline)
        if len(insert_kline_list) > 0:
            session.bulk_save_objects(insert_kline_list)
            session.commit()
            print('[插入历史数据成功]: 周期: {}, 共{}条'.format(freq, len(each_kline_list)))
            inserted_num = len(insert_kline_list)
        else:
            print('[没有历史数据插入]: 周期: {}, 共{}条'.format(freq, len(each_kline_list)))
    except Exception as ex:
        inserted_num = -1
        logger.error('[插入历史数据失败]: 周期: {}, 错误: {}'.format(freq, ex))
    finally:
        session.close()
    return inserted_num


def remove_his_klines(start_date, end_date, stockfullcode, freq, db_string=None):
    remove_result = -1
    if db_string is None:
        db_string = constant.DEFAULT_DB_WR_ENGINE_STRING
    session = get_db_session(db_string)
    id_stock = get_stock_id(stockfullcode)
    kline_model = _get_quote_model(freq=freq)
    try:
        to_delete_rows = session.query(kline_model).filter(kline_model.id_stock == id_stock).filter(
            kline_model.date_trade >= start_date).filter(kline_model.date_trade <= end_date)
        len_of_delete_row = len(to_delete_rows.all())
        if len_of_delete_row > 0:
            to_delete_rows.delete()
            session.commit()
            print(
                "[删除历史数据成功]: 品种 '{}', 周期: '{}' 在交易日从 '{}' 到 '{}' 分钟线, 共 {} 条数据".format(stockfullcode, freq, start_date,
                                                                                       end_date, len_of_delete_row))
            remove_result = len_of_delete_row
        else:
            remove_result = 0
            print("[没有需要删除的历史数据]: 品种 '{}', 周期: '{}' 在交易日从 '{}' 到 '{}' 分钟线, 共 {} 条数据".format(stockfullcode, freq,
                                                                                            start_date,
                                                                                            end_date,
                                                                                            len_of_delete_row))
    except Exception as ex:
        print("[删除历史数据失败]: 品种 '{}', 周期: '{}' 在交易日从 '{}' 到 '{}' 分钟线, 错误:{}".format(stockfullcode, freq, start_date,
                                                                                  end_date, ex))
    finally:
        session.close()
    return remove_result


def _get_quote_model(freq):
    quote_model = None
    if freq == '1m':
        quote_model = OneMinute
    elif freq == '5m':
        quote_model = FiveMinutes
    elif freq == '15m':
        quote_model = FifteenMinutes
    elif freq == '30m':
        quote_model = Halfhour
    elif freq == '60m':
        quote_model = OneHour
    elif freq == '1d':
        quote_model = OneDay
    else:
        print('[找不到对应频率的QuoteModel对象]: 传入频率字段:{}'.format(freq))
    return quote_model


@db_retry_on_exception(retry=5)
def get_tday_symbol_stocks(symbol, tday=None, db_engine_string=None):
    if db_engine_string is None:
        db_engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    tday_symbol_stock_list = list()
    if tday is None:
        tday = dt.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
    else:
        tday = pd.to_datetime(tday).to_pydatetime()
    db_session = get_db_session(db_engine_string)
    q = db_session.query(TdaySymbolStocksView.stockfullcode, TdaySymbolStocksView.val_OI)
    q = q.filter(TdaySymbolStocksView.tday == tday)
    q = q.filter(TdaySymbolStocksView.symbol == symbol)
    q = q.filter(TdaySymbolStocksView.val_OI > 0)
    q = q.order_by(desc(TdaySymbolStocksView.val_OI))
    q = q.order_by(desc(TdaySymbolStocksView.stockfullcode))
    for stock_oi in q.all():
        tday_symbol_stock_list.append(stock_oi)
    db_session.close()
    return tday_symbol_stock_list


def add_tday_symbol_stocks(symbol, tday=None, period=enumerator.ENUM_PERIOD.OneDay, market=None, db_engine_string=None):
    """
    增加交易日品种合约持仓量排名
    :param symbol:
    :param tday:
    :param period:
    :param market:
    :param db_engine_string:
    :return:
    """
    added_num = 0
    if not is_valid_tday(tday):
        return added_num
    if db_engine_string is None:
        db_engine_string = constant.DEFAULT_DB_WR_ENGINE_STRING
    if tday is None:
        tday = dt.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
    else:
        tday = pd.to_datetime(tday).to_pydatetime()
    # 获取下一个tday作为date_execute
    pre_tday = get_previous_tday(tday).to_pydatetime()
    calc_day_start = pre_tday.replace(hour=0, minute=0, second=0, microsecond=0) + dt.timedelta(hours=20)
    calc_day_end = tday.replace(hour=0, minute=0, second=0, microsecond=0) + dt.timedelta(hours=16)
    # 目前只有1分钟线和日线中有stockcode，只支持从这两个period获取OI数据
    if period == enumerator.ENUM_PERIOD.OneMinute:
        kline_view = OneMinuteView
    elif period == enumerator.ENUM_PERIOD.OneDay:
        kline_view = OneDayView
    else:
        kline_view = OneDayView
    db_session = get_db_session(db_engine_string)
    q = db_session.query(kline_view.id_stock, kline_view.stockfullcode, func.sum(kline_view.val_OI).label('total_oi'))
    q = q.filter(and_(kline_view.date_trade >= calc_day_start, kline_view.date_trade <= calc_day_end))
    if market is None:
        q = q.filter(or_(kline_view.market_name == 'DL', kline_view.market_name == 'SQ',
                         kline_view.market_name == 'ZZ', kline_view.market_name == 'SF',
                         kline_view.market_name == 'SE'))
    elif isinstance(market, list):
        q = q.filter(kline_view.market_name.in_(market))
    elif isinstance(market, str):
        q = q.filter(kline_view.market_name == market)
    elif isinstance(market, enumerator.ENUM_MARKET):
        q = q.filter(kline_view.market_name == market.name)
    else:
        raise TypeError(u'不支持的market类型{}'.format(type(market)))
    if symbol is not None:
        q = q.filter(kline_view.stockcode.like('{}%'.format(symbol)))
    q = q.filter(not_(kline_view.stockcode.like('%0001')))
    q = q.filter(not_(kline_view.stockcode.like('%0002')))
    q = q.filter(not_(kline_view.stockcode.like('%0003')))
    q = q.filter(not_(kline_view.stockcode.like('%0004')))
    q = q.filter(not_(kline_view.stockcode.like('%9001')))
    q = q.filter(not_(kline_view.stockcode.like('%0000')))
    q = q.filter(not_(kline_view.stockcode.like('%MI')))
    q = q.filter(not_(kline_view.stockcode.like('%PI')))
    q = q.filter(not_(kline_view.stockcode.like('%VI')))
    q = q.filter(not_(kline_view.stockcode.like('%9991')))
    q = q.filter(not_(kline_view.stockcode.like('%9992')))
    q = q.filter(kline_view.val_OI > 0)
    q = q.filter(kline_view.val_close > 0)
    q = q.group_by(kline_view.stockfullcode)
    q = q.order_by(desc(func.sum(kline_view.val_OI).label('total_oi')))
    q = q.order_by(desc(kline_view.stockfullcode))
    # 如果查询该交易日无有效的交易品种，没有则不插入该品种当日的TdaySymbol和TdaySymbolStocks的记录
    if len(q.all()) > 0:
        tday_symbol = TdaySymbol(tday=tday, symbol=symbol)
        db_session.add(tday_symbol)
        db_session.commit()
        logger.info(
            f'[TdaySymbol]增加信息: 品种:{tday_symbol}, 交易日:{tday}')
    else:
        db_session.close()
        return added_num
    # 如果该交易日当日有交易品种记录，回查先前的TdaySymbol记录，获取ID
    q2 = db_session.query(TdaySymbol)
    q2 = q2.filter(TdaySymbol.tday == tday)
    q2 = q2.filter(TdaySymbol.symbol == symbol)
    q2 = q2.one_or_none()
    tday_symbol_id = q2.id
    # 插入实际合约的记录
    for id_stock, stockfullcode, total_oi in q.all():
        if _get_future_symbol(stockfullcode) == symbol:
            tday_symbol_stock = TdaySymbolStocks(id_tday_symbol=tday_symbol_id, id_stock=id_stock, val_OI=total_oi)
            db_session.add(tday_symbol_stock)
            added_num += 1
            logger.info(
                f'[TdaySymbolStocks]增加信息: 品种:{symbol}, 交易日:{tday}，TdaySymbolStocks:{stockfullcode}')
    db_session.commit()
    db_session.close()
    return added_num


def del_tday_symbol_stocks(symbol, start, end, db_engine_string=None):
    """
    删除交易日品种相关持仓信息记录
    :param symbol:
    :param start:
    :param end:
    :param db_engine_string:
    :return:
    """
    del_num = -1
    if start is None:
        start = dt.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
    else:
        start = pd.to_datetime(start).to_pydatetime()
    if end is None:
        end = dt.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
    else:
        end = pd.to_datetime(end).to_pydatetime()
    if db_engine_string is None:
        db_engine_string = constant.DEFAULT_DB_WR_ENGINE_STRING
    try:
        db_session = get_db_session(db_engine_string)
        # 先删除tday_symbol_stock
        q1 = db_session.query(TdaySymbolStocksView)
        q1 = q1.filter(and_(TdaySymbolStocksView.tday >= start, TdaySymbolStocksView.tday <= end))
        q1 = q1.filter(TdaySymbolStocksView.symbol == symbol)
        del_tday_symbol_stock_list = list()
        del_tday_symbol_list = list()
        for tday_symbol_stock in q1.all():
            del_tday_symbol_stock_list.append(tday_symbol_stock.stockfullcode)
            q2 = db_session.query(TdaySymbolStocks).filter(TdaySymbolStocks.id == tday_symbol_stock.id)
            q2.delete()
        # 接着按照tday和symbol删除tday_symbol
        q3 = db_session.query(TdaySymbol)
        q3 = q3.filter(and_(TdaySymbol.tday >= start, TdaySymbol.tday <= end))
        q3 = q3.filter(TdaySymbol.symbol == symbol)
        for tday_symbol in q3.all():
            del_tday_symbol_list.append(tday_symbol.symbol)
            q4 = db_session.query(TdaySymbol).filter(TdaySymbol.id == tday_symbol.id)
            q4.delete()
        db_session.commit()
        db_session.close()
        logger.info(
            f'[TdaySymbol]删除信息: 品种:{symbol}, 开始交易日:{start}，结束交易日:{end}, TdaySymbolStocks:{del_tday_symbol_stock_list}, TdaySymbol:{del_tday_symbol_list}')
        del_num = len(del_tday_symbol_stock_list) + len(del_tday_symbol_list)
    except Exception as ex:
        logger.error(f'删除tday_symbol_stocks失败:\r\n{traceback.format_exc()}')
    return del_num


def get_ranked_stockfullcode(rank, symbol):
    """
    给定主力等级和品种，返回900X命名的stockfullcode
    :param rank:
    :param symbol:
    :return:
    """
    rank_stockfullcode = None
    for market_name, list_symbol in constant.SYMBOL_DICT.items():
        if symbol in list_symbol:
            return '{}{}900{}'.format(market_name, symbol, rank)
    return rank_stockfullcode


def get_ed_info_trading_period(stockfullcode, rank=None, start_date=constant.DATETIME_MIN,
                               end_date=constant.DATETIME_MAX):
    """
    构建EDInfo中实际主力合约在开始和结束时间
    获取数据库中记录的EDInfo信息，主力合约的开始时间是当前记录的date_ed + 20 hours，结束时间是下一条换约记录的date_ed + 16 hours
    :param ed_info_df:
    :param rank:
    :param start_date:
    :param end_date:
    :return:
    """
    # 定级
    rank = 1 if rank is None else rank
    # 获取品种
    symbol = _get_future_symbol(stockfullcode)
    # 获取ed_info
    ed_info_df = get_ed_info_df(get_ranked_stockfullcode(rank=rank, symbol=symbol))
    ed_info_df['rank'] = rank
    ed_info_df['symbol'] = ed_info_df['stockfullcode_main'].apply(_get_future_symbol)
    ed_info_df['main_contract'] = ed_info_df.apply(lambda x: x['stockfullcode_main'], axis=1)
    ed_info_df = ed_info_df[['rank', 'symbol', 'main_contract', 'date_ed', 'date_execute']].copy()
    trading_period_ed_info_list = list()
    trading_period_ed_info_df = pd.DataFrame()
    # 根据rank过滤
    if rank is not None:
        ed_info_df = ed_info_df[ed_info_df['rank'] == rank]
    ed_info_df = ed_info_df.sort_values(['rank', 'symbol', 'date_execute'])
    ed_info_df = ed_info_df.astype({
        'date_ed': 'datetime64[ns]',
        'date_execute': 'datetime64[ns]'
    })
    for symbol_rank, symbol_rank_df in ed_info_df.groupby(by=['symbol', 'rank']):
        symbol_rank_df_copy = symbol_rank_df.copy()
        # symbol_rank_df['main_contract'] = normalize_code_to_local(symbol_rank_df['main_contract'].values.tolist())
        symbol_rank_df_copy['stockfullcode'] = get_ranked_stockfullcode(symbol=symbol_rank[0], rank=symbol_rank[1])
        symbol_rank_df_copy['date_execute_start'] = symbol_rank_df_copy['date_execute'] + pd.to_timedelta('20:00:00')
        symbol_rank_df_copy['date_execute_end'] = symbol_rank_df_copy['date_execute'].shift(-1)
        # 将最后一条ed_info设置为当前日期
        symbol_rank_df_copy = symbol_rank_df_copy.fillna({'date_execute_end': pd.to_datetime(pd.datetime.now().date())})
        symbol_rank_df_copy['date_execute_end'] = symbol_rank_df_copy['date_execute_end'] + pd.to_timedelta('16:00:00')
        trading_period_ed_info_list.append(symbol_rank_df_copy)
    if len(trading_period_ed_info_list) > 0:
        trading_period_ed_info_df = pd.concat(trading_period_ed_info_list)
    # 过滤时间
    trading_period_ed_info_df = trading_period_ed_info_df[
        (trading_period_ed_info_df['date_execute_start'] >= start_date) & (
                trading_period_ed_info_df['date_execute_end'] <= end_date)]
    return trading_period_ed_info_df


def get_last_updated_kline_date(model, stocklist=None, db_string=None):
    """
    获取对应model表中的最新的行情
    :param model:
    :param stocklist:
    :param db_string:
    :return:
    """
    if db_string is None:
        db_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    session = get_db_session(db_string)
    update_date_dict = dict()
    if issubclass(model, KLineView):
        q = session.query(model.stockfullcode, func.max(model.date_trade).label('last_update_date'))
        if stocklist is not None:
            q = q.filter(model.stockfullcode.in_(stocklist))
        q = q.group_by(model.stockfullcode)
        for stockfullcode, last_update_date, in q:
            update_date_dict[stockfullcode] = last_update_date
    elif issubclass(model, KLine):
        q = session.query(model.id_stock, func.max(model.date_trade).label('last_update_date'))
        if stocklist is not None:
            stock_id_list = list()
            for stockfullcode in stocklist:
                stock_id = get_stock_id(stockfullcode)
                if stock_id is not None:
                    stock_id_list.append(stock_id)
            q = q.filter(model.id_stock.in_(stock_id_list))
        q = q.group_by(model.id_stock)
        for stock_id, last_update_date, in q:
            db_stock = to_stockfullcode(stock_id)
            if db_stock is not None:
                update_date_dict[str(db_stock)] = last_update_date
    session.close()
    return update_date_dict


def to_stockfullcode(key):
    stockcode_id_dict = get_stockcode_id_dict()
    stockfullcode_id_dict = get_stock_id_dict()
    if isinstance(key, int):
        return stockfullcode_id_dict.get(key)
    elif isinstance(key, str):
        id = stockcode_id_dict.get(key)
        if id is None:
            return None
        return stockfullcode_id_dict.get(id)
    return None


@db_retry_on_exception(retry=5)
def get_latest_main_contract_df(engine_string=None):
    return_dict = dict()
    if engine_string is None:
        engine_string = constant.DB_HIS_STRING
    conn = pyodbc.connect(engine_string)
    cursor = conn.cursor()
    sql = f"""
    SELECT t.* FROM 
        (SELECT stockfullcode, stockfullcode_main, date_execute FROM vw_edinfo WHERE stockfullcode like '%9001' ORDER BY date_execute DESC) t 
    GROUP BY stockfullcode
    """
    data_rows = cursor.execute(sql)
    for ed_info in data_rows:
        return_dict[ed_info.stockfullcode] = ed_info.stockfullcode_main
    return return_dict


@db_retry_on_exception(retry=5)
def get_latest_main_contract_dict(market=None, rank=1, engine_string=None):
    ed_info_dict = dict()
    if engine_string is None:
        engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    session = get_db_session(engine_string)
    q = session.query(EdInfoView.stockfullcode, func.max(EdInfoView.stockfullcode_main))
    q = q.filter(EdInfoView.stockfullcode.like('%900' + str(rank)))
    if market is not None:
        if isinstance(market, enumerator.ENUM_MARKET):
            q = q.filter(EdInfoView.stockfullcode.startswith(market.name))
        elif isinstance(market, str):
            q = q.filter(EdInfoView.stockfullcode.startswith(market))
        else:
            q = q.filter(EdInfoView.stockfullcode.startswith(str(market)))
    q = q.group_by(EdInfoView.stockfullcode)
    for stockfullcode, stockfullcode_main in q.all():
        ed_info_dict[stockfullcode] = stockfullcode_main
    session.close()
    return ed_info_dict


def get_latest_ed_info_df(stockfullcode=None, db_string=None):
    """
    按品种获取当前最新的ed_info
    :param stockfullcode:
    :param db_string:
    :return:
    """
    ed_info_list = _get_ed_info_list(stockfullcode=stockfullcode, db_string=db_string)
    ed_info_df = pd.DataFrame.from_records(ed_info_list)
    latest_ed_info_df = pd.DataFrame()
    latest_ed_info_list = list()
    if not ed_info_df.empty:
        for ed_stockfullcode, ed_stockfullcode_info_df in ed_info_df.groupby(ed_info_df.stockfullcode):
            latest_ed_info_list.append(ed_stockfullcode_info_df[
                                           ed_stockfullcode_info_df['date_execute'] == ed_stockfullcode_info_df[
                                               'date_execute'].max()])
    if len(latest_ed_info_list) > 0:
        latest_ed_info_df = pd.concat(latest_ed_info_list)
        latest_ed_info_df.reset_index(inplace=True, drop=True)
    return latest_ed_info_df


@db_retry_on_exception(retry=5)
def get_latest_edinfo_dict_by_market(market=None, engine_string=None):
    ed_info_dict = dict()
    if engine_string is None:
        engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    session = get_db_session(engine_string)
    q = session.query(EdInfoView.stockfullcode, func.max(EdInfoView.date_ed))
    if market is not None:
        if isinstance(market, enumerator.ENUM_MARKET):
            q = q.filter(EdInfoView.stockfullcode.startswith(market.name))
        elif isinstance(market, str):
            q = q.filter(EdInfoView.stockfullcode.startswith(market))
        else:
            q = q.filter(EdInfoView.stockfullcode.startswith(str(market)))
    q = q.group_by(EdInfoView.stockfullcode)
    for stockfullcode, date_ed in q.all():
        ed_info_dict[stockfullcode] = date_ed
    session.close()
    return ed_info_dict


def _get_ed_info_list(stockfullcode=None, start=constant.DATETIME_MIN, end=constant.DATETIME_MAX, db_string=None):
    '''
    获取数据库ed_info_list对象集合
    :param stockfullcode:
    :param start:
    :param end:
    :param db_string:
    :return:
    '''
    ed_info_list = list()
    if db_string is None:
        db_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    db_session = get_db_session(db_string)
    q = db_session.query(EdInfoView)
    if stockfullcode is not None:
        if isinstance(stockfullcode, list) and len(stockfullcode) > 0:
            q = q.filter(EdInfoView.stockfullcode.in_(stockfullcode))
        elif isinstance(stockfullcode, str):
            q = q.filter(EdInfoView.stockfullcode == stockfullcode)
    q = q.filter(EdInfoView.date_ed >= start)
    q = q.filter(EdInfoView.date_ed <= end)
    q = q.order_by(EdInfoView.date_ed)
    for edinfo in q.all():
        row = dict()
        row['id_stock'] = edinfo.id_stock
        row['stockfullcode'] = edinfo.stockfullcode
        row['id_main_stock'] = edinfo.id_main_stock
        row['stockfullcode_main'] = edinfo.stockfullcode_main
        row['date_ed'] = edinfo.date_ed
        row['date_execute'] = edinfo.date_execute
        row['sgbl'] = edinfo.sgbl
        row['fhbl'] = edinfo.fhbl
        row['pgbl'] = edinfo.pgbl
        row['pgjg'] = edinfo.pgjg
        row['preclose'] = edinfo.val_preclose,
        ed_info_list.append(row)
    db_session.close()
    return ed_info_list


def get_market_list(engine_string=None):
    global MARKET_LIST
    if not MARKET_LIST:
        if engine_string is None:
            engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
        session = get_db_session(engine_string)
        market_list = session.query(Market).all()
        MARKET_LIST = [m for m in market_list]
        session.close()
    return MARKET_LIST


def get_market_id(market_name, engine_string=None):
    market_list = get_market_list(engine_string=engine_string)
    for m in market_list:
        if m.market_name == market_name:
            return m.id_market
    return -1


def sync_marketid(source_engine_string, dest_engine_string, source_charset=None):
    src_session = get_db_session(source_engine_string)
    dst_session = get_db_session(dest_engine_string)
    src_max_id_market = src_session.query(func.max(Market.id_market).label('max_id_market')).one().max_id_market
    dst_max_id_market = dst_session.query(func.max(Market.id_market).label('max_id_market')).one().max_id_market
    if src_max_id_market is None:
        src_max_id_market = -1
    if dst_max_id_market is None:
        dst_max_id_market = -1
    updated_num = 0
    if src_max_id_market > dst_max_id_market:
        src_data_list = src_session.query(Market).filter(Market.id_market > dst_max_id_market).all()
        for src_data in src_data_list:
            if source_charset is not None:
                market_memo = src_data.memo.decode(source_charset)
            else:
                market_memo = src_data.memo
            new_market = Market(
                id_market=src_data.id_market,
                market_name=src_data.market_name,
                memo=market_memo
            )
            dst_session.add(new_market)
            updated_num += 1
        dst_session.commit()
    src_session.close()
    dst_session.close()
    return updated_num


def sync_stockid(source_engine_string, dest_engine_string, source_charset=None):
    src_session = get_db_session(source_engine_string)
    dst_session = get_db_session(dest_engine_string)
    src_max_id_stock = src_session.query(func.max(Stock.id_stock).label('max_id_stock')).one().max_id_stock
    dst_max_id_stock = dst_session.query(func.max(Stock.id_stock).label('max_id_stock')).one().max_id_stock
    if src_max_id_stock is None:
        src_max_id_stock = -1
    if dst_max_id_stock is None:
        dst_max_id_stock = -1
    updated_num = 0
    if src_max_id_stock > dst_max_id_stock:
        src_data_list = src_session.query(Stock).filter(Stock.id_stock > dst_max_id_stock).all()
        for src_data in src_data_list:
            stockname = ''
            if src_data.stockname is not None:
                stockname = src_data.stockname
            if source_charset is not None:
                stockname = stockname.decode(source_charset)
            new_stock = Stock(
                id_stock=src_data.id_stock,
                id_market=src_data.id_market,
                stockcode=src_data.stockcode,
                stockfullcode=src_data.stockfullcode,
                stockname=stockname
            )
            dst_session.add(new_stock)
            updated_num += 1
        dst_session.commit()
    src_session.close()
    dst_session.close()
    return updated_num


@db_retry_on_exception(retry=5)
def get_strategy_list(engine_string=None, load_cache=True):
    global STRATEGY_LIST
    if not STRATEGY_LIST or not load_cache:
        if engine_string is None:
            engine_string = constant.DB_TRADE_RD_ENGINE_STRING
        session = get_db_session(engine_string)
        strategy_list = session.query(Strategy).all()
        STRATEGY_LIST = [strategy for strategy in strategy_list]
        session.close()
    return STRATEGY_LIST


def get_strategy_id_dict(engine_string=None, load_cache=True):
    global STRATEGY_ID_DICT
    if not STRATEGY_ID_DICT:
        strategy_list = get_strategy_list(engine_string=engine_string, load_cache=load_cache)
        for strategy in strategy_list:
            STRATEGY_ID_DICT[strategy.name] = strategy.id
    return STRATEGY_ID_DICT


def get_strategy_id(strategy_name, load_cache=True, engine_string=None):
    if engine_string is None:
        engine_string = constant.DB_TRADE_RD_ENGINE_STRING
    strategy_id_dict = get_strategy_id_dict(engine_string=engine_string, load_cache=load_cache)
    return strategy_id_dict.get(strategy_name, None)


def get_stock_list(engine_string=None):
    global STOCK_LIST
    if not STOCK_LIST:
        if engine_string is None:
            engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
        session = get_db_session(engine_string)
        stocks = session.query(Stock).all()
        STOCK_LIST = [s for s in stocks]
        session.close()
    return STOCK_LIST


def get_stock_id_dict(engine_string=None):
    global STOCKFULLCODE_ID_DICT
    if not STOCKFULLCODE_ID_DICT:
        stock_list = get_stock_list(engine_string=engine_string)
        for stock in stock_list:
            STOCKFULLCODE_ID_DICT[stock.stockfullcode] = stock.id_stock
    return STOCKFULLCODE_ID_DICT


def get_stockcode_id_dict(engine_string=None):
    global STOCKCODE_ID_DICT
    if not STOCKCODE_ID_DICT:
        stock_list = get_stock_list(engine_string=engine_string)
        for stock in stock_list:
            STOCKCODE_ID_DICT[stock.stockcode] = stock.id_stock
    return STOCKCODE_ID_DICT


def get_stock_id(stockfullcode, engine_string=None):
    if engine_string is None:
        engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    stock_id_dict = get_stock_id_dict(engine_string)
    return stock_id_dict.get(stockfullcode, None)


def add_tday(tday_list, engine_string=None):
    """
    数据库插入交易日数据
    :param tday_list:
    :param engine_string:
    :return:
    """
    if engine_string is None:
        engine_string = constant.DEFAULT_DB_WR_ENGINE_STRING
    session = get_db_session(engine_string)
    tday_dict = [dict(tday=tday) for tday in tday_list]
    session.bulk_insert_mappings(Tday, tday_dict)
    session.commit()
    return len(tday_list)


def add_stock(stockfullcode, stockname='', engine_string=None):
    if engine_string is None:
        engine_string = constant.DEFAULT_DB_WR_ENGINE_STRING
    session = get_db_session(engine_string=engine_string)
    existed_stock = session.query(Stock).filter(Stock.stockfullcode == stockfullcode).all()
    id_stock = -1
    if not existed_stock:
        market_name, stockcode = split_market(stockfullcode)
        if not market_name:
            session.close()
            return -1
        id_market = get_market_id(market_name)
        if id_market < -1:
            session.close()
            return -1
        max_id_stock = session.query(func.max(Stock.id_stock).label('max_id_stock')).one().max_id_stock
        if max_id_stock is None:
            id_stock = 1
        else:
            id_stock = max_id_stock + 1
        new_stock = Stock(
            id_stock=id_stock,
            id_market=id_market,
            stockcode=stockcode,
            stockfullcode=stockfullcode,
            stockname=stockname
        )
        session.add(new_stock)
        session.commit()
    session.close()
    return id_stock


def remove_rt_klines(start_date, end_date, stockfullcode, period, db_string=None):
    if db_string is None:
        db_string = constant.RT03_DB_RT_WR_ENGINE_STRING
    session = get_db_session(db_string)
    id_stock = get_stock_id(stockfullcode)
    kline_model = OneMinute
    if period == '1m':
        kline_model = OneMinute
    elif period == '5m':
        kline_model = FiveMinutes
    to_delete_rows = session.query(kline_model).filter(kline_model.id_stock == id_stock).filter(
        kline_model.date_trade >= start_date).filter(kline_model.date_trade <= end_date)
    len_of_delete_row = len(to_delete_rows.all())
    if len_of_delete_row > 0:
        to_delete_rows.delete()
        print(u"[共删除]: 品种 '{}', 周期: '{}' 在交易日从 '{}' 到 '{}' 分钟线, 共 {} 条数据".format(stockfullcode, period, start_date,
                                                                                 end_date, len_of_delete_row))
        session.commit()
    session.close()


def add_rt_klines(kline_list, period, max_insert_num=800, db_string=None):
    '''
    往RT数据库中插入K线
    :param kline_list:
    :param period:
    :param max_insert_num:
    :param db_string:
    :return:
    '''
    insert_num = -1
    if db_string is None:
        db_string = constant.RT03_DB_RT_WR_ENGINE_STRING
    session = get_db_session(db_string)
    divided_kline_list = _split_list_by_divide_unit(list_data=kline_list, divide_unit=max_insert_num)
    try:
        for each_kline_list in divided_kline_list:
            insert_kline_list = list()
            for kline in each_kline_list:
                id_stock = get_stock_id(kline['contract'])
                if id_stock is None or id_stock < 0:
                    continue
                if period == '1m':
                    new_kline = OneMinute(
                        id_stock=id_stock,
                        date_trade=kline['date_trade'],
                        val_preclose=kline['pre_close'],
                        val_open=kline['open'],
                        val_high=kline['high'],
                        val_low=kline['low'],
                        val_close=kline['close'],
                        amt_amount=kline['amount'],
                        amt_volume=kline['vol'],
                        amt_wp=0,
                        amt_np=0,
                        val_OI=kline['oi'] if kline.get('oi', None) is not None else -1
                    )
                elif period == '5m':
                    new_kline = FiveMinutes(
                        id_stock=id_stock,
                        date_trade=kline['date_trade'],
                        val_preclose=kline['pre_close'],
                        val_open=kline['open'],
                        val_high=kline['high'],
                        val_low=kline['low'],
                        val_close=kline['close'],
                        amt_amount=kline['amount'],
                        amt_volume=kline['vol'],
                        amt_wp=0,
                        amt_np=0,
                        val_OI=kline['oi'] if kline.get('oi', None) is not None else -1
                    )
                insert_kline_list.append(new_kline)
            session.bulk_save_objects(insert_kline_list)
            session.commit()
            print('[插入数据库]: 周期: {}, 共{}条'.format(period, len(insert_kline_list)))
            insert_num += len(insert_kline_list)
    except Exception as ex:
        print('插入数据库出错: {}'.format(ex))
        insert_num = -1
    finally:
        session.close()
    return insert_num


# 根据市场名获取主力合约 (9001/9991)
def get_main_future_contract_code(market_name=None, engine_string=None, rank=0, include_expired=False):
    """
    获取主力合约的集合
    :param market_name:
    :param engine_string:
    :param rank: 0:9001, 1:9991, 2:9992
    :param include_expired: 是否需要获取旧合约和过期合约
    :return:
    """
    main_contract_list = list()
    market_name_list = list()
    if market_name is None:
        market_name_list = ['SF', 'DL', 'ZZ', 'SQ', 'SE']
    else:
        if isinstance(market_name, str):
            market_name_list.append(market_name)
        elif isinstance(market_name, list):
            market_name_list = market_name
    if engine_string is None:
        engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    session = get_db_session(engine_string)
    for each_market in market_name_list:
        q = session.query(EdInfoView.stockfullcode).filter(
            EdInfoView.stockfullcode.like('{}%'.format(each_market))).order_by(desc(EdInfoView.stockfullcode))
        if rank == 0:
            q = q.filter(EdInfoView.stockfullcode.like('%9001'))
        else:
            q = q.filter(EdInfoView.stockfullcode.like('%999' + str(rank)))
        main_contracts = q.distinct().all()
        for contract in main_contracts:
            str_contract = str(contract[0])
            if include_expired:
                main_contract_list.append(str_contract)
            else:
                if str_contract not in constant.EXPIRED_MAIN_CONTRACT:
                    main_contract_list.append(str_contract)
    session.close()
    return main_contract_list


@db_retry_on_exception(retry=5)
def get_main_future_contract(stockfullcode, tday=None, engine_string=None):
    if engine_string is None:
        engine_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    main_contract = None
    session = get_db_session(engine_string)
    q = session.query(EdInfoView).filter(EdInfoView.stockfullcode == stockfullcode)
    if tday is not None:
        tday = pd.to_datetime(tday).to_pydatetime()
        q = q.filter(EdInfoView.date_ed < tday)
    q = q.order_by(desc(EdInfoView.date_ed))
    edinfo = q.first()
    if edinfo:
        main_contract = str(edinfo.stockfullcode_main)
    session.close()
    return main_contract


@db_retry_on_exception(retry=5)
def get_ed_factor_info_org(stockfullcode=None, start='1900-01-01', end='2050-01-01', return_dict=False, db_string=None):
    if isinstance(start, str):
        start = dt.datetime.strptime(start, '%Y-%m-%d')
    if isinstance(end, str):
        end = dt.datetime.strptime(end, '%Y-%m-%d')
    assert isinstance(start, dt.datetime)
    assert isinstance(end, dt.datetime)
    if db_string is None:
        db_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    ed_info_dict = defaultdict(list)
    ed_info_list = []
    session = get_db_session(db_string)
    q = session.query(EdFactorInfoView)
    if stockfullcode is not None:
        if isinstance(stockfullcode, list) and len(stockfullcode) > 0:
            q = q.filter(EdFactorInfoView.stockfullcode.in_(stockfullcode))
        elif isinstance(stockfullcode, str):
            q = q.filter(EdFactorInfoView.stockfullcode == stockfullcode)
    q = q.filter(EdFactorInfoView.date_execute >= start)
    q = q.filter(EdFactorInfoView.date_execute <= end)
    q = q.order_by(EdFactorInfoView.date_execute)
    for edinfo in q.all():
        if return_dict:
            ed_info_dict[edinfo.stockfullcode].append(edinfo)
        else:
            ed_info_list.append(edinfo)
    session.close()
    if return_dict:
        return ed_info_dict
    return ed_info_list


def get_ed_factor_info_df(stockfullcode=None, start='1900-01-01', end='2050-01-01', db_string=None):
    edinfo_list = get_ed_factor_info_org(stockfullcode=stockfullcode, start=start, end=end, return_dict=False,
                                         db_string=None)
    data = [[edinfo.id_stock, edinfo.stockfullcode, edinfo.id_main_stock, edinfo.stockfullcode_main,
             edinfo.date_execute, edinfo.fhbl, edinfo.pgbl, edinfo.pgjg, edinfo.sgbl, edinfo.is_valid]
            for edinfo in edinfo_list]
    edinfo_df = pd.DataFrame(data=data, index=np.arange(0, len(edinfo_list)),
                             columns=['id_stock', 'stockfullcode', 'id_main_stock', 'stockfullcode_main',
                                      'date_execute', 'fhbl', 'pgbl', 'pgjg', 'sgbl', 'is_valid'])
    return edinfo_df


def add_ed_factor_info(ed_factor_info_df, db_string=None):
    if db_string is None:
        db_string = constant.DEFAULT_DB_WR_ENGINE_STRING
    dic_all_stockid = get_stock_id_dict()
    dic_stockid = dict([(code, dic_all_stockid[code]) for code in ed_factor_info_df['stockfullcode'].unique()])
    dic_main_stockid = dict([(code, dic_all_stockid[code]) for code in ed_factor_info_df['stockfullcode_main'].unique()
                             if code and not np.isnan(code)])
    dic_main_stockid[''] = None
    ed_factor_info_df['stockfullcode'].replace(dic_stockid, inplace=True)
    ed_factor_info_df['stockfullcode_main'].replace(dic_main_stockid, inplace=True)
    ed_factor_info_df['stockfullcode_main'].replace({np.nan: None}, inplace=True)
    ed_factor_info_df.rename(columns={'stockfullcode': 'id_stock', 'stockfullcode_main': 'id_main_stock'}, inplace=True)
    session = get_db_session(db_string)
    try:
        for _, grp in ed_factor_info_df[['id_stock', 'id_main_stock', 'date_execute', 'fhbl', 'pgbl', 'pgjg', 'sgbl',
                                         'is_valid']].groupby('id_stock', as_index=False):
            dic_values = grp.to_dict(orient='records')
            insert_stmt = insert(EdFactorInfo).values(dic_values)
            on_duplicate_key_stmt = insert_stmt.on_duplicate_key_update(
                fhbl=insert_stmt.inserted.fhbl,
                pgbl=insert_stmt.inserted.pgbl,
                pgjg=insert_stmt.inserted.pgjg,
                sgbl=insert_stmt.inserted.sgbl,
                is_valid=insert_stmt.inserted.is_valid)
            session.execute(on_duplicate_key_stmt)
        session.commit()
    except Exception as ex:
        logger.error('Upsert ed factor info failed, error:{}'.format(ex.message))
        session.rollback()
    finally:
        session.close()


@db_retry_on_exception(retry=5)
def get_ed_factor_org(stockfullcode=None, start='1900-01-01', end='2050-01-01', return_dict=False, db_string=None):
    if isinstance(start, str):
        start = dt.datetime.strptime(start, '%Y-%m-%d')
    if isinstance(end, str):
        end = dt.datetime.strptime(end, '%Y-%m-%d')
    assert isinstance(start, dt.datetime)
    assert isinstance(end, dt.datetime)
    if db_string is None:
        db_string = constant.DEFAULT_DB_RD_ENGINE_STRING
    ed_factor_dict = defaultdict(list)
    ed_factor_list = []
    session = get_db_session(db_string)
    q = session.query(EdFactorView)
    if stockfullcode is not None:
        if isinstance(stockfullcode, list) and len(stockfullcode) > 0:
            q = q.filter(EdFactorView.stockfullcode.in_(stockfullcode))
        elif isinstance(stockfullcode, str):
            q = q.filter(EdFactorView.stockfullcode == stockfullcode)
    q = q.filter(EdFactorView.date_execute >= start)
    q = q.filter(EdFactorView.date_execute <= end)
    q = q.order_by(EdFactorView.date_execute)
    for edinfo in q.all():
        if return_dict:
            ed_factor_dict[edinfo.stockfullcode].append(edinfo)
        else:
            ed_factor_list.append(edinfo)
    session.close()
    if return_dict:
        return ed_factor_dict
    return ed_factor_list


def get_ed_factor_df(stockfullcode=None, start='1900-01-01', end='2050-01-01', db_string=None):
    edfactor_list = get_ed_factor_org(stockfullcode=stockfullcode, start=start, end=end, return_dict=False,
                                      db_string=None)
    data = [[ed_factor.id_stock, ed_factor.stockfullcode, ed_factor.id_main_stock, ed_factor.stockfullcode_main,
             ed_factor.date_execute, ed_factor.factor, ed_factor.is_valid]
            for ed_factor in edfactor_list]
    ed_factor_df = pd.DataFrame(data=data, index=np.arange(0, len(edfactor_list)),
                                columns=['id_stock', 'stockfullcode', 'id_main_stock', 'stockfullcode_main',
                                         'date_execute', 'factor', 'is_valid'])
    return ed_factor_df


def add_ed_factor(ed_factor_df, db_string=None):
    if db_string is None:
        db_string = constant.DEFAULT_DB_WR_ENGINE_STRING
    dic_all_stockid = get_stock_id_dict()
    dic_stockid = dict([(code, dic_all_stockid[code]) for code in ed_factor_df['stockfullcode'].unique()])
    dic_main_stockid = dict([(code, dic_all_stockid[code]) for code in ed_factor_df['stockfullcode_main'].unique()
                             if code])
    dic_main_stockid[''] = None
    ed_factor_df['stockfullcode'].replace(dic_stockid, inplace=True)
    ed_factor_df['stockfullcode_main'].replace(dic_main_stockid, inplace=True)
    ed_factor_df['stockfullcode_main'].replace({np.nan: None}, inplace=True)
    ed_factor_df.rename(columns={'stockfullcode': 'id_stock', 'stockfullcode_main': 'id_main_stock'}, inplace=True)
    session = get_db_session(db_string)
    try:
        for _, grp in ed_factor_df[['id_stock', 'id_main_stock', 'date_execute', 'factor', 'is_valid'
                                    ]].groupby('id_stock', as_index=False):
            dic_values = grp.to_dict(orient='records')
            insert_stmt = insert(EdFactor).values(dic_values)
            on_duplicate_key_stmt = insert_stmt.on_duplicate_key_update(
                factor=insert_stmt.inserted.factor,
                is_valid=insert_stmt.inserted.is_valid)
            session.execute(on_duplicate_key_stmt)
        session.commit()
    except Exception as ex:
        logger.error('Upsert ed factor info failed, error:{}'.format(ex.message))
        session.rollback()
    finally:
        session.close()


@db_retry_on_exception(retry=5)
def add_ed_info(ed_info, add_execute_info=True):
    if not isinstance(ed_info, quote_info.EdInfo) and not isinstance(ed_info, EdInfoView):
        logger.error(u'add_ed_info只接受EdInfo的对象作为参数')
        return False
    id_continue_contract = get_stock_id(ed_info.stockfullcode)
    if id_continue_contract is None:
        logger.error(u'找不到{}的stock id！'.format(ed_info.stockfullcode))
        return False
    id_main_contract = get_stock_id(ed_info.stockfullcode_main)
    if id_main_contract is None:
        logger.error(u'找不到{}的stock id！'.format(ed_info.stockfullcode_main))
        return False
    session = get_db_session(engine_string=constant.DEFAULT_DB_WR_ENGINE_STRING)
    db_edinfo = EdInfo(
        id_stock=id_continue_contract,
        date_ed=ed_info.date_ed,
        fhbl=ed_info.fhbl,
        pgbl=ed_info.pgbl,
        pgjg=ed_info.pgjg,
        sgbl=ed_info.sgbl,
        flg_calc_factor=False,
        id_main_stock=id_main_contract
    )
    # check whether to add same edinfo again
    query_edinfo_row = session.query(EdInfo) \
        .filter(EdInfo.id_stock == id_continue_contract) \
        .filter(EdInfo.date_ed == ed_info.date_ed).filter(EdInfo.id_main_stock == id_main_contract)
    if len(query_edinfo_row.all()) == 0:
        session.add(db_edinfo)
    else:
        logger.info(u'无需更新，已存在[ed_info]记录为: id_stock: {}, date_ed: {}, id_main_stock: {}'.format(id_continue_contract,
                                                                                                 ed_info.date_ed,
                                                                                                 id_main_contract))
    if add_execute_info:
        db_ed_exe = EdExecuteInfo(
            id_stock=id_continue_contract,
            date_ed=ed_info.date_ed,
            date_execute=ed_info.date_execute,
            val_preclose=ed_info.preclose
        )
        # check whether to add same ed_execute_info again
        query_ed_execute_info = session.query(EdExecuteInfo) \
            .filter(EdExecuteInfo.id_stock == id_continue_contract) \
            .filter(EdExecuteInfo.date_ed == ed_info.date_ed) \
            .filter(EdExecuteInfo.date_execute == ed_info.date_execute)
        if len(query_ed_execute_info.all()) == 0:
            session.add(db_ed_exe)
        else:
            logger.info(u'无需更新，已存在[ed_execute_info]记录为: id_stock: {}, date_ed: {}, date_execute: {}'
                        .format(id_continue_contract, ed_info.date_ed, ed_info.date_execute))
    session.commit()
    session.close()
    return True


def del_ed_info(ed_info):
    """
    删除ed_info信息
    :param ed_info:
    :return:
    """
    if not isinstance(ed_info, EdInfo) and not isinstance(ed_info, EdInfoView):
        logger.error(u'del_ed_info只接受EdInfo的对象作为参数')
        return
    id_continue_contract = get_stock_id(ed_info.stockfullcode)
    if id_continue_contract is None:
        logger.error(u'找不到{}的stock id！'.format(ed_info.stockfullcode))
        return
    session = get_db_session(engine_string=constant.DEFAULT_DB_WR_ENGINE_STRING)
    session.query(EdInfo).filter(EdInfo.id_stock == id_continue_contract). \
        filter(EdInfo.date_ed == ed_info.date_ed).delete()
    session.query(EdExecuteInfo).filter(EdExecuteInfo.id_stock == id_continue_contract). \
        filter(EdExecuteInfo.date_ed == ed_info.date_ed).delete()
    session.commit()
    session.close()


@db_retry_on_exception(retry=5)
def get_tday_contracts_df(market_name=None, symbol=None, stockcode=None, stockfullcode=None,
                          start_tday=constant.DATETIME_MIN, end_tday=constant.DATETIME_MAX, dataframe=True):
    """
    读取交易日合约持仓量排名
    :param start_tday:
    :param end_tday:
    :param market_name:
    :param symbol:
    :param stockcode:
    :param stockfullcode:
    :param dataframe:
    :return:
    """
    tday_contracts_list = list()
    start_tday = pd.to_datetime(start_tday).strftime('%Y-%m-%d')
    end_tday = pd.to_datetime(end_tday).strftime('%Y-%m-%d')
    list_market_name = [market_name, ] if isinstance(market_name, str) else market_name
    list_symbol = [symbol, ] if isinstance(symbol, str) else symbol
    list_stockcode = [stockcode, ] if isinstance(stockcode, str) else stockcode
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    try:
        db_session = get_db_session(constant.DEFAULT_DB_RD_ENGINE_STRING)
        q = db_session.query(TdaySymbolStocksView.tday, TdaySymbolStocksView.market_name,
                             TdaySymbolStocksView.stockfullcode, TdaySymbolStocksView.stockcode,
                             TdaySymbolStocksView.symbol, TdaySymbolStocksView.val_OI)
        conditions = None
        if market_name is not None:
            conditions = TdaySymbolStocksView.market_name.in_(list_market_name)
        if symbol is not None:
            conditions = TdaySymbolStocksView.symbol.in_(list_symbol) if conditions is None else or_(conditions,
                                                                                                     TdaySymbolStocksView.symbol.in_(
                                                                                                         list_symbol))
        if stockcode is not None:
            conditions = TdaySymbolStocksView.stockcode.in_(list_stockcode) if conditions is None else or_(conditions,
                                                                                                           TdaySymbolStocksView.stockcode.in_(
                                                                                                               list_stockcode))
        if stockfullcode is not None:
            conditions = TdaySymbolStocksView.stockfullcode.in_(list_stockfullcode) if conditions is None else or_(
                conditions,
                TdaySymbolStocksView.stockfullcode.in_(
                    list_stockfullcode))
        if conditions is None:
            conditions = and_(TdaySymbolStocksView.tday >= start_tday, TdaySymbolStocksView.tday <= end_tday)
        else:
            conditions = and_(conditions, TdaySymbolStocksView.tday >= start_tday,
                              TdaySymbolStocksView.tday <= end_tday)
        q = q.filter(conditions).order_by(asc(TdaySymbolStocksView.tday))
        q = q.order_by(asc(TdaySymbolStocksView.symbol))
        q = q.order_by(desc(TdaySymbolStocksView.val_OI))
        tday_contracts_list = q.all()
        tday_contracts_list = [{key: getattr(each_item, key) for key in each_item._fields}
                               for each_item in tday_contracts_list]
    except Exception as ex:
        msg = f'读取合约持仓量记录失败:\r\n{traceback.format_exc()}'
        logger.error(msg)
    finally:
        db_session.close()
    if dataframe:
        result_df = pd.DataFrame.from_dict(tday_contracts_list)
        if not result_df.empty:
            result_df = result_df.astype({'tday': np.str}, copy=True)
        return result_df
    else:
        return tday_contracts_list


def _get_history_data_org_from_rt(stockfullcode, start_date=None, end_date=None, fields=None, freq='5m',
                                  ed_type=None, dataframe=True):
    """
    从实时数据库中获取行情信息，默认是rt03
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param jq_fields:
    :param freq:
    :param ed_type:
    :param dataframe:
    :return:
    """
    start_date = _to_datetime(start_date).strftime('%Y-%m-%d %H:%M:%S')
    end_date = _to_datetime(end_date).strftime('%Y-%m-%d %H:%M:%S')
    # 数据库中默认只有1分钟/5分钟的数据，所以如果小于等于1分钟，则取一分钟，如果大于等于1分钟，则取5分钟，更大级别的数据全部都用5分钟数据进行转化
    freq_num = convert_freq2num(freq_str=freq)
    view_table = t_VW_oneminute_history if freq_num <= 60 else t_VW_fiveminutes_history
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    # 建立数据库连接
    db_string = constant.RT03_DB_RT_RD_ENGINE_STRING
    try:
        db_session = get_db_session(db_string)
        q = db_session.query(view_table)
        conditions = None
        if list_stockfullcode is not None:
            conditions = view_table.columns['stockfullcode'].in_(list_stockfullcode)
        if conditions is None:
            conditions = and_(
                view_table.columns['date_trade'] >= start_date, view_table.columns['date_trade'] <= end_date)
        else:
            conditions = and_(conditions, view_table.columns['date_trade'] >= start_date,
                              view_table.columns['date_trade'] <= end_date)
        rt_kline_list = q.filter(conditions).all()
        rt_kline_list = [{key: getattr(rt_kline, key) for key in rt_kline._fields}
                         for rt_kline in rt_kline_list]
        rt_kline_df = pd.DataFrame.from_dict(rt_kline_list)
        if fields is not None:
            rt_kline_df = rt_kline_df[[fields]]
        return rt_kline_df
    except Exception as ex:
        print(
            f'读取实时数据库失败:品种:{list_stockfullcode}, 周期:{freq}, 开始:{start_date}, 结束:{end_date} \r\n{traceback.format_exc()}')


def calculate_contract_expire_data(year, month, day_of_week=calendar.FRIDAY, sequence=3):
    """
    通过传入的年份和月份，自动计算股指期货合约的交割日期和过期时间
    :param deliver_year:
    :param deliver_month:
    :param day_of_week:
    :param sequence:
    :return:
    """
    c = calendar.Calendar(firstweekday=day_of_week)
    # 根据year和month确定年和月
    month_cal = c.monthdatescalendar(year=year, month=month)
    # 不能直接获取指定顺位的日期， 法定交易日的指定会延后时间，所以这里需要根据get_tday_list判断
    first_day_of_week = \
        [day for week in month_cal for day in week if day.weekday() == day_of_week and day.month == month][0]
    # 拿到第一个星期X后，加14天
    target_day_of_week = first_day_of_week + pd.Timedelta(days=7 * (sequence - 1))
    # 获取交易日
    tday_list = get_tday_list()
    # 转化为datetime
    tday_list = pd.to_datetime(tday_list).tolist()
    # 从get_tday_list中找到大于或等于这个日期的即可
    day_of_week_tday_list = [item for item in tday_list if item.date() >= target_day_of_week]
    # 如果有，则进行相对性比较
    if len(day_of_week_tday_list) > 0:
        if target_day_of_week != day_of_week_tday_list[0].date():
            if day_of_week_tday_list[0].weekday_name == calendar.day_name[day_of_week]:
                target_day_of_week = day_of_week_tday_list[0].date()
    return target_day_of_week


class TestDBIO(unittest.TestCase):
    def setUp(self):
        setup_screen_logger()
        self.logger = logging.getLogger('root')

    def tearDown(self):
        pass

    def test_calculate_contract_expire_data(self):
        deliver_year = 2021
        deliver_month = 10
        day_of_week = calendar.MONDAY
        sequency = 2
        deliver_date = calculate_contract_expire_data(year=deliver_year, month=deliver_month, day_of_week=day_of_week,
                                                      sequence=sequency)
        pass

    def test_get_rt_kline_df(self):
        start_date = '2020-10-01'
        end_date = '2020-10-23'
        freq = '5m'
        stockfullcode = 'SH000300'
        rt_kline_df = _get_history_data_org_from_rt(stockfullcode=stockfullcode, start_date=start_date,
                                                    end_date=end_date, freq=freq)
        pass

    def test_get_tday_contracts_df(self):
        start_tday = '2014-01-01'
        end_tday = '2020-12-31'
        symbol = 'rb'
        market_name = 'SF'
        stockcode = 'a2009'
        stockfullcode = 'SFIF2008'
        tday_contract_df_tday = get_tday_contracts_df(start_tday=start_tday, end_tday=end_tday)
        tday_contract_df_symbol = get_tday_contracts_df(start_tday=start_tday, end_tday=end_tday, symbol=symbol)
        tday_contract_df_market = get_tday_contracts_df(start_tday=start_tday, end_tday=end_tday,
                                                        market_name=market_name)
        tday_contract_df_stockcode = get_tday_contracts_df(stockcode=stockcode)
        tday_contract_df_stockfullcode = get_tday_contracts_df(stockfullcode=stockfullcode)
        pass

    def test_get_tdays(self):
        tdays = get_tday_list('2020-01-01', '2022-12-31')
        self.assertEqual(len(tdays), 244)

    def test_get_previous_tday(self):
        previous_tday = get_previous_tday('2020-06-02')
        self.assertEqual('2020-06-01', previous_tday)

    def test_get_next_tday(self):
        next_tday = get_next_tday('2020-04-26')
        self.assertEqual('2020-06-03', next_tday)

    def test_get_future_info_dict(self):
        stockfullcode1 = 'SFIC9001'
        stockfullcode2 = 'ZZAP005'
        stockfullcode3 = 'DLau9001'

        future_volume_multiple_dict = _get_future_volume_multiple_dict()
        stockfullcode1_volumne_multiple = get_volume_multiple(stockfullcode=stockfullcode1)
        stockfullcode2_volumne_multiple = get_volume_multiple(stockfullcode=stockfullcode2)
        stockfullcode3_volumne_multiple = get_volume_multiple(stockfullcode=stockfullcode3)
        self.assertEqual(stockfullcode1_volumne_multiple, 200)
        self.assertEqual(stockfullcode2_volumne_multiple, 10)
        self.assertEqual(stockfullcode3_volumne_multiple, 1000)

    def test_get_ed_info_df(self):
        stockfullcode = 'ZZAP9001'
        ed_info_df1 = get_ed_info_df(stockfullcode=stockfullcode)
        ed_info_df2 = get_ed_info_df(stockfullcode=stockfullcode, start='2019-01-01', end='2019-12-31')
        self.assertEqual(len(ed_info_df1) > 0, True)
        self.assertEqual(len(ed_info_df2), 3)

    def test_get_future_main_ed_info(self):
        ed_info_df = get_ed_info_df()
        ed_info_df = ed_info_df[~ed_info_df.stockfullcode.str.contains('SH')]
        ed_info_df = ed_info_df[~ed_info_df.stockfullcode.str.contains('SZ')]
        ed_info_df = ed_info_df[ed_info_df.stockfullcode.str.contains('9001')]
        ed_info_df.sort_values(by=['stockfullcode', 'date_ed'], ascending=[True, True], inplace=True)
        ed_info_df.to_csv(r'D:\TempCSV\futures_ed_info_{}.csv'.format(dt.datetime.now().strftime('%Y%m%d')))

    def test_get_strategy_id(self):
        strategy_name = 'HKQSignalModel'
        id_strategy = get_strategy_id(strategy_name=strategy_name)
        pass

    def test_get_ed_factor_info_df(self):
        ed_factor_info_df = get_ed_factor_info_df(stockfullcode='SH600000')
        pass

    def test_get_main_future_contract_code(self):
        filter_symbol_lists = ['RM', 'TA', 'AL', 'AP', 'CU', 'MA', 'C', 'CF', 'I', 'M', 'NI', 'ZC',
                               'ZN', 'AU', 'OI', 'RB', 'RU', 'SR', 'Y', 'A', 'AG', 'B', 'BU', 'CS',
                               'EG', 'FG', 'FU', 'HC', 'J', 'JD', 'JM', 'L', 'P', 'PB', 'PP', 'SF',
                               'SM', 'V', 'CY', 'RS', 'SN', 'SP']
        return_stock_lists = list()
        main_future_codes = get_main_future_contract_code()
        for each_code in main_future_codes:
            symbol = _get_future_symbol(each_code)
            if symbol.upper() in filter_symbol_lists:
                return_stock_lists.append(each_code)
        pass

    def test_get_future_info(self):
        stockfullcode = 'SQrb1703'
        symbol = 'a'
        start = '2019-01-01'
        end = '2020-12-31'
        future_info_stockfullcode = get_future_info(stockfullcode=stockfullcode)
        future_info_symbol = get_future_info(symbol=symbol, open_date=start, expire_date=end)
        future_info_date_range = get_future_info(open_date=start, expire_date=end)
        pass

    def test_get_latest_ed_info_df(self):
        stockfullcode = ['SQrb9001', 'SFIF9002', 'DLj9003']
        get_latest_ed_info_df(stockfullcode=stockfullcode)
        pass

    def test_get_tday_df(self):
        tday_df = get_tday_df()
        pass

    def test_get_ed_info_trading_period(self):
        trading_period = get_ed_info_trading_period(stockfullcode='ZZAP9001')
        pass

    def test_correct_ed_info_df(self):
        ed_info_df = get_ed_info_df(stockfullcode='DLjm9001')
        ed_info_df = correct_ed_info_df(ed_info_df)
        pass

    def test_get_latest_edinfo_dict_by_market(self):
        market = enumerator.ENUM_MARKET.DL
        latest_ed_dict = get_latest_edinfo_dict_by_market(market=market)
        pass

    def test_get_last_updated_kline_date(self):
        latest_quote_df = get_last_updated_kline_date(model=OneDayView)
        pass

    def test_calc_main_contract(self):
        market = enumerator.ENUM_MARKET.SQ
        tday = '2020-12-30'
        main_contract_dict = calc_main_contract(tday=tday, market=market, freq='1d', field='oi')
        pass

    def test_add_edinfo(self):
        ed_info = quote_info.EdInfo(
            name='ZZPK9001',
            date_ed='2021-12-13',
            date_execute='2021-12-14',
            stockfullcode_main='ZZPK204',
            sgbl=-0.032289156627,
            preclose=0
        )
        logger.info('增加新的EdInfo - {}'.format(ed_info))
        succeed = add_ed_info(ed_info=ed_info)
        pass

    def test_get_trading_future_info(self):
        trading_contract_list = get_trading_contract_list(symbol='IF')
        pass

    def test_get_latest_main_contract_dict(self):
        main_contract_dict = get_latest_main_contract_dict(market='SF')
        pass

    def test_get_next_contract(self):
        current_stock = 'SFIF2104'
        next_stock = get_next_contract(current_stock)
        pass

    def test_del_tday_symbol_stocks(self):
        symbol = 'IF'
        start = '2021-05-26'
        end = '2021-05-26'
        result = del_tday_symbol_stocks(symbol=symbol, start=start, end=end)
        pass

    def test_add_tday_symbol_stocks(self):
        symbol = 'IF'
        tday = '2021-05-26'
        result = add_tday_symbol_stocks(symbol=symbol, tday=tday)
        pass

    def test_get_holiday_df(self):
        result = get_holiday_df()
        pass

    def test_get_latest_main_contract_df(self):
        result = get_latest_main_contract_df()
        pass
