# -*- coding:utf-8 -*-
import os
import gc
import copy
import unittest
import numpy as np
import pandas as pd
import logging
import datetime as dt
import pyodbc
import h5py as h5
import matplotlib.pyplot as plt
import rio3.datafeed.feed_rio as feed_rio
from tqdm import tqdm
from sqlalchemy import create_engine, or_, and_, desc, sql, asc
from sqlalchemy.sql import func
from collections import defaultdict

try:
    from QuantLib import *
except:
    print('QuantLib库为期权计算希腊字母的Python依赖, 请按需求本地安装， pip install QuantLib')
try:
    from rio3.datafeed.rlibs import COptionPrice
except:
    print('COptionPrice库为期权计算希腊字母的C++ DLL依赖, 请按需求本地安装')
from rio3.log.setup import setup_screen_logger, setup_file_logger
from datetime import timedelta
from rio3.base import constant
from rio3.datafeed.research.tushareIO import _get_trading_datetime_range, _get_trading_days
from rio3.datafeed.research.jqdata_sdkIO import normalize_code_to_local, normalize_code_to_jq, get_jq_api, \
    logout_jq_api, get_security_info
from rio3.datafeed.common import _get_cache_files_dir, convert_freq2num, _get_market, timer, \
    _get_security_type_by_market, get_resample_interval, _to_datetime, return_data, _get_h5_file_path, \
    _get_stock_code_from_h5_file, _fill_mid_and_midv, filter_quote_df, _calculate_zq_kdix, _get_tday, \
    calculate_mid_and_midv
from rio3.quote_core.quote_info import is_future
from rio3.quote_core.quote_filter import get_legal_period_data, get_trading_border_time, get_trading_time_border_diff
from rio3.datafeed.dbIO import get_tday_list, _get_history_data_org_from_rt, get_tday_df, db_retry_on_exception, \
    get_ed_info_df, get_db_session
from rio3.trade.models import db_rt_models

DEFAULT_FREQ = '5m'

KLINE_FREQ_LIST = ['1m', '5m', '15m', '30m', '60m', '1d']
TKLINE_FREQ_LIST = ['5s', '15s', '30s']
DEFAULT_RT_DB_FIELDS = ['id_stock', 'market_name', 'stockfullcode', 'dateInfo', 'dailyPreclose', 'date_trade',
                        'val_preclose', 'val_open', 'val_high', 'val_low', 'val_close', 'amt_amount', 'amt_volumn',
                        'amt_wp', 'amt_np', 'val_turnover', 'val_OI', 'flg_st', 'val_mid', 'val_midv', 'val_sp1',
                        'val_sv1', 'val_bp1', 'val_bv1', 'tday']

# 初始化分红日期
DEFAULT_DIVIDEND_DAY = 20161129

logger = logging.getLogger('root')


def _get_datafeed_parameters_dict(security_type, stockfullcode, start_date, end_date, cache_path, freq,
                                  cache_file_type=None):
    """
    通过传入参数, 返回datafeed数据需要的参数
    :param security_type:
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param cache_path:
    :param freq:
    :return:
    """
    start_date = _to_datetime(start_date)
    end_date = _to_datetime(end_date)
    start_date_num = int(start_date.strftime('%Y%m%d'))
    end_date_num = int(end_date.strftime('%Y%m%d'))
    start_time_num = int(start_date.strftime('%H%M%S'))
    end_time_num = int(end_date.strftime('%H%M%S'))
    # 定于最终返回的datafeed参数处理返回列表
    parameters_info_dict = dict()
    # 加入时间字段
    parameters_info_dict['start_date_num'] = start_date_num
    parameters_info_dict['end_date_num'] = end_date_num
    parameters_info_dict['start_time_num'] = start_time_num
    parameters_info_dict['end_time_num'] = end_time_num
    # 计算是否需要使用where语句
    parameters_info_dict['use_where'] = True if end_date_num - start_date_num <= 40000 else False

    # 根据security_type判断本地周期
    if security_type == 'options':
        # 目前期权的数据生成只有15m,非该周期直接返回空值
        if freq not in ['1m', '5m', '15m', '1d']:
            print(u'期权目前只支持返回最小周期为15m的本地行情')
        return_freq = freq
    # 其余的Security Type默认缓存的是1m和1d
    else:
        # 默认获取1分钟或是1d的基础数据，其余级别通过period_conversion_resample进行转化
        return_freq = freq if freq == '1d' or freq == '1m' else '1m'
    parameters_info_dict['freq'] = return_freq

    # 统一对stockfullcode进行数组转化处理
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    parameters_info_dict['stockfullcode'] = list_stockfullcode

    # 根据start/end_time_num 生成where_clause
    if start_time_num == 0 and end_time_num == 0:
        where_clause = "today>={} & today<={}".format(start_date_num, end_date_num)
    else:
        where_clause = "date>={} & date<={}".format(start_date_num, end_date_num)
    parameters_info_dict['where_clause'] = where_clause

    # 判断是nas/local_path/https
    from_local_cache = os.path.isdir(cache_path)
    # 如果cache_path是nas或者传入的是本地路径，则构造H5文件的读取路径
    if cache_path == 'nas' or from_local_cache:
        # 特殊处理期权数据,是在share路径下,其他的都还是nas/share
        default_cache_path = constant.H5_JQ_KLINE_DIR
        cache_root_path = default_cache_path if not from_local_cache else cache_path
        # 目前rio3的接口只支持读取K线
        cache_type_str = 'kline'
        local_data_msg = u"NAS缓存: '{}' ".format(default_cache_path) if cache_path == 'nas' else u"本地缓存: '{}' ".format(
            cache_path)
        h5_cache_files = list()
        # 循环构造stockfullcode的h5路径
        for each_code in list_stockfullcode:
            if not from_local_cache:
                # 如果cache_file_type不为空, 则判断读取的为非标准K线接口, 目前有期权的面板数据panel和波动率指标数据Index
                if cache_file_type is not None and security_type is not None:
                    cache_file_dir = os.path.join(cache_root_path, security_type, cache_file_type)
                else:
                    pass_security_type = security_type if security_type == 'options' else None
                    cache_file_dir = _get_cache_files_dir(cache_root_path=cache_root_path, stockfullcode=each_code,
                                                          is_local=False, pass_security_type=pass_security_type)
            else:
                cache_file_dir = cache_path
            # 如果cache_file_type不为空, 判断直接读取的是非K线, 直接拼接file_type文件名
            if cache_file_type is not None:
                cache_file_path = os.path.join(cache_file_dir, '{}_{}.h5'.format(each_code, cache_file_type))
            else:
                cache_file_path = _get_h5_file_path(source_path=cache_file_dir, stockfullcode=each_code,
                                                    h5_type=cache_type_str)
            h5_cache_files.append(cache_file_path)
        parameters_info_dict['cache_type'] = 'local_or_nas'
        parameters_info_dict['h5_cache_files'] = h5_cache_files
        logger.info(u"\n[数据源]：{}\n[标的/合约/代码]:'{}'".format(local_data_msg, ','.join(list_stockfullcode)))
    elif cache_path == 'http':
        parameters_info_dict['cache_type'] = 'http'
    else:
        logger.warning(u"传入的cache_path不正确: {}，请确认指定['nas', 'http', 'cache_path(实际本地路径)']三类参数中的一个".format(cache_path))
        parameters_info_dict['cache_type'] = 'invalid'
        parameters_info_dict['h5_cache_files'] = list()
    return parameters_info_dict


def _filter_cache_df(cache_df, start_date_num, end_date_num, start_time_num, end_time_num):
    """
    在Dataframe过滤具体的日期和时间
    :param cache_df:
    :param start_date_num:
    :param end_date_num:
    :param start_time_num:
    :param end_time_num:
    :return:
    """
    if start_time_num == 0 and end_time_num == 0:
        cache_df = cache_df[(cache_df['today'] >= start_date_num) & (cache_df['today'] <= end_date_num)]
    else:
        cache_df = cache_df[
            (cache_df['date'] >= start_date_num) & (cache_df['date'] <= end_date_num) & (
                    cache_df['sts'] >= start_time_num) & (cache_df['ets'] <= end_time_num)]
    return cache_df


def _get_options_data_org(stockfullcode, start_date=None, end_date=None, fields=None, freq='15m', cache_path='nas',
                          dataframe=True):
    """
    获取期权的合约数据，也包括标的数据
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param cache_path:
    :param dataframe:
    :return:
    """
    quote_df_dict = dict()
    quote_df = pd.DataFrame()
    datafeed_params_dict = _get_datafeed_parameters_dict(security_type='options', stockfullcode=stockfullcode,
                                                         start_date=start_date,
                                                         end_date=end_date, freq=freq, cache_path=cache_path)
    freq = datafeed_params_dict['freq']  # 获取频率
    # 如果cache_path是nas或者传入的是本地路径，则构造H5文件的读取路径
    if datafeed_params_dict['cache_type'] == 'local_or_nas':
        # 是否使用h5自己的过滤语句
        use_where = datafeed_params_dict['use_where']
        h5_cache_files = datafeed_params_dict['h5_cache_files']  # 获取缓存H5的具体路径
        for cache_file_path in h5_cache_files:
            # 通过H5路径文件名提取stockfullcode
            stock_code = _get_stock_code_from_h5_file(cache_file_path)
            # 读取相关stock code的HDF5缓存数据
            if h5.is_hdf5(cache_file_path):
                h5_file = pd.HDFStore(cache_file_path, mode='r')
                file_keys = h5_file.keys()
                if '/{}'.format(freq) in file_keys:
                    group_node = str(freq).lower()
                    # 如果start_time_num和end_time_num都是0，认为是按照tday查询
                    where_clause = datafeed_params_dict['where_clause']
                    # 读取H5数据
                    try:
                        if use_where:
                            cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields,
                                                   where=where_clause)
                        else:
                            cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields)
                        if not cache_df.empty:
                            cache_df = _filter_cache_df(cache_df=cache_df,
                                                        start_date_num=datafeed_params_dict['start_date_num'],
                                                        end_date_num=datafeed_params_dict['end_date_num'],
                                                        start_time_num=datafeed_params_dict['start_time_num'],
                                                        end_time_num=datafeed_params_dict['end_time_num'])
                    except Exception as ex:
                        logger.warning(u'品种[{}],级别[{}] 读取ZQ格式数据失败[{}]'.format(stock_code, freq, ex))
                        cache_df = return_data(return_frame=dataframe)
                    if not cache_df.empty:
                        # 去掉重复的数据
                        cache_df.drop_duplicates(
                            subset=['date', 'today', 'ets', 'open', 'high', 'low', 'close', 'amount', 'vol',
                                    'contract'],
                            inplace=True)
                        # 过滤N/A
                        cache_df.dropna(axis=0, subset=['open', 'high', 'low', 'close', 'amount', 'vol'], inplace=True)
                        if not cache_df.empty:
                            cache_df.reset_index(drop=True, inplace=True)
                            # 标记symbol和contract
                            cache_df['symbol'] = stock_code
                        # 默认组装字典格式的数据类型
                        quote_df_dict[stock_code] = cache_df
                h5_file.close()
    elif datafeed_params_dict['cache_type'] == 'http':
        logger.info(u'获取JQData缓存数据')
        quote_df_dict = _get_history_data_org_from_jq(stockfullcode=stockfullcode, start_date=start_date,
                                                      end_date=end_date, freq=freq, jq_fields=fields, dataframe=False)
    if dataframe:
        if len(quote_df_dict) > 0:
            quote_df = pd.concat(quote_df_dict.values())
            quote_df.reset_index(drop=True, inplace=True)
        return quote_df
    else:
        return quote_df_dict


def get_options_trading_info(underlying_name, info_type='summary', start_date=None, end_date=None, cache_path='nas'):
    """
    获取当前传入标的的合约信息和分红信息
    :param underlying_name:
    :param info_type: contracts/dividends
    :return: dataframe
    """
    # 设置默认时间
    start_date = int(pd.to_datetime(_get_tday(start_date)).strftime('%Y%m%d'))
    end_date = int(pd.to_datetime(_get_tday(end_date)).strftime('%Y%m%d'))
    list_info_type = [info_type, ] if isinstance(info_type, str) else info_type
    # 判断是否是本地缓存
    from_local_cache = os.path.isdir(cache_path)
    options_info_dict = dict()
    # 如果cache_path是nas或者传入的是本地路径，则构造H5文件的读取路径
    if cache_path == 'nas' or from_local_cache:
        if not from_local_cache:
            info_cache_file_dir = os.path.join(constant.H5_JQ_COMMON_DIR, 'options')
        else:
            info_cache_file_dir = cache_path
        for each_info_type in list_info_type:
            file_path = os.path.join(info_cache_file_dir, '{}.h5'.format(each_info_type))
            options_info_df = pd.DataFrame()
            if os.path.exists(file_path):
                options_info_df = pd.read_hdf(path_or_buf=file_path, key=underlying_name)
            # 过滤时间
            if not options_info_df.empty:
                options_info_df = options_info_df[
                    (options_info_df['tday'] >= start_date) & (options_info_df['tday'] <= end_date)]
            options_info_dict[each_info_type] = options_info_df
    elif cache_path == 'http':
        logger.warning('期权交易统计没有HTTP获取方式')
        for each_info_type in list_info_type:
            options_info_dict[each_info_type] = pd.DataFrame()
    if len(options_info_dict) == 1:
        return pd.concat(options_info_dict.values())
    return options_info_dict


@timer
def get_options_contract_data(stockfullcode, start_date=None, end_date=None, fields=None, freq='15m', cache_path='nas',
                              dataframe=True):
    """
    获取期权单合约/标的的原始数据，基本格式和股票/期货一致
    :param start_date: 起始日期 YYYYmmdd
    :param end_date: 结束日期 YYYYmmdd
    :param fields: 数据字段; close, open, high, low, volume, money
    :param freq: 可选1m, 5m, 15m；默认15m
    :param underlying: 标的资产；默认50ETF
    :param cache_path: 数据路径
    :return: 期权数据面板Dataframe
    """
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    list_fields = [fields, ] if isinstance(fields, str) else fields
    # 获取原始数据
    total_quote_df = _get_options_data_org(stockfullcode=list_stockfullcode, start_date=start_date, end_date=end_date,
                                           fields=list_fields, freq=freq, cache_path=cache_path, dataframe=True)
    # 如果没有数据，返回空
    if len(total_quote_df) == 0:
        return return_data(return_frame=dataframe)

    if not dataframe:
        total_quote_dict = dict()
        for contract, contract_df in total_quote_df.groupby(total_quote_df.contract):
            total_quote_dict[contract] = contract_df
        return total_quote_dict
    else:
        return total_quote_df


@db_retry_on_exception(retry=5)
def get_options_common_info_from_jq(underlying_name, info_type, is_trading=False, return_cache_version=False):
    """
    在线从JQData中获取期权的合约信息或者分红信息，可以选择是否以缓存版本返回
    :param underlying_name: 50ETF/300ETF
    :param info_type: contracts/dividends
    :param return_cache_version:
    :return:
    """
    local_underlying_code = constant.OPTIONS_UNDERLYING_CODE.get(underlying_name, None)
    jq_underlying_code = normalize_code_to_jq(local_underlying_code)
    options_info_df = pd.DataFrame()
    if info_type == 'contracts':
        # 一次获取最多3000条，这里用slice的方式获取数据，直到全部取完
        start_index = 0
        end_index = 2500
        contract_df = pd.DataFrame()
        contracts_list = list()
        while len(contract_df) > 0 or start_index == 0:
            q = get_jq_api().query(get_jq_api().opt.OPT_CONTRACT_INFO).filter(
                get_jq_api().opt.OPT_CONTRACT_INFO.underlying_type == underlying_name).slice(start_index, end_index)
            contract_df = get_jq_api().opt.run_query(q)
            if not contract_df.empty:
                contracts_list.append(contract_df)
            # 重新设置新的起始下标
            start_index = end_index
            # 重新设置新的结束下标
            end_index = start_index + 2500
        if len(contracts_list) > 0:
            contract_df = pd.concat(contracts_list)
            if is_trading:
                contract_df = contract_df[contract_df['contract_status'] == 'LIST']
            contract_df = format_future_options_contract(jq_contract_df=contract_df, whether_cache=return_cache_version)
        options_info_df = contract_df
    elif info_type == 'dividends':
        fund_dividend_code = jq_underlying_code.split('.')[0]
        q = get_jq_api().query(get_jq_api().finance.FUND_DIVIDEND).filter(
            (get_jq_api().finance.FUND_DIVIDEND.code == fund_dividend_code))
        dividends_df = get_jq_api().finance.run_query(q)
        options_info_df = dividends_df
    return options_info_df


def format_future_options_contract(jq_contract_df, whether_cache=False):
    # 如果合约为空，则直接返回
    if jq_contract_df.empty:
        return pd.DataFrame()
    format_contract_df = jq_contract_df.copy()
    # 代码中含有C或者P，直接替换会影响到品种中P或者C，所以先对code进行分解，变成合约+类型+行权价
    # 对code的字符串进行分解, CU1901C46000.XSGE 会分解成symbol:CU, stockcode:1901, con_type:C, exec_price:46000, market:.XSGE
    splited_code_df = format_contract_df['code'].str.extract(
        r'(?P<symbol>(\D*))(?P<stockcode>(\d*))(?P<con_type>(C|P|-C-|-P-))(?P<exec_price>(\d*))(?P<market>(\.\D*))')
    format_contract_df['exchange_stockcode'] = splited_code_df[['symbol', 'stockcode']].agg(''.join, axis=1)
    if whether_cache:
        # 找出配对标的合约名称
        # 根据C和P反向推出配对合约交易代码
        # 合约包含-C-或者-P-, 合约可以直接替换 例子：IO2010-P-5000.CCFX，这里只要把-P-替换为-C-就可以获取配对合约
        format_contract_df['pair_code'] = np.where(
            format_contract_df['code'].str.contains('-P-', case=True, regex=True),
            format_contract_df['code'].str.replace('-P-', '-C-', case=True, regex=True),
            format_contract_df['code'].str.replace('-C-', '-P-', case=True, regex=True))
        # 构造配对期
        splited_code_df['pair_con_type'] = np.where(
            splited_code_df['con_type'].str.contains('P', case=True, regex=True),
            splited_code_df['con_type'].str.replace('P', 'C', case=True, regex=True),
            splited_code_df['con_type'].str.replace('C', 'P', case=True, regex=True))
        format_contract_df['pair_code'] = splited_code_df[
            ['symbol', 'stockcode', 'pair_con_type', 'exec_price', 'market']].agg(''.join, axis=1)
        # 找出配对合约的信息
        pair_code_contract_df = format_contract_df['pair_code'].apply(
            lambda x: format_contract_df[format_contract_df['code'] == x].iloc[0])
        format_contract_df['pair_trading_code'] = pair_code_contract_df['trading_code']
        # 找出配对合约的行权价
        format_contract_df['pair_code_exercise_price'] = pair_code_contract_df['exercise_price']
        # 配对合约的合约单位
        format_contract_df['pair_code_contract_unit'] = pair_code_contract_df['contract_unit']
        # 找出配对合约是否调整
        format_contract_df['pair_code_is_adjust'] = pair_code_contract_df['is_adjust']
        # 找出配对合约的list_date
        format_contract_df['pair_code_list_date'] = pd.to_datetime(pair_code_contract_df['list_date']).dt.strftime(
            '%Y%m%d').astype(np.int64)
        # 找出配对合约的expire_date
        format_contract_df['pair_code_expire_date'] = pd.to_datetime(
            pair_code_contract_df['expire_date']).dt.strftime(
            '%Y%m%d').astype(np.int64)
        # 找出配对合约的last_trade_date
        format_contract_df['pair_code_last_trade_date'] = pd.to_datetime(
            pair_code_contract_df['last_trade_date']).dt.strftime('%Y%m%d').astype(np.int64)
        # 找出配对合约的exercise_date
        format_contract_df['pair_code_exercise_date'] = pair_code_contract_df['exercise_date']
    # 格式化所有的日期列为纯数字 list_date, expire_date, last_trade_date, exercise_date, delivery_date, delist_date
    format_contract_df['list_date'] = pd.to_datetime(format_contract_df['list_date']).dt.strftime("%Y%m%d").astype(
        np.int64)
    format_contract_df['expire_date'] = pd.to_datetime(format_contract_df['expire_date']).dt.strftime(
        "%Y%m%d").astype(np.int64)
    format_contract_df['last_trade_date'] = pd.to_datetime(format_contract_df['last_trade_date']).dt.strftime(
        "%Y%m%d").astype(
        np.int64)
    # 行权日固定有值得期权是欧式期权，不固定的是美式期权，JQ中有值的便是欧式期权，为None的是美式期权
    format_contract_df['exercise_type'] = np.where(format_contract_df['exercise_date'].isnull(), 'A', 'E')
    format_contract_df['pair_code_exercise_type'] = np.where(format_contract_df['pair_code_exercise_date'].isnull(),
                                                             'A', 'E')
    # contract_df['delist_date'] = pd.to_datetime(contract_df['delist_date']).dt.strftime("%Y%m%d").astype(np.int64)
    return format_contract_df


def get_options_common_info(underlying_name, info_type, is_trading=False, return_cache_version=True, cache_path='nas'):
    """
    获取当前传入标的的合约信息和分红信息
    :param underlying_name:
    :param info_type: contracts/dividends/summary
    :param is_trading: 过滤出正在交易的
    :param return_cache_version: 是否返回缓存的版本格式,包含了配对的合约
    :return: dataframe
    """
    list_info_type = [info_type, ] if isinstance(info_type, str) else info_type
    # 判断是否是本地缓存
    from_local_cache = os.path.isdir(cache_path)
    options_info_dict = dict()
    # 如果cache_path是nas或者传入的是本地路径，则构造H5文件的读取路径
    if cache_path == 'nas' or from_local_cache:
        if not from_local_cache:
            info_cache_file_dir = os.path.join(constant.H5_JQ_COMMON_DIR, 'options')
        else:
            info_cache_file_dir = cache_path
        for each_info_type in list_info_type:
            file_path = os.path.join(info_cache_file_dir, '{}.h5'.format(each_info_type))
            options_info_df = pd.DataFrame()
            if os.path.exists(file_path):
                options_info_df = pd.read_hdf(path_or_buf=file_path, key=underlying_name)
                if not options_info_df.empty:
                    # 如果只需要正在交易的合约,则进行过滤
                    if is_trading:
                        options_info_df = options_info_df[options_info_df['contract_status'] == 'LIST']
            options_info_dict[each_info_type] = options_info_df
    else:
        for each_info_type in list_info_type:
            jq_options_info_df = get_options_common_info_from_jq(underlying_name=underlying_name,
                                                                 info_type=each_info_type, is_trading=is_trading,
                                                                 return_cache_version=return_cache_version)
            options_info_dict[each_info_type] = jq_options_info_df
    if len(options_info_dict) == 1:
        return pd.concat(options_info_dict.values())
    return options_info_dict


def _get_options_trading_contract(underlying_name, tday_list):
    """
    获取每一个交易日正在交易的的期权品种，主要是判断是否在合约的挂牌日期和最后交易日内
    :param underlying:
    :param tday_list:
    :return:
    """
    tday_contract_df = pd.DataFrame()
    options_contract_df = get_options_common_info(underlying_name=underlying_name, info_type='contracts')
    if len(options_contract_df) == 0:
        return return_data(return_frame=True)
    # 选取当前交易日已经挂牌但还未到期的合约，为后续比价做准备
    tday_contract_list = list()
    for tday in tday_list:
        trading_contracts = options_contract_df[
            (options_contract_df.last_trade_date >= int(tday)) & (options_contract_df.list_date <= int(tday))]
        if not trading_contracts.empty:
            # 如果需要让返回的code是JQ格式
            trading_contracts_copy = trading_contracts.copy()
            trading_contracts_copy['tday'] = tday
            tday_contract_list.append(trading_contracts_copy)
        logger.info(f'[合约信息]: {underlying_name} 在 {tday} 有 {len(trading_contracts)} 个合约在交易')
    # 转化成平面数据，格式为
    if len(tday_contract_list) > 0:
        tday_contract_df = pd.concat(tday_contract_list)
        tday_contract_df['local_code'] = normalize_code_to_local(tday_contract_df['code'].tolist())
        tday_contract_df.rename(columns={'code': 'jq_code', 'local_code': 'code'}, inplace=True)
        tday_contract_df = tday_contract_df[['tday', 'code', 'jq_code', 'underlying_symbol']]
    return tday_contract_df


def get_options_index_data(underlying_name, start_date=None, end_date=None, fields=None, freq='15m', cache_path='nas',
                           dataframe=True):
    """
    获取期权面版数据生成的指标iv_index, 目前是各月份计算波动率指数的合约
    :param underlying_name:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param cache_path:
    :param dataframe:
    :return:
    """

    quote_index_df_dict = dict()
    quote_index_df = pd.DataFrame()
    datafeed_params_dict = _get_datafeed_parameters_dict(security_type='options', stockfullcode=underlying_name,
                                                         start_date=start_date, end_date=end_date, freq=freq,
                                                         cache_path=cache_path, cache_file_type='index')
    # 判断是否是本地缓存
    freq = datafeed_params_dict['freq']  # 获取频率
    # 如果cache_path是nas或者传入的是本地路径，则构造H5文件的读取路径
    if datafeed_params_dict['cache_type'] == 'local_or_nas':
        # 是否使用h5自己的过滤语句
        use_where = datafeed_params_dict['use_where']
        h5_cache_files = datafeed_params_dict['h5_cache_files']  # 获取缓存H5的具体路径
        for cache_file_path in h5_cache_files:
            # 读取相关stock code的HDF5缓存数据
            if h5.is_hdf5(cache_file_path):
                underlying_name = os.path.splitext(cache_file_path)[0]
                h5_file = pd.HDFStore(cache_file_path, mode='r')
                file_keys = h5_file.keys()
                if '/{}'.format(freq) in file_keys:
                    group_node = str(freq).lower()
                    where_clause = datafeed_params_dict['where_clause']
                    # 读取H5数据
                    try:
                        if use_where:
                            cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields,
                                                   where=where_clause)
                        else:
                            cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields)
                        if not cache_df.empty:
                            cache_df = _filter_cache_df(cache_df=cache_df,
                                                        start_date_num=datafeed_params_dict['start_date_num'],
                                                        end_date_num=datafeed_params_dict['end_date_num'],
                                                        start_time_num=datafeed_params_dict['start_time_num'],
                                                        end_time_num=datafeed_params_dict['end_time_num'])
                    except Exception as ex:
                        logger.warning(u'品种[{}],级别[{}] 读取ZQ格式数据失败[{}]'.format(underlying_name, freq, ex))
                        cache_df = pd.DataFrame()
                    if not cache_df.empty:
                        if not cache_df.empty:
                            cache_df.reset_index(drop=True, inplace=True)
                        # 默认组装字典格式的数据类型
                        quote_index_df_dict[underlying_name] = cache_df
                h5_file.close()
    elif datafeed_params_dict['cache_type'] == 'http':
        logger.info(u'实时计算指标数据')
        options_panel = get_options_underlying_data(start_date=start_date, end_date=end_date,
                                                    underlying_name=underlying_name, freq=freq)
        quote_index_df_dict[underlying_name] = get_call_put_by_month_df(panel_quote_df=options_panel)
    if dataframe:
        if len(quote_index_df_dict) > 0:
            quote_index_df = pd.concat(quote_index_df_dict.values())
            quote_index_df = quote_index_df.reset_index(drop=True)
        return quote_index_df
    else:
        return quote_index_df_dict


# 根据传入的period和时间范围,构造对应的H5 DataSet路径
def _get_cached_h5_nodes(freq, start_date, end_date):
    cached_node_list = list()
    tday_list = get_tday_list(start_dt=start_date, end_dt=end_date)
    if len(tday_list) > 0:
        tday_df = pd.DataFrame(data=tday_list, columns=['tday'])
        tday_df['cal_date'] = pd.to_datetime(tday_df['tday'])
        tday_df['dt_year'] = tday_df['tday'].dt.year
        tday_df['dt_month'] = tday_df['tday'].dt.month
        year_tday_iterator = tday_df.groupby(by=['dt_year'])
        for year, year_tday_df in year_tday_iterator:
            node_year = str(year)
            node_group_path = '/{}'.format(freq)
            for month, month_tday_df in year_tday_df.groupby(by=['dt_month']):
                # /time_interval/Y%M% exp: /freq/M201701
                node_dset_path = node_group_path + "/{}{}{}".format("M", node_year, str(month).zfill(2))
                cached_node_list.append(node_dset_path)
    return cached_node_list


@timer
def get_options_underlying_node_data(underlying_name, start_date=None, end_date=None, fields=None, freq='15m',
                                     cache_path='nas', dataframe=True):
    """
    获取按月存储的期权面板数据 = 合约数据集合 + 标的数据集合
    :param underlying_name:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param cache_path:
    :param dataframe:
    :return:
    """
    quote_df_dict = dict()
    quote_df = pd.DataFrame()
    datafeed_params_dict = _get_datafeed_parameters_dict(security_type='options', stockfullcode=underlying_name,
                                                         start_date=start_date,
                                                         end_date=end_date, freq=freq, cache_path=cache_path,
                                                         cache_file_type='panel_node')
    freq = datafeed_params_dict['freq']  # 获取频率
    if datafeed_params_dict['cache_type'] == 'local_or_nas':
        h5_cache_files = datafeed_params_dict['h5_cache_files']  # 获取缓存H5的具体路径
        for cache_file_path in h5_cache_files:
            # 读取相关stock code的HDF5缓存数据
            if h5.is_hdf5(cache_file_path):
                h5_file = pd.HDFStore(cache_file_path, mode='r')
                # 构造当前时间内需要访问的节点
                query_nodes_list = _get_cached_h5_nodes(start_date=start_date, end_date=end_date, freq=freq)
                # 获取当前缓存的现有节点
                current_file_keys = h5_file.keys()
                # 做交集
                actual_query_nodes = sorted(list(set(query_nodes_list).intersection(set(current_file_keys))))
                # 存放每个节点的数据
                total_nodes_df_list = list()
                # 循环actual_query_nodes，按节点读取缓存并拼接
                for query_node in actual_query_nodes:
                    # 如果start_time_num和end_time_num都是0，认为是按照tday查询
                    # 读取H5数据
                    try:
                        node_cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=query_node, columns=fields)
                        logger.info(u'[读取节点数据成功]: 品种:{},节点:{}'.format(underlying_name, query_node))
                        total_nodes_df_list.append(node_cache_df)
                    except Exception as ex:
                        logger.warning(u'[读取节点数据失败]: 品种:{},节点:{},错误:{}'.format(underlying_name, freq, ex))
                # 合成大面板数据
                cache_df = pd.DataFrame()
                if len(total_nodes_df_list) > 0:
                    logger.info(u'[拼接节点数据]: 品种:{}, 频率:[{}]'.format(underlying_name, freq))
                    cache_df = pd.concat(total_nodes_df_list)
                    if not cache_df.empty:
                        logger.info(u'[过滤数据]: 品种:{}, 频率:[{}]'.format(underlying_name, freq))
                        cache_df = _filter_cache_df(cache_df=cache_df,
                                                    start_date_num=datafeed_params_dict['start_date_num'],
                                                    end_date_num=datafeed_params_dict['end_date_num'],
                                                    start_time_num=datafeed_params_dict['start_time_num'],
                                                    end_time_num=datafeed_params_dict['end_time_num'])
                    if not cache_df.empty:
                        # 过滤N/A
                        cache_df.dropna(axis=0, subset=['open', 'high', 'low', 'close', 'amount', 'vol'], inplace=True)
                        # 重置index
                        if not cache_df.empty:
                            cache_df.reset_index(drop=True, inplace=True)
                            # 排序
                            cache_df.sort_values(by=['date', 'ets'], inplace=True)
                # 默认组装字典格式的数据类型
                quote_df_dict[underlying_name] = cache_df
                # 删除cache_df，回收内存
                logger.info(u'[清理缓存]: 品种:{}, 频率:[{}]'.format(underlying_name, freq))
                del cache_df
                gc.collect()
                # gc_cache(cache_df)
                # 关闭缓存文件
                h5_file.close()
    elif datafeed_params_dict['cache_type'] == 'http':
        logger.info(u'实时合成面板数据')
        if 'ETF' in underlying_name:
            quote_df_dict = _get_etf_options_underlying_core(underlying_name=underlying_name, start_date=start_date,
                                                             end_date=end_date, freq=freq, dataframe=dataframe)
        elif underlying_name == 'INDEX' or underlying_name == 'FUTURE':
            quote_df_dict = _get_future_options_underlying_core(underlying_name=underlying_name, start_date=start_date,
                                                                end_date=end_date, freq=freq, dataframe=dataframe)
    if dataframe:
        if len(quote_df_dict) > 0:
            quote_df = pd.concat(quote_df_dict.values())
            quote_df.reset_index(drop=True, inplace=True)
        return quote_df
    else:
        return quote_df_dict


@timer
def get_options_underlying_data(underlying_name, start_date=None, end_date=None, fields=None, freq='15m',
                                cache_path='nas', dataframe=True):
    """
    获取期权面板数据 = 合约数据集合 + 标的数据集合
    :param underlying_name:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param cache_path:
    :param dataframe:
    :return:
    """
    quote_df_dict = dict()
    quote_df = pd.DataFrame()
    datafeed_params_dict = _get_datafeed_parameters_dict(security_type='options', stockfullcode=underlying_name,
                                                         start_date=start_date,
                                                         end_date=end_date, freq=freq, cache_path=cache_path,
                                                         cache_file_type='panel')
    freq = datafeed_params_dict['freq']  # 获取频率
    if datafeed_params_dict['cache_type'] == 'local_or_nas':
        h5_cache_files = datafeed_params_dict['h5_cache_files']  # 获取缓存H5的具体路径
        for cache_file_path in h5_cache_files:
            # 读取相关stock code的HDF5缓存数据
            if h5.is_hdf5(cache_file_path):
                underlying_name = os.path.splitext(cache_file_path)[0]
                h5_file = pd.HDFStore(cache_file_path, mode='r')
                file_keys = h5_file.keys()
                if '/{}'.format(freq) in file_keys:
                    group_node = str(freq).lower()
                    # 如果start_time_num和end_time_num都是0，认为是按照tday查询
                    # 读取H5数据
                    try:
                        cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields)
                        if not cache_df.empty:
                            cache_df = _filter_cache_df(cache_df=cache_df,
                                                        start_date_num=datafeed_params_dict['start_date_num'],
                                                        end_date_num=datafeed_params_dict['end_date_num'],
                                                        start_time_num=datafeed_params_dict['start_time_num'],
                                                        end_time_num=datafeed_params_dict['end_time_num'])
                    except Exception as ex:
                        logger.warning(u'品种[{}],级别[{}] 读取ZQ格式数据失败[{}]'.format(underlying_name, freq, ex))
                        cache_df = pd.DataFrame()
                    if not cache_df.empty:
                        # 过滤N/A
                        cache_df.dropna(axis=0, subset=['open', 'high', 'low', 'close', 'amount', 'vol'], inplace=True)
                        if not cache_df.empty:
                            cache_df.reset_index(drop=True, inplace=True)
                        # 默认组装字典格式的数据类型
                        quote_df_dict[underlying_name] = cache_df
                h5_file.close()
    elif datafeed_params_dict['cache_type'] == 'http':
        logger.info(u'实时合成面板数据')
        if 'ETF' in underlying_name:
            quote_df_dict = _get_etf_options_underlying_core(underlying_name=underlying_name, start_date=start_date,
                                                             end_date=end_date, freq=freq, dataframe=dataframe)
        elif underlying_name == 'INDEX' or underlying_name == 'FUTURE':
            quote_df_dict = _get_future_options_underlying_core(underlying_name=underlying_name, start_date=start_date,
                                                                end_date=end_date, freq=freq, dataframe=dataframe)
    if dataframe:
        if len(quote_df_dict) > 0:
            quote_df = pd.concat(quote_df_dict.values())
            quote_df.reset_index(drop=True, inplace=True)
        return quote_df
    else:
        return quote_df_dict


def _get_future_options_underlying_core(underlying_name, start_date=None, end_date=None, freq='15m',
                                        greeks=['iv', 'delta', 'gamma', 'vega', 'theta', 'rho'], dataframe=True,
                                        cached_contract_path=None, cached_underlying_path=None, stockfullcode=None):
    """
    合成INDEX/FUTURE合约的面板数据
    FUTURE/INDEX注意标注不必要的代码，cache_path或者filter
    :param underlying_name: INDEX/FUTURE
    :param start_date:
    :param end_date:
    :param freq:
    :param cached_contract_path:
    :param greeks
    :param dataframe:
    :param stockfullcode:
    :return:
    """
    # 获取交易日列表
    tday_list = _get_trading_days(start_date=start_date, end_date=end_date, return_dataframe=False)
    if len(tday_list) == 0:
        return return_data(return_frame=dataframe)
    # 先获取合约信息表，分红信息
    options_trading_contract = _get_options_trading_contract(underlying_name=underlying_name, tday_list=tday_list)

    # 对options_trading_contract做code, jq_code, underlying_symbol的去重
    # 转化为本地格式,用来拿本地合约数据
    # 对品种进行去重，得到所有合约列表, JQ格式,用来后面取标的合约
    # 建立JQ和本地code对应的dataframe
    options_trading_contract_unique = options_trading_contract.drop_duplicates(
        subset=['code', 'jq_code', 'underlying_symbol']).copy()
    options_trading_contract_unique['local_underlying_symbol'] = normalize_code_to_local(
        options_trading_contract_unique['underlying_symbol'].tolist())

    # 根据获取的各交易日合约信息，获取本地原始数据
    logger.info(u'1. 读取 {} 具体合约的行情，开始时间:{}, 结束时间:{}, 频率: {}'.format(underlying_name, start_date, end_date, freq))
    # 由于商品期货的合约数量很多,这里先用本地合约cache对distinct_trading_contract做一次过滤
    cached_trading_contract, diff_trading_contract = get_stockfullcode_read_path(cache_type='options',
                                                                                 stockfullcode=
                                                                                 options_trading_contract_unique[
                                                                                     'code'].tolist())
    # 过滤出豆粕和铁矿
    cached_trading_contract = [item for item in cached_trading_contract if ('DLp' in item or 'DLi' in item)]

    # 过滤品种
    cached_trading_contract = [item for item in cached_trading_contract if ('DLp' in item or 'DLi' in item)]
    # cached_trading_contract = [item for item in cached_trading_contract if ('DLp' in item or 'DLi' in item or 'SQcu' in item or 'DLpg' in item)]

    if cached_contract_path is None:
        total_contract_df = get_options_contract_data(stockfullcode=cached_trading_contract, start_date=start_date,
                                                      end_date=end_date, freq=freq)
    else:
        total_contract_df = pd.read_hdf(cached_contract_path)
        total_contract_df = total_contract_df[total_contract_df['contract'].isin(cached_trading_contract)].copy()
        # total_contract_df.to_hdf(r'\\nas\share\HDF5\JQData\KlineH5\options\future_option_cache.h5', key='data')

    logger.info(u'2. 添加标的合约 {} 的行情'.format(underlying_name))
    # 添加标的合约的行情，如果underlying_type是INDEX，标的合约就是沪深300指数，如果是FUTURE，标的合约就是各期权合约对应的期货合约
    # 获取本地的标的合约代码格式
    if underlying_name == 'INDEX':
        underlying_local_code = _get_options_underlying_code(underlying_name=underlying_name)
        underlying_quote_df = get_history_data(stockfullcode=underlying_local_code, start_date=start_date,
                                               end_date=end_date, freq=freq, cache_path='http')
        # 如果不为空，Merge到合约上，用上标的的OHLC，Volume和Amount, 并命名为underlying_xxx
        if not underlying_quote_df.empty:
            underlying_quote_df = underlying_quote_df[
                ['date', 'ets', 'open', 'high', 'close', 'low', 'vol', 'amount', 'contract']].copy()
            underlying_quote_df['underlying_name'] = underlying_name
            underlying_quote_df.rename(
                columns={'high': 'underlying_high', 'close': 'underlying_close', 'low': 'underlying_low',
                         'open': 'underlying_open', 'vol': 'underlying_vol', 'amount': 'underlying_amount',
                         'contract': 'underlying_code'},
                inplace=True)
            total_contract_df = pd.merge(left=total_contract_df, right=underlying_quote_df, how='left',
                                         left_on=['date', 'ets'], right_on=['date', 'ets'])
    # 如果是商品期权合约，标的合约就是各期权合约对应的期货合约
    else:
        if cached_underlying_path is not None:
            underlying_quote_df = pd.read_hdf(path_or_buf=cached_underlying_path)
        else:
            # 从总的K线数据中获取所有合约列表
            # 从mapping_trading_contract_df中过滤出cached_trading_contract,提取JQ_code, 在线读取行情
            jq_cached_trading_contract = options_trading_contract_unique[
                options_trading_contract_unique['code'].isin(
                    cached_trading_contract)].copy().underlying_symbol.unique().tolist()
            # 因为期货的字段跟股票的不同，必须指定域
            jq_fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'open_interest']
            underlying_quote_df = get_history_data(stockfullcode=jq_cached_trading_contract,
                                                   start_date=start_date, end_date=end_date, freq=freq,
                                                   fields=jq_fields, cache_path='http')
        if not underlying_quote_df.empty:
            # Merge变量options_trading_contract_unique中的local_underlying_symbol
            total_contract_df = pd.merge(left=total_contract_df, right=options_trading_contract_unique, how='left',
                                         left_on=['contract'], right_on=['code'])

            underlying_quote_df.rename(
                columns={'today': 'underlying_today', 'high': 'underlying_high', 'close': 'underlying_close',
                         'low': 'underlying_low', 'open': 'underlying_open', 'vol': 'underlying_vol',
                         'amount': 'underlying_amount', 'contract': 'underlying_code'},
                inplace=True)
            underlying_quote_df = underlying_quote_df[
                ['date', 'sts', 'ets', 'underlying_today', 'underlying_high', 'underlying_close', 'underlying_low',
                 'underlying_open',
                 'underlying_vol', 'underlying_amount', 'underlying_code']]
            total_contract_df = pd.merge(left=total_contract_df, right=underlying_quote_df, how='left',
                                         left_on=['local_underlying_symbol', 'date', 'sts', 'ets'],
                                         right_on=['underlying_code', 'date', 'sts', 'ets'])

    # 去掉nan列
    # total_contract_df[pd.isnull(total_contract_df).any(axis=1)]
    total_contract_df.dropna(how='any', axis=0, inplace=True)

    # total_contract_df = pd.read_hdf(r'\\nas\share\HDF5\JQData\KlineH5\options\future_total_option_cache.h5', key='data')
    total_contract_df['underlying_name'] = underlying_name

    logger.info(u'3. 填充 {} 合约信息'.format(underlying_name))
    # 合并第一波代码合约信息
    options_contract_df = get_options_common_info(underlying_name=underlying_name, info_type='contracts')
    # 筛选出需要的列,需要用到exercise_date，如果为空，为英式期权
    options_contract_df = options_contract_df[
        ['code', 'trading_code', 'contract_type', 'exercise_price', 'contract_unit', 'expire_date', 'list_date',
         'last_trade_date', 'is_adjust', 'exercise_type', 'pair_code', 'pair_trading_code', 'pair_code_exercise_price',
         'pair_code_contract_unit', 'pair_code_is_adjust', 'pair_code_list_date', 'pair_code_expire_date',
         'pair_code_last_trade_date', 'pair_code_exercise_type']]

    # 由于股指期货和商品期货没有分红的概念，这里直接给一个默认值
    options_contract_df['dividend_date'] = pd.to_datetime(constant.DATETIME_MAX)
    options_contract_df['pair_code_dividend_date'] = pd.to_datetime(constant.DATETIME_MAX)

    # 转化pair_code和code，都变成本地代码格式，标准前缀
    options_contract_df['normalized_code'] = normalize_code_to_local(options_contract_df['code'].tolist())
    options_contract_df['pair_code'] = normalize_code_to_local(options_contract_df['pair_code'].tolist())
    # # 合并行情信息和合约信息
    options_contract_panel_df = pd.merge(left=total_contract_df, right=options_contract_df, how='left',
                                         left_on='contract',
                                         right_on='normalized_code')

    # 精确expire_date到小时线, 计算合约到期剩余天数
    options_contract_panel_df['expire_datetime'] = pd.to_datetime(options_contract_panel_df['expire_date'],
                                                                  format='%Y%m%d') + pd.Timedelta(hours=16)

    # 赋值具体交易时间列
    options_contract_panel_df['date_trade'] = pd.to_datetime(
        options_contract_panel_df['date'] * 1000000 + options_contract_panel_df['ets'],
        format='%Y%m%d%H%M%S')

    # 合约到期时间
    options_contract_panel_df['days_to_maturity'] = (options_contract_panel_df['expire_datetime'] -
                                                     options_contract_panel_df[
                                                         'date_trade']).dt.days + 1

    # 计算时间价值和内在价值
    logger.info(u'4. 计算时间价值和内在价值')
    calculate_time_value(options_contract_panel_df)

    # 测试数据
    options_contract_panel_df = options_contract_panel_df[
        (options_contract_panel_df['today'] > int(pd.to_datetime(start_date).strftime('%Y%m%d'))) & (
                options_contract_panel_df['today'] < int(pd.to_datetime(end_date).strftime('%Y%m%d')))]

    # 过滤Columns
    options_contract_panel_df = options_contract_panel_df[constant.ZQ_OPTIONS_PANEL_FORMAT_COLUMNS]

    # 计算希腊字母指标
    logger.info(u'5. 计算希腊字母指标: {}'.format(greeks))
    # 支持计算美式期权和英式期权的希腊字母，需要关键字exercise_type
    calculate_greeks_dll_np(options_contract_panel_df, greeks=greeks)

    # calculate_greeks_QL(options_contract_panel_df, greeks=greeks)

    return options_contract_panel_df


def _get_etf_options_underlying_core(underlying_name, start_date=None, end_date=None, freq='15m',
                                     greeks=['iv', 'delta', 'gamma', 'vega', 'theta', 'rho'], dataframe=True,
                                     cache_file_path=None):
    """
    合成ETF标的合约的面板数据
    :param underlying_name:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param cache_path:
    :param greeks
    :param dataframe:
    :return:
    """
    # 获取本地的标的合约代码格式
    underlying_local_code = _get_options_underlying_code(underlying_name=underlying_name)
    # 获取交易日列表
    tday_list = _get_trading_days(start_date=start_date, end_date=end_date, return_dataframe=False)
    if len(tday_list) == 0:
        return return_data(return_frame=dataframe)
    # 先获取合约信息表，分红信息
    options_trading_contract = _get_options_trading_contract(underlying_name=underlying_name, tday_list=tday_list)
    # 对品种进行去重，得到所有合约列表
    distinct_trading_contract = options_trading_contract.code.unique().tolist()
    # 根据获取的各交易日合约信息，获取本地原始数据
    logger.info(u'1. 读取 {} 具体合约的行情，开始时间:{}, 结束时间:{}, 频率:{}'.format(underlying_name, start_date, end_date, freq))
    if cache_file_path is None:
        total_contract_df = get_options_contract_data(stockfullcode=distinct_trading_contract, start_date=start_date,
                                                      end_date=end_date, freq=freq)
    else:
        total_contract_df = pd.read_hdf(cache_file_path)

    logger.info(u'2. 添加标的合约 {} 的行情'.format(underlying_name))
    # 添加标的合约的行情
    underlying_quote_df = get_options_contract_data(stockfullcode=underlying_local_code, start_date=start_date,
                                                    end_date=end_date, freq=freq)
    # 如果不为空，Merge到合约上，用上标的的OHLC，Volume和Amount, 并命名为underlying_xxx
    if not underlying_quote_df.empty:
        underlying_quote_df = underlying_quote_df[
            ['date', 'ets', 'open', 'high', 'close', 'low', 'vol', 'amount', 'contract']].copy()
        underlying_quote_df['underlying_name'] = underlying_name
        underlying_quote_df.rename(
            columns={'high': 'underlying_high', 'close': 'underlying_close', 'low': 'underlying_low',
                     'open': 'underlying_open', 'vol': 'underlying_vol', 'amount': 'underlying_amount',
                     'contract': 'underlying_code'},
            inplace=True)
        total_contract_df = pd.merge(left=total_contract_df, right=underlying_quote_df, how='left',
                                     left_on=['date', 'ets'], right_on=['date', 'ets'])

    logger.info(u'3. 填充 {} 合约信息'.format(underlying_name))
    # 填充分红日
    options_dividend_df = get_options_common_info(underlying_name=underlying_name, info_type='dividends')
    # ex_date可能会出现None, 去掉ex_date为None的记录
    options_dividend_df.dropna(axis=0, subset=['ex_date'], inplace=True)
    options_dividend_df['ex_date'] = pd.to_datetime(options_dividend_df['ex_date']).dt.strftime('%Y%m%d').astype(
        np.int64)
    # 根据目前拿到的分红日期表，通过查看合约生命周期，确定分红日期。分红日期前的价格需要置为初始价格，分红日期后的价格为调整价格

    # 合并第一波代码合约信息
    options_contract_df = get_options_common_info(underlying_name=underlying_name, info_type='contracts')
    # 筛选出需要的列
    options_contract_df = options_contract_df[
        ['code', 'trading_code', 'contract_type', 'exercise_price', 'contract_unit', 'expire_date', 'list_date',
         'last_trade_date', 'is_adjust', 'exercise_type', 'pair_code', 'pair_trading_code', 'pair_code_exercise_price',
         'pair_code_contract_unit', 'pair_code_is_adjust', 'pair_code_list_date', 'pair_code_expire_date',
         'pair_code_last_trade_date', 'pair_code_exercise_type']]

    ex_date_list = sorted(options_dividend_df.ex_date.tolist())
    # 结合分红列表确定每一个合约的分红时间，分红时间应大于等于list_date，小于等于last_trade_date
    options_contract_df['dividend_date'] = options_contract_df[['list_date', 'last_trade_date', 'is_adjust']].apply(
        lambda x: _get_dividend_date(x.list_date, x.last_trade_date, x.is_adjust, ex_date_list), axis=1)
    options_contract_df['pair_code_dividend_date'] = options_contract_df[
        ['pair_code_list_date', 'pair_code_last_trade_date', 'pair_code_is_adjust']].apply(
        lambda x: _get_dividend_date(x.pair_code_list_date, x.pair_code_last_trade_date, x.pair_code_is_adjust,
                                     ex_date_list), axis=1)

    # 转化pair_code和code，都变成本地代码格式，标准前缀
    options_contract_df['normalized_code'] = normalize_code_to_local(options_contract_df['code'].tolist())
    options_contract_df['pair_code'] = normalize_code_to_local(options_contract_df['pair_code'].tolist())
    # 合并行情信息和合约信息
    options_contract_panel_df = pd.merge(left=total_contract_df, right=options_contract_df, how='left',
                                         left_on='contract',
                                         right_on='normalized_code')

    # 精确expire_date到小时线, 计算合约到期剩余天数
    options_contract_panel_df['expire_datetime'] = pd.to_datetime(options_contract_panel_df['expire_date'],
                                                                  format='%Y%m%d') + pd.Timedelta(hours=16)

    # 赋值具体交易时间列
    options_contract_panel_df['date_trade'] = pd.to_datetime(
        options_contract_panel_df['date'] * 1000000 + options_contract_panel_df['ets'],
        format='%Y%m%d%H%M%S')

    # 合约到期时间
    options_contract_panel_df['days_to_maturity'] = (options_contract_panel_df['expire_datetime'] -
                                                     options_contract_panel_df['date_trade']).dt.days + 1

    # 配对信息
    price_unit = 0.001
    norm_unit = 10000
    # 行权价 & 是否为标准合约 & 合约单位
    options_contract_panel_df['pair_init_exercise_price'] = options_contract_panel_df['pair_trading_code'].str.slice(
        -5).astype(
        np.int64) * price_unit
    options_contract_panel_df['init_exercise_price'] = options_contract_panel_df['trading_code'].str.slice(-5).astype(
        np.int64) * price_unit
    # 通过判断分红日期，对赋值C/P的行权价，是否调整，合约单位
    # conditions = (options_contract_info_df['is_adjust'] == 1) & (
    #         (options_contract_info_df['dividend_date'] > options_contract_info_df['list_date']) & (
    #         options_contract_info_df['dividend_date'] <= options_contract_info_df['today']))
    # 对于所有adjust=1的合约，判定分红日之前的行情为初始价格，分红日以后的行情为调整价格, 本合约和配对合约都需要做改操作

    # 本合约调整
    code_conditions = (options_contract_panel_df['is_adjust'] == 1) & (
            options_contract_panel_df['dividend_date'] <= options_contract_panel_df['today'])
    options_contract_panel_df['exercise_price'] = np.where(code_conditions, options_contract_panel_df['exercise_price'],
                                                           options_contract_panel_df['init_exercise_price'])
    options_contract_panel_df['is_adjust'] = np.where(code_conditions, 1, 0)
    options_contract_panel_df['contract_unit'] = np.where(code_conditions, options_contract_panel_df['contract_unit'],
                                                          norm_unit)

    # 配对合约调整
    pair_code_conditions = (options_contract_panel_df['pair_code_is_adjust'] == 1) & (
            options_contract_panel_df['pair_code_dividend_date'] <= options_contract_panel_df['today'])
    options_contract_panel_df['pair_code_exercise_price'] = np.where(pair_code_conditions, options_contract_panel_df[
        'pair_code_exercise_price'], options_contract_panel_df['pair_init_exercise_price'])
    options_contract_panel_df['pair_code_is_adjust'] = np.where(pair_code_conditions, 1, 0)
    options_contract_panel_df['pair_code_contract_unit'] = np.where(pair_code_conditions, options_contract_panel_df[
        'pair_code_contract_unit'], norm_unit)

    # 计算时间价值和内在价值
    logger.info(u'4. 计算时间价值和内在价值')
    calculate_time_value(options_contract_panel_df)

    # 测试数据
    options_contract_panel_df = options_contract_panel_df[
        (options_contract_panel_df['today'] > int(pd.to_datetime(start_date).strftime('%Y%m%d'))) & (
                options_contract_panel_df['today'] < int(pd.to_datetime(end_date).strftime('%Y%m%d')))]

    # 过滤Columns
    # 因为新老数据的关系，可能存在(oi)open_interest，这里判断一下是否有该字段
    if 'oi' not in options_contract_panel_df.columns:
        filter_columns = constant.ZQ_OPTIONS_PANEL_FORMAT_COLUMNS.copy()
        filter_columns.remove('oi')
        options_contract_panel_df = options_contract_panel_df[filter_columns]
    else:
        options_contract_panel_df = options_contract_panel_df[constant.ZQ_OPTIONS_PANEL_FORMAT_COLUMNS]

    # 计算希腊字母指标
    logger.info(u'5. 计算希腊字母指标: {}'.format(greeks))
    calculate_greeks_dll_np(options_contract_panel_df, greeks=greeks)
    # calculate_greeks_QL(options_contract_panel_df, greeks=greeks)

    return options_contract_panel_df


def _cache_panel_data(panel_df, freq, export_path):
    """
    将Panel的数据按照频率-年-月存储
    :param panel_df:
    :return:
    """
    node_panel_df_dict = dict()
    if panel_df.empty:
        logger.warning('Panel数据为空，不做存储')
        return panel_df
    # 按照频率/年/年月进行节点拆分
    panel_df['date_trade'] = pd.to_datetime(panel_df['date'] * 1000000 + panel_df['ets'], format='%Y%m%d%H%M%S')
    panel_df['dt_year'] = panel_df['date_trade'].dt.year
    panel_df['dt_month'] = panel_df['date_trade'].dt.month
    year_df_interator = panel_df.groupby(by=['dt_year'])
    for year, year_panel_df in year_df_interator:
        node_year = str(year)
        node_group_path = '/{}'.format(freq)
        for month, month_panel_df in year_panel_df.groupby(by=['dt_month']):
            # /time_interval/Y%M% exp: /freq/M201701
            node_dset_path = node_group_path + "/{}".format("M" + node_year + str(month).zfill(2))
            month_panel_df.drop(['dt_year', 'dt_month'], axis=1, inplace=True)
            node_panel_df_dict[node_dset_path] = month_panel_df
    # 存储拆分后的节点
    for node_path, node_panel_df in sorted(node_panel_df_dict.items()):
        node_panel_df.to_hdf(export_path, node_path, complevel=9, complib='blosc')
        logger.info('[节点存储成功]: 频率:{}, 节点名: {}, 数据量:{}, 路径:{}'.format(freq, node_path, len(node_panel_df), export_path))


def _get_dividend_date(list_date, last_trade_date, is_adjust, ex_date_list):
    """
    通过判断当前期权合约的生命周期，确定分红时间，只对is_adjust为1的生效，为0的默认为20161129
    :param list_date:
    :param last_trade_date:
    :param is_adjust:
    :param ex_date_list:
    :return:
    """
    dividend_date = DEFAULT_DIVIDEND_DAY
    if is_adjust == 0:
        return dividend_date
    for ex_date in ex_date_list:
        if list_date <= ex_date <= last_trade_date:
            dividend_date = int(ex_date)
            break
    return dividend_date


def _get_options_underlying_code(underlying_name):
    """
    通过underlying_name获取本地命名的标的合约代码
    :param underlying_name:
    :return:
    """
    underlying_code = constant.OPTIONS_UNDERLYING_CODE.get(underlying_name, None)
    if underlying_code is not None:
        return underlying_code
    raise TypeError(f'不支持的标的{underlying_name}')


def calculate_time_value(options_quote_df):
    """
    计算期权的时间价值和内在价值
    :param options_quote_df:
    :return:
    """
    options_quote_df['intrinsic_value'] = np.where(options_quote_df['contract_type'] == 'CO',
                                                   options_quote_df['underlying_close'] - options_quote_df[
                                                       'exercise_price'],
                                                   options_quote_df['exercise_price'] - options_quote_df[
                                                       'underlying_close'])
    # 内在价值大于等于零, 处理所有intrinsic_value < 0，则等于0
    options_quote_df['intrinsic_value'] = np.where(options_quote_df['intrinsic_value'] < 0, 0,
                                                   options_quote_df['intrinsic_value'])
    options_quote_df['time_value'] = options_quote_df['close'] - options_quote_df['intrinsic_value']


@timer
def calculate_greeks_dll_np(options_quote_df, v=0.3, R=0.015, greeks=['iv', 'delta', 'gamma', 'vega', 'theta', 'rho']):
    """
    用C++ DLL计算希腊字母值 美式期权计算方法 没有固定的行权日期
    :param options_quote_df:
    0: underlying_close
    1: contract_type
    2: exercise_price
    3: close
    4: days_to_maturity
    5: expire_date
    :param v: 波动率
    :param R: 无风险利率
    :param greeks: ['iv', 'delta', 'gamma', 'vega', 'theta', 'rho']
    :return:
    """
    options_quote_copy = options_quote_df[
        ['underlying_close', 'contract_type', 'exercise_price', 'close', 'days_to_maturity', 'exercise_type']].copy()
    greeks_value_dict = defaultdict(list)
    list_greek = [greeks, ] if isinstance(greeks, str) else greeks
    list_greek = [item.lower() for item in list_greek]
    for nd_array_item in tqdm(np.array(options_quote_copy)):
        s = nd_array_item[0]  # 标的价格 underlying_close
        cp = 0 if nd_array_item[1] == 'CO' else 1  # contract_type 0: Call 1: Put
        k = nd_array_item[2]  # 行权价 exercise_price
        premium = nd_array_item[3]  # close
        t = float(nd_array_item[4])  # 过期时间 days_to_maturity
        exercise_type = str(nd_array_item[5])
        # 如果是美式期权
        if exercise_type == 'A':
            sigma = COptionPrice.iv_us(s=s, cp=cp, k=k, market_price=premium, t=t, v=v, r=R)
            # 填充希腊字母数值, 新版本支持一次性返回包含delta/gamma/vega/theta/rho的对象, 需要根据认购/认股计算
            greek_obj = COptionPrice.get_greeks(s=s, k=k, r=R, v=v, t=t, cp=cp)
            greeks_value_dict['iv'].append(sigma)
            greeks_value_dict['delta'].append(greek_obj.delta)
            greeks_value_dict['gamma'].append(greek_obj.gamma)
            greeks_value_dict['theta'].append(greek_obj.theta)
            greeks_value_dict['vega'].append(greek_obj.vega)
            greeks_value_dict['rho'].append(greek_obj.rho)
        # 如果是欧式期权
        else:
            sigma = COptionPrice.iv_eu(s=s, cp=cp, k=k, market_price=premium, t=t, v=v, r=R)
            # 填充希腊字母数值
            for greek_item in list_greek:
                if greek_item == 'iv':
                    greek_value = sigma
                elif greek_item == 'delta':
                    greek_value = COptionPrice.delta_eu(s=s, cp=cp, k=k, t=t, v=sigma, r=R)
                elif greek_item == 'gamma':
                    greek_value = COptionPrice.gamma_eu(s=s, k=k, t=t, v=sigma, r=R)
                elif greek_item == 'vega':
                    greek_value = COptionPrice.vega_eu(s=s, k=k, t=t, v=sigma, r=R)
                elif greek_item == 'theta':
                    greek_value = COptionPrice.theta_eu(s=s, cp=cp, k=k, t=t, v=sigma, r=R)
                elif greek_item == 'rho':
                    greek_value = COptionPrice.rho_eu(s=s, cp=cp, k=k, t=t, v=sigma, r=R)
                greeks_value_dict[greek_item].append(greek_value)
    # 合并入数据中
    for greek_key, greek_values in greeks_value_dict.items():
        options_quote_df[greek_key] = greek_values


@timer
def calculate_greeks_QL(options_quote_df, R=0.025, greeks='iv'):
    """
    在已经生成的期权数据面板上面计算希腊值；用QuantLib提供的接口计算
    0: underlying_close
    1: contract_type
    2: exercise_price
    3: close
    4: days_to_maturity
    5: expire_date
    :param options_quote_df: 期权数据面板
    :param R: 无风险利率
    :param greeks: 列表：要计算的希腊值 ['iv', 'delta', 'gamma', 'vega', 'theta', 'rho']
    :return: 带有希腊值的期权面板数据
    """
    # 当前期权合约数量统计
    options_quote_copy = options_quote_df[
        ['underlying_close', 'contract_type', 'exercise_price', 'close', 'days_to_maturity', 'expire_date']].copy()
    greeks_value_dict = defaultdict(list)
    list_greek = [greeks, ] if isinstance(greeks, str) else greeks
    list_greek = [item.lower() for item in list_greek]
    calendar = China()
    for nd_array_item in np.array(options_quote_copy):
        expire_date_obj = dt.datetime.strptime(str(nd_array_item[5]), "%Y%m%d")
        exercise = EuropeanExercise(Date(expire_date_obj.day, expire_date_obj.month, expire_date_obj.year))  # 设定期权到期日
        eval_date_obj = expire_date_obj - timedelta(days=int(nd_array_item[4]))
        eval_date = Date(eval_date_obj.day, eval_date_obj.month, eval_date_obj.year)  # 当前交易日
        Settings.instance().evaluationDate = eval_date
        payoff = PlainVanillaPayoff(Option.Call, nd_array_item[2]) if nd_array_item[1] == 'CO' else PlainVanillaPayoff(
            Option.Put, nd_array_item[2])
        option = EuropeanOption(payoff=payoff, exercise=exercise)
        S = QuoteHandle(SimpleQuote(nd_array_item[0]))  # 标的价格
        r = YieldTermStructureHandle(FlatForward(eval_date, R, Actual360()))  # 无风险利率
        sigma = BlackVolTermStructureHandle(BlackConstantVol(eval_date, calendar, 0.5, Actual365Fixed()))  # 猜测Sigma
        process = BlackScholesProcess(s0=S, riskFreeTS=r, volTS=sigma)
        engine = AnalyticEuropeanEngine(process)  # fits vanila european option
        option.setPricingEngine(engine)
        # calculate and set volatility
        try:
            iv = option.impliedVolatility(targetValue=nd_array_item[3], process=process)
        except RuntimeError:
            iv = 0.001
        sigma = BlackVolTermStructureHandle(BlackConstantVol(eval_date, calendar, iv, Actual365Fixed()))  # 设定Sigma
        process = BlackScholesProcess(s0=S, riskFreeTS=r, volTS=sigma)
        engine = AnalyticEuropeanEngine(process)  # fits vanila european option
        option.setPricingEngine(engine)
        year_days = 365
        for greek_item in list_greek:
            if greek_item == 'iv':
                greek_value = iv
            elif greek_item == 'delta':
                greek_value = option.delta()
            elif greek_item == 'gamma':
                greek_value = option.gamma()
            elif greek_item == 'vega':
                greek_value = option.vega() / 100
            elif greek_item == 'theta':
                greek_value = option.theta() / year_days
            elif greek_item == 'rho':
                greek_value = option.rho()
            greeks_value_dict[greek_item].append(greek_value)
    # 合并入数据中
    for greek_key, greek_values in greeks_value_dict.items():
        options_quote_df[greek_key] = greek_values


def get_call_put_by_month_df(panel_quote_df, greeks='iv'):
    """
    for back testing, 选择各月份计算波动率指数的合约
    :param panel_quote_df:
    :param greeks:
    :return:
    """
    # 筛选出实际的时间线
    if panel_quote_df.empty:
        return panel_quote_df
    # 最终返回带时间戳的指标数据
    quote_bar_index_list = list()
    for date_trade_tuple in panel_quote_df.groupby(['date', 'today', 'sts', 'ets']).size().index:
        logger.info(u"[计算月波动率指数]: 时间戳: {} {}, 希腊字母: {}".format(date_trade_tuple[0], date_trade_tuple[3], greeks))
        quote_bar_dict = dict()
        # 筛选出时间戳bar
        each_bar_quote_df = panel_quote_df[
            (panel_quote_df['date'] == date_trade_tuple[0]) & (panel_quote_df['ets'] == date_trade_tuple[3])].copy()
        put_month_filter_data = _get_call_put_by_month_core(bar_quote_df=each_bar_quote_df, contract_type='PO')
        call_month_filter_data = _get_call_put_by_month_core(bar_quote_df=each_bar_quote_df, contract_type='CO')

        quote_bar_dict['date'] = date_trade_tuple[0]
        quote_bar_dict['today'] = date_trade_tuple[1]
        quote_bar_dict['sts'] = date_trade_tuple[2]
        quote_bar_dict['ets'] = date_trade_tuple[3]
        quote_bar_dict['underlying_close'] = each_bar_quote_df['underlying_close'].unique()[0]

        # 计算call
        for i in range(len(put_month_filter_data)):
            index_name = 'p{}'.format(i + 1)
            quote_bar_dict[index_name] = 0 if put_month_filter_data[i].empty else np.mean(
                put_month_filter_data[i][greeks])
        # 计算put
        for i in range(len(call_month_filter_data)):
            index_name = 'c{}'.format(i + 1)
            quote_bar_dict[index_name] = 0 if call_month_filter_data[i].empty else np.mean(
                call_month_filter_data[i][greeks])

        # 补全列
        for each_index in range(4):
            index_name = 'c{}'.format(each_index + 1)
            if quote_bar_dict.get(index_name, None) is None:
                quote_bar_dict[index_name] = 0
            index_name = 'p{}'.format(each_index + 1)
            if quote_bar_dict.get(index_name, None) is None:
                quote_bar_dict[index_name] = 0

        # 计算diff
        quote_bar_dict['c1-c2'] = quote_bar_dict['c1'] - quote_bar_dict['c2']
        quote_bar_dict['c2-c3'] = quote_bar_dict['c2'] - quote_bar_dict['c3']
        quote_bar_dict['c3-c4'] = quote_bar_dict['c3'] - quote_bar_dict['c4']
        quote_bar_dict['c2-c4'] = quote_bar_dict['c2'] - quote_bar_dict['c4']

        quote_bar_dict['p1-p2'] = quote_bar_dict['p1'] - quote_bar_dict['p2']
        quote_bar_dict['p2-p3'] = quote_bar_dict['p2'] - quote_bar_dict['p3']
        quote_bar_dict['p3-p4'] = quote_bar_dict['p3'] - quote_bar_dict['p4']
        quote_bar_dict['p2-p4'] = quote_bar_dict['p2'] - quote_bar_dict['p4']
        quote_bar_index_list.append(quote_bar_dict)
    quote_bar_index_df = pd.DataFrame.from_dict(quote_bar_index_list)
    quote_bar_index_df = quote_bar_index_df[
        ['date', 'today', 'sts', 'ets', 'c1', 'c2', 'c3', 'c4', 'p1', 'p2', 'p3', 'p4', 'c1-c2', 'c2-c3', 'c3-c4',
         'c2-c4',
         'p1-p2', 'p2-p3', 'p3-p4', 'p2-p4']]
    return quote_bar_index_df


def _get_call_put_by_month_core(bar_quote_df, contract_type, is_adjust=0):
    """
    选择各月份计算波动率指数的合约的核心代码
    :param bar_quote_df:
    :param contract_type:
    :param is_adjust:
    :return:
    """
    # 筛选需要用的记录，转化为ndarray之后再进行循环计算
    # bar_quote_df_copy = bar_quote_df[[]]
    # 根据传入的contract_type，筛选数据
    filtered_bar_quote_df = bar_quote_df[
        (bar_quote_df['close'] > 0) & (bar_quote_df['contract_type'] == contract_type) & (
                bar_quote_df['is_adjust'] == is_adjust)].copy()
    filter_left_params = 0.99 if contract_type == 'CO' else 0.93
    filter_right_params = 1.07 if contract_type == 'CO' else 1.01
    unique_month_list = sorted(filtered_bar_quote_df['expire_date'].unique().tolist())
    month_df_list = list()
    for each_month in unique_month_list:
        each_month_df = filtered_bar_quote_df[filtered_bar_quote_df['expire_date'] == each_month].sort_values(
            by=['exercise_price']).copy()
        each_month_df = each_month_df[
            (each_month_df['exercise_price'] >= each_month_df['underlying_close'] * filter_left_params) &
            (each_month_df['exercise_price'] <= each_month_df['underlying_close'] * filter_right_params)]
        month_df_list.append(each_month_df.copy())
    return month_df_list


def get_realtime_data(stockfullcode, start_dt=constant.DATETIME_MIN, end_dt=constant.DATETIME_MAX, freq=DEFAULT_FREQ,
                      db_string=constant.DB_RT3_RD_STRING, is_latest=False, dataframe=True, ed_type='pre'):
    """
    获取SQLServer的实时行情数据
    :param stockfullcode:
    :param start_dt:
    :param end_dt:
    :param freq:
    :param db_string:
    :param is_latest:
    :param dataframe:
    :param ed_type:
    :return:
    """
    quote_total_dict = dict()
    # 获取原始字典排列数据
    quote_df_dict = _get_realtime_data_org2(stockfullcode=stockfullcode, start_dt=start_dt, end_dt=end_dt, freq=freq,
                                            db_string=db_string, is_latest=is_latest, dataframe=False)
    # 转化格式
    if len(quote_df_dict) > 0:
        for each_code, each_quote_df in quote_df_dict.items():
            if each_quote_df.empty:
                continue
            quote_df = each_quote_df.copy()
            # 转化格式
            quote_df = convert_rt2zq_format(rt_kline_df=quote_df, freq=freq, contract=each_code)
            # 过滤
            quote_df = filter_quote_df(quote_df=quote_df)
            # 计算kidx和didx
            _calculate_zq_kdix(quote_df=quote_df, freq_str=freq)
            # freq转化，如果freq不是1分钟或者1d，则需要进行转化
            quote_df = period_conversion_resample(quote_df=quote_df, freq_str=freq)
            # 获取复权信息表
            ed_quote_df = quote_df.copy()
            if ed_type is not None:
                ed_info_df = get_ed_info_df(stockfullcode=each_code, start=constant.DATETIME_MIN,
                                            end=constant.DATETIME_MAX)
                if not ed_info_df.empty:
                    ed_info_df['date_execute'] = pd.to_datetime(ed_info_df['date_execute'], format='%Y%m%d')
                    # 如果在指定查询周期内没有找edinfo，则不复权
                if ed_info_df is not None and not ed_info_df.empty:
                    # 添加ed_info_idx,用于复权运算,每次都重新计算排列
                    feed_rio.add_contract_idx(quote_df=ed_quote_df, ed_info_df=ed_info_df)
                    # 复权操作
                    ed_quote_df = feed_rio.calculate_ed_info(data_df=ed_quote_df, ed_info_df=ed_info_df,
                                                             ed_type=ed_type)
            quote_total_dict[each_code] = ed_quote_df
    if len(quote_total_dict.values()) > 0:
        print(u'[数量]: {} 条'.format(sum([len(value) for key, value in quote_total_dict.items()])))
    if not dataframe:
        return quote_total_dict
    total_quote_df = pd.concat(quote_total_dict.values()) if len(quote_total_dict.values()) > 0 else return_data(
        return_frame=dataframe)
    return total_quote_df


def _get_realtime_data_org2(stockfullcode, start_dt=constant.DATETIME_MIN, end_dt=constant.DATETIME_MAX,
                            columns=DEFAULT_RT_DB_FIELDS, freq=DEFAULT_FREQ, db_string=constant.DB_RT3_RD_STRING,
                            is_latest=False, dataframe=True):
    quote_df_dict = dict()
    start_dt = _to_datetime(start_dt)
    end_dt = _to_datetime(end_dt)
    start_dt_str = start_dt.strftime('%Y-%m-%d %H:%M:%S')
    end_dt_str = end_dt.strftime('%Y-%m-%d %H:%M:%S')
    # 多品种/合约查询
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    # 设置表格
    if freq == '5m':
        query_table = 'VW_fiveminutes_history'
    elif freq == '1m':
        query_table = 'VW_oneminute_history'
    elif freq == '1d':
        query_table = 'VW_oneday_history'
    where_clause = ''
    # 构成合约查询条件
    if list_stockfullcode is not None:
        where_clause += '''stockfullcode IN ({})'''.format(','.join(["'" + s + "'" for s in list_stockfullcode]))
    # 构成日期查询条件
    if not is_latest:
        where_clause += ''' AND date_trade BETWEEN '{}' AND '{}' '''.format(start_dt_str, end_dt_str)
    # 连接筛选列
    if columns is not None:
        filter_columns = ', '.join(columns)
    try:
        conn = pyodbc.connect(db_string)
        if is_latest:
            sql = f"""
            SELECT {filter_columns} FROM {query_table} AS t1 WHERE date_trade = (SELECT MAX(t2.date_trade) from {query_table} as t2 WHERE {where_clause} AND t1.id_stock = t2.id_stock)
            """
        else:
            sql = f"""
            SELECT {filter_columns} FROM {query_table} AS t1 WHERE {where_clause} ORDER BY date_trade ASC, stockfullcode ASC
            """
        # 将读出来的结果直接变为dataframe
        quote_df = pd.read_sql(sql=sql, con=conn, coerce_float=True)
        if dataframe:
            return quote_df
        else:
            if not quote_df.empty:
                for stock, stock_quote_df in quote_df.groupby(by=['stockfullcode']):
                    stock_quote_df_copy = stock_quote_df.copy()
                    stock_quote_df_copy.reset_index(inplace=True, drop=True)
                    quote_df_dict[stock] = stock_quote_df_copy
            return quote_df_dict
    except Exception as ex:
        print('读取实时行情失败: 合约:{}, 开始时间:{}, 结束时间:{}, 频率:{}, 错误:{}'.format(','.join(list_stockfullcode), start_dt, end_dt,
                                                                       freq, ex))


def _get_realtime_data_org(stockfullcode, start_dt=constant.DATETIME_MIN, end_dt=constant.DATETIME_MAX,
                           freq=DEFAULT_FREQ,
                           db_string=constant.RT03_DB_RT_RD_ENGINE_STRING, is_latest=False, dataframe=True):
    quote_df_dict = dict()
    start_dt = _to_datetime(start_dt)
    end_dt = _to_datetime(end_dt)
    # 多品种/合约查询
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    # 确定model_views
    view_model = get_real_time_view_model(freq=freq)
    try:
        session = get_db_session(engine_string=db_string)
        if not is_latest:
            q = session.query(view_model)
        else:
            q = session.query(view_model.columns['stockfullcode'], func.max(view_model.columns['date_trade']))
        conditions = None
        # 过滤stockfullcode
        if list_stockfullcode is not None:
            conditions = view_model.columns['stockfullcode'].in_(list_stockfullcode)
        # 判断是否是is_latest
        if not is_latest:
            if conditions is None:
                conditions = and_(view_model.columns['date_trade'] >= start_dt,
                                  view_model.columns['date_trade'] <= end_dt)
            else:
                conditions = and_(conditions, view_model.columns['date_trade'] >= start_dt,
                                  view_model.columns['date_trade'] <= end_dt)
            q = q.filter(conditions)
            quote_data_list = [{key: getattr(item, key) for key in item._fields} for item in q.all()]
            quote_df = pd.DataFrame.from_dict(quote_data_list)
        else:
            # q = q.filter(conditions).order_by(desc(view_model.columns['date_trade'])).limit(len(list_stockfullcode) * 10)
            # 如果需要获取最新的数据，先groupby stockfullcode，然后针对最小的date_trade做过滤
            if conditions is not None:
                grouped_result = q.filter(conditions).group_by(view_model.columns['stockfullcode']).all()
            else:
                grouped_result = q.group_by(view_model.columns['stockfullcode']).all()
            latest_tm_list = list()
            quote_data_list = list()
            # 统计当前所有stockfullcode的最新数据时间
            for each_pair in grouped_result:
                latest_tm_list.append(each_pair[1])
            # 如果不为空，则获取最小时间
            if len(latest_tm_list) > 0:
                min_tm = min(latest_tm_list)
                # 以最小时间单位查询数据
                if conditions is None:
                    q = session.query(view_model).filter(view_model.columns['date_trade'] >= min_tm)
                else:
                    q = session.query(view_model).filter(and_(conditions, view_model.columns['date_trade'] >= min_tm))
                quote_data_list = [{key: getattr(item, key) for key in item._fields} for item in q.all()]
            quote_df = pd.DataFrame.from_dict(quote_data_list)
            if not quote_df.empty:
                quote_df = quote_df.sort_values(by=['stockfullcode', 'date_trade'], ascending=[True, False])
                latest_data_list = list()
                # 筛选出最新的数据
                for each_stock, each_quote_df in quote_df.groupby(by=['stockfullcode']):
                    latest_data_list.append(
                        each_quote_df[each_quote_df['date_trade'] == each_quote_df['date_trade'].max()])
                quote_df = pd.concat(latest_data_list, sort=True)
        if dataframe:
            return quote_df
        else:
            if not quote_df.empty:
                for stock, stock_quote_df in quote_df.groupby(by=['stockfullcode']):
                    stock_quote_df_copy = stock_quote_df.copy()
                    stock_quote_df_copy.reset_index(inplace=True, drop=True)
                    quote_df_dict[stock] = stock_quote_df_copy
            return quote_df_dict
    except Exception as ex:
        print('读取实时行情失败: 合约:{}, 开始时间:{}, 结束时间:{}, 频率:{}, 错误:{}'.format(','.join(list_stockfullcode), start_dt, end_dt,
                                                                       freq, ex))


def get_real_time_view_model(freq):
    """
    根据频率确定对应的view_model
    :param freq:
    :return:
    """
    view_model = db_rt_models.t_VW_oneminute_history
    if freq == '1m':
        view_model = db_rt_models.t_VW_oneminute_history
    elif freq == '5m':
        view_model = db_rt_models.t_VW_fiveminutes_history
    return view_model


def get_history_data(stockfullcode, start_date=None, end_date=None, fields=None, freq=DEFAULT_FREQ,
                     ed_date=None, ed_type=None, cache_path='nas', dataframe=True):
    '''
    获取本地/NAS/JQSDK的行情数据，如果cache_path为http，需要调用logout_jq_api手动退出登录
    :param stockfullcode: 品种，标准/JQ格式皆可
    :param start_date: 开始时间
    :param end_date: 结束时间
    :param fields: 筛选域，默认['date', 'today', 'span', 'open', 'close', 'high', 'low', 'vol', 'amount', 'high_limit',
                        'low_limit', 'avg', 'pre_close', 'sts', 'ets', 'oi', 'factor', 'contract']
    :param freq: 行情频率 期货支持最小5秒开始的行情，股票支持最小1分钟开始的行情，数字 + 频率的字符串组合。
    如: 1s(秒), 1m(分钟), 1d(日), 1w(周期), 1M(月), 1Q(季度)， 如果只写单位，则默认数量为1
    :param ed_date: 目前版本无作用
    :param ed_type: 复权类型 pre/post/None
    :param cache_path: 指定数据获取方式 nas: 办公室网盘缓存 (常量参数) http: JQData云上数据(常量参数) path: 本地缓存路径 （变量参数）
    :param dataframe:
    :return:
    '''
    quote_total_dict = dict()
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    list_fields = [fields, ] if isinstance(fields, str) else fields
    # 由于期货数据目前还不能和JQData统一标准，所以需要分开两条路径获取,如果指定路径为http，则尝试从JQ获取主力合约数据，该数据不做复权处理
    list_future_stock = [stockcode for stockcode in list_stockfullcode if
                         is_future(stockcode) and (
                                 ('9001' in stockcode) or ('9002' in stockcode) or (
                                 '9003' in stockcode)) and cache_path != 'http']
    # 对list_stockfullcode求差集, 统计走JQ数据的stock list
    list_jq_data_stock = list(set(list_stockfullcode).difference(set(list_future_stock)))
    # 调用读取期货数据源接口
    if len(list_future_stock) > 0:
        quote_future_dict = feed_rio.get_history_data(stockfullcode=list_future_stock, start_date=start_date,
                                                      end_date=end_date, fields=list_fields, ed_type=ed_type, freq=freq,
                                                      cache_path=cache_path, dataframe=False)
        # 如果没有数据,返回空
        if len(quote_future_dict) > 0:
            for each_future_code, future_quote_df in quote_future_dict.items():
                quote_total_dict[each_future_code] = future_quote_df

    # 调用读取JQData数据源接口
    if len(list_jq_data_stock) > 0:
        # 对list_jq_data_stock进行本地和HTTP两个访问路径的审核
        if cache_path == 'auto':
            cached_contracts_list, http_contracts_list = get_stockfullcode_read_path(cache_type='stock',
                                                                                     stockfullcode=list_jq_data_stock)
            cached_quote_dict = _get_history_data_org(stockfullcode=cached_contracts_list, start_date=start_date,
                                                      end_date=end_date,
                                                      fields=list_fields, ed_type=ed_type,
                                                      freq=freq, cache_path='nas', dataframe=False)
            quote_jq_dict = _get_history_data_org(stockfullcode=http_contracts_list, start_date=start_date,
                                                  end_date=end_date,
                                                  fields=list_fields, ed_type=ed_type,
                                                  freq=freq, cache_path='http', dataframe=False)
            quote_jq_dict.update(cached_quote_dict)
        else:
            quote_jq_dict = _get_history_data_org(stockfullcode=list_jq_data_stock, start_date=start_date,
                                                  end_date=end_date,
                                                  fields=list_fields, ed_type=ed_type,
                                                  freq=freq, cache_path=cache_path, dataframe=False)
        # 如果没有数据，返回空
        if len(quote_jq_dict) > 0:
            for each_code, quote_df in quote_jq_dict.items():
                if not quote_df.empty:
                    quote_df = quote_df.copy()
                    quote_df = filter_quote_df(quote_df=quote_df)
                    # 填充mid和midv
                    _fill_mid_and_midv(quote_df=quote_df)
                    # 计算kidx和didx
                    _calculate_zq_kdix(quote_df=quote_df, freq_str=freq)
                    # freq转化，如果freq不是1分钟或者1d，则需要进行转化
                    # quote_df1 = period_conversion(quote_df=quote_df.copy(), freq_str=freq)
                    quote_df = period_conversion_resample(quote_df=quote_df, freq_str=freq)
                    if cache_path != 'http':
                        # 复权操作
                        # 判断是否是期货数据，如果是，则需要附上ed_info中的factor
                        quote_df = get_factor_data(org_da=quote_df, stockfullcode=each_code, cache_path=cache_path,
                                                   ed_type=ed_type)
                        quote_df = cal_ed_factor_data(quote_df=quote_df, ed_type=ed_type)
                quote_total_dict[each_code] = quote_df
    if len(quote_total_dict.values()) > 0:
        print(u'[数量]: {} 条'.format(sum([len(value) for key, value in quote_total_dict.items()])))
    if not dataframe:
        return quote_total_dict
    total_quote_df = pd.concat(quote_total_dict.values()) if len(quote_total_dict.values()) > 0 else return_data(
        return_frame=dataframe)
    return total_quote_df


def get_stockfullcode_read_path(cache_type, stockfullcode):
    """
    将传入的stockfullcode与本地缓存的合约做交集和差集，分别对应本地访问和http访问
    :param cache_type:
    :param stockfullcode:
    :return:
    """
    cache_contracts_list = get_cached_contracts_list(cache_type=cache_type)
    # 交集
    intersection_stockfullcode = list(set(stockfullcode).intersection(set(cache_contracts_list)))
    # 差集
    diff_stockfullcode = list(set(stockfullcode).difference(set(cache_contracts_list)))
    return intersection_stockfullcode, diff_stockfullcode


def get_factor_data(org_da, stockfullcode, ed_type=None, cache_path='nas'):
    """
    读取factor数据，拼接到行情数据上，主要针对期货品种
    :param org_da:
    :param stockfullcode:
    :param ed_type:
    :param cache_path:
    :return:
    """
    market = _get_market(stockfullcode)
    org_da_copy = org_da.copy()
    if ed_type is not None:
        if market in ['SQ', 'ZZ', 'DL', 'SE', 'SF']:
            ed_info_df = get_oi_ed_info_data(stockfullcode=stockfullcode, cache_path=cache_path)
            if not ed_info_df.empty:
                ed_info_df['cur_factor'] = ed_info_df['current_contract_close_price'] / ed_info_df[
                    'next_contract_close_price']
                ed_info_df['factor'] = ed_info_df.groupby('symbol')['cur_factor'].cumprod()
                ed_info_head_df = pd.DataFrame([{'today': 19000000, 'factor': 1.0}])
                df_edinfo_sel = ed_info_df[['next_tday', 'factor']].copy()
                df_edinfo_sel.rename(columns={'next_tday': 'today'}, inplace=True)
                df_edinfo_total = pd.concat([ed_info_head_df, df_edinfo_sel[ed_info_head_df.columns]])
                da_merge_zq_future_factor(org_da_copy, df_edinfo_total, dict_cols={'factor': 'factor'})
    return org_da_copy


def get_oi_ed_info_data(stockfullcode, cache_path='nas'):
    '''
    获取期货的ed_info数据
    :param stockfullcode:
    :param cache_path:
    :return:
    '''
    ed_info_df = pd.DataFrame()
    from_local_cache = os.path.isdir(cache_path)
    if cache_path == 'nas' or from_local_cache:
        cache_root_path = constant.H5_JQ_KLINE_DIR if not from_local_cache else cache_path
    else:
        logger.warning(u"传入的cache_path不正确: {}，请确认指定['nas', 'http', 'cache_path']三类参数中的一个".format(cache_path))
        return pd.DataFrame()
    # 循环构造stockfullcode的h5路径
    if not from_local_cache:
        cache_file_dir = _get_cache_files_dir(cache_root_path=cache_root_path, stockfullcode=stockfullcode,
                                              is_local=False)
    else:
        cache_file_dir = cache_path
    cache_file_path = _get_h5_file_path(source_path=cache_file_dir, stockfullcode=stockfullcode, h5_type='kline')
    try:
        ed_info_df = pd.read_hdf(path_or_buf=cache_file_path, key='edinfo')
    except Exception as ex:
        logger.warning(u'品种[{}], 读取Edinfo数据失败:{}'.format(stockfullcode, ex))
    return ed_info_df


def _period_conversion(quote_df, freq_str='5m'):
    """
    对ZQ格式的数据进行freq之间的转化，对于1m和1d，则直接返回相应数据。
    如果传入的基础Kline不是连续的，转化数据有时间戳的问题，目前已经不用了
    :param quote_df: ZQ格式的数据，具体列见ZQ_KLINE_FORMAT_COLUMNS
    :param freq_str: 传入要转化的频率字符串，分钟+m，注意必须是quote_df级别的倍数
    :return:
    """
    assert isinstance(quote_df, pd.DataFrame)
    if quote_df.empty:
        return pd.DataFrame()
    # 默认基础数据的级别是1分钟和1d,如果源数据是1分钟或者1d，则不需要进行转化
    if freq_str == '1m' or freq_str == '1d':
        return quote_df
    freq_minute_num = int(convert_freq2num(freq_str=freq_str) / 60)
    # 检查DateTrade是否在columns中，如果没有，需要手动用date和ets构造
    if 'DateTrade' not in quote_df.columns:
        quote_df['DateTrade'] = pd.to_datetime(quote_df['date'] * 1000000 + quote_df['ets'], format='%Y%m%d%H%M%S')
    t = quote_df['symbol'].shift(1)
    quote_df['symbol_first'] = (t != quote_df['symbol'])
    t = quote_df['symbol'].shift(-1)
    quote_df['symbol_last'] = (t != quote_df['symbol'])
    quote_df['rate'] = 1
    rate_v = quote_df['rate'].values
    symbol_first_v = quote_df['symbol_first'].values
    for i in range(len(quote_df)):
        if symbol_first_v[i]:
            pass
        elif rate_v[i - 1] == freq_minute_num:
            pass
        else:
            rate_v[i] = rate_v[i - 1] + 1
    quote_df['rate'] = rate_v
    quote_df['grp_flg'] = (quote_df['rate'] == freq_minute_num) | quote_df['symbol_last']
    quote_df['number'] = 1
    number_v = quote_df['number'].values
    grp_flg_v = quote_df['grp_flg'].values
    next_number = 1
    for i in range(len(quote_df)):
        number_v[i] = next_number
        if grp_flg_v[i]:
            next_number += 1
    quote_df['number'] = number_v
    df_symbol = quote_df.groupby('number', as_index=False)['symbol'].first()
    df_symbol['DateTrade'] = quote_df.groupby('number')['DateTrade'].last().values
    if 'open' in quote_df.columns:
        df_symbol['open'] = quote_df.groupby('number')['open'].first().values
    if 'close' in quote_df.columns:
        df_symbol['close'] = quote_df.groupby('number')['close'].last().values
    if 'high' in quote_df.columns:
        df_symbol['high'] = quote_df.groupby('number')['high'].max().values
    if 'low' in quote_df.columns:
        df_symbol['low'] = quote_df.groupby('number')['low'].min().values
    if 'vol' in quote_df.columns or 'volume' in quote_df.columns:
        df_symbol['vol'] = quote_df.groupby('number')['vol'].sum().values
    if 'amount' in quote_df.columns or 'money' in quote_df.columns:
        df_symbol['amount'] = quote_df.groupby('number')['amount'].sum().values
    if 'oi' in quote_df.columns or 'open_interest' in quote_df.columns:
        df_symbol['oi'] = quote_df.groupby('number')['oi'].last().values
    if 'today' in quote_df.columns:
        df_symbol['today'] = quote_df.groupby('number')['today'].last().values
    if 'factor' in quote_df.columns:
        df_symbol['factor'] = quote_df.groupby('number')['factor'].last().values
    if 'contract' in quote_df.columns:
        df_symbol['contract'] = quote_df.groupby('number')['contract'].last().values
    if 'mid' in quote_df.columns:
        df_symbol['mid'] = quote_df.groupby('number')['mid'].last().values
    if 'midv' in quote_df.columns:
        df_symbol['midv'] = quote_df.groupby('number')['midv'].last().values
    if 'symbol' in quote_df.columns:
        df_symbol['symbol'] = quote_df.groupby('number')['symbol'].last().values

    # 重新将zq格式的date，sts和ets通过DateTrade构造出来
    df_symbol['date'] = df_symbol.DateTrade.dt.strftime("%Y%m%d").astype(np.int64)
    span_value = convert_freq2num(freq_str=freq_str)
    df_symbol['span'] = span_value
    df_symbol['ets'] = df_symbol.DateTrade.dt.strftime('%H%M%S').astype(np.int64)
    df_symbol['sts'] = (df_symbol.DateTrade - pd.Timedelta(seconds=int(span_value))).dt.strftime('%H%M%S').astype(
        np.int64)

    df_symbol.drop(['number', 'DateTrade'], inplace=True, axis=1)
    # 重新计算kidx和didx
    _calculate_zq_kdix(quote_df=df_symbol, freq_str=freq_str)

    # 列排序
    df_symbol = df_symbol[constant.ZQ_KLINE_FORMAT_COLUMNS]

    return df_symbol


def period_conversion_options_panel(quote_df, freq_str='5m', filter_columns=constant.ZQ_OPTIONS_PANEL_FORMAT_COLUMNS):
    """
    对options panel数据进行分钟转化
    :param quote_df:
    :param freq_str:
    :param filter_columns:
    :return:
    """
    assert isinstance(quote_df, pd.DataFrame)
    if quote_df.empty:
        return return_data(return_frame=True)
    # 默认期权面板的基础数据的级别是1分钟和1d,如果源数据是1分钟或者1d，则不需要进行转化
    if freq_str == '1m':
        return quote_df
    quote_df_copy = quote_df.copy()
    resample_interval_type = get_resample_interval(freq_str=freq_str)
    freq_num = convert_freq2num(freq_str=freq_str)
    if 'DateTrade' not in quote_df_copy.columns:
        quote_df_copy['DateTrade'] = pd.to_datetime(quote_df_copy['date'] * 1000000 + quote_df_copy['ets'],
                                                    format='%Y%m%d%H%M%S')
    dict_how = dict()
    for each_column in quote_df_copy.columns:
        if each_column in constant.MERGE_OPTIONS_PANEL_KLINE_DICT.keys():
            dict_how[each_column] = constant.MERGE_OPTIONS_PANEL_KLINE_DICT[each_column]
    # 存放resample_quote_df的集合
    resample_quote_list = list()
    # 设置OI为-1
    quote_df_copy.fillna({'oi': -1}, inplace=True)
    # 按照groupby contract和today做比较
    for contract, contract_df in quote_df_copy.groupby(by=['contract']):
        contract_df_copy = contract_df.copy()
        # 排序
        contract_df_copy = contract_df_copy.sort_values(by=['DateTrade'])
        # 将DateTrade赋值到index, resample依赖于时间序列的index进行切分
        contract_df_copy.index = contract_df_copy.DateTrade
        period_quotes = contract_df_copy.resample('%s' % resample_interval_type, closed='right',
                                                  label='right').apply(dict_how)
        # 价格按照OHLC进行重采样
        period_quotes = period_quotes.dropna(how='any')
        resample_quote_list.append(period_quotes)
    resample_quote_df = pd.concat(resample_quote_list, sort=True)
    # 重新将zq格式的date，sts和ets通过DateTrade构造出来
    span_value = convert_freq2num(freq_str=freq_str)
    resample_quote_df['span'] = span_value
    resample_quote_df['ets'] = 0
    resample_quote_df['sts'] = 0
    # 修复周期转化的bug, resample转化1d的周期会让index + 1 天
    if freq_num < 86400:
        resample_quote_df['date'] = resample_quote_df.index.strftime("%Y%m%d").astype(np.int64)
        resample_quote_df['ets'] = resample_quote_df.index.strftime('%H%M%S').astype(np.int64)
        resample_quote_df['sts'] = (resample_quote_df.index - pd.Timedelta(seconds=int(span_value))).strftime(
            '%H%M%S').astype(np.int64)
    resample_quote_df['date'] = resample_quote_df['date'].astype(np.int64)
    resample_quote_df['today'] = resample_quote_df['today'].astype(np.int64)
    # 列排序
    resample_quote_df = resample_quote_df[quote_df.columns]
    # reset_index
    resample_quote_df.reset_index(drop=True, inplace=True)
    return resample_quote_df


# 分钟线之间的转化
def period_conversion_resample(quote_df, freq_str='5m', filter_columns=constant.ZQ_KLINE_FORMAT_COLUMNS):
    '''
    :param quote_df: zq格式的K线数据
    :param freq_str:
    :return:
    '''
    # 获取resample时间线
    assert isinstance(quote_df, pd.DataFrame)
    if quote_df.empty:
        return pd.DataFrame()
    # 默认基础数据的级别是1分钟和1d,如果源数据是1分钟或者1d，则不需要进行转化
    if freq_str == '1m':
        return quote_df
    quote_df = quote_df.copy()
    resample_interval_type = get_resample_interval(freq_str=freq_str)
    freq_num = convert_freq2num(freq_str=freq_str)
    if 'DateTrade' not in quote_df.columns:
        quote_df['DateTrade'] = pd.to_datetime(quote_df['date'] * 1000000 + quote_df['ets'], format='%Y%m%d%H%M%S')
    dict_how = dict()
    for each_column in quote_df.columns:
        if each_column in constant.MERGE_KLINE_DICT.keys():
            dict_how[each_column] = constant.MERGE_KLINE_DICT[each_column]

    # 将DateTrade赋值到index, resample依赖于时间序列的index进行切分
    quote_df.index = quote_df.DateTrade
    period_quotes = quote_df.resample('%s' % resample_interval_type, closed='right', label='right').apply(dict_how)
    # 价格按照OHLC进行重采样
    period_quotes = period_quotes.dropna(how='any')
    # 重新将zq格式的date，sts和ets通过DateTrade构造出来
    span_value = convert_freq2num(freq_str=freq_str)
    period_quotes['span'] = span_value
    period_quotes['ets'] = 0
    period_quotes['sts'] = 0
    # 修复周期转化的bug, resample转化1d的周期会让index + 1 天
    if freq_num < 86400:
        period_quotes['date'] = period_quotes.index.strftime("%Y%m%d").astype(np.int64)
        period_quotes['ets'] = period_quotes.index.strftime('%H%M%S').astype(np.int64)
        period_quotes['sts'] = (period_quotes.index - pd.Timedelta(seconds=int(span_value))).strftime(
            '%H%M%S').astype(np.int64)
    period_quotes['date'] = period_quotes['date'].astype(np.int64)
    period_quotes['today'] = period_quotes['today'].astype(np.int64)
    # 重新计算kidx和didx
    _calculate_zq_kdix(quote_df=period_quotes, freq_str=freq_str)
    # 列排序
    period_quotes = period_quotes[filter_columns]
    return period_quotes


def get_df_tday_datetime():
    """
    得到df_tday_tail的信息，含有tday_datetime列，today列，sts列。
    :return df_tday_tail: 含有tday_datetime列，today列，sts列。
    tday_datetime列的作用：从tday视角看的时间刻度的量化，最小分割单位时分钟。
    today：对应的today列
    sts：对应的sts列
    注意tday视角和自然日视角是不同的，1个tday一般最多可以包含有3个自然日的信息，但这3个自然日中的sts一定是不相同的。
    """
    df_tday = get_tday_df()
    df_tday['tday_end'] = df_tday['tday'].values + np.timedelta64(16, 'h')
    df_tday['tday_start'] = df_tday['tday_end'].shift(1)
    df_tday['today'] = df_tday['tday'].dt.strftime("%Y%m%d").astype(np.int64)
    df_tday['all'] = 1
    delta_s = [np.timedelta64(1, 'm') * x for x in range(60 * 24)]  # 按照1天有多少分钟产生时间戳
    df_delta = pd.DataFrame(delta_s, columns=['delta'])
    df_delta['all'] = 1
    df_tday_datetime = pd.merge(df_tday, df_delta, how='outer', on='all')
    df_tday_datetime['tday_datetime'] = df_tday_datetime['tday'] + df_tday_datetime['delta'] + np.timedelta64(1, 'm')
    df_tday_datetime['datetime_temp'] = df_tday_datetime['tday_start'] + df_tday_datetime['delta']
    df_tday_datetime['sts'] = df_tday_datetime['datetime_temp'].dt.hour * 10000 \
                              + df_tday_datetime['datetime_temp'].dt.minute * 100
    valid_cols = ['tday_datetime', 'today', 'sts']
    df_tday_datetime = df_tday_datetime[~df_tday_datetime['datetime_temp'].isnull()][valid_cols].copy()
    df_tday_datetime['sts'] = df_tday_datetime['sts'].astype(np.int64)
    return df_tday_datetime


def period_conversion_resample_sml(quote_df, freq_str='5m', filter_columns=constant.ZQ_KLINE_FORMAT_COLUMNS,
                                   dict_kline=constant.MERGE_KLINE_DICT):
    '''
    :param quote_df: zq格式的K线数据
    :param freq_str: 转化后的采样精度
    :param filter_columns: 输出的数据列
    :param dict_kline: resample的规则
    :return:
    '''
    # 获取resample时间线
    assert isinstance(quote_df, pd.DataFrame)
    if quote_df.empty:
        return pd.DataFrame()
    # 默认基础数据的级别是1分钟和1d,如果源数据是1分钟或者1d，则不需要进行转化
    if freq_str == '1m':
        return quote_df
    quote_df = quote_df.copy()
    resample_interval_type = get_resample_interval(freq_str=freq_str)
    quote_df['datetime'] = pd.to_datetime(quote_df['date'] * 1000000 + quote_df['ets'], format='%Y%m%d%H%M%S')
    if quote_df['span'].iat[0] < 86400:  # 表示是源数据是日内数据
        df_tday_datetime = get_df_tday_datetime()
        quote_df2 = pd.merge(quote_df, df_tday_datetime, how='left', on=['today', 'sts'])
    else:
        quote_df2 = quote_df.copy()
        quote_df2['tday_datetime'] = quote_df['datetime']

    dict_how = dict()
    for each_column in quote_df.columns:
        if each_column in dict_kline.keys():
            dict_how[each_column] = dict_kline[each_column]

    dict_how['datetime'] = 'last'
    # 将DateTrade赋值到index, resample依赖于时间序列的index进行切分
    quote_df2.index = quote_df2.tday_datetime
    period_quotes = quote_df2.resample('%s' % resample_interval_type, closed='right', label='right').apply(dict_how)
    # 价格按照OHLC进行重采样
    period_quotes = period_quotes.dropna(how='any')
    # 重新将zq格式的date，sts和ets通过DateTrade构造出来
    span_value = convert_freq2num(freq_str=freq_str)
    period_quotes['span'] = span_value
    period_quotes['date'] = period_quotes['date'].astype(np.int64)
    period_quotes['today'] = period_quotes['today'].astype(np.int64)
    period_quotes.reset_index(inplace=True)
    if span_value < 86400:  # 转化后是日内数据
        period_quotes['ets'] = period_quotes['datetime'].dt.strftime('%H%M%S').astype(np.int64)
        period_quotes['sts'] = (period_quotes['datetime'] - pd.Timedelta(seconds=int(span_value))).dt.strftime(
            '%H%M%S').astype(np.int64)
    else:
        period_quotes['ets'] = 0
        period_quotes['sts'] = 0
    # 重新计算kidx和didx
    _calculate_zq_kdix(quote_df=period_quotes, freq_str=freq_str)

    # 列排序
    out_cols = [x for x in filter_columns if x in period_quotes.columns]
    period_quotes = period_quotes[out_cols]
    return period_quotes


def cal_ed_factor_data(quote_df, ed_type=None):
    """
    将df_basic的数据转换为不复权，后复权和前复权的数据。
    股票：金额不变/成交量和价格变化
    :param quote_df: 字段中除了'factor'字段为后复权时的值，其他都是不复权时的值
    :param ed_type: 复权类型
    :return: 返回JQdata的实时的股票数据，拥有的列有：
    DateTrade: datetime64[ns]，时间段结束点
    open: float64，开盘价
    close：float64，收盘价
    high：float64，最高价
    low：float64，最低价
    volume：float64，成交量
    amount：float64，成交金额
    symbol：str，品种名字，支持股票，期货，指数，基金，lof基金，etf基金，分级基金A，分级基金B
    额外字段信息：
    'factor'：复权因子，当symbol是指数时，该项无用：
    ed_type为None的时候，factor一直等于1，不复权。
    ed_type为'post'的时候，factor的计算是从上市日开始，所以不同的起始时间，同一个时间点的复权因子不变。
    ed_type为'pre'的时候，结束点的复权因子是1，每次发生复权的时候，除以当时点的相对复权因子。相对复权因子可以有2种方法得到：
    A）通过公式计算得到。要知道分红比例，送股比例，送股价格，配股比例，配股价格，执行日收盘价。
    B）相对复权因子通过ed_type为'post'时保留的复权因子计算得到。相对复权因子=本交易日的复权因子 / 上交易日的复权因子。
    注意：复权方式价格改变，成交量改变，但成交金额不变。
    'high_limit'：涨停价，一般情况是上个交易日的收盘价110%，但刚上市的股票无此限制，可以认为是个特殊点。
    刚上市的股票有多种情况：或者是刚开始上市，或者是股权分置改革，或者是重大重组等事件。
    'low_limit'：跌停价，一般情况是上个交易日的收盘价90%.
    涨停价和跌停价在前复权和后复权的时候计算公式保持不变。在不复权的时候且正好是复权的时候，上个交易日的收盘价应该调整为
    上个交易日的收盘价 / 第二天的相对复权因子
    'vwap'：平均价格，相当于amount / volume
    'pre_close'：上个Bar的收盘价
    """
    if quote_df.empty:
        return pd.DataFrame()
    df = quote_df.copy()
    if ed_type is None:  # 不复权
        df['factor'] = 1.0
        return df
    elif ed_type == 'post':  # 后复权
        for col in ['open', 'close', 'high', 'low', 'high_limit', 'low_limit', 'pre_close', 'avg', 'vwap', 'mid',
                    'midv']:
            if col in df.columns:
                df[col] = round(quote_df[col] * quote_df['factor'], 2)
        for col in ['vol']:
            if col in df.columns:
                df[col] = round(quote_df[col] / quote_df['factor'], 0)
    elif ed_type == 'pre':
        df_factor_g = df.groupby('symbol', as_index=False)['factor'].last()
        df_factor_g.rename(columns={'factor': 'factor_last'}, inplace=True)
        df = pd.merge(df, df_factor_g, how='left', on='symbol')
        df['factor'] = (df['factor'] / df['factor_last'])
        for col in ['open', 'close', 'high', 'low', 'high_limit', 'low_limit', 'pre_close', 'avg', 'vwap', 'mid',
                    'midv']:
            if col in df.columns:
                df[col] = round(quote_df[col] * df['factor'], 2)
        for col in ['vol']:
            if col in df.columns:
                df[col] = round(quote_df[col] / df['factor'], 0)
        del df['factor_last']
    return df


def da_merge_zq_future_factor(lda, rda, dict_cols={'open': 'open'}):
    """
    merge 期货数据的复权因子
    :param lda: 要merge的df
    :param rda: merge 被merge的df
    :param dict_cols: 字典结构，key值是要merge的rda的列，value值是lda对应的添加的列名
    :return: lda增加新列
    """
    old_cols, new_cols = [], []
    for key in dict_cols:
        old_cols.append(key)
        new_cols.append(dict_cols[key])
        lda[dict_cols[key]] = rda[key].iloc[0]

    lda_merge_v = lda[new_cols].values
    rda_merge_v = rda[old_cols].values
    rda_date_v = (rda['today']).values
    date_end = 21 * (10 ** 6)
    rda_date_v = np.concatenate((rda_date_v, [date_end]))
    lda_date_v = (lda['today']).values

    j = 0
    step = 1000
    for i in range(len(lda)):
        cur_date = lda_date_v[i]
        while j < (len(rda) - step):
            if rda_date_v[j + step] <= cur_date:
                j += step
                continue
            else:
                while rda_date_v[j + 1] <= cur_date:
                    j += 1
                    continue
                else:
                    break
        else:
            while rda_date_v[j + 1] <= cur_date:
                j += 1
                continue
        lda_merge_v[i] = rda_merge_v[j]
    lda[new_cols] = lda_merge_v


def da_merge_zq(lda, rda, dict_cols={'open': 'open'}):
    """
    :param lda: 要merge的df
    :param rda: merge 被merge的df
    :param dict_cols: 字典结构，key值是要merge的rda的列，value值是lda对应的添加的列名
    :return: lda增加新列
    """
    old_cols, new_cols = [], []
    for key in dict_cols:
        old_cols.append(key)
        new_cols.append(dict_cols[key])
        lda[dict_cols[key]] = rda[key].iloc[0]

    lda_merge_v = lda[new_cols].values
    rda_merge_v = rda[old_cols].values
    rda_date_v = (rda['date'] * (10 ** 6) + rda['ets']).values
    date_end = 21 * (10 ** 12)
    rda_date_v = np.concatenate((rda_date_v, [date_end]))
    lda_date_v = (lda['date'] * (10 ** 6) + lda['ets']).values

    j = 0
    step = 1000
    for i in range(len(lda)):
        cur_date = lda_date_v[i]
        while j < (len(rda) - step):
            if rda_date_v[j + step] <= cur_date:
                j += step
                continue
            else:
                while rda_date_v[j + 1] <= cur_date:
                    j += 1
                    continue
                else:
                    break
        else:
            while rda_date_v[j + 1] <= cur_date:
                j += 1
                continue
        lda_merge_v[i] = rda_merge_v[j]
    lda[new_cols] = lda_merge_v


def convert_rt2zq_format(rt_kline_df, freq, contract):
    """
    转化实时数据为ZQ格式数据
    :param rt_kline_df:
    :param freq_str:
    :param contract:
    :return:
    """
    # 转化fre为秒
    span_value = convert_freq2num(freq_str=freq)
    # 构造tday
    market = _get_market(contract)
    if market != 'futures':
        quote_df = rt_kline_df.copy()
        quote_df.loc[:, 'tday'] = quote_df['date_trade']
    else:
        df_tday = get_tday_df(rt_kline_df.iloc[0]['date_trade'].date() - pd.Timedelta(days=10),
                              rt_kline_df.iloc[-1]['date_trade'].date() + pd.Timedelta(days=10)).copy()
        df_tday.loc[:, 'next_tday'] = df_tday['tday'].shift(-1)
        df_tday = df_tday.astype(np.datetime64)
        rt_kline_df.loc[:, 'date_trade_tmp'] = rt_kline_df['date_trade']
        if 'tday' in rt_kline_df.columns:
            rt_kline_df.drop(['tday'], axis=1, inplace=True)
        rt_kline_df.loc[:, 'date_trade_tmp'] = np.where(rt_kline_df['date_trade'].dt.hour < 6,
                                                        rt_kline_df['date_trade'] - pd.Timedelta(hours=6),
                                                        rt_kline_df['date_trade'])
        rt_kline_df.loc[:, 'date'] = rt_kline_df['date_trade_tmp'].dt.date.astype(np.datetime64)
        quote_df = pd.merge(rt_kline_df, df_tday, how='left', left_on='date', right_on='tday')

        quote_df.sort_index(inplace=True, ascending=False)
        quote_df.loc[:, 'tday'] = np.where((quote_df['date_trade_tmp'].dt.hour > 16), quote_df['next_tday'],
                                           quote_df['tday'])
    # 格式化tday为int
    quote_df['today'] = quote_df['tday'].dt.strftime('%Y%m%d')

    # 排序
    quote_df.sort_values(by=['date_trade'], inplace=True)

    # 定义类型
    quote_df = quote_df.astype(
        dtype={
            'amt_amount': np.float,
            'val_sp1': np.float,
            'val_sv1': np.float,
            'val_bp1': np.float,
            'val_bv1': np.float,
            'val_close': np.float,
            'val_high': np.float,
            'val_low': np.float,
            'val_open': np.float,
            'val_OI': np.float,
            'date_trade': np.datetime64,
            'amt_amount': np.float,
            'tday': np.int64
        })

    # 构造ZQ格式
    quote_df['date'] = quote_df['date_trade'].dt.strftime("%Y%m%d").astype(np.int64)
    quote_df['ets'] = quote_df['date_trade'].dt.strftime('%H%M%S').astype(np.int64)
    quote_df['sts'] = (quote_df['date_trade'] - pd.Timedelta(seconds=int(span_value))).dt.strftime('%H%M%S').astype(
        np.int64)

    # 重命名Columns
    quote_df.rename(columns=lambda c: c.lower(), inplace=True)
    quote_df.rename(columns={
        'val_sp1': 'Ask1Price',
        'val_sv1': 'Ask1Volume',
        'val_bp1': 'Bid1Price',
        'val_bv1': 'Bid1Volume',
        'amt_amount': 'amount',
        'amt_volumn': 'vol',
        'val_oi': 'oi',
        'val_open': 'open',
        'val_close': 'close',
        'val_high': 'high',
        'val_low': 'low'
    }, inplace=True)

    # 计算Mid和Midv
    calculate_mid_and_midv(quote_df=quote_df, close_column='close')

    # contracts
    quote_df['contract'] = contract
    quote_df['span'] = span_value

    # 赋值通用字段
    quote_df['high_limit'] = -1
    quote_df['low_limit'] = -1
    quote_df['avg'] = -1
    quote_df['pre_close'] = -1
    quote_df['factor'] = -1

    # 排序
    # 判断是否有mid或者midv
    sort_columns = copy.deepcopy(constant.ZQ_KLINE_ORG_COLUMNS)
    if 'mid' in quote_df.columns:
        sort_columns.extend(['mid'])
    if 'midv' in quote_df.columns:
        sort_columns.extend(['midv'])
    quote_df = quote_df[sort_columns]

    # 定义类型
    quote_df = quote_df.astype(constant.ZQ_KLINE_ORG_COLUMNS_DTYPES_DICT, copy=True)

    # 构造symbol
    quote_df['symbol'] = contract
    return quote_df


def convert_httpjq2zq_format(jq_kline_df, freq_str, contract, trading_dt=None):
    """
    转化单品种的JQKline数据为ZQ格式数据
    :param jq_kline_df:
    :param freq_str:
    :param tday:
    :param security_type:
    :return:
    """
    zq_kline_df = jq_kline_df.copy()
    span_value = convert_freq2num(freq_str=freq_str)
    zq_kline_df['span'] = span_value
    market = _get_market(contract)
    if market != 'futures':
        dt_year = zq_kline_df.index.year
        dt_month = zq_kline_df.index.month
        dt_day = zq_kline_df.index.day
        dt_hour = zq_kline_df.index.hour
        dt_minute = zq_kline_df.index.minute
        zq_kline_df['date'] = dt_year * 10000 + dt_month * 100 + dt_day
        zq_kline_df['ets'] = dt_hour * 10000 + dt_minute * 100
        zq_kline_df['sts'] = np.where(freq_str == '1d', 0,
                                      np.where(zq_kline_df['ets'] % 10000 == 0, zq_kline_df['ets'] - 4100,
                                               zq_kline_df['ets'] - 100))
    else:
        zq_kline_df['date'] = zq_kline_df.index.strftime("%Y%m%d").astype(np.int64)
        zq_kline_df['ets'] = zq_kline_df.index.strftime('%H%M%S').astype(np.int64)
        zq_kline_df['sts'] = (zq_kline_df.index - pd.Timedelta(seconds=int(span_value))).strftime('%H%M%S').astype(
            np.int64)
    if trading_dt is None:
        zq_kline_df['today'] = zq_kline_df['date']
    # 如果trading_dt不为空，表示传入交易时间表，然后做交易时间的merge
    else:
        trading_dt_copy = trading_dt.copy()
        trading_dt_copy.rename(columns={'date': 'today'}, inplace=True)
        rda_year = trading_dt_copy.start_trading_dt.dt.year
        rda_month = trading_dt_copy.start_trading_dt.dt.month
        rda_day = trading_dt_copy.start_trading_dt.dt.day
        rda_hour = trading_dt_copy.start_trading_dt.dt.hour
        rda_minute = trading_dt_copy.start_trading_dt.dt.minute
        trading_dt_copy['date'] = rda_year * 10000 + rda_month * 100 + rda_day
        trading_dt_copy['ets'] = rda_hour * 10000 + rda_minute * 100
        da_merge_zq(zq_kline_df, trading_dt_copy, dict_cols={'today': 'today'})
    # 将把header全部lowercase
    zq_kline_df.columns = zq_kline_df.columns.str.lower()
    if 'open_interest' not in zq_kline_df.columns:
        zq_kline_df.loc[:, 'open_interest'] = -1
    zq_kline_df['open_interest'] = np.where(zq_kline_df['open_interest'].isnull(), -1, zq_kline_df['open_interest'])
    # 修改column name
    zq_kline_df = zq_kline_df.rename(columns={'volume': 'vol', 'open_interest': 'oi', 'money': 'amount'})

    # 将contract的jq格式转成标准模式
    if 'contract' in zq_kline_df.columns:
        zq_kline_df['contract'] = normalize_code_to_local(zq_kline_df['contract'].tolist())

    # 添加没有的列
    for extra_column in ['high_limit', 'low_limit', 'avg', 'pre_close', 'factor']:
        if extra_column not in zq_kline_df.columns:
            zq_kline_df[extra_column] = -1

    # 排序
    zq_kline_df = zq_kline_df[constant.ZQ_KLINE_ORG_COLUMNS]

    # 定义类型
    zq_kline_df = zq_kline_df.astype(constant.ZQ_KLINE_ORG_COLUMNS_DTYPES_DICT, copy=True)

    return zq_kline_df


def _get_history_data_org_from_jq(stockfullcode, start_date=None, end_date=None, jq_fields=None, freq=DEFAULT_FREQ,
                                  ed_type=None, dataframe=True):
    """
    从JQData SDK获取行情数据，返回数据源等同于_get_history_info_org
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param jq_fields:
    :param freq:
    :param ed_type:
    :param dataframe:
    :return:
    """
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    jq_quote_df_dict = dict()
    # 转化stockfullcode格式为jq
    jq_stockfullcode = normalize_code_to_jq(list_stockfullcode)
    # 根据stockfullcode的个数，决定处理返回的数据
    freq_num = convert_freq2num(freq)
    if jq_fields is None:
        if freq_num < 86400:
            jq_fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'open_interest']
        else:
            jq_fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'factor',
                         'high_limit', 'low_limit', 'avg', 'pre_close', 'open_interest']
    # 根据start/end_date获取
    trading_dt_df = _get_trading_datetime_range()
    try:
        # 先尝试默认Panel格式获取，该方式适用于股票/指数等交易时间固定的品种
        try:
            jq_data = get_jq_api().get_price(
                security=jq_stockfullcode, start_date=start_date, end_date=end_date,
                frequency=freq, fields=jq_fields, fq=ed_type, panel=True)
        # 如果失败了，则将模式变成dataframe，该方式适用于混杂商品期货/股票/股指期货等交易时间不一致的品种
        except Exception as ex:
            logger.warning('传入品种交易时间不一致，改为用dataframe格式获取: 错误:{}'.format(ex))
            jq_data = get_jq_api().get_price(
                security=jq_stockfullcode, start_date=start_date, end_date=end_date,
                frequency=freq, fields=jq_fields, fq=ed_type, panel=False)
        if not jq_data.empty:
            fields_concat_list = list()
            # 判断返回的数据类型，如果是Panel, 说明stockfullcode是多个，则需要组装平面数据
            if isinstance(jq_data, pd.Panel):
                for field in jq_fields:
                    field_df = jq_data[field]
                    multi_col = pd.MultiIndex.from_product([list(field_df.columns), [field]], names=['code', 'field'])
                    field_df.columns = multi_col
                    fields_concat_list.append(field_df)
                if len(fields_concat_list) > 0 and len(fields_concat_list) == len(jq_fields):
                    concat_fields_df = pd.concat(fields_concat_list, axis=1)
                if not concat_fields_df.empty:
                    stock_array = set(
                        [stock_field_tuple[0] for stock_field_tuple in concat_fields_df.columns.to_list()])
                    for stock_key in stock_array:
                        # 获取每一个品种的数据
                        each_stock_kline_df = concat_fields_df[stock_key].copy()
                        each_stock_kline_df.loc[:, 'contract'] = stock_key
                        contract_local_code = normalize_code_to_local(stock_key)
                        zq_data = convert_httpjq2zq_format(jq_kline_df=each_stock_kline_df, freq_str=freq,
                                                           contract=contract_local_code, trading_dt=trading_dt_df)
                        zq_data.loc[:, 'symbol'] = contract_local_code
                        jq_quote_df_dict[contract_local_code] = zq_data
            # 如果不是Panel，则认为是Dataframe，直接赋值
            else:
                concat_fields_df = jq_data
                if not concat_fields_df.empty:
                    for each_stock, each_stock_kline_df in concat_fields_df.groupby(concat_fields_df.code):
                        each_stock_kline_df_copy = each_stock_kline_df.copy()
                        each_stock_kline_df_copy = each_stock_kline_df_copy.set_index('time')
                        each_stock_kline_df_copy.loc[:, 'contract'] = each_stock
                        contract_local_code = normalize_code_to_local(each_stock)
                        zq_data = convert_httpjq2zq_format(jq_kline_df=each_stock_kline_df_copy, freq_str=freq,
                                                           contract=contract_local_code, trading_dt=trading_dt_df)
                        zq_data.loc[:, 'symbol'] = contract_local_code
                        jq_quote_df_dict[contract_local_code] = zq_data
    except Exception as ex:
        msg = "调用JQData API出错，错误信息:{}".format(ex)
        print(msg)
        logger.error(msg)
    if dataframe:
        quote_df = pd.DataFrame()
        if len(jq_quote_df_dict) > 0:
            quote_df = pd.concat(jq_quote_df_dict.values())
            quote_df.reset_index(drop=True, inplace=True)
        return quote_df
    else:
        return jq_quote_df_dict


def _get_tick_data_org_from_jq(stockfullcode, start_date=None, end_date=None, tick_fields=None, skip=True,
                               dataframe=True):
    """
    从JQData SDK获取行情数据，返回数据源等同于_get_history_info_org
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param jq_fields:
    :param freq:
    :param ed_type:
    :param dataframe:
    :return:
    """
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    jq_tick_df_dict = dict()
    # 转化stockfullcode格式为jq
    jq_stockfullcode = normalize_code_to_jq(list_stockfullcode)
    # 根据stockfullcode的个数，决定处理返回的数据
    tick_fields = tick_fields if tick_fields is not None else ['time', 'current', 'high', 'low', 'volume', 'money',
                                                               'a1_v', 'a1_p', 'b1_v', 'b1_p']
    try:
        # jq的接口不支持一次获取多个标的，需要循环调用获取
        for each_stock in jq_stockfullcode:
            # 根据each_stock的quote_type，决定tick_fields
            tick_fields_copy = tick_fields.copy()
            market = _get_market(each_stock)
            security_type = _get_security_type_by_market(market)
            # 如果是获取期货的tick，域值需要在加上持仓: position
            if security_type in ['SF', 'SQ', 'ZZ', 'DL']:
                tick_fields_copy.append('position')
            jq_tick_df = get_jq_api().get_ticks(security=each_stock, start_dt=start_date, end_dt=end_date,
                                                fields=tick_fields, skip=skip)
            if not jq_tick_df.empty:
                # 赋值contract标记
                jq_tick_df['contract'] = each_stock
                jq_tick_df_dict[each_stock] = jq_tick_df
    except Exception as ex:
        logger.error(u"调用JQData API出错，错误信息:{}".format(ex))
    if dataframe:
        quote_df = pd.DataFrame()
        if len(jq_tick_df_dict) > 0:
            quote_df = pd.concat(jq_tick_df_dict.values())
            quote_df.reset_index(drop=True, inplace=True)
        return quote_df
    else:
        return jq_tick_df_dict


def get_ticks(stockfullcode, start_date=None, end_date=None, tick_fields=None, skip=True, cache_path='http',
              dataframe=True):
    """
    获取JQ格式的Tick数据
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param fields:
    :param skip: 默认为True，过滤掉无成交变化的tick数据；当指定skip=False时，返回的tick数据会保留从2019年6月25日以来无成交有盘口变化的tick数据
    :param cache_path:
    :param dataframe:
    :return:
    """
    tick_df_dict = dict()
    start_date = _to_datetime(start_date)
    end_date = _to_datetime(end_date)
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    # 判断是否是本地缓存
    from_local_cache = os.path.isdir(cache_path)
    # 如果cache_path是nas或者传入的是本地路径，则构造h5文件的读取路径
    if cache_path == 'nas' or from_local_cache:
        tick_df_dict = dict()
    # 获取JQData线上数据
    else:
        tick_df_dict = _get_tick_data_org_from_jq(stockfullcode=list_stockfullcode, start_date=start_date,
                                                  end_date=end_date, tick_fields=tick_fields, skip=skip,
                                                  dataframe=False)
    if dataframe:
        tick_df = pd.DataFrame()
        if len(tick_df_dict) > 0:
            tick_df = pd.concat(tick_df_dict.values())
            tick_df = tick_df.reset_index(drop=True)
        return tick_df
    else:
        return tick_df_dict


def _get_history_data_org(stockfullcode, start_date=None, end_date=None, fields=None, freq=DEFAULT_FREQ, ed_type=None,
                          cache_path='nas', dataframe=True):
    """
    获取ZQ_KLINE_ORG_COLUMNS格式的数据，不复权，不做Period Conversion，仅返回1d或者1m的数据
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param ed_type:
    :param cache_path:
    :param dataframe:
    :return:
    """
    # ToDo 暂时不支持获取分钟级别以下的数据
    if freq in TKLINE_FREQ_LIST:
        return return_data(return_frame=dataframe)
    quote_df_dict = dict()
    quote_df = pd.DataFrame()
    start_date = _to_datetime(start_date)
    end_date = _to_datetime(end_date)
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    # 判断是否是本地缓存
    from_local_cache = os.path.isdir(cache_path)
    # 如果cache_path是nas或者传入的是本地路径，则构造h5文件的读取路径
    if cache_path == 'nas' or from_local_cache:
        local_data_msg = u'[NAS缓存]:{}'.format(
            constant.H5_JQ_KLINE_DIR) if cache_path == 'nas' else u"本地 '{}' 缓存".format(
            cache_path)
        start_date_num = int(start_date.strftime('%Y%m%d'))
        end_date_num = int(end_date.strftime('%Y%m%d'))
        start_time_num = int(start_date.strftime('%H%M%S'))
        end_time_num = int(end_date.strftime('%H%M%S'))
        # 是否使用h5自己的过滤语句
        use_where = False
        # 如果开始结束时间差超过4年，则不启用where筛选条件
        if end_date_num - start_date_num <= 40000:
            use_where = True
        # 获得freq的秒数
        freq_seconds = convert_freq2num(freq_str=freq)
        # 根据读取path和period的级别，决定cache_path的root_dir和缓存文件的后缀名
        if cache_path == 'nas' or from_local_cache:
            if freq_seconds >= 60:
                cache_root_path = constant.H5_JQ_KLINE_DIR if not from_local_cache else cache_path
                cache_type_str = 'kline'
            elif freq < 60:
                cache_root_path = constant.H5_JQ_TKLINE_DIR if not from_local_cache else cache_path
                cache_type_str = 'tkline'
        else:
            logger.warning(u"传入的cache_path不正确: {}，请确认指定['nas', 'http', 'cache_path']三类参数中的一个".format(cache_path))
            return return_data(return_frame=dataframe)
        h5_cache_files = list()
        # 循环构造stockfullcode的h5路径
        for each_code in list_stockfullcode:
            if not from_local_cache:
                cache_file_dir = _get_cache_files_dir(cache_root_path=cache_root_path, stockfullcode=each_code,
                                                      is_local=False)
            else:
                cache_file_dir = cache_path
            cache_file_path = _get_h5_file_path(source_path=cache_file_dir, stockfullcode=each_code,
                                                h5_type=cache_type_str)
            h5_cache_files.append(cache_file_path)
        logger.info(u"\n[数据源]：{}\n[代码]:'{}'".format(local_data_msg, ','.join(list_stockfullcode)))
        for cache_file_path in h5_cache_files:
            # 通过H5路径文件名提取stockfullcode
            stock_code = _get_stock_code_from_h5_file(cache_file_path)
            market_name = _get_market(stock_code)
            # 特别处理行业指数，行业指数直接读对应的节点，不做周期转化
            # 默认获取1分钟或是1d的基础数据，其余级别通过period_conversion_resample进行转化
            read_freq = freq if market_name == 'SI' else (freq if freq == '1d' or freq == '1m' else '1m')
            # 读取相关stock code的HDF5缓存数据
            if h5.is_hdf5(cache_file_path):
                h5_file = pd.HDFStore(cache_file_path, mode='r')
                file_keys = h5_file.keys()
                if '/{}'.format(read_freq) in file_keys:
                    group_node = str(read_freq).lower()
                    # 如果start_time_num和end_time_num都是0，认为是按照tday查询
                    if start_time_num == 0 and end_time_num == 0:
                        where_clause = "today>={} & today<={}".format(start_date_num, end_date_num)
                    else:
                        where_clause = "date>={} & date<={}".format(start_date_num, end_date_num)
                    # 读取H5数据
                    try:
                        if use_where:
                            cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields,
                                                   where=where_clause)
                        else:
                            cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields)
                        if not cache_df.empty:
                            # 用dataframe作过滤
                            if start_time_num == 0 and end_time_num == 0:
                                cache_df = cache_df[
                                    (cache_df['today'] >= start_date_num) & (
                                            cache_df['today'] <= end_date_num)]
                            else:
                                cache_df = cache_df[
                                    (cache_df['date'] >= start_date_num) & (
                                            cache_df['date'] <= end_date_num) & (
                                            cache_df['sts'] >= start_time_num) & (
                                            cache_df['ets'] <= end_time_num)]
                    except Exception as ex:
                        logger.warning(u'品种[{}],级别[{}] 读取ZQ格式数据失败[{}]'.format(stock_code, read_freq, ex))
                        cache_df = pd.DataFrame()
                    if not cache_df.empty:
                        # 过滤N/A
                        cache_df.dropna(axis=0, subset=['open', 'high', 'low', 'close', 'amount', 'vol'], inplace=True)
                        if not cache_df.empty:
                            cache_df.reset_index(drop=True, inplace=True)
                            # 标记symbol和contract
                            cache_df['symbol'] = stock_code
                            if market_name in ['SH', 'SZ']:
                                cache_df['contract'] = stock_code
                                # 当频率不是1m，并且ed_type不是None的时候，需要从1d中获取复权因子，但等权数据本身并没有1d的线
                                if read_freq != '1d' and '/1d' in file_keys:
                                    cache_df = _merge_stock_daily_factor(df_minute=cache_df,
                                                                         cache_path=cache_file_path)
                        # 默认组装字典格式的数据类型
                        quote_df_dict[stock_code] = cache_df
                h5_file.close()
    # 如果是http，则从JQ获取数据
    elif cache_path == 'http':
        logger.info(u'获取JQData缓存数据')
        quote_df_dict = _get_history_data_org_from_jq(stockfullcode=stockfullcode, start_date=start_date,
                                                      end_date=end_date, freq=freq, jq_fields=fields, ed_type=ed_type,
                                                      dataframe=False)
    # 如果是rt，默认读取的是rt03上的实时数据
    elif cache_path == 'rt':
        logger.info(u'获取SQLServer实时数据库数据')
        rt_quote_df_dict = _get_history_data_org_from_rt(stockfullcode=stockfullcode, start_date=start_date,
                                                         end_date=end_date, freq=freq, fields=fields, ed_type=ed_type,
                                                         dataframe=False)
        # 分合约转化ZQ格式
        for each_stock, stock_quote_df in rt_quote_df_dict.groupby(by=['stockfullcode']):
            stock_quote_df_copy = stock_quote_df.copy()
            stock_quote_df_copy.reset_index(inplace=True, drop=True)
            stock_quote_df_copy = convert_rt2zq_format(rt_kline_df=stock_quote_df_copy, freq=freq, contract=each_stock)
            quote_df_dict[each_stock] = stock_quote_df_copy
    if dataframe:
        if len(quote_df_dict) > 0:
            quote_df = pd.concat(quote_df_dict.values())
            quote_df.reset_index(drop=True, inplace=True)
        return quote_df
    else:
        return quote_df_dict


def _merge_stock_daily_factor(df_minute, cache_path, merge_columns=['factor']):
    '''
    把日线的factor合并到分钟线中
    :param df_minute:
    :param df_daily:
    :param merge_columns:
    :return:
    '''
    df_minute_copy = df_minute.copy()
    for col in merge_columns:
        if col in df_minute_copy.columns:
            del df_minute_copy[col]
    try:
        df_daily = pd.read_hdf(path_or_buf=cache_path, key='1d', columns=['today', 'factor'])
        grp_columns = ['today']
        df_minute_copy = pd.merge(df_minute_copy, df_daily[grp_columns + merge_columns], how='left', on=grp_columns)
    except Exception as ex:
        logger.error(u"读取{}的factor出错，请检查是否有日线，Error: {}".format(cache_path, ex))
        return df_minute
    return df_minute_copy


def merge_contract_info(quote_df, type, merge_columns=['stockfullcode', 'stockcode', 'open_date', 'expire_date'],
                        cache_path='nas'):
    """
    为标的合约附上合约信息
    :param quote_df: 行情
    :param types: futures/stock
    :param merge_columns: 需要merge的columns
    :param cache_path: 本地缓存merge/在线SDKmerge
    :return:
    """
    if 'contract' not in quote_df.columns:
        logger.warning("[合约信息合成失败]: 当前数据为空或数据没有 'contract' 字段 ")
        return quote_df
    securtiy_info_df = get_security_info(types=type, cache_path=cache_path)
    if not securtiy_info_df.empty and not quote_df.empty:
        securtiy_info_df_copy = securtiy_info_df[merge_columns]
        quote_df = pd.merge(left=quote_df, right=securtiy_info_df_copy, how='left', left_on='contract',
                            right_on='stockfullcode')
    else:
        logger.warning("[合约信息合成失败]: 合约信息为空 ")
    return quote_df


def get_cached_contracts_list(cache_type=None):
    """
    根据cache_type读取所有
    :param cache_type:
    :return:
    """
    cache_contracts_list = list()
    if cache_type is None:
        list_cache_type = ['futures', 'stock', 'options']
    else:
        list_cache_type = [cache_type, ] if isinstance(cache_type, str) else cache_type
    # 合约列表的缓存文件路径
    cached_contracts_file_path = r'\\nas\share\HDF5\JQData\Common\cached_contracts.h5'
    if os.path.exists(cached_contracts_file_path):
        for cache_type in list_cache_type:
            cached_contract_df = pd.read_hdf(path_or_buf=cached_contracts_file_path, key=cache_type)
            cache_contracts_list.extend(cached_contract_df.contracts.values.tolist())
    return sorted(cache_contracts_list)


def cache_cached_contracts_list():
    """
    记录当前NAS缓存的合约列表，当get_history_data的cache_path设置为auto的时候，为本地和在线获取历史数据做分类
    :return:
    """
    # 缓存合约的存放目录
    cache_kline_path_dict = {
        'futures': {'root_dir': r'\\nas\share\HDF5\ZQ_H5\CacheHDF5\KlineH5',
                    'sub_dirs': ['commodity_future', 'index_future']},
        'stock': {'root_dir': r'\\nas\share\HDF5\JQData\KlineH5', 'sub_dirs': ['index', 'stock']},
        'options': {'root_dir': r'\\nas\share\HDF5\JQData\KlineH5\options',
                    'sub_dirs': ['SH', 'SF', 'SQ', 'ZZ', 'DL', 'SE']}
    }
    cached_contract_dict = dict()
    # 合约列表的缓存文件路径
    cached_contracts_file_path = r'\\nas\share\HDF5\JQData\Common\cached_contracts.h5'
    for cache_type, cache_dir_dict in cache_kline_path_dict.items():
        root_dir = cache_dir_dict['root_dir']
        total_contract_list = list()
        for sub_dir in cache_dir_dict['sub_dirs']:
            scan_dir = os.path.join(root_dir, sub_dir)
            contract_list = list()
            for root, dir_names, file_names in os.walk(scan_dir):
                for file_name in file_names:
                    contract_arr = file_name.split('_')
                    if len(contract_arr) > 1:
                        contract_list.append(contract_arr[0])
            total_contract_list.extend(contract_list)
        cached_contract_dict[cache_type] = total_contract_list
        logger.info(f'[缓存合约列表]: 类型:{cache_type}, 目录:{scan_dir}, 合约数量: {len(total_contract_list)}')
    # 开始缓存合约列表
    for cached_type, cached_contract_list in cached_contract_dict.items():
        contract_df = pd.DataFrame(data=cached_contract_list, columns=['contracts'])
        contract_df.to_hdf(path_or_buf=cached_contracts_file_path, key=cached_type, complevel=9, complib='blosc',
                           format='table')
        logger.info(f'[缓存成功]: 类型:{cached_type}, 路径:{cached_contracts_file_path}, 合约数量: {len(cached_contract_list)}')


def gc_cache(gc_cache):
    list_gc_cache = [gc_cache, ] if isinstance(gc_cache, str) else gc_cache
    for gc_cache in list_gc_cache:
        del gc_cache
    gc.collect()


class TestFeedIO(unittest.TestCase):
    def setUp(self):
        setup_screen_logger()
        setup_file_logger()
        self.logger = logging.getLogger('root')

    def tearDown(self):
        pass

    def test_get_cached_contracts_list(self):
        cached_contracts_list = get_cached_contracts_list()
        pass

    def test_cache_cached_contracts_list(self):
        """
        Task: 更新本地有缓存的合约列表
        :return:
        """
        cache_cached_contracts_list()
        pass

    def test_get_history_data_auto(self):
        """
        当cache_path = 'auto' 的时候，自动判断应该读取本地缓存还是http缓存
        :return:
        """
        list_stockfullcode = ['SFIF9001', 'SH510050', 'SH000300', 'SH600000']
        start_date = '2020-01-01'
        end_date = '2022-12-31'
        cache_path = r'D:\workspace\python_workspace\svn03\rio3\scripts\cache_data\zq'
        #cache_path = 'auto'
        data_df = get_history_data(stockfullcode=list_stockfullcode, start_date=start_date, end_date=end_date,
                                   freq='5m', cache_path=cache_path, dataframe=True)
        pass

    def test_get_rt_history_data(self):
        stockfullcode = ['SH000300']
        freq = '30m'
        start_dt = '2020-08-10'
        end_dt = '2020-12-01'
        cache_path = 'rt'
        data = get_history_data(stockfullcode=stockfullcode, freq=freq, start_date=start_dt, end_date=end_dt,
                                cache_path=cache_path)
        pass

    def test_get_redis_history_data(self):
        stockfullcode = ['DLjm9001']
        freq = '1h'
        start_dt = '2020-2-10'
        end_dt = '2021-12-01'
        # cache_path = 'redis_py'
        cache_path = 'redis://:redisxy123456R88@publicpystrategy.redis.rds.aliyuncs.com:6401/4'
        # cache_path = 'redis://:redisxy123456R88@r-uf6fb5d61707be84pd.redis.rds.aliyuncs.com:6401/0'
        ed_type = 'post'
        data1 = get_history_data(stockfullcode=stockfullcode, freq=freq, start_date=start_dt, end_date=end_dt,
                                 dataframe=False, cache_path=cache_path, ed_type=None)
        data2 = get_history_data(stockfullcode=stockfullcode, freq=freq, start_date=start_dt, end_date=end_dt,
                                 dataframe=False, cache_path=cache_path, ed_type=ed_type)
        pass

    def test_merge_contract_info(self):
        # stockfullcode = ['SFIF9001', 'SQrb9001', 'DLjd9001', 'ZZMA9001']
        stockfullcode = ['SH600000', 'SZ000001']
        start = '2019-04-21'
        end = '2020-12-10'
        freq = '1m'
        quote_df = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                    dataframe=True, ed_type='post')
        quote_contract_info_df = merge_contract_info(quote_df=quote_df, type='stock')
        pass

    def test_get_etf_options_contract_data(self):
        start = '2019-04-21'
        end = '2021-12-10'
        stockfullcode = ['SH510050']
        freq = '15m'
        # local_path = r'D:\TempCSV\options_data'
        data_local = get_options_contract_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq)
        data_local = get_options_contract_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq)

        pass

    def test_get_index_options_contract_data(self):
        start = '2019-04-01'
        end = '2020-12-31'
        stockfullcode = 'SFIO2106-P-5400'
        freq = '5m'
        # 需要测试NAS路径
        local_path = r'D:\TempCSV\HDF5\JQData\KlineH5\options\SF'
        data_local = get_options_contract_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                               cache_path=local_path)
        pass

    def test_get_legal_future_history_data(self):
        start = '2010-01-01'
        end = '2020-12-30'
        stockfullcode = 'SQrb9001'
        freq = '30m'
        data_nas = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                    ed_type=None)
        legal_data_not_call_auction = get_legal_period_data(data_df=data_nas, stockfullcode=stockfullcode,
                                                            is_call_auction=False)
        legal_data_call_auction = get_legal_period_data(data_df=data_nas, stockfullcode=stockfullcode,
                                                        is_call_auction=True)

        pass

    def test_calculate_ed_info(self):
        start = '2010-01-01'
        end = '2020-12-30'
        stockfullcode = 'SQrb9001'
        freq = '30m'
        data_nas = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                    ed_type=None)
        ed_info = get_ed_info_df(stockfullcode=stockfullcode, start=constant.DATETIME_MIN, end=constant.DATETIME_MAX)
        feed_rio.calculate_ed_info(data_df=data_nas, ed_info_df=ed_info, ed_type='post')
        pass

    def test_get_sw_history_data(self):
        start = '2015-01-01'
        end = '2021-12-28'
        # stockfullcode = ['SH600887', 'SH600703', 'SH601318', 'SZ000423', 'SZ002304', 'SZ002007', 'SZ000568']
        stockfullcode = ['SI801010']
        freq = '1m'
        # local_dir = r'D:\TempCSV\stock_data\SI_1m'
        local_dir = 'nas'
        data_local = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                      ed_type='post', cache_path=local_dir)
        data_local_local_post = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                                 ed_type='pre', dataframe=True)
        # data_local = period_conversion_resample(data_local, '1d')
        # data_local['close'].plot()
        # plt.show()
        data_local_local_post = period_conversion_resample(data_local_local_post, '1d')
        data_local_local_pre = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                                ed_type='pre')

        data_http_3m = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='3m',
                                        ed_type='post', cache_path='http')
        pass

    def test_get_stock_history_data(self):
        start = '2015-01-01'
        end = '2021-10-19'
        # stockfullcode = ['SH600887', 'SH600703', 'SH601318', 'SZ000423', 'SZ002304', 'SZ002007', 'SZ000568']
        stockfullcode = ['SH600000', 'SZ399905', 'SFIF9001', 'SQrb9001']
        stockfullcode = ['SH000905']
        freq = '1d'
        # local_dir = r'\\nas\share\HDF5\JQData\KlineH5\stock\SI_1m'
        local_dir = 'nas'
        data_local = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                      ed_type=None, cache_path='http')
        data_local_local_post = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                                 ed_type='pre', dataframe=True)
        # data_local = period_conversion_resample(data_local, '1d')
        # data_local['close'].plot()
        # plt.show()
        data_local_local_post = period_conversion_resample(data_local_local_post, '1d')
        data_local_local_pre = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                                ed_type='pre')

        data_http_3m = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='3m',
                                        ed_type='post', cache_path='http')
        pass

    def test_get_etf_history_data(self):
        start = '2020-01-21'
        end = '2020-05-22 20:00:00'
        stockfullcode = ['SH510300']
        freq = '1m'
        data_local_local_post = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                                 ed_type='pre', dataframe=True, cache_path='http')
        pass

    def test_get_diff_period_history_data(self):
        start = '2020-04-21'
        end = '2020-06-15'
        stockfullcode = ['SQrb9001']
        freqs = ['5m']
        result_dict = dict()
        data = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='5m', ed_type='post')
        # result_dict[freq] = data
        pass

    def test_get_index_history_data(self):
        start = '2019-05-06'
        end = '2019-05-10'
        stockfullcode = ['SH000300']
        # stockfullcode = DEFAULT_INDEX_STOCK
        freq = '5m'
        data_post = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                     ed_type='post')
        pass

    def test_get_ed_stock_history_data(self):
        start = '2018-01-01'
        end = '2020-02-05'
        stockfullcode = ['SH600887', 'SH600703', 'SH601318', 'SH999999']
        freq = '1m'
        data_pre = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                    ed_type='pre')
        data_post = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                     ed_type='post')
        pass

    def test_get_quote_trade_time(self):
        start = '2010-01-01'
        end = '2020-12-31'
        stockfullcode = ['SQrb9001']
        freq = '5m'
        data = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq, ed_type='pre',
                                dataframe=True)
        # data = get_trading_border_time(quote_df=data)
        data = get_trading_time_border_diff(quote_df=data)
        pass

    def test_get_future_contract_history_data(self):
        contract_list = ['M2001.XDCE', 'M2003.XDCE', 'M2005.XDCE', 'M2007.XDCE', 'M2009.XDCE', 'M2011.XDCE',
                         'I2002.XDCE',
                         'I2003.XDCE', 'I2005.XDCE', 'I2007.XDCE', 'I2009.XDCE', 'I2010.XDCE', 'I2101.XDCE',
                         'M2101.XDCE',
                         'I2102.XDCE', 'M2103.XDCE', 'I2105.XDCE', 'M2105.XDCE']
        start = '2019-11-01'
        end = '2020-12-12'
        fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'open_interest']
        # fields = None
        contract_quote_df = get_history_data(stockfullcode=contract_list, freq='5m', start_date=start, end_date=end,
                                             fields=fields, cache_path='http')
        pass

    def test_get_pca_stock_history_data(self):
        start = '2019-10-01'
        end = '2020-12-31'
        list_stockfullcode = ['pca001']
        freq = '1m'
        # cache_path = 'nas'
        cache_path = r'\\nas\share\HDF5\ExternalData\zyw_data\pca'
        for stockfullcode in list_stockfullcode:
            data_pre = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                        dataframe=True, cache_path=cache_path)
            pass

    def test_get_future_history_data(self):
        start = '2011-08-01'
        end = '2021-08-31'
        list_stockfullcode = ['SFIF9001']
        # list_stockfullcode = ['SFIF9001', 'SFIC9001', 'SFIH9001', 'DLi9001', 'DLj9001', 'SQrb9001', 'SQru9001',
        #                       'ZZAP9001', 'SQni9001', 'DLp9001', 'SQfu9001', 'ZZSR9001', 'ZZTA9001', 'ZZMA9001']
        freq = '1m'
        # cache_path = r'X:\HDF5\ZQ_H5\CacheHDF5\KlineH5\backup\20210317\index_future\SF'
        cache_path = 'nas'
        for stockfullcode in list_stockfullcode:
            data_pre = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                        ed_type=None, dataframe=True, cache_path=cache_path)
            print(len(data_pre))
            # data_pre.to_hdf(r'D:\TempCSV\future_data\new_ed_klines\{}__5m_post_kline.h5'.format(stockfullcode),
            #                 key='data')
        data_post = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                     ed_type='post')
        pass

    def test_compare_history_data_columns(self):
        start = '2019-10-01'
        end = '2020-12-31'
        list_stockfullcode = ['DLj9001']
        # list_stockfullcode = ['SFIF9001', 'SFIC9001', 'SFIH9001', 'DLi9001', 'DLj9001', 'SQrb9001', 'SQru9001',
        #                       'ZZAP9001', 'SQni9001', 'DLp9001', 'SQfu9001', 'ZZSR9001', 'ZZTA9001', 'ZZMA9001']
        freq = '15m'
        # cache_path = r'D:\TempCSV\future_data\secondary_contract\SQ'
        cache_path = 'nas'
        for stockfullcode in list_stockfullcode:
            data_pre = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq=freq,
                                        ed_type='post', dataframe=True, cache_path=cache_path)

        pass

    def test_compare_freq_history_data(self):
        start = '2019-10-01'
        end = '2020-12-31'
        list_stockfullcode = ['DLj9001']
        freq = '1m'
        cache_path = 'nas'
        for stockfullcode in list_stockfullcode:
            data_df = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='1m',
                                       ed_type='post', dataframe=True, cache_path=cache_path)
            year_data_df = period_conversion_resample(quote_df=data_df, freq_str='3d')
            day_data_df = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='1d',
                                           ed_type='post', dataframe=True, cache_path=cache_path)
            week_data_df = period_conversion_resample(quote_df=day_data_df, freq_str='1w')
            pass

    def test_compare_return_value(self):
        org_csv_df = pd.read_csv(r'D:\TempCSV\future_data\org.csv')
        org_csv_df['pre_close'] = org_csv_df['close'].shift(1)
        org_csv_df.dropna(axis=0, how='any', inplace=True)
        org_csv_df['returns'] = org_csv_df['close'] / org_csv_df['pre_close'] - 1
        org_csv_df['DateTrade'] = pd.to_datetime(org_csv_df['date'] * 1000000 + org_csv_df['ets'],
                                                 format='%Y%m%d%H%M%S')
        org_csv_df = org_csv_df[['DateTrade', 'close', 'returns']]

        old_csv_df = pd.read_csv(r'D:\TempCSV\future_data\old.csv')
        old_csv_df['pre_close'] = old_csv_df['close'].shift(1)
        old_csv_df.dropna(axis=0, how='any', inplace=True)
        old_csv_df['returns'] = old_csv_df['close'] / old_csv_df['pre_close'] - 1
        old_csv_df['DateTrade'] = pd.to_datetime(old_csv_df['date'] * 1000000 + old_csv_df['ets'],
                                                 format='%Y%m%d%H%M%S')
        old_csv_df = old_csv_df[['DateTrade', 'close', 'returns']]
        new_csv_df = pd.read_csv(r'D:\TempCSV\future_data\new_kline_correct.csv')
        new_csv_df['pre_close'] = new_csv_df['close'].shift(1)
        new_csv_df.dropna(axis=0, how='any', inplace=True)
        new_csv_df['returns'] = new_csv_df['close'] / new_csv_df['pre_close'] - 1
        new_csv_df['DateTrade'] = pd.to_datetime(new_csv_df['date'] * 1000000 + new_csv_df['ets'],
                                                 format='%Y%m%d%H%M%S')
        new_csv_df = new_csv_df[['DateTrade', 'close', 'returns']]
        merge_result1 = pd.merge(left=old_csv_df, right=org_csv_df, how='left', left_on='DateTrade',
                                 right_on='DateTrade')
        merge_result2 = pd.merge(left=new_csv_df, right=org_csv_df, how='left', left_on='DateTrade',
                                 right_on='DateTrade')
        merge_result3 = pd.merge(left=old_csv_df, right=new_csv_df, how='left', left_on='DateTrade',
                                 right_on='DateTrade')
        # merge_result1.to_csv(r'D:\TempCSV\future_data\result1.csv')
        # merge_result2.to_csv(r'D:\TempCSV\future_data\result2.csv')
        merge_result3.to_csv(r'D:\TempCSV\future_data\result3.csv')
        pass

    def test_get_ed_future_history_data(self):
        pass

    def test_get_http_history_data(self):
        start = '2014-01-01'
        end = '2021-12-02'
        stockfullcode = 'OI1301.XZCE'
        df = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='30m',
                              ed_type=None, cache_path='http')
        df_jq = get_jq_api().get_price(security='RB9999.XSGE', start_date=start, end_date=end, frequency='3m')
        logout_jq_api()
        pass

    def test_get_all_stocks_klines(self):
        start = '2016-01-01'
        end = '2020-12-31'
        jq_security_codes = get_jq_api().get_all_securities(types=['stock']).index.tolist()
        local_security_codes = normalize_code_to_local(jq_security_codes)
        get_history_data(stockfullcode=local_security_codes, start_date=start, end_date=end, freq='30m',
                         ed_type='pre', cache_path='nas')
        pass

    def test_get_local_cache_history_data(self):
        start = '2010-01-01'
        end = '2020-12-31'
        stockfullcode = ['SH600000', 'SZ000001']
        cache_path = r'D:\TempCSV'

        df1 = get_history_data(stockfullcode='SFIF9001', start_date=start, end_date=end, freq='1m',
                               ed_type='post', cache_path=cache_path)
        df2 = get_history_data(stockfullcode='SH601320', start_date=start, end_date=end, freq='3m',
                               ed_type=None, cache_path=cache_path)
        pass

    def test__get_trading_date_range(self):
        start = '2019-01-01'
        end = '2020-12-31'
        tday_trading_datetime_data = _get_trading_datetime_range(start_date=start, end_date=end)
        pass

    def test_period_conversion(self):
        start = '2015-01-01'
        end = '2018-01-02'
        stockfullcode = ['SH600000']
        df_1m = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='1m',
                                 ed_type=None)
        self.assertEqual(240, len(df_1m))
        df_5m = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='5m',
                                 ed_type=None)
        self.assertEqual(48, len(df_5m))
        df_15m = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='15m',
                                  ed_type=None)
        self.assertEqual(16, len(df_15m))
        df_30m = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='30m',
                                  ed_type=None)
        self.assertEqual(8, len(df_30m))
        df_60m = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='60m',
                                  ed_type=None)
        self.assertEqual(4, len(df_60m))
        df_1d = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='1d',
                                 ed_type=None)
        self.assertEqual(1, len(df_1d))
        df_10m = get_history_data(stockfullcode=stockfullcode, start_date=start, end_date=end, freq='10m',
                                  ed_type=None)
        self.assertEqual(24, len(df_10m))

    def test_get_oi_ed_info_data(self):
        stockfullcode = 'SFIF9001'
        data = get_oi_ed_info_data(stockfullcode=stockfullcode)
        pass

    def test_compare_ws_and_jq_future_data(self):
        ws_data = pd.read_hdf(r'D:\TempCSV\kline_compare\jq_SFIF9001_1m_post_old2.h5')
        jq_data = pd.read_hdf(r'D:\TempCSV\kline_compare\jq_SFIF9001_1m_post_new.h5')
        ws_data = ws_data[['date', 'today', 'open', 'close', 'high', 'low', 'vol', 'amount', 'sts', 'ets', 'contract']]
        jq_data = jq_data[['date', 'today', 'open', 'close', 'high', 'low', 'vol', 'amount', 'sts', 'ets', 'contract']]
        ws_data['date'] = ws_data['date'].astype(np.int64)
        ws_data['ets'] = ws_data['ets'].astype(np.int64)
        ws_data['DateTrade'] = pd.to_datetime(ws_data['date'] * 1000000 + ws_data['ets'],
                                              format='%Y%m%d%H%M%S')
        jq_data['date'] = jq_data['date'].astype(np.int64)
        jq_data['ets'] = jq_data['ets'].astype(np.int64)
        jq_data['DateTrade'] = pd.to_datetime(jq_data['date'] * 1000000 + jq_data['ets'],
                                              format='%Y%m%d%H%M%S')
        compare_result = pd.merge(left=ws_data, right=jq_data, left_on='DateTrade', right_on='DateTrade')
        pass

    def test_get_jq_ticks_data(self):
        stockfullcode_stock = 'SH600000'
        stockfullcode_future = 'SFIC2103'
        start_dt = '2021-01-05 09:00:00'
        end_dt = '2021-01-05 16:00:00'
        stockfullcode = ['I9999.XDCE']
        stock_ticks_data = get_ticks(stockfullcode=stockfullcode, start_date=start_dt, end_date=end_dt,
                                     cache_path='http')
        future_ticks_data = get_ticks(stockfullcode=stockfullcode_future, start_date=start_dt, end_date=end_dt,
                                      cache_path='http')
        pass

    def test_get_etf_options_underlying_core(self):
        start_date = '2019-11-01'
        end = '2020-05-15'
        underlying_name = '300ETF'
        freq = '15m'
        cache_file_path = r'D:\TempCSV\options_data\50ETF_contracts.h5'
        result = _get_etf_options_underlying_core(start_date=start_date, end_date=end, underlying_name=underlying_name,
                                                  freq=freq, cache_file_path=cache_file_path)
        pass

    def test_get_future_options_underlying_core(self):
        """
        Task:Option Step3.2: 将股指期权合约行情整合成面板数据
        :return:
        """
        start_date = '2019-11-01 09:00:00'
        end = '2020-11-30 16:00:00'
        list_underlying_name = ['FUTURE']
        list_freqs = ['5m']
        # cache_file_path = None
        cached_contract_path = r'D:\TempCSV\options_data\dlm_dli_cache.h5'

        # cache_file_path = r'\\nas\share\HDF5\JQData\KlineH5\options\future_option_cache.h5'
        # cache_file_path = None
        for underlying_name in list_underlying_name:
            for freq in list_freqs:
                result = _get_future_options_underlying_core(start_date=start_date, end_date=end,
                                                             underlying_name=underlying_name,
                                                             freq=freq, cache_file_path=cached_contract_path)
                data_columns = ['date', 'today', 'sts', 'ets']
                result.to_hdf(r'\\nas\share\HDF5\JQData\KlineH5\options\panel\{}_panel.h5'.format(underlying_name),
                              key=freq,
                              complevel=9, complib='blosc')
        pass

    def test_export_etf_options_etf_options_underlying_panel(self):
        """
        Task:Option Step3.1: 将ETF期权合约行情整合成面板数据,按年月节点存储
        :return:
        """
        start_date = '2017-01-01'
        end = '2020-12-31'

        # list_underlying_name = ['300ETF', '50ETF']
        # list_freqs = ['5m', '15m']
        list_underlying_name = ['300ETF']
        list_freqs = ['1m', '5m', '15m', '1d']
        cache_file_path = None
        store_panel = True
        # cache_file_path = r'\\nas\share\HDF5\JQData\KlineH5\options\300ETF_contracts.h5'
        for underlying_name in list_underlying_name:
            for freq in list_freqs:
                result = _get_etf_options_underlying_core(start_date=start_date, end_date=end,
                                                          underlying_name=underlying_name,
                                                          freq=freq, cache_file_path=cache_file_path)
                # 按全数据面板存:
                full_panel_export_path = r'\\nas\share\HDF5\JQData\KlineH5\options\panel\{}_panel.h5'.format(
                    underlying_name)
                result.to_hdf(full_panel_export_path, key=freq, complevel=9, complib='blosc')

                # 按月节点面板数据寸:
                node_panel_export_path = r'\\nas\share\HDF5\JQData\KlineH5\options\panel\{}_panel_node.h5'.format(
                    underlying_name)
                _cache_panel_data(panel_df=result, freq=freq, export_path=node_panel_export_path)
        pass

    def test_export_index_future_options_underlying_panel(self):
        """
        Task:Option Step3.2: 将股指期权合约行情整合成面板数据
        FUTURE/INDEX注意标注不必要的代码，cache_path或者filter
        :return:
        """
        start_date = '2019-10-01'
        end = '2021-12-31'
        list_underlying_name = ['INDEX']
        list_freqs = ['5m', '15m']
        cached_contract_path = None
        # cache_file_path = r'D:\TempCSV\options_data\index_option_contract.h5'
        for underlying_name in list_underlying_name:
            for freq in list_freqs:
                result = _get_future_options_underlying_core(start_date=start_date, end_date=end,
                                                             underlying_name=underlying_name,
                                                             freq=freq, cached_contract_path=cached_contract_path)
                # 按全数据面板存:
                full_panel_export_path = r'\\nas\share\HDF5\JQData\KlineH5\options\panel\{}_panel.h5'.format(
                    underlying_name)
                result.to_hdf(full_panel_export_path, key=freq, complevel=9, complib='blosc')

                # 按月节点面板数据寸:
                node_panel_export_path = r'\\nas\share\HDF5\JQData\KlineH5\options\panel\{}_panel_node.h5'.format(
                    underlying_name)
                _cache_panel_data(panel_df=result, freq=freq, export_path=node_panel_export_path)
        pass

    def test_export_commodity_future_options_underlying_panel(self):
        """
        Task:Option Step3.3: 将商品期权合约行情整合成面板数据
        FUTURE/INDEX注意标注不必要的代码，cache_path或者filter
        :return:
        """
        start_date = '2019-10-01'
        end = '2020-12-31'
        list_underlying_name = ['FUTURE']
        list_freqs = ['5m']
        cache_file_path = None
        cached_contract_path = r'D:\TempCSV\options_data\dlm_dli_cache.h5'
        cached_underlying_path = r'D:\TempCSV\options_data\dlm_dli_underlying_cache.h5'
        for underlying_name in list_underlying_name:
            for freq in list_freqs:
                result = _get_future_options_underlying_core(start_date=start_date, end_date=end,
                                                             underlying_name=underlying_name,
                                                             freq=freq, cached_contract_path=cached_contract_path,
                                                             cached_underlying_path=cached_underlying_path)
                data_columns = ['date', 'today', 'sts', 'ets']
                result.to_hdf(r'\\nas\share\HDF5\JQData\KlineH5\options\panel\{}_panel.h5'.format(underlying_name),
                              key=freq,
                              complevel=9, complib='blosc')
        pass

    def test_export_etf_options_underlying_panel(self):
        start_date = '2017-01-01'
        end = '2020-12-31'
        list_underlying_name = ['50ETF', '300ETF']
        # list_freqs = ['5m', '15m']
        list_freqs = ['1m', '5m', '15m']
        # cache_file_path = r'D:\TempCSV\options_data\50ETF_contracts.h5'
        cache_file_path = None
        for underlying_name in list_underlying_name:
            for freq in list_freqs:
                result = _get_etf_options_underlying_core(start_date=start_date, end_date=end,
                                                          underlying_name=underlying_name,
                                                          freq=freq, cache_file_path=cache_file_path)
                data_columns = ['date', 'today', 'sts', 'ets']
                # result.to_hdf(
                #     r'\\nas\share\HDF5\JQData\KlineH5\options\panel\{}_panel_table.h5'.format(underlying_name),
                #     key=freq,
                #     complevel=9,
                #     complib='blosc',
                #     format='table', data_columns=data_columns)
                result.to_hdf(r'\\nas\share\HDF5\JQData\KlineH5\options\panel\{}_panel.h5'.format(underlying_name),
                              key=freq,
                              complevel=9, complib='blosc')
        pass

    def test_get_options_underlying_data_freq(self):
        """
        测试获取面板数据
        :return:
        """
        start_date = '2017-01-02'
        end = '2021-12-31'
        # INDEX/FUTURE/50ETF/300ETF
        underlying_name = 'FUTURE'
        # underlying_name = '50ETF'
        freq = '5m'
        # local_path = r'D:\TempCSV\options_data\local_panel'
        local_path = 'nas'
        local_lzy_cache_path = r'D:\TempCSV\options_data\greeks\option_panel_50ETF_5m_20170102_20200602.h5'
        # nas_result = get_options_underlying_data(start_date=start_date, end_date=end, underlying_name=underlying_name,
        #                                          freq=freq)
        local_result = get_options_underlying_data(start_date=start_date, end_date=end, underlying_name=underlying_name,
                                                   freq=freq, cache_path=local_path)
        local_result['date_trade'] = pd.to_datetime(local_result['date'] * 1000000 + local_result['ets'],
                                                    format='%Y%m%d%H%M%S')
        pass

    def test_get_options_underlying_data_daily(self):
        """
        测试获取面板数据
        :return:
        """
        start_date = '2017-01-02'
        end = '2021-12-31'
        # INDEX/FUTURE/50ETF/300ETF
        underlying_name = 'INDEX'
        # underlying_name = '50ETF'
        freq = '1d'
        # local_path = r'D:\TempCSV\options_data'
        local_path = 'nas'
        local_lzy_cache_path = r'D:\TempCSV\options_data\greeks\option_panel_50ETF_5m_20170102_20200602.h5'
        # nas_result = get_options_underlying_data(start_date=start_date, end_date=end, underlying_name=underlying_name,
        #                                          freq=freq)
        local_result = get_options_underlying_data(start_date=start_date, end_date=end, underlying_name=underlying_name,
                                                   freq=freq, cache_path=local_path)
        local_result['date_trade'] = pd.to_datetime(local_result['date'] * 1000000 + local_result['ets'],
                                                    format='%Y%m%d%H%M%S')
        pass

    def test_get_options_underlying_data_by_diff_greeks(self):
        start_date = '2020-01-01'
        end = '2020-06-30'
        underlying_name = '50ETF'
        freq = '15m'
        cpp_local_path = r'D:\TempCSV\options_data\greeks\cpp'
        quantlib_local_path = r'D:\TempCSV\options_data\greeks\quantlib'
        cpp_result = get_options_underlying_data(start_date=start_date, end_date=end, underlying_name=underlying_name,
                                                 freq=freq, cache_path=cpp_local_path)
        quantlib_result = get_options_underlying_data(start_date=start_date, end_date=end,
                                                      underlying_name=underlying_name,
                                                      freq=freq, cache_path=quantlib_local_path)
        cpp_iv_result = cpp_result.groupby(cpp_result.today).agg({'iv': 'mean'}).plot()
        plt.show()
        quantlib_iv_result = quantlib_result.groupby(quantlib_result.today).agg({'iv': 'mean'}).plot()
        plt.show()
        pass

    def test_get_call_put_month_data(self):
        """
        Task:Option Step4: 计算期权面板数据的Iv Index数据
        :return:
        """
        start_date = '2017-01-01'
        end = '2020-12-31'
        list_underlying_name = ['50ETF', '300ETF']
        # list_underlying_name = ['50ETF']
        list_freqs = ['15m']
        # list_freqs = ['5m']
        # cache_path = r'D:\TempCSV\options_data\greeks'
        cache_path = 'nas'
        for underlying_name in list_underlying_name:
            for freq in list_freqs:
                options_panel = get_options_underlying_data(start_date=start_date, end_date=end,
                                                            underlying_name=underlying_name,
                                                            freq=freq, cache_path=cache_path)
                data_columns = ['date', 'today', 'sts', 'ets']
                call_put_month_data = get_call_put_by_month_df(panel_quote_df=options_panel)
                call_put_month_data.to_hdf(
                    r'\\nas\share\HDF5\JQData\KlineH5\options\index\{}_index.h5'.format(underlying_name), key=freq,
                    complevel=9,
                    complib='blosc',
                    format='table', data_columns=data_columns)
            pass

    def test_get_options_index_data(self):
        start_date = '2017-01-01'
        end = '2020-12-16'
        underlying_name = '300ETF'
        freq = '15m'
        local_path = r'D:\TempCSV\options_data'
        nas_options_index = get_options_index_data(start_date=start_date, end_date=end,
                                                   underlying_name=underlying_name,
                                                   freq=freq)
        local_options_index = get_options_index_data(start_date=start_date, end_date=end,
                                                     underlying_name=underlying_name,
                                                     freq=freq, cache_path=local_path)

        pass

    def test_get_options_common_df(self):
        cache_path = 'http'
        underlying_name = 'FUTURE'
        info_type = ['contracts']
        is_trading = True
        return_cache_version = False
        local_result = get_options_common_info(underlying_name=underlying_name, cache_path='nas', is_trading=is_trading,
                                               info_type=info_type)
        http_result = get_options_common_info(underlying_name=underlying_name, cache_path=cache_path,
                                              info_type=info_type, is_trading=is_trading,
                                              return_cache_version=return_cache_version)
        pass

    def test_get_options_trading_info(self):
        start_date = '2010-01-01'
        end_date = '2021-07-30'
        underlying_name = '300ETF'
        cache_path = r'D:\TempCSV\options_data'
        # df_nas = get_options_trading_info(start_date=start_date, end_date=end_date, underlying_name=underlying_name)
        df_local = get_options_trading_info(start_date=start_date, end_date=end_date, underlying_name=underlying_name,
                                            cache_path=cache_path)
        pass

    def test_get_cached_h5_nodes(self):
        start = '2019-02-28'
        end = '2020-12-31'
        freq = '15m'
        nodes = _get_cached_h5_nodes(start_date=start, end_date=end, freq=freq)
        pass

    def test_get_options_underlying_node_data(self):
        start = '2021-02-08'
        end = '2021-04-01'
        freq = '1d'
        underlying_name = '50ETF'
        full_panel_path = r'D:\TempCSV\options_data\node_store'
        node_panel_path = r'\\nas\share\HDF5\JQData\KlineH5\options\panel'
        # cache_path = 'nas'
        node_panel_df = get_options_underlying_node_data(underlying_name=underlying_name, start_date=start,
                                                         end_date=end,
                                                         freq=freq, cache_path=node_panel_path)
        full_panel_df = get_options_underlying_data(underlying_name=underlying_name, start_date=start, end_date=end,
                                                    freq=freq, cache_path=full_panel_path)
        print('node_cache:{}, full_panel:{}'.format(len(node_panel_df), len(full_panel_df)))
        self.assertEqual(len(node_panel_df), len(full_panel_df))

        pass

    def test_period_conversion_options_panel(self):
        start_date = '2017-01-01'
        end = '2021-12-31'
        # INDEX/FUTURE/50ETF/300ETF
        # underlying_name = 'INDEX'
        underlying_name = '300ETF'
        freq = '15m'
        local_path = r'D:\TempCSV\options_data'
        # local_path = 'nas'
        local_result = get_options_underlying_data(start_date=start_date, end_date=end, underlying_name=underlying_name,
                                                   freq=freq, cache_path=local_path)
        period_local_result = period_conversion_options_panel(quote_df=local_result, freq_str='60m')
        print(
            'org_freq_quote_num: {}, period_convert_quote_num: {}'.format(len(local_result), len(period_local_result)))
        period_local_result.to_hdf(r'{}\{}_{}_panel.h5'.format(local_path, underlying_name, '60m'), key='60m')

    def test_get_realtime_data(self):
        stockfullcode = ['SH000300', 'SH000905', 'SFIC2203', 'SFIC2209', 'SFIF2203', 'SFIF2209']
        freq = '1m'
        start_dt = '2022-02-22'
        end_dt = '2022-02-23'
        is_latest = True
        dataframe = True
        rt_quote_df = get_realtime_data(stockfullcode=stockfullcode, freq=freq, start_dt=start_dt, end_dt=end_dt,
                                        is_latest=is_latest, dataframe=dataframe)
        stock_rt_quote_df = get_realtime_data(stockfullcode=['SH000300', 'SH000905'], freq=freq, start_dt=start_dt,
                                              end_dt=end_dt,
                                              is_latest=is_latest, dataframe=dataframe)
        rt_quote_df = period_conversion_resample(quote_df=rt_quote_df, freq_str='30m')
        pass

    def test_get_realtime_data_latest(self):
        stockfullcode = ['SFIC2105', 'SQrb9001', 'SFIF2105', 'SFIF2109', 'SH000300', 'SH000905']
        # stockfullcode = ['SFIC2105', 'SFIC2112', 'SFIF2105', 'SFIF2112']
        rt_quote_df = get_realtime_data(stockfullcode=stockfullcode, freq='1m', is_latest=False, start_dt='2022-03-01', end_dt='2099-01-01')
        pass

    def test_get_realtime_data_latest(self):
        stockfullcode = ['SH510500', 'SZ159922', 'SH510580']
        # stockfullcode = ['SFIC2105', 'SFIC2112', 'SFIF2105', 'SFIF2112']
        rt_quote_df = get_realtime_data(stockfullcode=stockfullcode, freq='1m', is_latest=True)
        pass

    def test_get_history_data_rt(self):
        stockfullcode = ['SH000300', 'SH000905']
        freq = '1m'
        start_dt = '2021-03-18'
        end_dt = '2021-03-23'
        dataframe = True
        rt_quote_df = get_history_data(stockfullcode=stockfullcode, freq=freq, start_date=start_dt, end_date=end_dt,
                                       dataframe=dataframe, cache_path='rt')
        pass


if __name__ == "__main__":
    unittest.main()
