# -*- coding:utf-8 -*-
import datetime as dt
import logging
import os
import json
import copy
import sys
import unittest
import pandas as pd
import numpy as np
import h5py as h5
import re
import matplotlib.pyplot as plt
from rio3.base import regex
from rio3.datafeed.common import FREQ_REGEX
from rio3.quote_core.quote_info import BasicQuoteInfo
from collections import defaultdict
from rio3.quote_core.quote_info import is_future
from rio3.base import constant, enumerator
from rio3.log.setup import setup_screen_logger, setup_file_logger
from rio3.datafeed.dbIO import get_ed_info_df, check_ed_info_df, get_tday_list, get_previous_tday, get_next_tday, \
    get_ed_info_org
from rio3.datafeed.common import _to_datetime, convert_freq2num, _get_cache_files_dir, _get_h5_file_path, \
    _get_stock_code_from_h5_file, filter_quote_df, _fill_mid_and_midv, _calculate_zq_kdix, timer, get_resample_interval, \
    return_data
from rio3.datafeed.redis_feed import get_redis_history_data_org
from rio3.quote_core.quote_filter import get_legal_period_data, convert_dt_to_tday
from rio3.utils.date import from_oadate

logger = logging.getLogger('root')

# 震球的Kline基础数据列，严格按照顺序存储
ZQ_KLINE_COLUMNS = ['date', 'today', 'span', 'open', 'close', 'high', 'low', 'mid', 'midv',
                    'vol', 'amount', 'sts', 'ets', 'oi', 'contract', 'contract_idx']

DEFAULT_FREQ = '5m'

PERIOD2TEXT_DICT = {
    'tick': 'tick',
    '5s': 'fiveseconds',
    '15s': 'fifteenseconds',
    '30s': 'thirtyseconds',
    '1m': 'oneminute',
    '5m': 'fiveminutes',
    '15m': 'fifteenminutes',
    '30m': 'halfhour',
    '60m': 'onehour',
    '1d': 'oneday'
}


# 分钟线之间的转化
def period_conversion(kline_df, freq):
    '''
    :param kline_df
    :param interval_sec:
    :return:
    '''
    # 获取resample时间线
    resample_interval_type = get_resample_interval(freq)
    dict_how = {
        'Tday': constant.MERGE_KLINE_DICT['Tday'],
        'Volume': constant.MERGE_KLINE_DICT['Volume'],
        'Open': constant.MERGE_KLINE_DICT['Open'],
        'High': constant.MERGE_KLINE_DICT['High'],
        'Low': constant.MERGE_KLINE_DICT['Low'],
        'Close': constant.MERGE_KLINE_DICT['Close'],
        'Wp': constant.MERGE_KLINE_DICT['Wp'],
        'Np': constant.MERGE_KLINE_DICT['Np'],
        'Preclose': constant.MERGE_KLINE_DICT['Preclose'],
        'Dailypreclose': constant.MERGE_KLINE_DICT['Dailypreclose'],
        'Sp1': constant.MERGE_KLINE_DICT['SP1'],
        'Sv1': constant.MERGE_KLINE_DICT['SV1'],
        'Bp1': constant.MERGE_KLINE_DICT['BP1'],
        'Bv1': constant.MERGE_KLINE_DICT['BV1'],
        'Mid': constant.MERGE_KLINE_DICT['Mid'],
        'Midv': constant.MERGE_KLINE_DICT['Midv']}
    if 'Amount' in kline_df.columns:
        dict_how['Amount'] = constant.MERGE_KLINE_DICT['Amount']
    if 'Marketposition' in kline_df.columns:
        dict_how['Marketposition'] = constant.MERGE_KLINE_DICT['Marketposition']
    period_minutes = kline_df.resample('%s' % resample_interval_type, closed='right', label='right').apply(dict_how)
    # 价格按照OHLC进行重采样
    period_minutes = period_minutes.dropna(how='any')
    return period_minutes


def _convert_freq2text(freq_str: str):
    """
    将rio3传入的freq_str转成English的文本命名方式
    :param freq_str:
    :return:
    """
    # 如果传入单独的频率周期，
    if freq_str in ['m', 'd', 'w', 'M', 'Q']:
        freq_str = '1{}'.format(freq_str)
    freq_text = PERIOD2TEXT_DICT.get(freq_str, None)
    if freq_text is None:
        logger.warning(TypeError(f'不支持的freq_str: {freq_str}的转化类型, 期货品种缓存频率直读只支持5s/15s/30s/1m/5m/15m/30m/60m/1d'))
        # 如果PERIOD2TEXT_DICT没有现有数据的支持类型，直接获取1分钟进行转化
        logger.info(f'使用默认1分钟周期/日线周期进行数据转化')
        # 判断freq_str的单位，如果是1d以上的，都按照1d作为最小单位取
        # 将freq_str等于单个周期单位的转化为标准周期的转化
        freq_num_unit = re.match(FREQ_REGEX, freq_str)
        if freq_num_unit is not None:
            freq_unit = freq_num_unit.group('unit')
            # 1d以上的周期，取1d的线作为基础数据
            if freq_unit in ['d', 'w', 'M', 'Q']:
                freq_text = 'oneday'
            else:
                freq_text = 'oneminute'
        else:
            freq_text = 'oneminute'
    return freq_text


# 将本地历史数据转化成震球研究基础数据
def convert_da2zq_data(data_df, stockfullcode, freq):
    '''
    震球数据的标准columns
    date | tday | didx | kidx | grid | span | open | close | high | low | mid | midv | vol | amount | sts | ets | oi | contract
           date     today  didx  kidx  grid  span    open   close    high     low     mid       midv     vol    sts    ets      oi  contract
    0  20100104  20100104     0     1   101    30  4523.0  4523.0  4523.0  4523.0  4522.5  4522.0135  1766.0  85930  90000  912462  SQrb1005
    1  20100104  20100104    60     2   102    30  4528.0  4535.0  4536.0  4525.0  4535.0  4534.0075  6154.0  90030  90100  914126  SQrb1005
    2  20100104  20100104   120     3   103    30  4537.0  4534.0  4540.0  4532.0  4533.5  4533.1391  9812.0  90100  90130  914606  SQrb1005

    date : 自然日期
            转换成时间算法 ：da['Date'] = pd.to_datetime(da.date,format='%Y%m%d')
    today : 交易日期
            股指期货和date 一致
            商品期货 晚上21点 到次日15点

    span:   间隔的秒数

    open,high,low,close : 价格

    vol: 成交量

    mid:  (ask1_price + bid1_price)/2

    midv:  (ask1_price * ask1_vol + bid1_price * bid1_vol) / (ask1_vol + bid1_vol)

    sts : start time stamp
            sts_A = 1000000 + da.sts
            da['sts_time'] = pd.to_datetime(sts_A,format = '1%H%M%S')
    ets : end time stamp
            ets_A = 1000000 + da.ets
            da['ets_time'] = pd.to_datetime(ets_A,format = '1%H%M%S')

    oi :  open interest (持仓手数)

    contract : 对应的实际合约

    :param stockfullcode:
    :param period:
    :param start:
    :param end:
    :return:
    '''
    # 默认获取1分钟线以上的kline数据
    # # 过滤数据
    # if period.value < 1800:
    #     da_df = get_legal_period_data(data_df=da_df, stockfullcode=stockfullcode, period=period)
    # da_df = da_filter_onebar(da=da_df)

    # reset index以保证序列正确
    da_df = data_df.copy().reset_index()
    span_num = int(convert_freq2num(freq_str=freq))
    if not da_df.empty:
        da_df['date'] = da_df.DateTrade.dt.strftime("%Y%m%d").astype(np.int64)
        da_df['span'] = span_num
        da_df['ets'] = da_df.DateTrade.dt.strftime('%H%M%S').astype(np.int64)
        da_df['sts'] = (da_df.DateTrade - pd.Timedelta(seconds=span_num)).dt.strftime('%H%M%S').astype(
            np.int64)

        # 获取ed_info
        ed_info_df = get_ed_info_df(stockfullcode=stockfullcode, start=constant.DATETIME_MIN,
                                    end=constant.DATETIME_MAX)
        if not ed_info_df.empty:
            ed_info_df['date_execute'] = pd.to_datetime(ed_info_df['date_execute'], format='%Y%m%d')
            # 添加ed_info_idx,用于复权运算,每次都重新计算排列
            add_contract_idx(quote_df=da_df, ed_info_df=ed_info_df)
        else:
            da_df['stockfullcode_main'] = stockfullcode
            da_df['index'] = -1

        # 将把header全部lowercase
        da_df.columns = da_df.columns.str.lower()

        # 修改column name
        da_df = da_df.rename(
            columns={'volume': 'vol', 'marketposition': 'oi', 'stockfullcode_main': 'contract',
                     'index': 'contract_idx'})

        # 将contract转化成str
        da_df['contract'] = da_df['contract'].map(str).map(str.strip)

        # 删掉无用的columns
        if 'datetrade' in da_df.columns:
            da_df.drop(['datetrade'], inplace=True, axis=1)

        # 对tday进行重命名
        da_df = da_df.rename(columns={'tday': 'today'})

        # 对tday进行格式化
        da_df['today'] = pd.to_datetime(da_df['today']).dt.strftime('%Y%m%d').astype(np.int64)

        # 计算kidx和didx
        _calculate_zq_kdix(quote_df=da_df, freq_str=freq)

        # 填充标准格式
        da_df['symbol'] = stockfullcode
        da_df['high_limit'] = float(-1)
        da_df['low_limit'] = float(-1)
        da_df['pre_close'] = float(-1)
        da_df['avg'] = float(-1)
        da_df['factor'] = float(-1)
        # 排序和过滤掉不需要缓存的列
        da_df = da_df[constant.ZQ_KLINE_FORMAT_COLUMNS]
        # 定义格式
        da_df = da_df.astype(constant.ZQ_KLINE_ORG_COLUMNS_DTYPES_DICT, copy=True)

    return da_df


def get_history_data(stockfullcode, start_date=None, end_date=None, fields=None, freq=DEFAULT_FREQ, ed_date=None,
                     ed_type=None, cache_path='nas', dataframe=True):
    '''
    获取本地/NAS的行情数据
    :param stockfullcode: 品种，标准格式
    :param start_date: 开始时间
    :param end_date: 结束时间
    :param fields: 筛选域，默认['date', 'today', 'span', 'open', 'close', 'high', 'low', 'vol', 'amount', 'high_limit',
                        'low_limit', 'avg', 'pre_close', 'sts', 'ets', 'oi', 'factor', 'contract']
    :param freq: 行情频率 最大支持1d, 最小支持1m，默认是5m
    :param ed_date: 目前版本无作用
    :param ed_type: 复权类型 pre/post/None
    :param cache_path: 指定数据获取方式 nas: 办公室网盘缓存 (常量参数) http: JQData云上数据(常量参数) path: 本地缓存路径 （变量参数）
    :param dataframe:
    :return:
    '''
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    list_fields = [fields, ] if isinstance(fields, str) else fields
    # 首先获取原始数据
    quote_dict = _get_history_data_org(stockfullcode=list_stockfullcode, start_date=start_date, end_date=end_date,
                                       fields=list_fields, freq=freq, cache_path=cache_path, dataframe=False)
    # 如果没有数据，返回空
    if len(quote_dict) == 0:
        return pd.DataFrame()

    for each_code, quote_df in quote_dict.items():
        if not quote_df.empty:
            quote_df_copy = quote_df.copy()
            quote_df_copy = filter_quote_df(quote_df=quote_df_copy)
            # 填充mid和midv
            _fill_mid_and_midv(quote_df=quote_df_copy)
            # 计算kidx和didx
            _calculate_zq_kdix(quote_df=quote_df_copy, freq_str=freq)
            # 期货数据频率是固定的，所以不做从1m开始的转化
            # 如果传入的cache_path是redis地址，则不做resample转化
            if re.match(pattern=regex.REDIS_CONNECT_URL_FORMAT, string=cache_path) is None:
                quote_df_copy = period_conversion_resample(quote_df=quote_df_copy, freq_str=freq)
            # 获取复权信息表
            ed_quote_df = quote_df_copy
            if ed_type is not None:
                ed_info_df = get_ed_info_df(stockfullcode=each_code, start=constant.DATETIME_MIN,
                                            end=constant.DATETIME_MAX)
                if not ed_info_df.empty:
                    ed_info_df['date_execute'] = pd.to_datetime(ed_info_df['date_execute'], format='%Y%m%d')
                    # 如果在指定查询周期内没有找edinfo，则不复权
                if ed_info_df is not None and not ed_info_df.empty:
                    # 添加ed_info_idx,用于复权运算,每次都重新计算排列
                    add_contract_idx(quote_df=ed_quote_df, ed_info_df=ed_info_df)
                    # 如果cache_path = redis连接字符串, 永远都取前复权的数据
                    ed_type = 'pre' if re.match(pattern=regex.REDIS_CONNECT_URL_FORMAT,
                                                string=cache_path) is not None else ed_type
                    # 复权操作
                    ed_quote_df = calculate_ed_info(data_df=ed_quote_df, ed_info_df=ed_info_df, ed_type=ed_type)
            # 填充标准格式
            ed_quote_df['high_limit'] = float(-1)
            ed_quote_df['low_limit'] = float(-1)
            ed_quote_df['pre_close'] = float(-1)
            ed_quote_df['avg'] = float(-1)
            ed_quote_df['factor'] = float(-1)
            # 排序和过滤掉不需要缓存的列
            ed_quote_df = ed_quote_df[constant.ZQ_KLINE_FORMAT_COLUMNS]
            # 定义格式
            ed_quote_df = ed_quote_df.astype(constant.ZQ_KLINE_ORG_COLUMNS_DTYPES_DICT, copy=True)
            quote_dict[each_code] = ed_quote_df
    if not dataframe:
        return quote_dict
    total_quote_df = pd.concat(quote_dict.values())
    return total_quote_df


def attach_contract(quote_df=None):
    """
    默认存储的ZQ格式数据中，contract可能不准确，保留外部调用函数，重新校正
    :param quote_df:
    :return:
    """
    pass


def _get_history_data_org(stockfullcode, start_date=None, end_date=None, fields=None, freq=DEFAULT_FREQ,
                          cache_path='nas', dataframe=True):
    """
    获取旧版本格式的期货ZQ_KLINE_ORG_COLUMNS的数据，不复权
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param cache_path:
    :param dataframe:
    :return:
    """
    quote_df_dict = dict()
    quote_df = pd.DataFrame()
    start_date = _to_datetime(start_date)
    end_date = _to_datetime(end_date)
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    # 判断是否是本地缓存
    from_local_cache = os.path.isdir(cache_path)
    # 对于9001/9002/9003的数据，http是不能获取的，所以用户传入auto或是nas的时候，都置为nas
    cache_path = 'nas' if cache_path == 'http' or cache_path == 'auto' or cache_path == 'nas' else cache_path
    # 如果cache_path是nas或者传入的是本地路径，则构造h5文件的读取路径
    if cache_path == 'nas' or from_local_cache:
        freq_text = _convert_freq2text(freq_str=freq)
        # 根据freq_num判断nas_path的kline或者tkline路径
        freq_num = convert_freq2num(freq_str=freq)
        nas_path = constant.ZQ_H5_TKLINE_DIR if freq_num < 60 else constant.ZQ_H5_KLINE_DIR
        local_data_msg = u'[NAS缓存]:{}'.format(nas_path) if cache_path == 'nas' else u"本地 '{}' 缓存".format(cache_path)
        start_date_num = int(start_date.strftime('%Y%m%d'))
        end_date_num = int(end_date.strftime('%Y%m%d'))
        start_time_num = int(start_date.strftime('%H%M%S'))
        end_time_num = int(end_date.strftime('%H%M%S'))
        # 是否使用h5自己的过滤语句
        use_where = False
        # 如果开始结束时间差超过4年，则不启用where筛选条件
        if end_date_num - start_date_num <= 40000:
            use_where = True
        # 根据读取path和period的级别，决定cache_path的root_dir和缓存文件的后缀名
        if cache_path == 'nas' or from_local_cache:
            if freq_num >= 60:
                cache_root_path = nas_path if not from_local_cache else cache_path
                cache_type_str = 'kline'
            elif freq_num < 60:
                cache_root_path = nas_path if not from_local_cache else cache_path
                cache_type_str = 'tkline'
        else:
            logger.warning(u"传入的cache_path不正确: {}，请确认指定['nas', 'cache_path']两类参数中的一个".format(cache_path))
            return return_data(return_frame=dataframe)
        h5_cache_files = list()
        # 循环构造stockfullcode的h5路径
        for each_code in list_stockfullcode:
            if not from_local_cache:
                cache_file_dir = _get_cache_files_dir(cache_root_path=cache_root_path, stockfullcode=each_code,
                                                      is_local=False)
            else:
                cache_file_dir = cache_path
            cache_file_path = _get_h5_file_path(source_path=cache_file_dir, stockfullcode=each_code,
                                                h5_type=cache_type_str)
            h5_cache_files.append(cache_file_path)
        logger.info(u"\n[数据源]：{}\n[代码]:'{}'".format(local_data_msg, ','.join(list_stockfullcode)))
        for cache_file_path in h5_cache_files:
            # 通过H5路径文件名提取stockfullcode
            stock_code = _get_stock_code_from_h5_file(cache_file_path)
            # 读取相关stock code的HDF5缓存数据
            if h5.is_hdf5(cache_file_path):
                h5_file = pd.HDFStore(cache_file_path, mode='r')
                file_keys = h5_file.keys()
                if '/{}'.format(freq_text) in file_keys:
                    group_node = str(freq_text).lower()
                    # 如果start_time_num和end_time_num都是0，认为是按照tday查询
                    if start_time_num == 0 and end_time_num == 0:
                        where_clause = "today>={} & today<={}".format(start_date_num, end_date_num)
                    else:
                        where_clause = "date>={} & date<={}".format(start_date_num, end_date_num)
                    # 读取H5数据
                    try:
                        if use_where:
                            cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields,
                                                   where=where_clause)
                        else:
                            cache_df = pd.read_hdf(path_or_buf=cache_file_path, key=group_node, columns=fields)
                        if not cache_df.empty:
                            # 用dataframe作过滤
                            if start_time_num == 0 and end_time_num == 0:
                                cache_df = cache_df[
                                    (cache_df['today'] >= start_date_num) & (
                                            cache_df['today'] <= end_date_num)]
                            else:
                                cache_df = cache_df[
                                    (cache_df['date'] >= start_date_num) & (
                                            cache_df['date'] <= end_date_num) & (
                                            cache_df['sts'] >= start_time_num) & (
                                            cache_df['ets'] <= end_time_num)]
                    except Exception as ex:
                        logger.warning(u'品种[{}],级别[{}] 读取ZQ格式数据失败[{}]'.format(stock_code, freq, ex))
                        cache_df = pd.DataFrame()
                    if not cache_df.empty:
                        # 过滤N/A
                        cache_df = cache_df.dropna(axis=0, subset=['open', 'high', 'low', 'close', 'amount', 'vol'])
                        if not cache_df.empty:
                            cache_df = cache_df.reset_index(drop=True)
                            # 标记symbol和contract
                            cache_df['symbol'] = stock_code
                        # 默认组装字典格式的数据类型
                        quote_df_dict[stock_code] = cache_df
                h5_file.close()
    # 修改为判断cache_path是否符合redis连接字符串格式 2021-02-03
    elif re.match(pattern=regex.REDIS_CONNECT_URL_FORMAT, string=cache_path) is not None:
        quote_df_dict = get_redis_history_data_org(stockfullcode=list_stockfullcode, start_dt=start_date,
                                                   end_dt=end_date, freq=freq, redis_url=cache_path, dataframe=False)
    elif cache_path == 'http':
        logger.info(u'目前不支持从JQData获取期货缓存数据')

    if dataframe:
        if len(quote_df_dict) > 0:
            quote_df = pd.concat(quote_df_dict.values())
            quote_df = quote_df.reset_index(drop=True)
        return quote_df
    else:
        return quote_df_dict


# 分钟线之间的转化
def period_conversion_resample(quote_df, freq_str='5m', filter_columns=constant.ZQ_KLINE_FORMAT_COLUMNS):
    '''
    :param quote_df: zq格式的K线数据
    :param freq_str:
    :return:
    '''
    # 获取resample时间线
    assert isinstance(quote_df, pd.DataFrame)
    if quote_df.empty:
        return pd.DataFrame()
    # 默认基础数据的级别是1分钟和1d,如果源数据是1分钟或者1d，则不需要进行转化
    if freq_str == '1m':
        return quote_df
    quote_df = quote_df.copy()
    resample_interval_type = get_resample_interval(freq_str=freq_str)
    freq_num = convert_freq2num(freq_str=freq_str)
    if 'DateTrade' not in quote_df.columns:
        quote_df['DateTrade'] = pd.to_datetime(quote_df['date'] * 1000000 + quote_df['ets'], format='%Y%m%d%H%M%S')
    dict_how = dict()
    for each_column in quote_df.columns:
        if each_column in constant.MERGE_KLINE_DICT.keys():
            dict_how[each_column] = constant.MERGE_KLINE_DICT[each_column]

    # 将DateTrade赋值到index, resample依赖于时间序列的index进行切分
    quote_df.index = quote_df.DateTrade
    period_quotes = quote_df.resample('%s' % resample_interval_type, closed='right', label='right').apply(dict_how)
    # 价格按照OHLC进行重采样
    period_quotes = period_quotes.dropna(how='any')
    # 重新将zq格式的date，sts和ets通过DateTrade构造出来
    span_value = convert_freq2num(freq_str=freq_str)
    period_quotes['span'] = span_value
    period_quotes['ets'] = 0
    period_quotes['sts'] = 0
    # 修复周期转化的bug, resample转化1d的周期会让index + 1 天
    if freq_num < 86400:
        period_quotes['date'] = period_quotes.index.strftime("%Y%m%d").astype(np.int64)
        period_quotes['ets'] = period_quotes.index.strftime('%H%M%S').astype(np.int64)
        period_quotes['sts'] = (period_quotes.index - pd.Timedelta(seconds=int(span_value))).strftime(
            '%H%M%S').astype(np.int64)
    period_quotes['date'] = period_quotes['date'].astype(np.int64)
    period_quotes['today'] = period_quotes['today'].astype(np.int64)
    # 重新计算kidx和didx
    _calculate_zq_kdix(quote_df=period_quotes, freq_str=freq_str)
    return period_quotes


def calculate_ed_info(data_df, ed_info_df=None, ed_type=None, contract_idx_column='contract_idx',
                      price_column_list=['open', 'close', 'high', 'low', 'mid', 'midv'], vol_column_list=['vol'],
                      amount_column_list=['amount'], vol_column='vol'):
    """
    # 处理震球数据格式的复权算法
    :param da: 行情数据，DataFrame格式，必须含有赋权表索引的字段，对应到相关的ed_info_df的index
    :param ed_info_df: 复权信息数据，DataFrame格式
    :param ed_type: 复权类型，支持前复权和后复权
    :param contract_idx_column: da含有的赋权表索引的字段名
    :param price_column_list: 价格列表赋权信息
    :param vol_column_list: 复权类型，成交量列表赋权信息
    :param amount_column_list: 复权类型，成交金额列表赋权信息
    :param vol_column: 缺省的成交量名
    """
    assert isinstance(data_df, pd.DataFrame)
    # 如果data_df或ed_info_df有一个为空，则直接返回数据
    if data_df.empty:
        return data_df

    if ed_info_df is not None:
        assert isinstance(ed_info_df, pd.DataFrame)
        if ed_info_df.empty:
            return data_df
        else:
            add_contract_idx(quote_df=data_df, ed_info_df=ed_info_df)
    else:
        stockfullcode = data_df.iloc[0].symbol
        ed_info_df = get_ed_info_df(stockfullcode=stockfullcode, start=constant.DATETIME_MIN, end=constant.DATETIME_MAX)
        if not ed_info_df.empty:
            ed_info_df['date_execute'] = pd.to_datetime(ed_info_df['date_execute'], format='%Y%m%d')
            # 如果在指定查询周期内没有找edinfo，则不复权
        if ed_info_df is not None and not ed_info_df.empty:
            # 添加ed_info_idx,用于复权运算,每次都重新计算排列
            add_contract_idx(quote_df=data_df, ed_info_df=ed_info_df)

    if ed_type is not None:
        price_c1_v, price_c0_v, vol_c1_v = calculate_da_ed_factors(data_df, ed_info_df,
                                                                   ed_type=ed_type,
                                                                   contract_idx_column=contract_idx_column)
        future_en = is_future(str(ed_info_df['stockfullcode'].iloc[0]))

        internal_price_column_set = set(price_column_list) & set(data_df.columns)
        internal_vol_column_set = set(vol_column_list) & set(data_df.columns)
        internal_amount_column_set = set(amount_column_list) & set(data_df.columns)

        if not future_en:
            for amount_column in internal_amount_column_set:
                data_df[amount_column] /= data_df[vol_column]

        # 取消复权操作对Amount的修改 2020-09-28
        # for price_column in (internal_price_column_set | internal_amount_column_set):
        # for price_column in internal_price_column_set:
        #     data_df[price_column] = price_c1_v * data_df[price_column].values + price_c0_v
        # # 对vol进行复权操作
        # data_df[vol_column] = vol_c1_v * data_df[vol_column].values

        # 恢复复权操作 修改Amount, 不修改vol的模式，需要推广Fixed版本复权的时候再做改动 2020-10-14
        for price_column in (internal_price_column_set | internal_amount_column_set):
            data_df[price_column] = price_c1_v * data_df[price_column].values + price_c0_v
        if not future_en:
            for vol_column in internal_vol_column_set:
                data_df[vol_column] = vol_c1_v * data_df[vol_column].values

            for amount_column in internal_amount_column_set:
                data_df[amount_column] = data_df[amount_column].values * data_df[amount_column].values
    # 删除contract_idx字段
    if 'contract_idx' in data_df.columns:
        data_df.drop(['contract_idx'], inplace=True, axis=1)
    # 删除Tday字段
    if 'Tday' in data_df.columns:
        data_df.drop(['Tday'], inplace=True, axis=1)
    return data_df


def add_contract_idx(quote_df, ed_info_df):
    """
    添加contract_idx，每次都重新计算
    :param self:
    :param quote_df:
    :param ed_info_df:
    :return:
    """
    ed_info_df = check_ed_info_df(quote_df, ed_info_df)
    if 'contract_idx' in quote_df.columns:
        del quote_df['contract_idx']
    data_merge(td=quote_df, da=ed_info_df, td_date_column='Tday', da_date_column='date_execute',
               merge_column_list=None)
    if 'contract' in quote_df.columns:
        quote_df.drop(['contract'], inplace=True, axis=1)
    quote_df.rename(columns={'index': 'contract_idx', 'stockfullcode_main': 'contract'}, inplace=True)


def double_reset(da):
    """
    :param da: 可以时任意的DataFrame
    功能： 将da的index列两次reset_index
    """
    for temp_column in ['index', 'level_0']:
        if temp_column in da.columns:
            del da[temp_column]
    da.reset_index(inplace=True)
    da.reset_index(inplace=True)


def data_merge(td, da, td_date_column='date_closeposition', da_date_column='DateTrade',
               merge_column_list=['profit_perfect', 'Tday'], td_sort_en=False, da_sort_en=False,
               rename_column_list=None):
    """
    功能：将期货数据要merge的列根据时间merge到交易数据上，缺省列是波动率和'Tday'
    td 和 da的时间频率不匹配，
    将td的td_date_column列和 da的da_date_column 列同步，
    保证da的时间与td.DateTrade 发生之前的最近一个对应的时间戳对齐（保证无未来）
    同时如果rename_column_list不是None，merge到的将是新的列名

    :param td: 交易数据
    :param da: 行情数据
    :param td_date_column: 交易数据的时间列
    :param da_date_column: 行情数据的时间列
    :param merge_column_list: 要merge的列
    :param td_sort_en: merge之前td是否要sort
    :param da_sort_en: merge之前是否要sort
    :param rename_column_list: merge之后的新列名，名字若不改则用以前的列名
    """

    reset_occur = False
    if merge_column_list is None:
        merge_column_list = ['index', 'stockfullcode_main']
        double_reset(da)
        reset_occur = True

    if rename_column_list is None:
        td_merge_column_list = merge_column_list
    else:
        td_merge_column_list = rename_column_list

    if td_sort_en:
        td.sort_values(td_date_column, inplace=True)

    if da_sort_en:
        da.sort_values(da_date_column, inplace=True)

    da_merge_v = da[merge_column_list].values
    for i in range(len(merge_column_list)):
        td[td_merge_column_list[i]] = da_merge_v[0][i]
    td_merge_v = td[td_merge_column_list].values
    da_date_v = da[da_date_column].values
    date_end = np.datetime64('2030-01-01')
    da_date_v = np.concatenate((da_date_v, [date_end]))
    td_date_v = td[td_date_column].values

    j = 0
    step = 1000
    for i in range(len(td)):
        cur_date = td_date_v[i]
        while j < (len(da) - step):
            if da_date_v[j + step] <= cur_date:
                j += step
                continue
            else:
                while da_date_v[j + 1] <= cur_date:
                    j += 1
                    continue
                else:
                    break
        else:
            while da_date_v[j + 1] <= cur_date:
                j += 1
                continue
        td_merge_v[i] = da_merge_v[j]
    if len(td) == 1:
        for i in range(len(merge_column_list)):
            td[td_merge_column_list[i]] = td_merge_v[0][i]
    else:
        td[td_merge_column_list] = td_merge_v

    if reset_occur:
        for temp_column in ['index', 'level_0']:
            if temp_column in da.columns:
                del da[temp_column]


def calculate_da_ed_factors(
        data_df,
        ed_info_df,
        ed_type=None,
        contract_idx_column='contract_idx'):
    """
    :param data_df: 行情数据，DataFrame格式，必须含有'contract_idx'字段，对应到相关的ed_info_df的index
    :param ed_info_df: 复权信息数据，DataFrame格式
    :param ed_type: 复权类型，支持前复权和后复权
    :param contract_idx_column: da含有的赋权表索引的字段名
    :return: price_c1_v, price_c0_v, vol_c1_v 和da等长的价格和成交量系数
    """
    if ed_type is None:
        return data_df
    fhbl_v = ed_info_df['fhbl'].astype(np.float64).values
    pgbl_v = ed_info_df['pgbl'].astype(np.float64).values
    pgjg_v = ed_info_df['pgjg'].astype(np.float64).values
    sgbl_v = ed_info_df['sgbl'].astype(np.float64).values
    future_en = is_future(str(ed_info_df['stockfullcode'].iloc[0]))

    price_c1 = 1.0
    price_c0 = 0.0
    vol_c1 = 1.0

    price_c1_v = np.array(len(data_df) * [price_c1])
    price_c0_v = np.array(len(data_df) * [price_c0])
    vol_c1_v = np.array(len(data_df) * [vol_c1])

    group_s = data_df.groupby(contract_idx_column)[data_df.columns[0]].count()
    group_contract_idx = group_s.index.values
    group_pos = np.append(np.array([0]), group_s.cumsum().values)

    # 因为da的数据的group_contract_idx可能会有不连续的情况，
    # 这样需要两个指针，一个是cur_da_group_contract_idx，指向da中的group_contract_idx
    # 一个是ed_info_ptr，指向ed_info_df中的index

    # first = True
    if ed_type == 'post':  # 后复权处理
        da_group_idx = 1  # 第0段不需要赋值，指向group的指针初始化到下个要赋值的da的段号
        for ed_info_ptr in range(group_contract_idx[0] + 1, group_contract_idx[-1] + 1):
            temp = (1 + sgbl_v[ed_info_ptr] + pgbl_v[ed_info_ptr])
            price_c1 = temp * price_c1
            price_c0 = temp * price_c0 - pgbl_v[ed_info_ptr] * pgjg_v[ed_info_ptr] + fhbl_v[ed_info_ptr] * 0.9
            if not future_en:
                vol_c1 /= (1 + pgbl_v[ed_info_ptr] + sgbl_v[ed_info_ptr])  # 上面5行是每次后复权基本运算
            cur_da_group_contract_idx = group_contract_idx[da_group_idx]  # 即da中的当前group_contract_idx
            if cur_da_group_contract_idx == ed_info_ptr:  # cur_da_group_contract_idx和ed_info_ptr同步时da每段才能前进
                cur_pos_start = group_pos[da_group_idx]
                cur_pos_end = group_pos[da_group_idx + 1]
                price_c1_v[cur_pos_start: cur_pos_end] = price_c1
                price_c0_v[cur_pos_start: cur_pos_end] = price_c0
                if not future_en:
                    vol_c1_v[cur_pos_start: cur_pos_end] = vol_c1
                da_group_idx += 1  # 保持同步，非常关键，只有da每段赋值后才能前进

    elif ed_type == 'pre':  # 前复权处理，与后复权相比，循环是反过来走的
        da_group_idx = len(group_contract_idx) - 2  # 最后1段不需要赋值，指向group的指针初始化到倒数第2段赋值的da的段号
        for ed_info_ptr in range(group_contract_idx[-1] - 1, group_contract_idx[0] - 1, -1):
            temp = 1.0 / (1 + sgbl_v[ed_info_ptr + 1] + pgbl_v[ed_info_ptr + 1])
            price_c1 = temp * price_c1
            price_c0 = temp * (price_c0 + pgbl_v[ed_info_ptr + 1] * pgjg_v[ed_info_ptr + 1] -
                               fhbl_v[ed_info_ptr + 1] * 0.9)
            if not future_en:
                vol_c1 *= (1 + pgbl_v[ed_info_ptr + 1] + sgbl_v[ed_info_ptr + 1])  # 上面5行是每次前复权基本运算
            cur_da_group_contract_idx = group_contract_idx[da_group_idx]  # 即da中的当前group_contract_idx
            if cur_da_group_contract_idx == ed_info_ptr:  # cur_da_group_contract_idx和ed_info_ptr同步时da每段才能前进
                cur_pos_start = group_pos[da_group_idx]
                cur_pos_end = group_pos[da_group_idx + 1]
                price_c1_v[cur_pos_start: cur_pos_end] = price_c1
                price_c0_v[cur_pos_start: cur_pos_end] = price_c0
                if not future_en:
                    vol_c1_v[cur_pos_start: cur_pos_end] = vol_c1
                da_group_idx -= 1  # 保持同步，非常关键，只有da每段赋值后才能前进

    return price_c1_v, price_c0_v, vol_c1_v


def get_history_data_csharp(stockfullcode, start_date=None, end_date=None, fields=None, freq=DEFAULT_FREQ, ed_date=None,
                            ed_type=None, columns=None, cache_path='nas', dataframe=True):
    """
    获取通用格式的历史数据，股票/期货通用
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param ed_date:
    :param ed_type:
    :param cache_path:
    :param dataframe:
    :return:
    """
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    list_fields = [fields, ] if isinstance(fields, str) else fields
    # 首先获取原始数据
    quote_dict = _get_history_data_csharp_org(stockfullcode=list_stockfullcode, start_date=start_date,
                                              end_date=end_date, fields=list_fields, freq=freq, cache_path=cache_path,
                                              columns=columns, dataframe=False)
    # 如果没有数据，返回空
    if len(quote_dict) == 0:
        return return_data(return_frame=dataframe)
    # 如果不复权
    if ed_type is None:
        # 返回结果
        if not dataframe:
            return quote_dict
        total_quote_df = pd.concat(quote_dict.values())
        return total_quote_df

    # 复权信息存储变量
    ed_info_dict = dict()
    # 复权操作
    if ed_type == 'post':
        for tmp_quote_df in quote_dict.values():
            if not tmp_quote_df.empty and len(tmp_quote_df) > 0:
                first_quote_date = pd.to_datetime(
                    tmp_quote_df.iloc[0].DateTrade.replace(hour=0, minute=0, second=0, microsecond=0))
                last_quote_date = pd.to_datetime(
                    tmp_quote_df.iloc[-1].DateTrade.replace(hour=0, minute=0, second=0, microsecond=0))
                # 设置默认的first_post_ed_date
                first_post_ed_date = get_previous_tday(first_quote_date)
                if ed_date is not None and ed_date <= first_quote_date:
                    first_post_ed_date = ed_date
                ed_info_dict = get_ed_info_org(stockfullcode=list_stockfullcode,
                                               start=first_post_ed_date.strftime('%Y-%m-%d'),
                                               end=last_quote_date.strftime('%Y-%m-%d'), return_dict=True)
    elif ed_type == 'pre':
        for tmp_quote_df in quote_dict.values():
            first_quote_date = pd.to_datetime(
                tmp_quote_df.iloc[0].DateTrade.replace(hour=0, minute=0, second=0, microsecond=0))
            last_quote_date = pd.to_datetime(
                tmp_quote_df.iloc[-1].DateTrade.replace(hour=0, minute=0, second=0, microsecond=0))
            # 获取下一个交易日
            end_date = get_next_tday(last_quote_date)
            if ed_date is not None and ed_date >= last_quote_date:
                end_date = ed_date

            start_date = get_previous_tday(first_quote_date)

            ed_info_dict = get_ed_info_org(stockfullcode=list_stockfullcode, start=start_date.strftime('%Y-%m-%d'),
                                           end=end_date.strftime('%Y-%m-%d'), return_dict=True)
            if len(ed_info_dict) > 0:
                for tmp_stockfullcode in ed_info_dict.keys():
                    ed_info_dict[tmp_stockfullcode].sort(key=lambda x: x.date_ed, reverse=True)
    ed_quote_dict = defaultdict(list)

    for tmp_stockfullcode, tmp_quote_df in quote_dict.items():
        tmp_ed_info_list = ed_info_dict.get(tmp_stockfullcode, None)
        # 如果在指定查询周期内没有找edinfo，则不复权
        tmp_ed_type = ed_type
        if tmp_ed_info_list is None:
            tmp_ed_type = None
        ed_quote_df = convert_orgQuote_to_edQuote(org_quote_df=tmp_quote_df, ed_info_list=tmp_ed_info_list,
                                                  ed_type=tmp_ed_type)
        ed_quote_dict[tmp_stockfullcode] = ed_quote_df
    # 返回结果
    if not dataframe:
        return ed_quote_dict
    # 判断返回数据
    total_quote_df = pd.concat(ed_quote_dict.values())
    return total_quote_df


def convert_orgQuote_to_edQuote(org_quote_df, ed_info_list, ed_type=None):
    if ed_type is None:
        return org_quote_df
    if ed_type == 'pre':
        ed_info_list.sort(key=lambda e: e.date_ed, reverse=False)
        for ed_info in ed_info_list:
            if ed_info.is_future:
                filtered_quote_df = org_quote_df[
                    org_quote_df.DateTrade < ed_info.date_ed + dt.timedelta(hours=20)].copy()
            elif not ed_info.is_future:
                filtered_quote_df = org_quote_df[org_quote_df.DateTrade < ed_info.date_ed].copy()
            filtered_quote_df['Close'] = BasicQuoteInfo.get_pre_ed_price(filtered_quote_df['Close'], ed_info)
            filtered_quote_df['Open'] = BasicQuoteInfo.get_pre_ed_price(filtered_quote_df['Open'], ed_info)
            filtered_quote_df['High'] = BasicQuoteInfo.get_pre_ed_price(filtered_quote_df['High'], ed_info)
            filtered_quote_df['Low'] = BasicQuoteInfo.get_pre_ed_price(filtered_quote_df['Low'], ed_info)
            filtered_quote_df['Mid'] = BasicQuoteInfo.get_pre_ed_price(filtered_quote_df['Mid'], ed_info)
            filtered_quote_df['Midv'] = BasicQuoteInfo.get_pre_ed_price(filtered_quote_df['Midv'], ed_info)
            ed_quote_df = BasicQuoteInfo.get_pre_ed_volume(
                filtered_quote_df['Volume'], ed_info) * BasicQuoteInfo.get_pre_ed_price(
                filtered_quote_df['Amount'] / filtered_quote_df['Volume'], ed_info)
            filtered_quote_df['Amount'] = ed_quote_df
            filtered_quote_df['Volume'] = BasicQuoteInfo.get_pre_ed_volume(filtered_quote_df['Volume'], ed_info)
            # 删掉复权前的分钟线
            org_quote_df = org_quote_df[~org_quote_df.DateTrade.isin(filtered_quote_df.DateTrade)]
            # org_quote_df = org_quote_df.drop(org_quote_df.index[filtered_quote_df.index], inplace=True)
            # 插入合并后的分钟线
            org_quote_df = pd.concat([org_quote_df, filtered_quote_df])
    elif ed_type == 'post':
        ed_info_list.sort(key=lambda e: e.date_ed, reverse=True)
        for ed_info in ed_info_list:
            if ed_info.is_future:
                filtered_quote_df = org_quote_df[
                    org_quote_df.DateTrade > ed_info.date_ed + dt.timedelta(hours=20)].copy()
            elif not ed_info.is_future:
                filtered_quote_df = org_quote_df[org_quote_df.DateTrade > ed_info.date_ed].copy()
            filtered_quote_df['Close'] = BasicQuoteInfo.get_post_ed_price(filtered_quote_df['Close'], ed_info)
            filtered_quote_df['Open'] = BasicQuoteInfo.get_post_ed_price(filtered_quote_df['Open'], ed_info)
            filtered_quote_df['High'] = BasicQuoteInfo.get_post_ed_price(filtered_quote_df['High'], ed_info)
            filtered_quote_df['Low'] = BasicQuoteInfo.get_post_ed_price(filtered_quote_df['Low'], ed_info)
            filtered_quote_df['Mid'] = BasicQuoteInfo.get_post_ed_price(filtered_quote_df['Mid'], ed_info)
            filtered_quote_df['Midv'] = BasicQuoteInfo.get_post_ed_price(filtered_quote_df['Midv'], ed_info)
            ed_quote_df = BasicQuoteInfo.get_post_ed_volume(filtered_quote_df['Volume'],
                                                            ed_info) * BasicQuoteInfo.get_post_ed_price(
                filtered_quote_df['Amount'] / filtered_quote_df['Volume'], ed_info)
            filtered_quote_df['Amount'] = ed_quote_df
            filtered_quote_df['Volume'] = BasicQuoteInfo.get_post_ed_volume(filtered_quote_df['Volume'], ed_info)
            # 删掉复权前的分钟线
            org_quote_df = org_quote_df[~org_quote_df.DateTrade.isin(filtered_quote_df.DateTrade)]
            org_quote_df = pd.concat([org_quote_df, filtered_quote_df])
    else:
        raise NotImplementedError(u'不支持的复权模式{}'.format(ed_type))
    # 按照DateTrade排序
    org_quote_df = org_quote_df.sort_values(by=['DateTrade'])
    return org_quote_df


# 从D20190102或M200901分离出时间信息
def _get_dset_node_time(node_path):
    dset_name = os.path.basename(node_path)
    return int(dset_name[1:])


def _get_history_data_csharp_org(stockfullcode, start_date=None, end_date=None, fields=None, freq=DEFAULT_FREQ,
                                 columns=None, cache_path='nas', dataframe=True):
    """
    获取通用格式的H5数据，不复权
    :param stockfullcode:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param cache_path:
    :param dataframe:
    :return:
    """
    quote_df_dict = dict()
    quote_df = pd.DataFrame()
    start_date = _to_datetime(start_date)
    end_date = _to_datetime(end_date)
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    # 判断是否是本地缓存
    from_local_cache = os.path.isdir(cache_path)
    # 如果cache_path是nas或者传入的是本地路径，则构造h5文件的读取路径
    if cache_path == 'nas' or from_local_cache:
        freq_num = convert_freq2num(freq_str=freq)
        if freq == 'tick':
            nas_path = constant.COMMON_H5_TICK_DIR
        else:
            # 根据freq_num判断nas_path的kline或者tkline路径
            nas_path = constant.COMMON_H5_TKLINE_DIR if 0 < freq_num < 60 else constant.COMMON_H5_KLINE_DIR
        local_data_msg = u'[NAS缓存]:{}'.format(nas_path) if cache_path == 'nas' else u"本地 '{}' 缓存".format(cache_path)
        # 设置缓存格式的Index列名
        cache_index_column = 'DateTrade'
        # 根据读取path和period的级别，决定cache_path的root_dir和缓存文件的后缀名
        if cache_path == 'nas' or from_local_cache:
            if freq_num >= 60:
                cache_root_path = nas_path if not from_local_cache else cache_path
                cache_type_str = 'kline'
            elif 0 < freq_num <= 60:
                cache_root_path = nas_path if not from_local_cache else cache_path
                cache_type_str = 'tkline'
            elif freq_num == -1:
                cache_root_path = nas_path if not from_local_cache else cache_path
                cache_type_str = 'tick'
                cache_index_column = 'Time'
        else:
            logger.warning(u"传入的cache_path不正确: {}，请确认指定['nas', 'cache_path']两类参数中的一个".format(cache_path))
            return return_data(return_frame=dataframe)
        h5_cache_files = list()
        # 循环构造stockfullcode的h5路径
        for each_code in list_stockfullcode:
            if not from_local_cache:
                cache_file_dir = _get_cache_files_dir(cache_root_path=cache_root_path, stockfullcode=each_code,
                                                      is_local=False)
            else:
                cache_file_dir = cache_path
            cache_file_path = _get_h5_file_path(source_path=cache_file_dir, stockfullcode=each_code,
                                                h5_type=cache_type_str)
            h5_cache_files.append(cache_file_path)
        print(u"\n[数据源]：{}\n[代码]:'{}'\n[频率]:'{}'".format(local_data_msg, ','.join(list_stockfullcode), freq))
        for cache_file_path in h5_cache_files:
            # 每一个H5品种缓存的dataframe集合
            stock_code_df = pd.DataFrame()
            # 通过H5路径文件名提取stockfullcode
            stock_code = _get_stock_code_from_h5_file(cache_file_path)
            # 读取相关stock code的HDF5缓存数据
            concat_list = list()
            if h5.is_hdf5(cache_file_path):
                h5_file_id = h5.File(cache_file_path, mode='r')
                query_node_list = _get_cache_h5_nodes(period=freq, start_date=start_date, end_date=end_date)
                for each_query_node in query_node_list:
                    if _check_element_exist(h5_file_id, each_query_node):
                        cache_df = _get_dset_data(h5_file=h5_file_id, node_path=each_query_node, columns=columns,
                                                  set_index=False)
                        # 源Tick数据文件中没有Tday, 直接用节点的日期线补充
                        if cache_type_str == 'tick':
                            cache_df['Tday'] = pd.to_datetime(_get_dset_node_time(each_query_node), format='%Y%m%d')
                        concat_list.append(cache_df)
                h5_file_id.close()
            # 将读出来的节点连接
            if len(concat_list) > 0:
                stock_code_df = pd.concat(concat_list)
            if not stock_code_df.empty:
                # 读取HDF5数据是分DataSet获取的，append之后的index是乱序并且有重复，需要重新reset_index的值
                stock_code_df = stock_code_df.reset_index(drop=True)
                # 根据cache_index_column对时间进行转化
                if cache_type_str == 'tick':
                    stock_code_df[cache_index_column] = stock_code_df[cache_index_column].apply(from_oadate)
                else:
                    stock_code_df[cache_index_column] = pd.to_datetime(stock_code_df[cache_index_column],
                                                                       format='%Y%m%d%H%M%S')
                    # Tick级别以上的数据都带有交易日字段，直接转化即可
                    stock_code_df['Tday'] = pd.to_datetime(stock_code_df['Tday'], format='%Y%m%d')
                # 做具体的时间过滤和Filter函数的过滤
                stock_code_df = filter_cache_df_data(cache_df=stock_code_df, filter_column='Tday',
                                                     start_time=start_date, end_time=end_date, filters=None)
            quote_df_dict[stock_code] = stock_code_df
    else:
        logger.info(u'目前不支持从JQData获取期货缓存数据')

    if dataframe:
        if len(quote_df_dict) > 0:
            quote_df = pd.concat(quote_df_dict.values())
            quote_df = quote_df.reset_index(drop=True)
        return quote_df
    else:
        return quote_df_dict


def filter_cache_df_data(cache_df, filter_column=None, start_time=None, end_time=None, filters=None, compare_num=False):
    """
    默认读取DataSet整个内容，需要根据传入的时间对记录进行筛选
    遍历传入的filter，根据条件进行query筛选
    :param cache_df:
    :param filter_column:
    :param start_time:
    :param end_time:
    :param filters:
    :param compare_num:
    :return:
    """
    if start_time is not None and end_time is not None:
        # 如果compare_num为True, 说明原始数据是纯数据，这时候需要对传入时间进行int转化
        if compare_num:
            start_time = int(pd.to_datetime(start_time, format='%Y%m%d%H%M%S'))
            end_time = int(pd.to_datetime(end_time, format='%Y%m%d%H%M%S'))
        cache_df = cache_df[(cache_df[filter_column] >= start_time) & (cache_df[filter_column] <= end_time)]
    # 外部表达式过滤
    if filters is not None:
        if isinstance(cache_df, pd.DataFrame):
            for filter in filters:
                cache_df = cache_df.query(filter)
    return cache_df


def _get_cache_h5_nodes(period, start_date, end_date):
    """
    根据传入的period和时间范围,构造对应的H5 DataSet路径
    :param period:
    :param start_date:
    :param end_date:
    :return:
    """
    cache_node_list = list()
    tday_df = pd.DataFrame(data=get_tday_list(start_dt=start_date, end_dt=end_date), columns=['tday'])
    year_tday_iterator = tday_df.groupby(tday_df['tday'].map(lambda t: t.year))
    # 需要支持Tick数据和Kline数据两个周期的节点构造
    if period == 'tick':
        for i in range(len(tday_df)):
            tday_dt = pd.to_datetime(tday_df.iloc[i].tday).to_pydatetime()
            # 2017/D20170918
            cache_node = r'/{}/D{}'.format(tday_dt.year, tday_dt.date().strftime('%Y%m%d'))
            cache_node_list.append(cache_node)
    # 如果是Kline频率
    else:
        period_node = _convert_freq2text(freq_str=period)
        for year, year_tday_df in year_tday_iterator:
            node_year = str(year)
            # period exp: /oneminute
            node_group_path = '/{}'.format(period_node)
            for month, month_tday_df in year_tday_df.groupby(year_tday_df['tday'].map(lambda t: t.month)):
                # /period/%Y%M exp: /oneminute/M201701
                node_dset_path = node_group_path + '/{}{}{}'.format('M', node_year, str(month).zfill(2))
                cache_node_list.append(node_dset_path)
    return cache_node_list


def _check_element_exist(parent, path):
    """
    检测一个h5节点是否存在
    :param parent:
    :param path:
    :return:
    """
    check_exist_element = parent.get(path, getclass=True)
    if check_exist_element is None or check_exist_element == 'default':
        return False
    else:
        return True


# 获取一个h5文件指定节点的内容，以dataframe包装返回
def _get_dset_data(h5_file, node_path, columns=None, set_index=True):
    df_data = None
    try:
        if _check_element_exist(h5_file, node_path):
            data_set = h5_file[node_path]
            # 支持按指定column选取dataset的值
            if columns:
                tuple_columns = tuple(columns)
                data_set_value = data_set[tuple_columns]
            else:
                data_set_value = data_set[()]
            df_data = pd.DataFrame(data_set_value)
            if set_index:
                # 检测DataSet中存储的Index的值，将其设置成转化df的Index
                get_index_str = data_set.attrs.get("Index")
                if get_index_str is not None:
                    df_data = df_data.set_index(get_index_str.split(","))
    except Exception as ex:
        print('[Error]: h5_file:{}, node_path:{}, ex:{}'.format(h5_file, node_path, ex))
    return df_data


def reduce_mem_usage(props):
    # 计算当前内存
    start_mem_usg = props.memory_usage().sum() / 1024 ** 2
    print("Memory usage of the dataframe is :", start_mem_usg, "MB")

    # 哪些列包含空值，空值用-999填充。why：因为np.nan当做float处理
    NAlist = []
    for col in props.columns:
        # 这里只过滤了objectd格式，如果你的代码中还包含其他类型，请一并过滤
        if props[col].dtypes != object:

            print("**************************")
            print("columns: ", col)
            print("dtype before", props[col].dtype)

            # 判断是否是int类型
            isInt = False
            mmax = props[col].max()
            mmin = props[col].min()

            # Integer does not support NA, therefore Na needs to be filled
            if not np.isfinite(props[col]).all():
                NAlist.append(col)
                props[col].fillna(-999, inplace=True)  # 用-999填充

            # test if column can be converted to an integer
            asint = props[col].fillna(0).astype(np.int64)
            result = np.fabs(props[col] - asint)
            result = result.sum()
            if result < 0.01:  # 绝对误差和小于0.01认为可以转换的，要根据task修改
                isInt = True

            # make interger / unsigned Integer datatypes
            if isInt:
                if mmin >= 0:  # 最小值大于0，转换成无符号整型
                    if mmax <= 255:
                        props[col] = props[col].astype(np.uint8)
                    elif mmax <= 65535:
                        props[col] = props[col].astype(np.uint16)
                    elif mmax <= 4294967295:
                        props[col] = props[col].astype(np.uint32)
                    else:
                        props[col] = props[col].astype(np.uint64)
                else:  # 转换成有符号整型
                    if mmin > np.iinfo(np.int8).min and mmax < np.iinfo(np.int8).max:
                        props[col] = props[col].astype(np.int8)
                    elif mmin > np.iinfo(np.int16).min and mmax < np.iinfo(np.int16).max:
                        props[col] = props[col].astype(np.int16)
                    elif mmin > np.iinfo(np.int32).min and mmax < np.iinfo(np.int32).max:
                        props[col] = props[col].astype(np.int32)
                    elif mmin > np.iinfo(np.int64).min and mmax < np.iinfo(np.int64).max:
                        props[col] = props[col].astype(np.int64)
            else:  # 注意：这里对于float都转换成float16，需要根据你的情况自己更改
                props[col] = props[col].astype(np.float16)

            print("dtype after", props[col].dtype)
            print("********************************")
    print("___MEMORY USAGE AFTER COMPLETION:___")
    mem_usg = props.memory_usage().sum() / 1024 ** 2
    print("Memory usage is: ", mem_usg, " MB")
    print("This is ", 100 * mem_usg / start_mem_usg, "% of the initial size")
    return props, NAlist


class FeedRioTest(unittest.TestCase):
    def setUp(self):
        setup_screen_logger()
        setup_file_logger()
        self.logger = logging.getLogger('root')
        pass

    def tearDown(self):
        pass

    def test_get_history_data(self):
        stockfullcode = 'SFIF9001'
        start_date = '2019-01-01'
        end_date = '2020-12-31'
        freq = '1d'
        ed_type = 'pre'
        data_df = get_history_data(stockfullcode=stockfullcode, freq=freq, start_date=start_date, end_date=end_date,
                                   ed_type=ed_type)
        pass

    def test_get_history_data_csharp_tick(self):
        stockfullcode = 'SFIH9001'
        start_date = '2021-08-01'
        end_date = '2021-12-31'
        freq = 'tick'
        cache_path = 'nas'
        # cache_path = r'D:\TempCSV\jq_data\data\CacheHDF5\TickH5\commodity_future\DL'
        history_data = _get_history_data_csharp_org(stockfullcode=stockfullcode, start_date=start_date,
                                                    end_date=end_date,
                                                    freq=freq, cache_path=cache_path)
        pass

    def test_get_history_data_csharp(self):
        stockfullcode = 'ZZMA9001'
        start_date = '2021-01-01'
        end_date = '2021-12-31'
        freq = '1m'
        ed_type = None
        cache_path_local = r'D:\data\KlineH5\commodity_future\ZZ'
        cache_path_nas = 'nas'
        history_data_local = get_history_data_csharp(stockfullcode=stockfullcode, start_date=start_date, end_date=end_date,
                                               freq=freq, ed_type=ed_type, cache_path=cache_path_local)
        history_data_nas = get_history_data_csharp(stockfullcode=stockfullcode, start_date=start_date, end_date=end_date,
                                                     freq=freq, ed_type=ed_type, cache_path=cache_path_nas)
        zq_history_data = convert_da2zq_data(data_df=history_data_local, freq=freq, stockfullcode=stockfullcode)
        tday_history_data = convert_dt_to_tday(history_data_local)

        legal_history_data = get_legal_period_data(data_df=history_data, stockfullcode=stockfullcode, freq=freq)
        pass

    def test_convert_freq2text(self):
        freq = '1M'
        freq_text = _convert_freq2text(freq_str=freq)
        pass

    def test_get_cache_h5_nodes(self):
        freq = 'tick'
        tick_nodes = _get_cache_h5_nodes(period=freq, start_date='2010-01-01', end_date='2020-12-31')
        kline_nodes = _get_cache_h5_nodes(period='1m', start_date='2010-01-01', end_date='2020-12-31')
        pass

    def test_get_history_data_csharp_tick1(self):
        freq = 'tick'
        start_tday = '2021-01-01'
        end_tday = '2021-03-25'
        stockfullcode = 'DLi2106'
        cache_path = 'D:\TempCSV\jq_data\data\CacheHDF5\TickH5\commodity_future\DL'
        tick_df = get_history_data_csharp(freq=freq, start_date=start_tday, end_date=end_tday,
                                          stockfullcode=stockfullcode, cache_path=cache_path)
        pass

    def test_compare_history_data_csharp(self):
        stockfullcode = 'ZZAP9001'
        start_date = '2021-01-01'
        end_date = '2021-12-31'
        freq = '1m'
        ed_type = None
        cache_path = r'D:\TempCSV\jq_data\data\CacheHDF5\KlineH5\index_future\SF'
        history_data = get_history_data_csharp(stockfullcode=stockfullcode, start_date=start_date, end_date=end_date,
                                               freq=freq, ed_type=ed_type, cache_path=cache_path)

        history_data_dt = history_data.copy()
        history_data_dt['DateTrade'] = pd.to_datetime(history_data_dt['DateTrade']).dt.strftime("%Y%m%d%H%M%S").astype(
            np.float)

        pass

    def test_compare_wstock_and_jq_data(self):
        list_stockfullcode = ['ZZMA9001']
        list_freq = ['1m', '1d']
        start_date = '2021-05-20'
        end_date = '2021-05-20'
        jq_cache_path = r'D:\TempCSV\jq_data\data\CacheHDF5\KlineH5\commodity_future\ZZ'
        for stockfullcode in list_stockfullcode:
            for freq in list_freq:
                wstock_data = get_history_data_csharp(stockfullcode=stockfullcode, freq=freq, start_date=start_date,
                                                      end_date=end_date, cache_path='nas')
                jq_data = get_history_data_csharp(stockfullcode=stockfullcode, freq=freq, start_date=start_date,
                                                  end_date=end_date, cache_path=jq_cache_path)
                pass


if __name__ == "__main__":
    unittest.main()
