# -*- coding:utf-8 -*-
import unittest
import datetime as dt
import pandas as pd
from rio3.log.setup import setup_screen_logger
from sqlalchemy import PrimaryKeyConstraint, func, literal_column, Table
from sqlalchemy import create_engine, Column, Integer, String, Text, ForeignKey, Index, Date, DateTime, DECIMAL, \
    Boolean, BigInteger, select, case, Float, distinct, BINARY, text, and_
from sqlalchemy.ext import compiler
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import relationship, Session, aliased, Query
from sqlalchemy.schema import DDLElement
from sqlalchemy.sql import table
from sqlalchemy.dialects.mysql import TINYINT
from rio3.utils.num import float_equal, round
from rio3.quote_core import quote_info
from sqlalchemy.dialects.mssql import BIT, SMALLDATETIME, MONEY

Base = declarative_base()
metadata = Base.metadata


class CreateView(DDLElement):
    def __init__(self, name, selectable):
        self.name = name
        self.selectable = selectable


class DropView(DDLElement):
    def __init__(self, name):
        self.name = name


@compiler.compiles(CreateView)
def compile(element, compiler, **kw):
    return "CREATE VIEW %s AS %s" % (element.name, compiler.sql_compiler.process(element.selectable))


@compiler.compiles(DropView)
def compile(element, compiler, **kw):
    return "DROP VIEW %s" % (element.name)


def view(name, metadata, selectable):
    t = table(name)

    if isinstance(selectable, Query):
        selectable = selectable.subquery()

    for c in selectable.c:
        c._make_proxy(t)

    CreateView(name, selectable).execute_at('after-create', metadata)
    DropView(name).execute_at('before-drop', metadata)
    return t


class Strategy(Base):
    __tablename__ = 'lookup_strategy'
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    name = Column(String(50), primary_key=False, nullable=False)
    alias = Column(String(10), primary_key=False, nullable=True)


class Market(Base):
    __tablename__ = 'lookup_marketid'
    id_market = Column(Integer, primary_key=True, nullable=False, autoincrement=False)
    market_name = Column(String(16), nullable=False)
    memo = Column(Text(1024), nullable=True)

    def __repr__(self):
        return u"<LookUpMarketId(id_market={}, market_name='{}', memo='{}')>".format(self.id_market, self.market_name,
                                                                                     self.memo)

    def __str__(self):
        return "id_market={}, market_name={}".format(self.id_market, self.market_name)

    def __eq__(self, other):
        assert isinstance(other, Market)
        if self.id_market == other.id_market and self.market_name == other.market_name:
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


class Stock(Base):
    __tablename__ = 'lookup_stockid'
    id_stock = Column(Integer, primary_key=True, autoincrement=False)
    id_market = Column(Integer, ForeignKey('lookup_marketid.id_market'), nullable=False)
    stockcode = Column(String(16), primary_key=False, nullable=False)
    stockfullcode = Column(String(16), primary_key=False, unique=True, nullable=False)
    stockname = Column(String(64), primary_key=False, nullable=True)

    market = relationship('Market')

    __table_args__ = (
        PrimaryKeyConstraint('id_stock'),
    )

    def __repr__(self):
        return u"<LookUpStockId(id_stock={}, id_market={}, stockcode='{}', stockfullcode='{}', stockname='{}')>".format(
            self.id_stock, self.id_market, self.stockcode, self.stockfullcode, self.stockname)

    def __str__(self):
        return "id_stock={}, stockfullcode={}".format(
            self.id_stock, self.stockfullcode)

    def __eq__(self, other):
        if self.id_stock == other.id_stock and self.id_market == other.id_market and self.stockcode == other.stockcode \
                and self.stockfullcode == other.stockfullcode:
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


Market.stocks = relationship('Stock', order_by=Stock.stockcode)

_stock_view = view("VW_lookup_stockid", Base.metadata,
                   select([
                       Stock.id_stock.label('id_stock'),
                       Stock.id_market.label('id_market'),
                       Stock.stockfullcode.label('stockfullcode'),
                       Stock.stockcode.label('stockcode'),
                       Stock.stockname.label('stockname'),
                       Market.market_name.label('market_name')]). \
                   select_from(
                       Stock.__table__.join(Market, Stock.id_market == Market.id_market)
                   ))


class StockView(Base):
    __table__ = _stock_view
    __table_args__ = (
        PrimaryKeyConstraint('id_stock'),
    )

    def __repr__(self):
        return "<StockView(id_stock={}, id_market={}, market_name='{}', stockcode='{}', " \
               "stockfullcode='{}', stockname='{}')>".format(
            self.id_stock, self.id_market, self.market_name, self.stockcode, self.stockfullcode, self.stockname)

    def __str__(self):
        return "market_name='{}', stockcode='{}', stockfullcode='{}'".format(
            self.market_name, self.stockcode, self.stockfullcode)

    def __eq__(self, other):
        if self.id_stock == other.id_stock and self.id_market == other.id_market and self.stockcode == other.stockcode and \
                self.stockfullcode == other.stockfullcode and self.market_name == other.market_name:
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


class Tday(Base):
    __tablename__ = 'lookup_tday'
    tday = Column(Date, primary_key=True)

    def __repr__(self):
        return "<Tday(tday='{}')>".format(self.tday.strftime('%Y%m%d'))

    def __str__(self):
        return "{}".format(self.tday.strftime('%Y%m%d'))

    def __eq__(self, other):
        return self.tday == other.tday

    def __ne__(self, other):
        return not self.__eq__(other)


class Holiday(Base):
    __tablename__ = 'lookup_holiday'
    holiday = Column(Date, primary_key=True)

    def __repr__(self):
        return "<Holiday(holiday='{}')>".format(self.holiday.strftime('%Y%m%d'))

    def __str__(self):
        return "{}".format(self.holiday.strftime('%Y%m%d'))

    def __eq__(self, other):
        return self.holiday == other.holiday

    def __ne__(self, other):
        return not self.__eq__(other)


class EdInfo(Base):
    __tablename__ = 'ed_info'
    id_stock = Column(Integer, ForeignKey('lookup_stockid.id_stock'), primary_key=True)
    date_ed = Column(DateTime, primary_key=True)
    fhbl = Column(DECIMAL(5, 3), nullable=False, default=0)
    pgbl = Column(DECIMAL(5, 3), nullable=False, default=0)
    pgjg = Column(DECIMAL(10, 3), nullable=False, default=0)
    sgbl = Column(DECIMAL(18, 12), nullable=False, default=0)
    flg_calc_factor = Column(Boolean, nullable=False, default=False)
    id_main_stock = Column(Integer, ForeignKey('lookup_stockid.id_stock'), nullable=True)
    # __table_args__ = (Index('idx_edinfo_idstock_dateed', "id_stock", "date_ed", unique=True),)

    stock = relationship('Stock', foreign_keys=id_stock)
    stock_main = relationship('Stock', foreign_keys=id_main_stock)

    def __repr__(self):
        return "<EdInfo(stock='{}', stock_main='{}', date_ed='{}', fhbl={}, pgbl={}, pgjg={}, sgbl={})>".format(
            self.stock.stockfullcode, self.stock_main.stockfullcode, self.date_ed.strftime('%Y%m%d'),
            self.fhbl, self.pgbl, self.pgjg, self.sgbl)

    def __str__(self):
        return "stock='{}', stock_main='{}', date_ed='{}', fhbl={}, pgbl={}, pgjg={}, sgbl={}".format(
            self.stock.stockfullcode, self.stock_main.stockfullcode, self.date_ed.strftime('%Y%m%d'),
            self.fhbl, self.pgbl, self.pgjg, self.sgbl)

    def __eq__(self, other):
        assert isinstance(other, EdInfo)
        return self.id_stock == other.id_stock and self.date_ed == other.date_ed and float_equal(self.fhbl,
                                                                                                 other.fhbl) and \
               float_equal(self.pgbl, other.pgbl) and float_equal(self.pgjg, other.pgjg) and float_equal(self.sgbl,
                                                                                                         other.sgbl) and \
               self.flg_calc_factor == other.flg_calc_factor and self.id_main_stock == other.id_main_stock

    def __ne__(self, other):
        return not self.__eq__(other)


class EdInfoEbs(Base):
    __tablename__ = 'ed_info_ebs'

    id_stock = Column(ForeignKey(u'lookup_stockid.id_stock'), nullable=False, index=True)
    date_record = Column(DateTime, nullable=False)
    date_execute = Column(DateTime, nullable=False, index=True)
    fhbl = Column(DECIMAL(20, 8), nullable=False)
    pgbl = Column(DECIMAL(20, 8), nullable=False)
    pgjg = Column(DECIMAL(20, 8), nullable=False)
    sgbl = Column(DECIMAL(20, 8), nullable=False)
    date_update = Column(DateTime, nullable=False)
    object_id = Column(String(255), primary_key=True, unique=True, server_default=text("''"))
    rec_type = Column(TINYINT(1), nullable=False)
    is_del = Column(BINARY(1), nullable=False, server_default=text("'0'"))

    lookup_stockid = relationship('Stock', foreign_keys=id_stock)

    def __repr__(self):
        return "<EdInfoEbs(stock='{}', date_execute='{}', fhbl={}, pgbl={}, pgjg={}, sgbl={})>".format(
            self.lookup_stockid.stockfullcode, self.date_execute.strftime('%Y%m%d'),
            self.fhbl, self.pgbl, self.pgjg, self.sgbl)

    def __str__(self):
        return "stock='{}', date_execute='{}', fhbl={}, pgbl={}, pgjg={}, sgbl={}".format(
            self.lookup_stockid.stockfullcode, self.date_execute.strftime('%Y%m%d'),
            self.fhbl, self.pgbl, self.pgjg, self.sgbl)

    def __eq__(self, other):
        assert isinstance(other, EdInfoEbs)
        return self.id_stock == other.id_stock and self.date_execute == other.date_execute and \
               self.date_record == other.date_record and float_equal(self.fhbl, other.fhbl) and \
               float_equal(self.pgbl, other.pgbl) and float_equal(self.pgjg, other.pgjg) and \
               float_equal(self.sgbl, other.sgbl) and self.date_update == other.date_update and \
               self.object_id == other.object_id and self.rec_type == other.rec_type and \
               self.is_del == other.is_del

    def __ne__(self, other):
        return not self.__eq__(other)


class EdExecuteInfo(Base):
    __tablename__ = 'ed_execute_info'
    id_stock = Column(Integer, ForeignKey('lookup_stockid.id_stock'), primary_key=True)
    date_ed = Column(DateTime, primary_key=True)
    date_execute = Column(DateTime, nullable=False)
    val_preclose = Column(DECIMAL(10, 4), nullable=True, default=0)
    # __table_args__ = (Index('idx_edexecuteinfo_idstock_dateed', "id_stock", "date_ed", unique=True),)

    stock = relationship('Stock', foreign_keys=[id_stock])

    def __repr__(self):
        return "<EdExecuteInfo(stock='{}', date_ed='{}', date_execute='{}')>".format(
            self.stock.stockfullcode, self.date_ed.strftime('%Y%m%d'), self.date_execute.strftime('%Y%m%d'))

    def __str__(self):
        return "stock='{}', date_ed='{}', date_execute='{}'".format(
            self.stock.stockfullcode, self.date_ed.strftime('%Y%m%d'), self.date_execute.strftime('%Y%m%d'))

    def __eq__(self, other):
        assert isinstance(other, EdExecuteInfo)
        return self.id_stock == other.id_stock and self.date_ed == other.date_ed and \
               self.date_execute == other.date_execute

    def __ne__(self, other):
        return not self.__eq__(other)


class EdFactor(Base):
    __tablename__ = 'ed_factor'
    __table_args__ = (
        Index('ed_factor_stock_exec_date_unique_idx', 'id_stock', 'date_execute'),
    )

    id_stock = Column(ForeignKey(u'lookup_stockid.id_stock'), primary_key=True, nullable=False, index=True)
    id_main_stock = Column(ForeignKey(u'lookup_stockid.id_stock'), index=True)
    date_execute = Column(Date, primary_key=True, nullable=False)
    factor = Column(DECIMAL(20, 8), nullable=False)
    is_valid = Column(TINYINT(4), nullable=False)

    lookup_stockid = relationship(u'Stock', primaryjoin='EdFactor.id_main_stock == Stock.id_stock')
    lookup_stockid1 = relationship(u'Stock', primaryjoin='EdFactor.id_stock == Stock.id_stock')


class EdFactorInfo(Base):
    __tablename__ = 'ed_factor_info'
    __table_args__ = (
        Index('ed_factor_info_stock_date_execute_unique', 'id_stock', 'date_execute'),
    )

    id_stock = Column(ForeignKey(u'lookup_stockid.id_stock'), primary_key=True, nullable=False, index=True)
    id_main_stock = Column(ForeignKey(u'lookup_stockid.id_stock'), index=True)
    date_execute = Column(Date, primary_key=True, nullable=False)
    fhbl = Column(DECIMAL(20, 8), nullable=False)
    pgbl = Column(DECIMAL(20, 8), nullable=False)
    pgjg = Column(DECIMAL(20, 8), nullable=False)
    sgbl = Column(DECIMAL(20, 8), nullable=False)
    is_valid = Column(TINYINT(1), nullable=False)

    lookup_stockid = relationship(u'Stock', primaryjoin='EdFactorInfo.id_main_stock == Stock.id_stock')
    lookup_stockid1 = relationship(u'Stock', primaryjoin='EdFactorInfo.id_stock == Stock.id_stock')


stock_alias1 = aliased(Stock)
stock_alias2 = aliased(Stock)

_edinfo_view = view(
    "VW_edinfo", Base.metadata,
    select([EdInfo.id_stock.label('id_stock'),
            stock_alias1.stockfullcode.label('stockfullcode'),
            case([(stock_alias2.id_stock != None, stock_alias2.id_stock), ],
                 else_=stock_alias1.id_stock).label('id_main_stock'),
            case([(stock_alias2.stockfullcode != None, stock_alias2.stockfullcode), ],
                 else_=stock_alias1.stockfullcode).label('stockfullcode_main'),
            EdInfo.date_ed.label('date_ed'),
            case([(EdExecuteInfo.date_execute != None, EdExecuteInfo.date_execute), ],
                 else_=EdInfo.date_ed).label('date_execute'),
            EdInfo.fhbl.label('fhbl'), EdInfo.pgbl.label('pgbl'), EdInfo.pgjg.label('pgjg'),
            EdInfo.sgbl.label('sgbl'), EdInfo.flg_calc_factor.label('flg_calc_factor'),
            EdExecuteInfo.val_preclose.label('val_preclose')
            ]). \
        select_from(
        EdInfo.__table__. \
            outerjoin(stock_alias1, EdInfo.id_stock == stock_alias1.id_stock). \
            outerjoin(stock_alias2, EdInfo.id_main_stock == stock_alias2.id_stock). \
            outerjoin(EdExecuteInfo,
                      and_(EdInfo.id_stock == EdExecuteInfo.id_stock, EdInfo.date_ed == EdExecuteInfo.date_ed))
    )
)


class EdInfoView(Base):
    __table__ = _edinfo_view
    __table_args__ = (
        PrimaryKeyConstraint('id_stock', 'date_ed'),
    )

    def __repr__(self):
        return "<EdInfoView(stock='{}', stock_main='{}', date_ed='{}', date_execute='{}', fhbl={}, pgbl={}, pgjg={}, sgbl={})>".format(
            self.stockfullcode, self.stockfullcode_main, self.date_ed.strftime('%Y%m%d'),
            self.date_execute.strftime('%Y%m%d'), self.fhbl, self.pgbl, self.pgjg, self.sgbl)

    def __str__(self):
        return "stock='{}', stock_main='{}', date_ed='{}', date_execute='{}', fhbl={}, pgbl={}, pgjg={}, sgbl={}".format(
            self.stockfullcode, self.stockfullcode_main, self.date_ed.strftime('%Y%m%d'),
            self.date_execute.strftime('%Y%m%d'), self.fhbl, self.pgbl, self.pgjg, self.sgbl)

    def __eq__(self, other):
        assert isinstance(other, EdInfoView)
        return self.id_stock == other.id_stock and self.date_ed == other.date_ed and float_equal(self.fhbl,
                                                                                                 other.fhbl) and \
               float_equal(self.pgbl, other.pgbl) and float_equal(self.pgjg, other.pgjg) and float_equal(self.sgbl,
                                                                                                         other.sgbl) and \
               self.flg_calc_factor == other.flg_calc_factor and self.id_main_stock == other.id_main_stock and \
               self.date_execute == other.date_execute

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_ed_info(self):
        ed_info = quote_info.EdInfo(
            name=self.stockfullcode,
            date_ed=self.date_ed,
            date_execute=self.date_execute,
            stockfullcode_main=self.stockfullcode_main,
            fhbl=self.fhbl,
            pgbl=self.pgbl,
            pgjg=self.pgjg,
            sgbl=self.sgbl,
            id_stock=self.id_stock,
            id_stock_main=self.id_main_stock
        )
        return ed_info


_edinfo_ebs_view = view(
    "VW_edinfo_ebs", Base.metadata,
    select([EdInfoEbs.id_stock.label('id_stock'),
            stock_alias1.stockfullcode.label('stockfullcode'),
            EdInfoEbs.date_record.label('date_record'),
            EdInfoEbs.date_execute.label('date_execute'),
            EdInfoEbs.fhbl.label('fhbl'),
            EdInfoEbs.pgbl.label('pgbl'),
            EdInfoEbs.pgjg.label('pgjg'),
            EdInfoEbs.sgbl.label('sgbl'),
            EdInfoEbs.rec_type.label('rec_type'),
            EdInfoEbs.object_id.label('object_id')]).
        select_from(
        EdInfoEbs.__table__.outerjoin(stock_alias1, EdInfoEbs.id_stock == stock_alias1.id_stock)).
        where(EdInfoEbs.is_del == 0)
)


class EdInfoEbsView(Base):
    __table__ = _edinfo_ebs_view
    __table_args__ = (
        PrimaryKeyConstraint('object_id'),
    )

    def __repr__(self):
        return "<EdInfoEbsView(stock='{}', date_record='{}', date_execute='{}', " \
               "fhbl={}, pgbl={}, pgjg={}, sgbl={}, rec_type={}, object_id={})>".format(
            self.stockfullcode, self.date_record.strftime('%Y%m%d'), self.date_execute.strftime('%Y%m%d'),
            self.fhbl, self.pgbl, self.pgjg, self.sgbl, self.rec_type, self.object_id)

    def __str__(self):
        return "stock='{}', date_record='{}', date_execute='{}', " \
               "fhbl={}, pgbl={}, pgjg={}, sgbl={}, rec_type={}, object_id={}".format(
            self.stockfullcode, self.date_record.strftime('%Y%m%d'), self.date_execute.strftime('%Y%m%d'),
            self.fhbl, self.pgbl, self.pgjg, self.sgbl, self.rec_type, self.object_id)

    def __eq__(self, other):
        assert isinstance(other, EdInfoEbsView)
        return self.id_stock == other.id_stock and self.date_record == other.date_record and \
               self.date_execute == other.date_execute and float_equal(self.fhbl, other.fhbl) and \
               float_equal(self.pgbl, other.pgbl) and float_equal(self.pgjg, other.pgjg) and \
               float_equal(self.sgbl, other.sgbl) and self.rec_type == other.rec_type and \
               self.object_id == other.object_id

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_ed_info(self):
        ed_info = quote_info.EdInfoEbs(
            id_stock=self.id_stock,
            stockfullcode=self.stockfullcode,
            date_record=self.date_record,
            date_execute=self.date_execute,
            fhbl=self.fhbl,
            pgbl=self.pgbl,
            pgjg=self.pgjg,
            sgbl=self.sgbl,
            rec_type=self.rec_type,
            object_id=self.object_id
        )
        return ed_info


class FutureInfo(Base):
    __tablename__ = 'lookup_future_info'
    id_stock = Column(Integer, ForeignKey('lookup_stockid.id_stock'), primary_key=True, nullable=False)
    symbol = Column(String(16), primary_key=False, nullable=True)
    name = Column(String(20), primary_key=False, nullable=True)
    open_date = Column(Date, primary_key=False, nullable=True)
    expire_date = Column(Date, primary_key=False, nullable=True)
    start_deliver_date = Column(Date, primary_key=False, nullable=True)
    end_deliver_date = Column(Date, primary_key=False, nullable=True)
    long_margin_ratio = Column(DECIMAL(10, 3), primary_key=False, nullable=True)
    short_margin_ratio = Column(DECIMAL(10, 3), primary_key=False, nullable=True)
    volume_multiple = Column(Integer, primary_key=False, nullable=True)
    create_date = Column(Date, primary_key=False, nullable=True)
    max_limit_order_volume = Column(Integer, primary_key=False, nullable=True)
    min_limit_order_volume = Column(Integer, primary_key=False, nullable=True)
    max_market_order_volume = Column(Integer, primary_key=False, nullable=True)
    min_market_order_volume = Column(Integer, primary_key=False, nullable=True)
    price_tick = Column(DECIMAL(10, 3), primary_key=False, nullable=True)
    update_date = Column(DateTime, primary_key=False, nullable=False)

    __table_args__ = (
        PrimaryKeyConstraint('id_stock'),
    )

    def __repr__(self):
        return u"<StockInfo(id_stock={}, symbol='{}', name='{}', open_date='{}', expire_date='{}', " \
               u"start_deliver_date='{}', end_deliver_date='{}', long_margin_ratio={}, short_margin_ratio={}, " \
               u"volume_multiple={}, create_date='{}', max_limit_order_volume={}, min_limit_order_volume={}, " \
               u"max_market_order_volume={}, min_market_order_volume={}, price_tick={}, update_date='{}')>".format(
            self.id_stock, self.symbol, self.name, self.open_date.strftime('%Y%m%d'),
            self.expire_date.strftime('%Y%m%d'),
            self.start_deliver_date.strftime('%Y%m%d'), self.end_deliver_date.strftime('%Y%m%d'),
            self.long_margin_ratio,
            self.short_margin_ratio, self.volume_multiple, self.create_date.strftime('%Y%m%d'),
            self.max_limit_order_volume,
            self.min_limit_order_volume, self.max_market_order_volume, self.min_market_order_volume, self.price_tick,
            self.update_date.strftime('%Y%m%d %H:%M:%S'))

    def __str__(self):
        return u"" \
               u"id_stock={}, 品种={}, 合约名={}, 上市日={}, 到期日={}, " \
               u"开始交割日={}, 结束交割日={}, 多头保证金率={}, 到期日={}, " \
               u"乘数={}, create_date={}, 限价单最大下单量={}, 限价单最小下单量={}, " \
               u"市价单最大下单量={}, 市价单最小下单量={}, 最小变动价位={}, 更新日期={}".format(
            self.id_stock, self.symbol, self.name, self.open_date.strftime('%Y%m%d'),
            self.expire_date.strftime('%Y%m%d'),
            self.start_deliver_date.strftime('%Y%m%d'), self.end_deliver_date.strftime('%Y%m%d'),
            self.long_margin_ratio,
            self.short_margin_ratio, self.volume_multiple, self.create_date.strftime('%Y%m%d'),
            self.max_limit_order_volume,
            self.min_limit_order_volume, self.max_market_order_volume, self.min_market_order_volume, self.price_tick,
            self.update_date.strftime('%Y%m%d %H:%M:%S'))

    def __eq__(self, other):
        if self.id_stock == other.id_stock and self.symbol == other.symbol and self.open_date == other.open_date \
                and self.expire_date == other.expire_date and self.start_deliver_date == other.start_deliver_date \
                and self.end_deliver_date == other.end_deliver_date and self.long_margin_ratio == other.long_margin_ratio \
                and self.short_margin_ratio == other.short_margin_ratio and self.volume_multiple == other.volume_multiple \
                and self.create_date == other.create_date and self.max_limit_order_volume == other.max_limit_order_volume \
                and self.min_limit_order_volume == other.min_limit_order_volume \
                and self.max_market_order_volume == other.max_market_order_volume \
                and self.min_market_order_volume == other.min_market_order_volume and self.price_tick == other.price_tick:
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


_futureinfo_view = view(
    "VW_lookup_future_info", Base.metadata,
    select([FutureInfo.id_stock.label('id_stock'),
            Stock.stockfullcode.label('stockfullcode'),
            FutureInfo.symbol.label('symbol'),
            FutureInfo.name.label('name'),
            FutureInfo.create_date.label('create_date'),
            FutureInfo.open_date.label('open_date'),
            FutureInfo.expire_date.label('expire_date'),
            FutureInfo.start_deliver_date.label('start_deliver_date'),
            FutureInfo.end_deliver_date.label('end_deliver_date'),
            FutureInfo.long_margin_ratio.label('long_margin_ratio'),
            FutureInfo.short_margin_ratio.label('short_margin_ratio'),
            FutureInfo.volume_multiple.label('volume_multiple'),
            FutureInfo.max_limit_order_volume.label('max_limit_order_volume'),
            FutureInfo.min_limit_order_volume.label('min_limit_order_volume'),
            FutureInfo.max_market_order_volume.label('max_market_order_volume'),
            FutureInfo.min_market_order_volume.label('min_market_order_volume'),
            FutureInfo.price_tick.label('price_tick'),
            FutureInfo.update_date.label('update_date'),
            ]). \
        select_from(
        FutureInfo.__table__.outerjoin(Stock, FutureInfo.id_stock == Stock.id_stock))
)


class FutureInfoView(Base):
    __table__ = _futureinfo_view
    __table_args__ = (
        PrimaryKeyConstraint('id_stock'),
    )

    def __repr__(self):
        return u"<FutureInfoView(id_stock={}, stockfullcode='{}', symbol='{}', name='{}', open_date='{}', expire_date='{}', " \
               u"start_deliver_date='{}', end_deliver_date='{}', long_margin_ratio={}, short_margin_ratio={}, " \
               u"volume_multiple={}, create_date='{}', max_limit_order_volume={}, min_limit_order_volume={}, " \
               u"max_market_order_volume={}, min_market_order_volume={}, price_tick={}, update_date='{}')>".format(
            self.id_stock, self.stockfullcode, self.symbol, self.name, self.open_date.strftime('%Y%m%d'),
            self.expire_date.strftime('%Y%m%d'),
            self.start_deliver_date.strftime('%Y%m%d'), self.end_deliver_date.strftime('%Y%m%d'),
            self.long_margin_ratio,
            self.short_margin_ratio, self.volume_multiple, self.create_date.strftime('%Y%m%d'),
            self.max_limit_order_volume,
            self.min_limit_order_volume, self.max_market_order_volume, self.min_market_order_volume, self.price_tick,
            self.update_date.strftime('%Y%m%d %H:%M:%S')
        )

    def __str__(self):
        return u"id_stock={}, symbol={}, name={}, open_date={}, expire_date={}, start_deliver_date={}, " \
               u"end_deliver_date={}, long_margin_ratio={}, short_margin_ratio={}, volume_multiple={}, " \
               u"create_date={}, max_limit_order_volume={}, min_limit_order_volume={}, max_market_order_volume={}, " \
               u"min_market_order_volume={}, price_tick={}, update_date={}, stockfullcode={}".format(
            self.id_stock, self.symbol, self.name, self.open_date.strftime('%Y%m%d'),
            self.expire_date.strftime('%Y%m%d'),
            self.start_deliver_date.strftime('%Y%m%d'), self.end_deliver_date.strftime('%Y%m%d'),
            self.long_margin_ratio,
            self.short_margin_ratio, self.volume_multiple, self.create_date.strftime('%Y%m%d'),
            self.max_limit_order_volume,
            self.min_limit_order_volume, self.max_market_order_volume, self.min_market_order_volume, self.price_tick,
            self.update_date.strftime('%Y%m%d %H:%M:%S'), self.stockfullcode
        )

    def __eq__(self, other):
        if self.id_stock == other.id_stock and self.symbol == other.symbol \
                and pd.to_datetime(self.open_date) == pd.to_datetime(other.open_date) \
                and pd.to_datetime(self.expire_date) == pd.to_datetime(other.expire_date) \
                and pd.to_datetime(self.start_deliver_date) == pd.to_datetime(other.start_deliver_date) \
                and pd.to_datetime(self.end_deliver_date) == pd.to_datetime(other.end_deliver_date) \
                and float_equal(self.long_margin_ratio, other.long_margin_ratio) \
                and float_equal(self.short_margin_ratio, other.short_margin_ratio) \
                and self.volume_multiple == other.volume_multiple \
                and pd.to_datetime(self.create_date) == pd.to_datetime(other.create_date) \
                and self.max_limit_order_volume == other.max_limit_order_volume \
                and self.min_limit_order_volume == other.min_limit_order_volume \
                and self.max_market_order_volume == other.max_market_order_volume \
                and self.min_market_order_volume == other.min_market_order_volume \
                and float_equal(self.price_tick, other.price_tick):
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


_ed_factor_info_view = view(
    "VW_ed_factor_info", Base.metadata,
    select([EdFactorInfo.id_stock.label('id_stock'),
            stock_alias1.stockfullcode.label('stockfullcode'),
            case([(EdFactorInfo.id_main_stock != None, EdFactorInfo.id_main_stock), ],
                 else_=EdFactorInfo.id_stock).label('id_main_stock'),
            case([(EdFactorInfo.id_main_stock != None, stock_alias2.stockfullcode), ],
                 else_=stock_alias1.stockfullcode).label('stockfullcode_main'),
            EdFactorInfo.date_execute.label('date_execute'),
            EdFactorInfo.fhbl.label('fhbl'),
            EdFactorInfo.pgbl.label('pgbl'),
            EdFactorInfo.pgjg.label('pgjg'),
            EdFactorInfo.sgbl.label('sgbl'),
            EdFactorInfo.is_valid.label('is_valid')]).
        select_from(
        EdFactorInfo.__table__.outerjoin(stock_alias1, EdFactorInfo.id_stock == stock_alias1.id_stock).
            outerjoin(stock_alias2, EdFactorInfo.id_main_stock == stock_alias2.id_stock)
    )
)


class EdFactorInfoView(Base):
    __table__ = _ed_factor_info_view
    __table_args__ = (
        PrimaryKeyConstraint('id_stock', 'date_execute'),
    )


_ed_factor_view = view(
    "VW_ed_factor", Base.metadata,
    select([EdFactor.id_stock.label('id_stock'),
            stock_alias1.stockfullcode.label('stockfullcode'),
            case([(EdFactor.id_main_stock != None, EdFactor.id_main_stock), ],
                 else_=EdFactor.id_stock).label('id_main_stock'),
            case([(EdFactor.id_main_stock != None, stock_alias2.stockfullcode), ],
                 else_=stock_alias1.stockfullcode).label('stockfullcode_main'),
            EdFactor.date_execute.label('date_execute'),
            EdFactor.factor.label('factor'),
            EdFactor.is_valid.label('is_valid')]).
        select_from(
        EdFactor.__table__.outerjoin(stock_alias1, EdFactor.id_stock == stock_alias1.id_stock).
            outerjoin(stock_alias2, EdFactor.id_main_stock == stock_alias2.id_stock)
    )
)


class EdFactorView(Base):
    __table__ = _ed_factor_view
    __table_args__ = (
        PrimaryKeyConstraint('id_stock', 'date_execute'),
    )


class TdaySymbol(Base):
    __tablename__ = 'lookup_tday_symbol_id'
    id = Column(BigInteger, primary_key=True, nullable=False, autoincrement=True)
    tday = Column(Date, nullable=True)
    symbol = Column(String(16), primary_key=False, nullable=True)

    def __repr__(self):
        return "<TdaySymbol(id={}, tday='{}', symbol='{}')>".format(self.id, self.tday.strftime('%Y-%m-%d'),
                                                                    self.symbol)

    def __str__(self):
        return "id={}, tday={}, symbol={}".format(self.id, self.tday.strftime('%Y-%m-%d'), self.symbol)

    def __eq__(self, other):
        assert isinstance(other, TdaySymbol)
        return self.id == other.id and self.tday == other.tday and self.symbol == other.symbol

    def __ne__(self, other):
        return not self.__eq__(other)


class TdaySymbolStocks(Base):
    __tablename__ = 'lookup_tday_symbol_stocks'
    id = Column(BigInteger, primary_key=True, nullable=False, autoincrement=True)
    id_tday_symbol = Column(BigInteger, ForeignKey('lookup_tday_symbol_id.id'),
                            primary_key=True, nullable=False, autoincrement=False)
    id_stock = Column(Integer, ForeignKey('lookup_stockid.id_stock'),
                      primary_key=True, nullable=False, autoincrement=False)
    val_OI = Column(BigInteger, nullable=False, default=0)

    def __eq__(self, other):
        assert isinstance(other, TdaySymbolStocks)
        return self.id == other.id and self.id_tday_symbol == other.id_tday_symbol \
               and self.id_stock == other.id_stock and self.val_OI == other.val_OI

    def __ne__(self, other):
        return not self.__eq__(other)


_tday_symbol_stocks_view = view('vw_lookup_tday_symbol_stocks', Base.metadata,
                                select([
                                    TdaySymbolStocks.id.label('id'),
                                    TdaySymbolStocks.id_stock.label('id_stock'),
                                    TdaySymbol.tday.label('tday'),
                                    Market.market_name.label('market_name'),
                                    Stock.stockfullcode.label('stockfullcode'),
                                    Stock.stockcode.label('stockcode'),
                                    TdaySymbol.symbol.label('symbol'),
                                    TdaySymbolStocks.val_OI.label('val_OI')
                                ]).select_from(
                                    TdaySymbolStocks.__table__.
                                        outerjoin(Stock, TdaySymbolStocks.id_stock == Stock.id_stock).
                                        outerjoin(TdaySymbol, TdaySymbolStocks.id_tday_symbol == TdaySymbol.id).
                                        outerjoin(Market, Stock.id_market == Market.id_market)
                                ))


class TdaySymbolStocksView(Base):
    __table__ = _tday_symbol_stocks_view
    __mapper_args__ = {
        'primary_key': [_tday_symbol_stocks_view.c.tday, _tday_symbol_stocks_view.c.stockfullcode]
    }
    __table_args__ = (
        PrimaryKeyConstraint('tday', 'stockfullcode'),
    )

    def __repr__(self):
        return "<TdaySymbolStocksView(id={}, id_stock='{}', tday='{}' stockfullcode='{}', stockcode='{}', symbol='{}', val_OI='{}')>".format(
            self.id, self.id_stock, self.tday.strftime('%Y-%m-%d'), self.stockfullcode, self.stockcode, self.symbol,
            self.val_OI)

    def __str__(self):
        return "id={}, id_stock={}, tday='{}' stockfullcode='{}', stockcode='{}', symbol='{}', val_OI='{}' ".format(
            self.id, self.id_stock, self.tday.strftime('%Y-%m-%d'), self.stockfullcode, self.stockcode, self.symbol,
            self.val_OI)

    def __eq__(self, other):
        assert isinstance(other, TdaySymbolStocksView)
        return self.id == other.id and self.id_stock == other.id_stock and self.tday == other.tday \
               and self.stockfullcode == other.stockfullcode and self.stockcode == other.stockcode \
               and self.symbol == other.symbol and self.val_OI == other.val_OI

    def __ne__(self, other):
        return not self.__eq__(other)


class HisKLine(object):
    """
    历史数据库的基础K线结构
    """

    @declared_attr
    def id_stock(cls):
        return Column(Integer, ForeignKey('lookup_stockid.id_stock'), primary_key=True)

    date_trade = Column(DateTime, primary_key=True)
    tday = Column(Date, nullable=True)
    val_preclose = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_open = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_high = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_low = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_close = Column(DECIMAL(10, 4), nullable=False, default=0)
    amt_amount = Column(DECIMAL(19, 4), nullable=False, default=0)
    amt_volume = Column('amt_volumn', BigInteger, nullable=False, default=0)
    amt_wp = Column(BigInteger, nullable=False, default=0)
    amt_np = Column(BigInteger, nullable=False, default=0)
    val_OI = Column(BigInteger, nullable=False, default=0)
    val_sp1 = Column(DECIMAL(10, 4), nullable=True, default=0)
    val_sv1 = Column(BigInteger, nullable=True, default=0)
    val_bp1 = Column(DECIMAL(10, 4), nullable=True, default=0)
    val_bv1 = Column(BigInteger, nullable=True, default=0)
    val_mid = Column(DECIMAL(10, 4), nullable=True, default=0)
    val_midv = Column(DECIMAL(10, 4), nullable=True, default=0)

    @declared_attr
    def stock(cls):
        return relationship(Stock, primaryjoin=lambda: Stock.id_stock == cls.id_stock)


class HisKLineView(object):
    @property
    def date(self):
        return self.date_trade

    @property
    def tday(self):
        return self.tday

    @property
    def o(self):
        return self.val_open

    @property
    def h(self):
        return self.val_high

    @property
    def l(self):
        return self.val_low

    @property
    def c(self):
        return self.val_close

    @property
    def v(self):
        return self.amt_volumn

    @property
    def amt(self):
        return self.amt_amount

    @property
    def oi(self):
        return self.val_OI

    @property
    def sp1(self):
        return self.val_sp1

    @property
    def sv1(self):
        return self.val_sv1

    @property
    def bp1(self):
        return self.val_bp1

    @property
    def bv1(self):
        return self.val_bv1

    @property
    def mid(self):
        return self.val_mid

    @property
    def midv(self):
        return self.val_midv

    def to_quote_info(self):
        bqi = quote_info.BasicQuoteInfo(
            name=self.stockfullcode,
            date=self.date,
            tday=self.tday,
            o=self.val_open,
            h=self.val_high,
            l=self.val_low,
            c=self.val_close,
            amt=self.amt_amount,
            oi=self.val_OI,
            v=self.amt_volumn,
            sp1=self.val_sp1,
            sv1=self.val_sv1,
            bp1=self.val_bp1,
            bv1=self.val_bv1,
            mid=self.val_mid,
            midv=self.val_midv
        )
        return bqi


class KLine(object):
    """
    实时数据库的基础K线结构
    """
    @declared_attr
    def id_stock(cls):
        return Column(Integer, ForeignKey('lookup_stockid.id_stock'), primary_key=True)

    date_trade = Column(DateTime, primary_key=True)
    tday = Column(Date, nullable=True)
    val_preclose = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_open = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_high = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_low = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_close = Column(DECIMAL(10, 4), nullable=False, default=0)
    amt_amount = Column(DECIMAL(19, 4), nullable=False, default=0)
    amt_volume = Column('amt_volumn', BigInteger, nullable=False, default=0)
    amt_wp = Column(BigInteger, nullable=False, default=0)
    amt_np = Column(BigInteger, nullable=False, default=0)
    val_OI = Column(BigInteger, nullable=True, default=0)
    val_sp1 = Column(DECIMAL(10, 4), nullable=True, default=0)
    val_sv1 = Column(BigInteger, nullable=True, default=0)
    val_bp1 = Column(DECIMAL(10, 4), nullable=True, default=0)
    val_bv1 = Column(BigInteger, nullable=True, default=0)
    val_mid = Column(DECIMAL(10, 4), nullable=True, default=0)
    val_midv = Column(DECIMAL(10, 4), nullable=True, default=0)

    @declared_attr
    def stock(cls):
        return relationship(Stock, primaryjoin=lambda: Stock.id_stock == cls.id_stock)


class KLineView(object):
    @property
    def date(self):
        return self.date_trade

    @property
    def tday(self):
        return self.tday

    @property
    def o(self):
        return self.val_open

    @property
    def h(self):
        return self.val_high

    @property
    def l(self):
        return self.val_low

    @property
    def c(self):
        return self.val_close

    @property
    def v(self):
        return self.amt_volumn

    @property
    def amt(self):
        return self.amt_amount

    @property
    def oi(self):
        return self.val_OI

    @property
    def sp1(self):
        return self.val_sp1

    @property
    def sv1(self):
        return self.val_sv1

    @property
    def bp1(self):
        return self.val_bp1

    @property
    def bv1(self):
        return self.val_bv1

    @property
    def mid(self):
        return self.val_mid

    @property
    def midv(self):
        return self.val_midv

    def to_quote_info(self):
        bqi = quote_info.BasicQuoteInfo(
            name=self.stockfullcode,
            date=self.date,
            tday=self.tday,
            o=self.val_open,
            h=self.val_high,
            l=self.val_low,
            c=self.val_close,
            amt=self.amt_amount,
            oi=self.val_OI,
            v=self.amt_volumn,
            sp1=self.val_sp1,
            sv1=self.val_sv1,
            bp1=self.val_bp1,
            bv1=self.val_bv1,
            mid=self.val_mid,
            midv=self.val_midv
        )
        return bqi


class OneMinute(KLine, Base):
    __tablename__ = 'oneminute_history'


class FiveMinutes(KLine, Base):
    __tablename__ = 'fiveminutes_history'


class FifteenMinutes(KLine, Base):
    __tablename__ = 'fifteenminutes_history'


class Halfhour(KLine, Base):
    __tablename__ = 'halfhour_history'


class OneHour(KLine, Base):
    __tablename__ = 'onehour_history'


class OneDay(KLine, Base):
    __tablename__ = 'oneday_history'
    flg_st = Column(Boolean, nullable=False, default=False)


_oneday_view = view(
    "VW_oneday_history", Base.metadata,
    select([
        OneDay.id_stock.label('id_stock'),
        Market.market_name.label('market_name'),
        Stock.stockfullcode.label('stockfullcode'),
        Stock.stockcode.label('stockcode'),
        func.DATE(OneDay.date_trade).label('dateInfo'),
        OneDay.date_trade.label('date_trade'),
        # since tday is added based on new structure, tday infomation should come from datetrade. This is only for Oneday Kline
        case([(OneDay.tday != literal_column('0000-00-00') or OneDay.tday != None, OneDay.tday), ],
             else_=func.DATE(OneDay.date_trade)).label('tday'),
        OneDay.val_preclose.label('val_preclose'),
        OneDay.val_open.label('val_open'),
        OneDay.val_high.label('val_high'),
        OneDay.val_low.label('val_low'),
        OneDay.val_close.label('val_close'),
        OneDay.amt_amount.label('amt_amount'),
        OneDay.amt_volume.label('amt_volumn'),
        OneDay.amt_wp.label('amt_wp'),
        OneDay.amt_np.label('amt_np'),
        literal_column('0').label('val_turnover'),
        OneDay.flg_st.label('flg_st'),
        OneDay.val_OI.label('val_OI'),
        OneDay.val_preclose.label('dailyPreclose'),
        case([(OneDay.val_sp1 != literal_column('0') or OneDay.val_sp1 != None, OneDay.val_sp1), ],
             else_=OneDay.val_close).label('val_sp1'),
        case([(OneDay.val_sv1 != literal_column('0') or OneDay.val_sv1 != None, OneDay.val_sv1), ],
             else_=OneDay.amt_volume).label('val_sv1'),
        case([(OneDay.val_bp1 != literal_column('0') or OneDay.val_bp1 != None, OneDay.val_bp1), ],
             else_=OneDay.val_close).label('val_bp1'),
        case([(OneDay.val_bv1 != literal_column('0') or OneDay.val_bv1 != None, OneDay.val_bv1), ],
             else_=OneDay.amt_volume).label('val_bv1'),
        case([(OneDay.val_mid != literal_column('0') or OneDay.val_mid != None, OneDay.val_mid), ],
             else_=OneDay.val_close).label('val_mid'),
        case([(OneDay.val_midv != literal_column('0') or OneDay.val_midv != None, OneDay.val_midv), ],
             else_=OneDay.val_close).label('val_midv'),
    ]). \
        select_from(
        OneDay.__table__. \
            outerjoin(Stock, OneDay.id_stock == Stock.id_stock). \
            outerjoin(Market, Stock.id_market == Market.id_market)
    )
)


class OneDayView(KLineView, Base):
    __table__ = _oneday_view
    __table_args__ = (
        PrimaryKeyConstraint('id_stock', 'date_trade'),
    )


_oneminute_view = view(
    "VW_oneminute_history", Base.metadata,
    select([
        OneMinute.id_stock.label('id_stock'),
        Market.market_name.label('market_name'),
        Stock.stockfullcode.label('stockfullcode'),
        Stock.stockcode.label('stockcode'),
        func.DATE(OneMinute.date_trade).label('dateInfo'),
        OneMinute.date_trade.label('date_trade'),
        OneMinute.tday.label('tday'),
        case([(OneDay.val_preclose != None, OneDay.val_preclose), ],
             else_=literal_column('0')).label('dailyPreclose'),
        OneMinute.val_preclose.label('val_preclose'),
        OneMinute.val_open.label('val_open'),
        OneMinute.val_high.label('val_high'),
        OneMinute.val_low.label('val_low'),
        OneMinute.val_close.label('val_close'),
        OneMinute.amt_amount.label('amt_amount'),
        OneMinute.amt_volume.label('amt_volumn'),
        OneMinute.amt_wp.label('amt_wp'),
        OneMinute.amt_np.label('amt_np'),
        OneMinute.val_OI.label('val_OI'),
        literal_column('0').label('val_turnover'),
        case([(OneDay.flg_st != None, OneDay.flg_st), ],
             else_=literal_column('0')).label('flg_st'),
        OneMinute.val_sp1.label('val_sp1'),
        OneMinute.val_sv1.label('val_sv1'),
        OneMinute.val_bp1.label('val_bp1'),
        OneMinute.val_bv1.label('val_bv1'),
        case([(OneMinute.val_mid != literal_column('0'), OneMinute.val_mid), ],
             else_=OneMinute.val_close).label('val_mid'),
        case([(OneMinute.val_midv != literal_column('0'), OneMinute.val_midv), ],
             else_=OneMinute.val_close).label('val_midv'),
    ]). \
        select_from(
        OneMinute.__table__. \
            outerjoin(OneDay,
                      and_(
                          func.DATE(OneMinute.date_trade) == OneDay.date_trade,
                          OneMinute.id_stock == OneDay.id_stock)). \
            outerjoin(Stock, OneMinute.id_stock == Stock.id_stock). \
            outerjoin(Market, Stock.id_market == Market.id_market)
    )
)


class OneMinuteView(KLineView, Base):
    __table__ = _oneminute_view
    __table_args__ = (
        PrimaryKeyConstraint('id_stock', 'date_trade'),
    )


# rt03上的db_rt03数据库 1分钟行情表
t_VW_oneminute_history = Table(
    'VW_oneminute_history', metadata,
    Column('id_stock', Integer, nullable=False),
    Column('market_name', String(16, 'Chinese_PRC_CI_AS')),
    Column('stockfullcode', String(16, 'Chinese_PRC_CI_AS')),
    Column('dateInfo', Date),
    Column('dailyPreclose', MONEY, nullable=False),
    Column('date_trade', SMALLDATETIME, nullable=False),
    Column('val_preclose', MONEY, nullable=False),
    Column('val_open', MONEY, nullable=False),
    Column('val_high', MONEY, nullable=False),
    Column('val_low', MONEY, nullable=False),
    Column('val_close', MONEY, nullable=False),
    Column('amt_amount', MONEY, nullable=False),
    Column('amt_volumn', BigInteger, nullable=False),
    Column('amt_wp', BigInteger, nullable=False),
    Column('amt_np', BigInteger, nullable=False),
    Column('val_turnover', Float(24)),
    Column('val_OI', BigInteger),
    Column('flg_st', BIT, nullable=False),
    Column('stockcode', String(16, 'Chinese_PRC_CI_AS')),
    Column('val_mid', MONEY, nullable=False),
    Column('val_midv', MONEY, nullable=False),
    Column('val_sp1', MONEY, nullable=False),
    Column('val_sv1', BigInteger, nullable=False),
    Column('val_bp1', MONEY, nullable=False),
    Column('val_bv1', BigInteger, nullable=False),
    Column('tday', Date)
)

# rt03上的db_rt03数据库 5分钟行情表
t_VW_fiveminutes_history = Table(
    'VW_fiveminutes_history', metadata,
    Column('id_stock', Integer, nullable=False),
    Column('market_name', String(16, 'Chinese_PRC_CI_AS')),
    Column('stockfullcode', String(16, 'Chinese_PRC_CI_AS')),
    Column('dateInfo', Date),
    Column('dailyPreclose', MONEY),
    Column('date_trade', SMALLDATETIME, nullable=False),
    Column('val_preclose', MONEY, nullable=False),
    Column('val_open', MONEY, nullable=False),
    Column('val_high', MONEY, nullable=False),
    Column('val_low', MONEY, nullable=False),
    Column('val_close', MONEY, nullable=False),
    Column('amt_amount', MONEY, nullable=False),
    Column('amt_volumn', BigInteger, nullable=False),
    Column('amt_wp', BigInteger, nullable=False),
    Column('amt_np', BigInteger, nullable=False),
    Column('val_turnover', Float(24)),
    Column('val_OI', BigInteger),
    Column('flg_st', BIT),
    Column('val_mid', MONEY, nullable=False),
    Column('val_midv', MONEY, nullable=False),
    Column('val_sp1', MONEY, nullable=False),
    Column('val_sv1', BigInteger, nullable=False),
    Column('val_bp1', MONEY, nullable=False),
    Column('val_bv1', BigInteger, nullable=False),
    Column('tday', Date)
)
