# -*- coding:utf-8 -*-
import unittest
import datetime as dt
import pandas as pd
from rio3.log.setup import setup_screen_logger
from sqlalchemy import PrimaryKeyConstraint, func, literal_column, Table
from sqlalchemy import create_engine, Column, Integer, String, Text, ForeignKey, Index, Date, DateTime, DECIMAL, \
    Boolean, BigInteger, select, case, Float, distinct, BINARY, text, and_
from sqlalchemy.ext import compiler
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import relationship, Session, aliased, Query
from sqlalchemy.schema import DDLElement
from sqlalchemy.sql import table
from sqlalchemy.dialects.mysql import TINYINT
from rio3.utils.num import float_equal, round
from rio3.quote_core import quote_info
from sqlalchemy.dialects.mssql import BIT, MONEY, SMALLDATETIME, SMALLMONEY

Base = declarative_base()
metadata = Base.metadata


class CreateView(DDLElement):
    def __init__(self, name, selectable):
        self.name = name
        self.selectable = selectable


class DropView(DDLElement):
    def __init__(self, name):
        self.name = name


@compiler.compiles(CreateView)
def compile(element, compiler, **kw):
    return "CREATE VIEW %s AS %s" % (element.name, compiler.sql_compiler.process(element.selectable))


@compiler.compiles(DropView)
def compile(element, compiler, **kw):
    return "DROP VIEW %s" % (element.name)


def view(name, metadata, selectable):
    t = table(name)

    if isinstance(selectable, Query):
        selectable = selectable.subquery()

    for c in selectable.c:
        c._make_proxy(t)

    CreateView(name, selectable).execute_at('after-create', metadata)
    DropView(name).execute_at('before-drop', metadata)
    return t


class Stock(Base):
    __tablename__ = 'lookup_stockid'
    id_stock = Column(Integer, primary_key=True, autoincrement=False)
    id_market = Column(Integer, ForeignKey('lookup_marketid.id_market'), nullable=False)
    stockcode = Column(String(16), primary_key=False, nullable=False)
    stockfullcode = Column(String(16), primary_key=False, unique=True, nullable=False)
    stockname = Column(String(64), primary_key=False, nullable=True)

    market = relationship('Market')

    __table_args__ = (
        PrimaryKeyConstraint('id_stock'),
    )


class Market(Base):
    __tablename__ = 'lookup_marketid'
    id_market = Column(Integer, primary_key=True, nullable=False, autoincrement=False)
    market_name = Column(String(16), nullable=False)
    memo = Column(Text(1024), nullable=True)


class KLineRT(object):
    """
    实时数据库的基础K线结构
    """

    @declared_attr
    def id_stock(cls):
        return Column(Integer, ForeignKey('lookup_stockid.id_stock'), primary_key=True)

    date_trade = Column(DateTime, primary_key=True)
    val_preclose = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_open = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_high = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_low = Column(DECIMAL(10, 4), nullable=False, default=0)
    val_close = Column(DECIMAL(10, 4), nullable=False, default=0)
    amt_amount = Column(DECIMAL(19, 4), nullable=False, default=0)
    amt_volume = Column('amt_volumn', BigInteger, nullable=False, default=0)
    amt_wp = Column(BigInteger, nullable=False, default=0)
    amt_np = Column(BigInteger, nullable=False, default=0)
    val_OI = Column(BigInteger, nullable=True, default=0)

    @declared_attr
    def stock(cls):
        return relationship(Stock, primaryjoin=lambda: Stock.id_stock == cls.id_stock)


class OneMinute(KLineRT, Base):
    __tablename__ = 'oneminute_history'


class FiveMinutes(KLineRT, Base):
    __tablename__ = 'fiveminutes_history'
