# -*- coding:utf-8 -*-
import re
import datetime as dt
import unittest
import redis
import json
import logging
import numpy as np
import pandas as pd
from rio3.log.setup import setup_screen_logger, setup_file_logger
from collections import defaultdict
from rio3.base.object import BaseObject
from rio3.base import constant
from rio3.base.enumerator import ENUM_PERIOD
from rio3.datafeed.common import _get_future_symbol, _to_datetime, convert_freq2num, return_data, calculate_mid_and_midv
from rio3.datafeed.dbIO import get_volume_multiple, get_tday_df
from rio3.apis.sina_api import get_latest_kline
from rio3.datafeed.research.tushareIO import _get_previous_tday

STRATEGY_INFO_DICT = {}
WEIGHT_DICT_CACHE = {}
ACCOUNT_INFO_DICT = {}

REDIS_MD_PREFIX = 'MARKETDATA'

logger = logging.getLogger('root')


class RedisIO(BaseObject):
    def __init__(self, name, redis_string=None):
        super(RedisIO, self).__init__(name=name)
        self.redis_string = redis_string
        if self.redis_string is None:
            self.redis_string = constant.DEFAULT_REDIS_STRING
        # self.redis_client = redis.from_url(self.redis_string)
        self.redis_client = redis.Redis(host=constant.DEFAULT_REDIS_HOST, port=constant.DEFAULT_REDIS_PORT,
                                        password=constant.DEFAULT_REDIS_PASSWD, decode_responses=True)
        self.redis_md_prefix = 'MARKETDATA'
        self.redis_strategy_info_prefix = "STRATEGYINFO"
        self.redis_account_info_prefix = "ACCOUNTINFO"

    @staticmethod
    def to_tday(date_trade):
        assert isinstance(date_trade, dt.datetime)
        if date_trade.hour < 8 and date_trade.weekday == 5:  # 周六凌晨
            return date_trade.date() + dt.timedelta(days=2)
        if date_trade.hour > 20 and date_trade.weekday == 4:  # 周五晚上
            return date_trade.date() + dt.timedelta(days=3)
        if date_trade.hour > 20 and date_trade.weekday <= 3:  # 周一到周四晚上
            return date_trade.date() + dt.timedelta(days=1)
        return date_trade.date()

    @staticmethod
    def get_tday(date_trade):
        tday = pd.to_datetime(date_trade.date())
        # 晚上
        if date_trade.hour > constant.COMMODITY_DAY_CUTOFF_HOUR:
            if tday.weekday() == 4:  # 周五晚上
                return tday + dt.timedelta(days=3)
            else:  # 其他晚上
                return tday + dt.timedelta(days=1)
        # 周六凌晨
        if date_trade.hour < 8 and tday.weekday() == 5:
            return tday + dt.timedelta(days=2)
        return tday

    def get_strategy_info_dict(self):
        global STRATEGY_INFO_DICT
        if not STRATEGY_INFO_DICT:
            strategy_keys = self.redis_client.keys(
                '{}:*'.format(self.redis_strategy_info_prefix))
            with self.redis_client.pipeline() as pipe:
                for strategy_key in strategy_keys:
                    pipe.hgetall(strategy_key)
                strategy_info_list = pipe.execute()
                STRATEGY_INFO_DICT = dict([('STRATEGYINFO:{}:{}:{}:{}'.format(
                    info['strategy_name'], re.match(('\D+'), info['stockfullcode'][2:]).group(),
                    info['period'], info['model_id']), info) for info in strategy_info_list])
        return STRATEGY_INFO_DICT

    def get_account_info_dict(self):
        if not ACCOUNT_INFO_DICT:
            account_keys = self.redis_client.keys(
                '{}:*'.format(self.redis_account_info_prefix))
            for account_key in account_keys:
                account_info = self.redis_client.hgetall(account_key)
                ACCOUNT_INFO_DICT[account_key] = account_info
        return ACCOUNT_INFO_DICT

    def get_account_info_df(self, account_name=None):
        account_info_df = pd.DataFrame()
        data_list = list()
        account_info_dict = self.get_account_info_dict()
        for key, info in account_info_dict.items():
            # 如果传入的account_name不是None，则返回相应账号信息
            data_row = list()
            if account_name is not None:
                if str(account_name) == info['account_name']:
                    data_row.append(info['id'])
                    data_row.append(info['account_num'])
                    data_row.append(info['account_name'])
                    data_row.append(info['memo'])
                    data_row.append(info['is_valid'])
                    data_list.append(data_row)
                    break
            # 如果传入的account_name是None，则返回整个数据
            else:
                data_row.append(info['id'])
                data_row.append(info['account_num'])
                data_row.append(info['account_name'])
                data_row.append(info['memo'])
                data_row.append(info['is_valid'])
                data_list.append(data_row)
        if len(data_list) > 0:
            account_info_df = pd.DataFrame(
                data=data_list,
                columns=[
                    'id',
                    'account_num',
                    'account_name',
                    'memo',
                    'is_valid'])
        return account_info_df

    def get_model_id_dict(self):
        '''
        返回一个dict， key=strategy ，value=[modelid,modelid]
        '''
        strategy_info_dict = self.get_strategy_info_dict()
        strategy_dict = defaultdict(list)
        for key, info in strategy_info_dict.items():
            strategy_name = info['strategy_name']
            model_id = info['model_id']
            if model_id not in strategy_dict[strategy_name]:
                strategy_dict[strategy_name].append(model_id)
        return strategy_dict

    def get_strategy_by_model_id(self, model_id):
        '''
        根据model_id 返回strategy_name
        :return:
        '''
        strategy_model_id_dict = self.get_model_id_dict()
        modelid_strategy_name = None
        for strategy_name, model_id_list in strategy_model_id_dict.items():
            if model_id in model_id_list:
                modelid_strategy_name = strategy_name
                break
        return modelid_strategy_name

    def get_model_def(self, model_id):
        '''
        返回一个modeldef 的冬冬，初步定义为string，也可以是json
        '''
        strategy_info_dict = self.get_strategy_info_dict()
        for key, info in strategy_info_dict.items():
            if str(model_id) == info['model_id']:
                return info
        return None

    def get_model_def_symbols(self, model_id):
        '''
        返回一个model_id定义的品种集合
        :param model_id:
        :return:
        '''
        strategy_info_dict = self.get_strategy_info_dict()
        symbol_set = set()
        for key, info in strategy_info_dict.items():
            if str(model_id) == info['model_id']:
                symbol = _get_future_symbol(info['stockfullcode'])
                if symbol is not None:
                    symbol_set.add(symbol)
        return sorted(symbol_set)

    def get_model_stock_weight_dict(self, model_id):
        '''
        获得策略品种相对权重
        '''
        if str(model_id) in WEIGHT_DICT_CACHE:
            return WEIGHT_DICT_CACHE[str(model_id)]
        model_weight_dict = {}
        strategy_info_dict = self.get_strategy_info_dict()
        for key, info in strategy_info_dict.items():
            if str(model_id) != info['model_id']:
                continue
            is_valid = bool(info['is_valid'])
            if not is_valid:
                continue
            volume = int(info['piece_volume'])
            stockfullcode = info['stockfullcode']
            weight = float(info['weight'])
            volume_multiple = get_volume_multiple(stockfullcode)
            kline = get_latest_kline(stockfullcode)
            if kline is None:
                print(u'找不到{}的最新价格'.format(stockfullcode))
                continue
            price = kline.val_close
            amt = volume * volume_multiple * price * weight
            model_weight_dict[stockfullcode] = amt
        total_amt = sum(model_weight_dict.values())
        for stockfullcode, amt in model_weight_dict.items():
            model_weight_dict[stockfullcode] = amt / total_amt
        WEIGHT_DICT_CACHE[str(model_id)] = model_weight_dict
        return model_weight_dict

    def get_model_interval(self, model_id):
        '''
        输入模型id，获得模型使用的bar 的秒数
        如果模型使用的是一分钟，那么返回60。5分钟，那么返回300。 以此类推
        '''
        mid = self.get_model_def(model_id)
        if mid is None:
            return 0
        if 'period' not in mid:
            return 0
        prd = int(mid['period'])
        return prd

    def get_stock_position_by_model_id(self, model_id):
        """
        根据model_id读取股票持仓情况
        """
        keys = self.redis_client.keys(
            'POSITIONSTOCK:*:{}:LAST'.format(model_id))
        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for key in keys:
                pipe.hgetall(key)
            list_posi = pipe.execute()
        return [posi for posi in list_posi if float(posi['Position']) != 0]

    def get_holding_stock_by_model_id(self, model_id):
        list_posi = self.get_stock_position_by_model_id(model_id)
        return [posi['Contract'] for posi in list_posi]

    def reset_stock_list(self, strategy_name, model_id, stock_list):
        """
        更新model_id对应的stock_list
        """
        key = 'STOCKLIST:{}:{}'.format(strategy_name, model_id)
        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            pipe.delete(key)
            pipe.sadd(key, *stock_list)
            pipe.execute()
        pass

    def get_account_symbol_weight_config_df(self, investor_id=None):
        """
        读取账户品种配置数据
        :param investor_id:
        :return:
        """
        map_id_keys = self.redis_client.hgetall('MAPIDKEY:{}'.format('ACCOUNTSYMBOLWEIGHTCONFIG'))
        # 获取redis_keys
        valid_keys = list(map_id_keys.values())
        # 通过pipeline读取数据
        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for key in valid_keys:
                pipe.hgetall(key)
            data = pipe.execute()
        return_data_df = pd.DataFrame(data)
        return_data_df.dropna(how='any', inplace=True)
        if not return_data_df.empty:
            return_data_df = return_data_df.astype(
                {'id': np.int,
                 'account_num': np.int,
                 }, copy=True)
        return return_data_df

    def get_account_position_setup_auto(
            self,
            list_account=None):
        """
        读取账户策略仓位自动配置信息
        :param list_account:str or list, optional
        :return:dict
        """
        if isinstance(list_account, str):
            list_account = [list_account, ]
        elif isinstance(list_account, list) and len(list_account) == 0:
            list_account = None

        list_hash_id = self.redis_client.keys('ACCOUNTPOSITIONSETUPAUTO*')

        if list_account is not None and len(list_account) > 0:
            list_hash_id = [hash_id for hash_id in list_hash_id if hash_id.split(':')[1] in list_account]

        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for hash_id in list_hash_id:
                pipe.hgetall(hash_id)
            list_setup = pipe.execute()
        df_setup = pd.DataFrame(list_setup)
        df_setup[['strategy_name', 'symbol', 'period', 'model_id']] = df_setup['model_name'].str.split(':', expand=True)
        df_setup[['piece', 'period']] = df_setup[[
            'piece', 'period']].astype(np.int64)
        return df_setup.copy()

    def get_futures_position(self, list_model_name, tday=None):
        """
        读取期货策略tday的所有信号和tday前的持仓信息(由于全部读取数据量太大, 所以先筛选后按需读取)
        :param list_model_name: str or list, strategy_name:symbol:period:model_id, eg:CLuoModel:AP:300:260179
        :param tday: str or date
        :return:dict
        """
        if tday is None:
            tday = dt.date.today()
        else:
            tday = pd.to_datetime(tday).date()

        last_tday = _get_previous_tday(tday)
        last_tday = pd.to_datetime(last_tday, format='%Y%m%d')

        if isinstance(list_model_name, str):
            list_model_name = [list_model_name, ]

        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for key in list_model_name:
                pipe.lrange('POSITION:' + key, 0, -1)
            list_hash_id = pipe.execute()
        col_name = [
            'strategy_name',
            'symbol',
            'period',
            'model_id',
            'date_trade',
            'hash_id']
        df_hash_id_info = pd.DataFrame([dict(zip(col_name, hash_id.split(
            ':')[1:] + [hash_id])) for sub_list in list_hash_id for hash_id in set(sub_list)])
        df_hash_id_info['date_trade'] = pd.to_datetime(
            df_hash_id_info['date_trade'], format='%Y%m%d%H%M%S')
        df_hash_id_info['date'] = df_hash_id_info['date_trade'].dt.date
        today_index = (
                              df_hash_id_info['date_trade'] < pd.to_datetime(tday) +
                              pd.Timedelta(
                                  hours=16)) & (
                              df_hash_id_info['date_trade'] > pd.to_datetime(last_tday) +
                              pd.Timedelta(
                                  hours=16))
        prev_index = df_hash_id_info['date_trade'] < pd.to_datetime(
            last_tday) + pd.Timedelta(hours=16)
        df_today_sig_id = df_hash_id_info[today_index]
        df_before_today_sig_id = df_hash_id_info[prev_index]
        df_prev_holding_id = df_before_today_sig_id.sort_values(
            'date_trade').groupby(['model_id', 'symbol'], as_index=False).last()

        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for hash_id in df_today_sig_id['hash_id']:
                pipe.hgetall(hash_id)
            list_today_sig = pipe.execute()
        df_today_sig = pd.DataFrame(list_today_sig)
        df_today_sig = df_today_sig.astype(
            {
                'Close': np.float64,
                'LastActionTime': np.datetime64,
                'Period': np.int,
                'Position': np.float64,
                'UpdateTime': np.datetime64,
                'Weight': np.float64,
                'ZqPosition': np.float64},
            copy=True)

        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for hash_id in df_prev_holding_id['hash_id']:
                pipe.hgetall(hash_id)
            list_prev_holding = pipe.execute()
        df_prev_holding = pd.DataFrame(list_prev_holding)
        df_prev_holding = df_prev_holding.astype(
            {
                'Close': np.float64,
                'LastActionTime': np.datetime64,
                'Period': np.int,
                'Position': np.float64,
                'UpdateTime': np.datetime64,
                'Weight': np.float64,
                'ZqPosition': np.float64},
            copy=True)
        df_prev_holding = df_prev_holding[df_prev_holding['Position'] != 0]
        return df_prev_holding, df_today_sig

    def get_last_position(self, list_model_name):
        if isinstance(list_model_name, str):
            list_model_name = [list_model_name, ]
        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for key in list_model_name:
                pipe.hgetall('POSITION:{}:LAST'.format(key))
            list_last_position = pipe.execute()
            for posi in list_last_position:
                posi['symbol'] = re.match(r'(\D+)', posi['Contract']).group()
        df_last_position = pd.DataFrame(list_last_position)
        df_last_position = df_last_position.astype(
            {
                # 'Close': np.float64,
                # 'LastActionTime': dt.datetime,
                # 'Period': np.int,
                'Position': np.float,
                # 'UpdateTime': np.datetime64,
                'Weight': np.float
                # 'ZqPosition': np.float64
            })
        df_last_position['LastActionTime'] = pd.to_datetime(df_last_position['LastActionTime'])
        # return df_last_position.query('Position != 0')
        return df_last_position

    def get_manual_order_position(self, account_name_list=None):
        if isinstance(account_name_list, str):
            account_name_list = [account_name_list, ]
        all_redis_key_list = self.redis_client.keys('MANUALTRADERECORD:*')
        if account_name_list is not None:
            filter_key_list = [key for key in all_redis_key_list if key.split(':')[1] in account_name_list]
        else:
            filter_key_list = all_redis_key_list
        if len(filter_key_list) <= 0:
            return pd.DataFrame()
        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for key in filter_key_list:
                pipe.hgetall(key)
            manual_position_list = pipe.execute()
        if len(manual_position_list) <= 0:
            return pd.DataFrame()
        ret_df = pd.DataFrame.from_dict(manual_position_list)
        ret_df = ret_df.astype({
            'volume': np.int64
        }, copy=True)
        return ret_df.query('volume != 0')

    def get_pbc_investor_trading_info(self, investor_id=None):
        """
        获取持仓奖励配置用户实时的TradingInfo
        :param investor_id:
        :return:
        """
        total_position_df = pd.DataFrame()
        list_investor_id = list()
        if investor_id is not None:
            if isinstance(investor_id, str):
                list_investor_id.append(investor_id)
            elif isinstance(investor_id, list):
                list_investor_id = investor_id
        key_pattern = 'TRADINGINFO'
        trading_keys = list()
        if len(list_investor_id) > 0:
            [trading_keys.append('{}:{}'.format(key_pattern, each_investor_id)) for each_investor_id in
             list_investor_id]
        else:
            trading_keys = self.redis_client.keys(pattern='{}:*'.format(key_pattern))
        responses = list()
        with self.redis_client.pipeline() as pipe:
            pipe.multi()
            for key in trading_keys:
                pipe.get(key)
            responses = pipe.execute()
        # 根据实时信息构造总资金，可用资金，多头持仓，空头持仓，保证金占用，保证金比列，合约乘数，结算价，结算价更新日期
        total_position_df_list = list()
        # responses去掉None
        responses = [item for item in responses if item is not None]
        for each_data in responses:
            try:
                json_trading_info = json.loads(each_data)
                json_investor_id = json_trading_info['InvestorId']
                # 获取PositionList, 获取多头持仓，空头持仓
                positions_list = json_trading_info.get('PositionList', [])
                positions_df = pd.DataFrame(data=positions_list)
                positions_df = get_position_info_df(investor_id=json_investor_id, positions_df=positions_df, filter_zero=False)
                # 获取BonusContractList，获取合约乘数，保证金比列，结算价，结算价更新日期
                summary_df = positions_df.copy()
                bonus_contract_list = json_trading_info.get('BonusContractList', [])
                if bonus_contract_list is not None and len(bonus_contract_list) > 0:
                    bonus_contract_df = pd.DataFrame(data=bonus_contract_list)
                    # Merge信息
                    summary_df = pd.merge(left=positions_df, right=bonus_contract_df, left_on='StockCode',
                                          right_on='StockCode', how='outer')
                # 统一AccountNum
                if summary_df.empty:
                    summary_df = pd.DataFrame(data=[[json_investor_id, '', 0, 0, 0]],
                                              columns=['InvestorId', 'StockCode', 'LongVolume', 'ShortVolume',
                                                       'UseMargin'])
                summary_df['InvestorId'] = json_investor_id
                summary_df['Balance'] = json_trading_info['Balance']
                summary_df['Available'] = json_trading_info['Balance'] - json_trading_info['Margin']
                total_position_df_list.append(summary_df)
            except Exception as ex:
                print('读取账号 {} 的CTP信息有错误'.format(json_trading_info['InvestorId'], ex))
        if len(total_position_df_list) > 0:
            total_position_df = pd.concat(total_position_df_list, sort=True)
            total_position_df.reset_index(drop=True, inplace=True)
            # 对Position中没有的列，置为0
            total_position_df.fillna(
                {'LongVolume': 0, 'ShortVolume': 0, 'LongMarginRate': -1, 'ShortMarginRate': -1, 'UseMargin': 0},
                inplace=True)
        return total_position_df


def get_position_info_df(investor_id, positions_df, filter_zero):
    '''
    获取单个账号的净头寸
    :param account_num:
    :param positions_df:
    :return:
    '''
    net_position_df = pd.DataFrame(data=list(),
                                   columns=['InvestorId', 'StockCode', 'LongVolume', 'ShortVolume', 'UseMargin'])
    if positions_df.empty:
        return net_position_df
    net_position_df = positions_df.copy()
    net_position_df['InvestorId'] = investor_id
    net_position_df['LongVolume'] = net_position_df['TdLongVolume'] + net_position_df['YdLongVolume']
    net_position_df['ShortVolume'] = net_position_df['TdShortVolume'] + net_position_df['YdShortVolume']
    net_position_df = net_position_df[['InvestorId', 'StockCode', 'LongVolume', 'ShortVolume', 'UseMargin']]
    # 如果filter_zero为0
    if filter_zero:
        net_position_df = net_position_df[~(net_position_df[['LongVolume', 'ShortVolume']] == 0).all(axis=1)]
    return net_position_df


def get_redis_history_data_org(stockfullcode, freq, start_dt=None, end_dt=None, dataframe=True,
                               redis_url=constant.BACKUP_REDIS_URL):
    """
    从指定redis数据源读取实时数据
    :param stockfullcode:
    :param freq:
    :param start_dt:
    :param end_dt:
    :param dataframe:
    :param redis_url: redis/redis_backup
    :return:
    """
    redis_pool = None
    try:
        redis_pool = redis.ConnectionPool.from_url(url=redis_url, decode_responses=True)
    except Exception as ex:
        logger.warning('请检查redis_url格式: redis_url:{}, error:{}'.format(redis_url, ex))
        return return_data(dataframe)
    # 初始化redis客户端
    rClient = redis.StrictRedis(connection_pool=redis_pool)
    # 设置column列表
    columns = constant.ZQ_KLINE_ORG_COLUMNS.copy()
    columns.extend(['symbol', 'mid', 'midv'])
    quote_total_dict = dict()
    total_quote_df = pd.DataFrame()
    list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
    start_dt = _to_datetime(start_dt)
    end_dt = _to_datetime(end_dt)
    # 转化fre为秒
    span_value = convert_freq2num(freq_str=freq)
    # 用字典存储stockfullcode, idx_key
    stock_idx_key_dict = dict()
    for each_stock in list_stockfullcode:
        idx_key = '{}:{}:{}'.format(REDIS_MD_PREFIX, each_stock, span_value)
        stock_idx_key_dict[each_stock] = idx_key
    # 全数字化时间戳
    start_score = start_dt.strftime('%Y%m%d%H%M%S')
    end_score = end_dt.strftime('%Y%m%d%H%M%S')
    lua_body = '''
        local isKeyValid = redis.call('EXISTS', KEYS[1])
        local r = {}
        if isKeyValid == 1 then
            local hashKeyList = redis.call('ZRANGEBYSCORE', KEYS[1], ARGV[1], ARGV[2])
            for k,v in pairs(hashKeyList) do
                local hashValue = redis.call('HGETALL', v)
                local hashData = {}
                if hashValue then
                    for idx = 1, #hashValue, 2 do
                        hashData[hashValue[idx]] = hashValue[idx + 1]
                    end
                    r[k] = cjson.encode(hashData)
                end
            end
        end
        return r
    '''
    lua_inst = rClient.register_script(lua_body)
    for each_stock, each_idx_key in stock_idx_key_dict.items():
        lua_result = lua_inst(keys=[each_idx_key], args=[start_score, end_score])
        if lua_result is None or len(lua_result) == 0:
            quote_total_dict[each_stock] = pd.DataFrame()
            continue
        # 将lua_list中的item变为JSON格式
        json_obj_list = [json.loads(item) for item in lua_result]
        # 转化为dataframe
        quote_df = pd.DataFrame.from_records(data=json_obj_list)

        # 日期格式转化
        quote_df['DateTrade'] = pd.to_datetime(quote_df['DateTrade'])

        # 构造tday
        if span_value < 86400:
            df_tday = get_tday_df(quote_df.iloc[0]['DateTrade'].date() - pd.Timedelta(days=10),
                                  quote_df.iloc[-1]['DateTrade'].date() + pd.Timedelta(days=10)).copy()
            df_tday.loc[:, 'next_tday'] = df_tday['tday'].shift(-1)
            df_tday = df_tday.astype(np.datetime64)
            quote_df.loc[:, 'date_trade_tmp'] = quote_df['DateTrade']
            quote_df.loc[:, 'date_trade_tmp'] = np.where(quote_df['DateTrade'].dt.hour < 6,
                                                         quote_df['DateTrade'] - pd.Timedelta(hours=6),
                                                         quote_df['DateTrade'])
            quote_df.loc[:, 'date'] = quote_df['date_trade_tmp'].dt.date.astype(np.datetime64)
            quote_df = pd.merge(quote_df, df_tday, how='left', left_on='date', right_on='tday')

            quote_df.sort_index(inplace=True, ascending=False)
            quote_df.loc[:, 'tday'] = np.where((quote_df['date_trade_tmp'].dt.hour > 16), quote_df['next_tday'],
                                               quote_df['tday'])
            # 格式化tday为int
            quote_df['today'] = quote_df['tday'].dt.strftime('%Y%m%d')
        else:
            quote_df['tday'] = quote_df['DateTrade'].dt.strftime('%Y%m%d')
            quote_df['today'] = quote_df['DateTrade'].dt.strftime('%Y%m%d')
        # 排序
        quote_df.sort_values(by=['DateTrade'], inplace=True)

        # 字段读出来带有逗号,需要去除
        quote_df['Amount'] = quote_df['Amount'].str.replace(',', '')
        quote_df['Ask1Price'] = quote_df['Ask1Price'].str.replace(',', '')
        quote_df['Bid1Price'] = quote_df['Bid1Price'].str.replace(',', '')
        quote_df['Open'] = quote_df['Open'].str.replace(',', '')
        quote_df['Low'] = quote_df['Low'].str.replace(',', '')
        quote_df['High'] = quote_df['High'].str.replace(',', '')
        quote_df['Close'] = quote_df['Close'].str.replace(',', '')
        quote_df['OpenInterest'] = quote_df['OpenInterest'].str.replace(',', '')

        # 定义类型
        quote_df = quote_df.astype(
            dtype={
                'Amount': np.float,
                'Ask1Price': np.float,
                'Ask1Volume': np.float,
                'Bid1Price': np.float,
                'Bid1Volume': np.float,
                'Close': np.float,
                'High': np.float,
                'Low': np.float,
                'Open': np.float,
                'OpenInterest': np.float,
                'UpdateTime': np.datetime64,
                'Volume': np.float,
                'tday': np.int64
            })

        # 计算Mid和Midv
        calculate_mid_and_midv(quote_df)

        # 构造ZQ格式
        quote_df['date'] = quote_df['DateTrade'].dt.strftime("%Y%m%d").astype(np.int64)
        quote_df['ets'] = quote_df['DateTrade'].dt.strftime('%H%M%S').astype(np.int64)
        quote_df['sts'] = (quote_df['DateTrade'] - pd.Timedelta(seconds=int(span_value))).dt.strftime('%H%M%S').astype(
            np.int64)

        # 重命名Columns
        quote_df.rename(columns=lambda c: c.lower(), inplace=True)
        quote_df.rename(columns={'volume': 'vol', 'openinterest': 'oi'}, inplace=True)

        # 构造symbol, contracts
        quote_df['symbol'] = each_stock
        quote_df['contract'] = each_stock
        quote_df['span'] = span_value

        # 赋值通用字段
        quote_df['high_limit'] = -1
        quote_df['low_limit'] = -1
        quote_df['avg'] = -1
        quote_df['pre_close'] = -1
        quote_df['factor'] = -1

        # 排序

        quote_df = quote_df[columns]

        # 加入字典
        quote_total_dict[each_stock] = quote_df
    if not dataframe:
        return quote_total_dict
    total_quote_df = pd.concat(quote_total_dict.values()) if len(quote_total_dict.values()) > 0 else return_data(
        return_frame=dataframe)
    return total_quote_df


def get_instrument_field_from_redis(security_list=['Options', 'SpotOption', 'Futures']):
    """
    从redis获取所有的合约信息
    :return:
    """
    inst_list = []
    try:
        client = redis.from_url(constant.DEFAULT_REDIS_STRING, decode_responses=True)
        all_keys = client.smembers('FUTURESINFO:KEYS')
        with client.pipeline() as pipe:
            for key in all_keys:
                arr = key.split(':')
                if arr[1] not in security_list:
                    continue
                pipe.get(key)
            inst_jsons = pipe.execute()
            for inst_json in inst_jsons:
                inst_dict = json.loads(inst_json)
                inst_list.append(inst_dict)
    except Exception as ex:
        logger.info('[查询redis合约信息失败]: 命名空间:{}, 数量: {}'.format(ex))
    return inst_list



def get_trade_period_json(cache_path='redis'):
    if cache_path == 'redis':
        redis_url = constant.DEFAULT_REDIS_STRING
    elif cache_path == 'redis_backup':
        redis_url = constant.BACKUP_REDIS_URL
    else:
        logger.warning('请指定cache_path读取源, redis/redis_backup ')
        return None

    client = redis.from_url(redis_url, decode_responses=True)
    return client.get('SYMBOLTRADINGPERIOD')


class RedisIoTest(unittest.TestCase):
    def setUp(self):
        setup_screen_logger()
        setup_file_logger()
        self.logger = logging.getLogger('root')
        self.rdio = RedisIO('rdio')

    def tearDown(self):
        pass

    def test_get_last_position(self):
        lp = self.rdio.get_last_position(['CKCommodityModel:AP:60:270115', 'CKCommodityModel:AP:60:270116'])
        pass

    def test_get_history_data_org(self):
        quote_list = self.rdio.get_history_data_org(
            'SQrb9001',
            period=ENUM_PERIOD.ThirtySeconds,
            start='2019-03-05',
            end='2019-03-06')
        pass

    def test_get_model_id_dict(self):
        model_id_dict = self.rdio.get_model_id_dict()
        pass

    def test_get_strategy_by_model_id(self):
        model_id = '740476'
        strategy_name = self.rdio.get_strategy_by_model_id(model_id)
        pass

    def test_get_model_def(self):
        model_id = '999999'
        model_def_info = self.rdio.get_model_def(model_id)
        pass

    def test_get_model_def_symbols(self):
        model_id = '999999'
        model_id_symbols = self.rdio.get_model_def_symbols(model_id)
        pass

    def test_get_model_interval(self):
        model_id = 260234
        strategy_info_dict = self.rdio.get_strategy_info_dict()
        model_interval = self.rdio.get_model_interval(model_id=model_id)
        pass

    def test_get_account_df(self):
        account_name = '12312312'
        account_dict = self.rdio.get_account_info_dict()
        account_info = self.rdio.get_account_info_df(account_name=account_name)
        pass

    def test_get_account_position_setup_auto(self):
        config_df = self.rdio.get_account_position_setup_auto()
        pass

    def test_get_futures_position(self):
        position_df = self.rdio.get_futures_position()
        pass

    def test_get_manual_order_position(self):
        self.rdio.get_manual_order_position(account_name_list=['109090007'])
        pass

    def test_get_account_symbol_weight_config_df(self):
        data_df = self.rdio.get_account_symbol_weight_config_df()
        pass

    def test_get_redis_history_data_org(self):
        stockfullcode = ['DLa9001']
        freq = '30m'
        start_dt = '2020-07-10 09:35:00'
        end_dt = '2020-10-18 16:00:00'
        quote_df = get_redis_history_data_org(stockfullcode=stockfullcode, freq=freq, start_dt=start_dt, end_dt=end_dt,
                                              dataframe=False)
        pass

    def test_get_trade_period_json(self):
        json_str = get_trade_period_json()
        json_text = json.loads(json_str)
        pass

    def test_get_pbc_investor_trading_info(self):
        data = self.rdio.get_pbc_investor_trading_info()
        pass

    def test_get_instrument_field_from_redis(self):
        result = get_instrument_field_from_redis(security_list=['Futures'])
        pass

    def test_check_redis_market_latest(self):
        future_redis_pool = redis.ConnectionPool.from_url(constant.DEFAULT_REDIS_STRING, decode_responses=True, db=0)
        r_client = redis.StrictRedis(connection_pool=future_redis_pool)
        check_stock_code = [
            'DLi9001',
            'DLj9001',
            'DLjd9001',
            'DLjm9001',
            'DLp9001',
            'DLpp9001',
            'SEsc9001',
            'SFIC9001',
            'SFIF9001',
            'SFIH9001',
            'SQau9001',
            'SQbu9001',
            'SQfu9001',
            'SQni9001',
            'SQrb9001',
            'SQru9001',
            'SQzn9001',
            'ZZAP9001',
            'ZZMA9001',
            'ZZSR9001',
            'ZZTA9001',
            'ZZZC9001'
        ]

        for each_stock in check_stock_code:
            redis_period_zset_key = 'MARKETDATA:{}:{}'.format(each_stock, 300)
            start_date_num = 20210204200000
            end_date_num = 20210205160000
            zset_range_list = r_client.zrangebyscore(redis_period_zset_key, start_date_num, end_date_num,
                                                     withscores=True,
                                                     score_cast_func=int)
            if len(zset_range_list) > 0:
                print("enough kline: latest: {} ".format(zset_range_list[-1]))
            else:
                print("lack of kline: latest: {} ".format(zset_range_list[-1]))
