# coding: utf-8
from sqlalchemy import Column, DateTime, Float, ForeignKey, String, Table, text
from sqlalchemy.dialects.mysql import BIGINT, DATETIME, INTEGER, LONGTEXT
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
metadata = Base.metadata


class DjangoMigration(Base):
    __tablename__ = 'django_migrations'

    id = Column(INTEGER(11), primary_key=True)
    app = Column(String(255), nullable=False)
    name = Column(String(255), nullable=False)
    applied = Column(DATETIME(fsp=6), nullable=False)


class ProductProperty(Base):
    __tablename__ = 'product_property'

    id = Column(BIGINT(20), primary_key=True)
    createdate = Column(DateTime)
    creator = Column(BIGINT(20))
    modifieddate = Column(DateTime)
    modifier = Column(BIGINT(20))
    optlock = Column(BIGINT(20))
    status = Column(INTEGER(11), nullable=False)
    description = Column(String(255))
    name = Column(String(255))
    product_category = Column(INTEGER(11), nullable=False)


class Stock(Base):
    __tablename__ = 'stock'

    id = Column(BIGINT(20), primary_key=True)
    jq_stockfullcode = Column(String(255))
    stockfullcode = Column(String(255), unique=True)
    stockname = Column(String(255))


class User(Base):
    __tablename__ = 'user'

    id = Column(BIGINT(20), primary_key=True)
    createdate = Column(DateTime)
    creator = Column(BIGINT(20))
    modifieddate = Column(DateTime)
    modifier = Column(BIGINT(20))
    optlock = Column(BIGINT(20))
    status = Column(INTEGER(11), nullable=False)
    email = Column(String(255), nullable=False)
    identity_number = Column(String(255), nullable=False)
    identity_type = Column(INTEGER(11), nullable=False)
    password = Column(String(100), nullable=False)
    phone_number = Column(String(255))
    real_name = Column(String(255), nullable=False)
    user_name = Column(String(255))
    user_type = Column(INTEGER(11), nullable=False)


t_vw_fund_netvalue = Table(
    'vw_fund_netvalue', metadata,
    Column('id', BIGINT(20), server_default=text("'0'")),
    Column('fund_name', String(255)),
    Column('fund_status', INTEGER(11)),
    Column('fund_property', String(255)),
    Column('nav_date', DateTime),
    Column('unit_nav', Float(asdecimal=True)),
    Column('accumulative_nav', Float(asdecimal=True)),
    Column('product_property_id', BIGINT(20)),
    Column('product_id', BIGINT(20))
)


t_vw_user2product = Table(
    'vw_user2product', metadata,
    Column('user_id', BIGINT(20)),
    Column('product_inv_id', BIGINT(20)),
    Column('real_name', String(255)),
    Column('name', String(255)),
    Column('user_type', INTEGER(11)),
    Column('email', String(255)),
    Column('status', INTEGER(11)),
    Column('identity_type', INTEGER(11))
)


class MarketDatum(Base):
    __tablename__ = 'market_data'

    id = Column(BIGINT(20), primary_key=True)
    amount = Column(Float(asdecimal=True))
    close = Column(Float(asdecimal=True))
    date_trade = Column(DateTime)
    high = Column(Float(asdecimal=True))
    low = Column(Float(asdecimal=True))
    open = Column(Float(asdecimal=True))
    oi = Column(Float(asdecimal=True))
    volume = Column(Float(asdecimal=True))
    id_stock = Column(ForeignKey('stock.id'), index=True)

    stock = relationship('Stock')


class Product(Base):
    __tablename__ = 'product'

    id = Column(BIGINT(20), primary_key=True)
    createdate = Column(DateTime)
    creator = Column(BIGINT(20))
    modifieddate = Column(DateTime)
    modifier = Column(BIGINT(20))
    optlock = Column(BIGINT(20))
    status = Column(INTEGER(11), nullable=False)
    date_found = Column(DateTime)
    description = Column(LONGTEXT)
    name = Column(String(255))
    prodcut_manager = Column(String(255))
    product_type = Column(INTEGER(11), nullable=False)
    scope = Column(String(800))
    sort_key = Column(BIGINT(20))
    product_property_id = Column(ForeignKey('product_property.id'), index=True)
    ref_id_stock = Column(ForeignKey('stock.id'), index=True)

    product_property = relationship('ProductProperty')
    stock = relationship('Stock')
    users = relationship('User', secondary='m2m_user_products')


class Article(Base):
    __tablename__ = 'article'

    id = Column(BIGINT(20), primary_key=True)
    createdate = Column(DateTime)
    creator = Column(BIGINT(20))
    modifieddate = Column(DateTime)
    modifier = Column(BIGINT(20))
    optlock = Column(BIGINT(20))
    status = Column(INTEGER(11), nullable=False)
    article_type = Column(INTEGER(11), nullable=False)
    content = Column(LONGTEXT)
    publish_date = Column(DateTime, nullable=False)
    title = Column(String(255))
    product_id = Column(ForeignKey('product.id'), index=True)

    product = relationship('Product')


t_m2m_user_products = Table(
    'm2m_user_products', metadata,
    Column('user_id', ForeignKey('user.id'), nullable=False, index=True),
    Column('product_inv_id', ForeignKey('product.id'), nullable=False, index=True)
)


class UploadFile(Base):
    __tablename__ = 'upload_file'

    id = Column(BIGINT(20), primary_key=True)
    createdate = Column(DateTime)
    creator = Column(BIGINT(20))
    modifieddate = Column(DateTime)
    modifier = Column(BIGINT(20))
    optlock = Column(BIGINT(20))
    status = Column(INTEGER(11), nullable=False)
    original_file_name = Column(String(255))
    store_file_name = Column(String(255))
    upload_type = Column(INTEGER(11), nullable=False)
    article_id = Column(ForeignKey('article.id'), index=True)
    product_id = Column(ForeignKey('product.id'), index=True)

    article = relationship('Article')
    product = relationship('Product')


class Netvalue(Base):
    __tablename__ = 'netvalue'

    id = Column(BIGINT(20), primary_key=True)
    createdate = Column(DateTime)
    creator = Column(BIGINT(20))
    modifieddate = Column(DateTime)
    modifier = Column(BIGINT(20))
    optlock = Column(BIGINT(20))
    status = Column(INTEGER(11), nullable=False)
    date_publish = Column(DateTime)
    value = Column(Float(asdecimal=True))
    product_id = Column(ForeignKey('product.id'), index=True)
    uploadfile_id = Column(ForeignKey('upload_file.id'), index=True)
    unitvalue = Column(Float(asdecimal=True))

    product = relationship('Product')
    uploadfile = relationship('UploadFile')
