import ctypes
import ctypes.wintypes
import datetime
import os
import re
import psutil
import threading
import time
import dateutil.parser as parser
import win32api
import win32con
import win32gui
import win32process
import json
from rio3.datafeed.redis_feed import get_trade_period_json
from rio3.base.lookup_table import FUTURE_LEGAL_PERIOD


WM_COPYDATA = 0x004A
WM_HEARTBEAT = 0x0401
DD_INIT = 0x0
DD_SUBSCRIBE = 0x1
DD_UNSUBSCRIBE = 0x2
DD_KLINE = 0x3

class COPYDATASTRUCT(ctypes.Structure):
    _fields_ = [
        ('dwData', ctypes.wintypes.LPARAM),
        ('cbData', ctypes.wintypes.DWORD),
        ('lpData', ctypes.c_void_p)
    ]


PCOPYDATASTRUCT = ctypes.POINTER(COPYDATASTRUCT)


class InitInfo(ctypes.Structure):
    _fields_ = [
        ('Name', ctypes.c_char * 30),
        ('ProcessId', ctypes.c_int32),
        ('Hwnd', ctypes.wintypes.HANDLE)
    ]


class SubscribeInfo(ctypes.Structure):
    _fields_ = [
        ('StockFullCode', ctypes.c_char * 20),
        ('Period', ctypes.c_int32)
    ]


class CKLine(ctypes.Structure):
    _fields_ = [
        ('StockFullCode', ctypes.c_char * 20),
        ('Open', ctypes.c_double),
        ('High', ctypes.c_double),
        ('Low', ctypes.c_double),
        ('Close', ctypes.c_double),
        ('Volume', ctypes.c_double),
        ('Amount', ctypes.c_double),
        ('OpenInterest', ctypes.c_longlong),
        ('Timestamp', ctypes.c_double),
        ('UpdateTime', ctypes.c_double),
        ('Period', ctypes.c_int32),
        ('AskPrice1', ctypes.c_double),
        ('AskVolume1', ctypes.c_double),
        ('BidPrice1', ctypes.c_double),
        ('BidVolume1', ctypes.c_double),
    ]
PCKLINE = ctypes.POINTER(CKLine)


class KLine(object):
    def __init__(self, c_kline):
        self.stockfullcode = c_kline.StockFullCode.decode()
        self.open = c_kline.Open
        self.high = c_kline.High
        self.low = c_kline.Low
        self.close = c_kline.Close
        self.volume = c_kline.Volume
        self.amount = c_kline.Amount
        self.open_interest = c_kline.OpenInterest
        self.timestamp = self.from_oa_date(c_kline.Timestamp)
        self.update_time = self.from_oa_date(c_kline.UpdateTime)
        self.period = c_kline.Period
        self.ask_price1 = c_kline.AskPrice1
        self.ask_volume1 = c_kline.AskVolume1
        self.bid_price1 = c_kline.BidPrice1
        self.bid_volume1 = c_kline.BidVolume1

    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30)
    @classmethod
    def from_oa_date(cls, oa):
        return cls.OLE_TIME_ZERO + datetime.timedelta(oa)

    def __str__(self):
        return f'{self.stockfullcode},{self.timestamp},{self.period},{self.open},{self.high},{self.low},{self.close},' \
            f'{self.volume},{self.amount},{self.open_interest},{self.update_time},{self.ask_price1},' \
            f'{self.ask_volume1},{self.bid_price1},{self.bid_volume1}'

    def to_dict(self):
        date_str = self.timestamp.year * 10000 + self.timestamp.month * 100 + self.timestamp.day
        time_str = self.timestamp.hour * 10000 + self.timestamp.minute * 100 + self.timestamp.second
        mid = (self.ask_price1 + self.bid_price1) / 2
        if (self.ask_volume1 + self.bid_volume1) > 0:
            midv = (self.ask_price1 * self.ask_volume1 + self.bid_price1 * self.bid_volume1) / (self.ask_volume1 + self.bid_volume1)
        else:
            midv = mid
            print('ERROR, VOL == 0 @ %d' % date_str)
        return {'symbol': self.stockfullcode,
                'date': date_str,
                'today': date_str,
                'ets': time_str,
                'open': self.open,
                'close': self.close,
                'high': self.high,
                'low': self.low,
                'vol': self.volume,
                'amount': self.amount,
                'mid': mid,
                'midv': midv,
                'dt': self.timestamp}


class ErrorMsg(object):
    def __init__(self, id, msg):
        self.error_id = id
        self.error_msg = msg


def get_procs(pname):
    procs = []
    for proc in psutil.process_iter():
        try:
            if proc.name() == pname:
                procs.append(proc)
        except:
            pass
    return procs


class MainWindowFinder(object):
    def __init__(self):
        self.pid = 0
        self.main_handle = 0

    @staticmethod
    def is_main_window(hwnd):
        return win32gui.GetWindow(hwnd, 4) == 0 and win32gui.IsWindowVisible(hwnd)

    def enum_windows_callback(self, hwnd, windows):
        ids = win32process.GetWindowThreadProcessId(hwnd)
        if ids[1] != self.pid or not self.is_main_window(hwnd):
            return 1
        self.main_handle = hwnd
        windows.append(hwnd)
        return 0

    def find_main_window(self, pid):
        self.pid = pid
        self.main_handle = 0
        try:
            win32gui.EnumWindows(self.enum_windows_callback, [])
        except Exception as ex:
            pass
        return self.main_handle


class IpcClient:
    def __init__(self, model_name, OnErrorCallback, interval=15, timeout=60, filter_legal_time=True,
                 process_name='MarketDataReceiver.exe'):
        message_map = {
            win32con.WM_COPYDATA: self.OnCopyData,
            WM_HEARTBEAT: self.OnHeartbeat
        }
        self.OnError = OnErrorCallback
        wc = win32gui.WNDCLASS()
        wc.lpfnWndProc = message_map
        wc.lpszClassName = 'IPCWindowClass'
        hinst = wc.hInstance = win32api.GetModuleHandle(None)
        classAtom = win32gui.RegisterClass(wc)
        self.hwnd = win32gui.CreateWindow(
            classAtom,
            "ipc client",
            0,
            0,
            0,
            win32con.CW_USEDEFAULT,
            win32con.CW_USEDEFAULT,
            0,
            0,
            hinst,
            None)
        self.pid = os.getpid()
        self.server_proc = None
        self.server_main_hwnd = 0
        self.model_name = model_name
        self.is_init = False
        self.last_heartbeat_time = datetime.datetime.min
        self.heartbeat_interval = interval
        self.heartbeat_timeout = timeout
        self.check_heartbeat_timer = None
        self.send_heartbeat_timer = None
        self.callback_dict = {}
        self.filter_legal_time = filter_legal_time
        self.process_name = process_name
        if self.filter_legal_time:
            self.legal_period_dict = IpcClient._get_legal_trade_time_dict()
            self.re_symbol = re.compile('^[a-zA-Z]+')

    def OnHeartbeat(self, hwnd, msg, wparam, lparam):
        self.last_heartbeat_time = datetime.datetime.now()
        return 1

    def OnCopyData(self, hwnd, msg, wparam, lparam):
        try:
            pCDS = ctypes.cast(lparam, PCOPYDATASTRUCT)
            if pCDS.contents.dwData != DD_KLINE:
                return 0
            pCKline = ctypes.cast(pCDS.contents.lpData, PCKLINE)
            kline = KLine(pCKline.contents)
            if self.filter_legal_time:
                is_legal = False
                symbol = self.re_symbol.match(kline.stockfullcode).group()
                legal_time_list = self.legal_period_dict.get(symbol)
                if legal_time_list is not None:
                    for legal_time in legal_time_list:
                        if kline.period != 1800:
                            if legal_time[0] < kline.timestamp <= legal_time[1]:
                                is_legal = True
                                break
                        else:
                            if legal_time[0].hour == 10 and legal_time[0].minute == 30:
                                if legal_time[0] <= kline.timestamp <= legal_time[1]:
                                    is_legal = True
                                    break
                            else:
                                if legal_time[0] < kline.timestamp <= legal_time[1]:
                                    is_legal = True
                                    break
                if not is_legal:
                    return 1
            key = (kline.stockfullcode, kline.period)
            callbacks = self.callback_dict.get(key)
            if callbacks:
                for callback in callbacks:
                    callback(kline)
            return 1
        except Exception as ex:
            print(ex.args)
            return 0

    def PumpMessages(self):
        win32gui.PumpMessages()

    def send_heartbeat(self):
        try:
            win32api.PostMessage(self.server_main_hwnd, WM_HEARTBEAT, self.pid, 0)
            self.send_heartbeat_timer = threading.Timer(self.heartbeat_interval, self.send_heartbeat)
            self.send_heartbeat_timer.setDaemon(True)
            self.send_heartbeat_timer.start()
        except Exception as ex:
            print(ex.args)

    def check_heartbeat(self):
        if (datetime.datetime.now() - self.last_heartbeat_time).total_seconds() > self.heartbeat_timeout:
            err = ErrorMsg(5, 'heartbeat timeout')
            self.OnError(err)
            self.send_heartbeat_timer.cancel()
            self.check_heartbeat_timer.cancel()
            print('Try Reconnect')
            ret = self.Init()
            if ret:
                self.last_heartbeat_time = datetime.datetime.now()
                if len(self.callback_dict) > 0:
                    tmp_dict = self.callback_dict
                    self.callback_dict = {}
                    for (code, period), callbacks in tmp_dict.items():
                        for callback in callbacks:
                            self.SubscribeMarketData(code, period, callback)
                self.send_heartbeat_timer = threading.Timer(self.heartbeat_interval, self.send_heartbeat)
                self.send_heartbeat_timer.setDaemon(True)
                self.send_heartbeat_timer.start()
        self.check_heartbeat_timer = threading.Timer(self.heartbeat_interval, self.check_heartbeat)
        self.check_heartbeat_timer.setDaemon(True)
        self.check_heartbeat_timer.start()

    def Init(self):
        while True:
            time.sleep(1)
            procs = get_procs(self.process_name)
            if len(procs) == 0:
                err = ErrorMsg(1, 'Server Not Found')
                self.OnError(err)
                continue
            if len(procs) > 1:
                err = ErrorMsg(2, 'Multi Server Processes')
                self.OnError(err)
                continue
            self.server_proc = procs[0]
            finder = MainWindowFinder()
            self.server_main_hwnd = finder.find_main_window(self.server_proc.pid)
            if self.server_main_hwnd == 0:
                err = ErrorMsg(3, 'Server Window Not Found')
                self.OnError(err)
                continue
            init_info = InitInfo()
            model_name = self.model_name
            if isinstance(model_name, str):
                model_name = model_name.encode()
            init_info.Name = model_name
            init_info.ProcessId = self.pid
            init_info.Hwnd = self.hwnd
            p_init_info = ctypes.byref(init_info)
            cds = COPYDATASTRUCT()
            cds.dwData = DD_INIT
            cds.cbData = ctypes.sizeof(init_info)
            cds.lpData = ctypes.cast(p_init_info, ctypes.c_void_p)
            ret, lpdw_res = win32gui.SendMessageTimeout(self.server_main_hwnd, WM_COPYDATA, 0, cds, win32con.SMTO_NORMAL, 3000)
            if ret == 0:
                err = ErrorMsg(4, 'Init Failed, Timeout.')
                self.OnError(err)
                continue
            else:
                self.is_init = True
                self.send_heartbeat_timer = threading.Timer(self.heartbeat_interval, self.send_heartbeat)
                self.check_heartbeat_timer = threading.Timer(self.heartbeat_interval, self.check_heartbeat)
                self.send_heartbeat_timer.setDaemon(True)
                self.check_heartbeat_timer.setDaemon(True)
                self.send_heartbeat_timer.start()
                self.check_heartbeat_timer.start()
                self.last_heartbeat_time = datetime.datetime.now()
                return True

    def SubscribeMarketData(self, stockfullcode, period, callback):
        try:
            if not self.is_init and not self.Init():
                if not self.Init():
                    return False
            b_stockfullcode = stockfullcode
            if isinstance(stockfullcode, str):
                b_stockfullcode = stockfullcode.encode()
            sub_info = SubscribeInfo()
            sub_info.StockFullCode = b_stockfullcode
            sub_info.Period = period
            p_sub_info = ctypes.byref(sub_info)
            cds = COPYDATASTRUCT()
            cds.dwData = DD_SUBSCRIBE
            cds.cbData = ctypes.sizeof(sub_info)
            cds.lpData = ctypes.cast(p_sub_info, ctypes.c_void_p)
            ret, lpdw_res = win32gui.SendMessageTimeout(self.server_main_hwnd, WM_COPYDATA, self.pid, cds, win32con.SMTO_NORMAL, 60000)
            if ret == 0:
                print(f'IPC: 订阅{stockfullcode}/{period}失败')
                return False
            else:
                self.callback_dict.setdefault((stockfullcode, period), [])
                self.callback_dict[(stockfullcode, period)].append(callback)
                print(f'IPC: 订阅{stockfullcode}/{period}成功')
                return True
        except Exception as ex:
            print(ex.args)
            return False

    @staticmethod
    def _get_legal_trade_time_dict():
        legal_time_json = get_trade_period_json(cache_path='redis')
        try:
            legal_time_dict = json.loads(legal_time_json)
        except:
            legal_time_dict = FUTURE_LEGAL_PERIOD
        today = datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        tmp_dict = {}
        for symbol, legal_time_list in legal_time_dict.items():
            tmp_dict.setdefault(symbol, [])
            for legal_time_dict in legal_time_list:
                date_start = parser.parse(legal_time_dict['date_start'])
                date_end = parser.parse(legal_time_dict['date_end'])
                if date_start <= today <= date_end:
                    time_start = parser.parse(legal_time_dict['time_start'])
                    time_end = parser.parse(legal_time_dict['time_end'])
                    if time_end.hour <= 4:
                        time_end = time_end + datetime.timedelta(days=1)
                    tmp_dict[symbol].append((time_start, time_end))
        return tmp_dict


def on_market_data(kline):
    print(kline)
    pass


def on_error(err):
    print(err.error_msg)
    pass


# 支持的周期: [30, 60, 300, 900]
def start_md_client():
    client = IpcClient('TestModel', on_error)
    client.Init()
    client.SubscribeMarketData('SQrb9001', 30, on_market_data)
    client.SubscribeMarketData('DLi9001', 30, on_market_data)
    client.SubscribeMarketData('ZZSR9001', 30, on_market_data)
    client.PumpMessages()


def get_timeout_secs():
    now = datetime.datetime.now()
    if 5 < now.hour <= 15:
        timeout = (parser.parse('15:05:00') - now).total_seconds()
    elif 15 < now.hour <= 23:
        timeout = (parser.parse('02:35:00') - now + datetime.timedelta(days=1)).total_seconds()
    else:
        timeout = (parser.parse('02:35:00') - now).total_seconds()
    timeout = timeout if timeout > 0 else 10
    return timeout


if __name__ == '__main__':
    md_th = threading.Thread(target=start_md_client)
    md_th.start()
    md_th.join(timeout=get_timeout_secs())