import time
import numpy as np
import pyodbc
import json
import redis
import socket
import threading
import pandas as pd
import datetime as dt
from enum import Enum
import logging
import re
import traceback
from urllib.parse import urlparse

TDAY_CACHE = None
DB_HIS_STRING = ';'.join([
    'DRIVER={MySQL ODBC 5.3 ANSI Driver}',
    'SERVER=rm-uf605nn2go0mj7q49o.mysql.rds.aliyuncs.com',
    'PORT=3306',
    'DATABASE=db_his',
    'UID=sqlro',
    'PWD=XYsqlrd654321a88'
])


class EnumPublishType(Enum):
    Position = 1
    BarStatus = 2


def _rebuild_db_str(url, db):
    url_res = urlparse(url)
    if isinstance(db, int):
        return f'{url_res.scheme}://{url_res.netloc}/{db}'
    return f'{url_res.scheme}://{url_res.netloc}'


class PublishWorker(object):
    def __init__(self, strategy_name, pub_type,
                 db_str='redis://:redisxy123456R88@backupstrategynode.redis.rds.aliyuncs.com:6401', db=1,
                 hb_db=0, hb_prefix='HEARTBEAT', hb_suffix=None):
        """
        :param strategy_name: str,策略名
        :param pub_type: EnumPublishType,发布信号类型.Position会验证信号值是否变化,信号值未改变时不发布新记录.BarStatus不验证信号值,每个bar都应有一条记录
        :param db_str: str, redis连接字符串
        :param db: int, redis数据库编号,默认发布到db1
        """
        db_str = _rebuild_db_str(db_str, db)
        hb_db_str = _rebuild_db_str(db_str, hb_db)
        self.strategy_name = strategy_name
        self.pub_type = pub_type
        self.heartbeat_client = redis.from_url(hb_db_str, decode_responses=True)
        self.heartbeat_prefix = hb_prefix
        self.heartbeat_suffix = hb_suffix
        thread_safe_pool = redis.BlockingConnectionPool.from_url(db_str, decode_responses=True)
        self.publish_client = redis.StrictRedis(connection_pool=thread_safe_pool)
        self.publish_lua_inst = PublishWorker.register_lua(self.publish_client, pub_type)
        self.heartbeat_thread = threading.Thread(target=self.send_heartbeat)
        self.heartbeat_thread.setDaemon(True)
        self.heartbeat_thread.start()
        if pub_type is EnumPublishType.BarStatus:
            self.key_prefix = 'BARSTATUS'
        elif pub_type is EnumPublishType.Position:
            self.key_prefix = 'POSITION'
        else:
            self.key_prefix = 'POSITION'

    def publish(self, stockfullcode, action_time, signal, price=0, ref=None, overwrite=False):
        """
        发布单个策略信号
        :param stockfullcode: str, eg; SFIF9001
        :param action_time: py_datetime or pandas Timestamp
        :param signal: float, 信号值
        :param price: float, 产生信号时的价格
        :param ref: dict, 自定义数据
        :param overwrite: boolean, 是否覆盖已有记录
        :return: 成功发布信号的条数
        """
        now_str = dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        key = f'{self.key_prefix}:{self.strategy_name}:{stockfullcode}:{action_time.strftime("%Y%m%d%H%M%S")}'
        ref = ref if ref else '{}'
        ref_str = json.dumps(ref) if isinstance(ref, dict) else ref
        return self.publish_lua_inst(keys=[key], args=[1 if overwrite else 0, now_str, signal, price, ref_str])

    def batch_publish(self, df, overwrite=False):
        """
        批量发布多个策略信号
        :param df: pandas.DataFrame, 必需包含['stockfullcode', 'action_time', 'signal']列, 可选['price', 'ref']列, 如没有price则写0, dtype同publish方法
        :param overwrite: boolean, 是否覆盖已有记录
        :return: 成功发布信号的条数
        """
        keys = []
        now_str = dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        args = []
        df['action_time'] = pd.to_datetime(df['action_time'])
        df = df.sort_values('action_time')
        if 'price' not in df.columns:
            df['price'] = 0
        if 'ref' not in df.columns:
            df['ref'] = '{}'
        for row in df[['stockfullcode', 'action_time', 'signal', 'price', 'ref']].to_numpy():
            key = f'{self.key_prefix}:{self.strategy_name}:{row[0]}:{row[1].strftime("%Y%m%d%H%M%S")}'
            keys.append(key)
            args.append(row[2])
            args.append(row[3])
            ref = row[4]
            args.append(json.dumps(ref) if isinstance(ref, dict) else ref)
        pub_cnt = 0
        batch_cnt = 10000
        len_keys = len(keys)
        for i in range(0, len_keys, batch_cnt):
            if i + batch_cnt < len_keys:
                tmp_keys = keys[i: i + batch_cnt]
                tmp_args = [1 if overwrite else 0, now_str] + args[3 * i: 3 * (i + batch_cnt)]
            else:
                tmp_keys = keys[i: len_keys]
                tmp_args = [1 if overwrite else 0, now_str] + args[3 * i: 3 * len_keys]
            pub_cnt += self.publish_lua_inst(keys=tmp_keys, args=tmp_args)
            print(pub_cnt)
        return pub_cnt

    def send_heartbeat(self):
        while True:
            self.heartbeat_client.set(f'{self.heartbeat_prefix}:{self.strategy_name}'
                                      f'{f"_{self.heartbeat_suffix}" if self.heartbeat_suffix else ""}',
                                      dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
            time.sleep(15)

    def clear_signal(self):
        """
        删除该策略所有历史记录
        :return:
        """
        list_keys = self.publish_client.keys(f'{self.key_prefix}:{self.strategy_name}*')
        if list_keys is not None and len(list_keys) > 0:
            self.publish_client.delete(*list_keys)

    @staticmethod
    def register_lua(client, pub_type):
        if pub_type is EnumPublishType.BarStatus:
            lua_body = '''local overwrite = ARGV[1]
local update_time = ARGV[2]
local cnt = 0
for i = 1, #KEYS do
    local key = KEYS[i]
    local sig = ARGV[3 + 3 * (i - 1)]
    local price = ARGV[4 + 3 * (i - 1)]
    local ref_str = ARGV[5 + 3 * (i - 1)]
    local split_str = {}
    key:gsub('[^:]+', function(s) table.insert(split_str, s) end)
    local model_key = split_str[1] .. ':' .. split_str[2] .. ':' .. split_str[3]
    local if_pass = true
    if overwrite ~= '0' then
        if_pass = false
    else
        local exist = redis.call('ZSCORE', model_key, key)
        if not exist then
            if_pass = false
        end
    end
    if not if_pass then
        local strategy_name, stockfullcode, action_time = split_str[2], split_str[3], split_str[4]
        local at_str = string.sub(action_time, 1, 4) .. '-' .. string.sub(action_time, 5, 6) .. '-' ..
            string.sub(action_time, 7, 8) .. ' ' .. string.sub(action_time, 9, 10) .. ':' ..
            string.sub(action_time, 11, 12) .. ':' .. string.sub(action_time, 13, 14)
        redis.call('ZADD', model_key, tonumber(action_time), key)
        local ret = redis.call('HMSET', key, 'stockfullcode', stockfullcode, 'position', sig, 'action_time', at_str, 'update_time', update_time, 'price', price, 'ref', ref_str)
        if ret.ok then
            cnt = cnt + 1
            local pubKey = 'PUBSUB:' .. split_str[1] .. ':' .. strategy_name .. ':' .. stockfullcode
            redis.call('PUBLISH', pubKey, key .. ':' .. sig .. ':' .. price .. ':' .. ref_str)
        end
    end
end
return cnt'''
        elif pub_type is EnumPublishType.Position:
            lua_body = '''local overwrite = ARGV[1]
local update_time = ARGV[2]
local cnt = 0
local ret
for i = 1, #KEYS do
    local key = KEYS[i]
    local sig = ARGV[3 + 3 * (i - 1)]
    local price = ARGV[4 + 3 * (i - 1)]
    local ref_str = ARGV[5 + 3 * (i - 1)]
    local split_str = {}
    key:gsub('[^:]+', function(s) table.insert(split_str, s) end)
    local model_key = split_str[1] .. ':' .. split_str[2] .. ':' .. split_str[3]
    local if_pass = true
    if overwrite ~= '0' then
        if_pass = false
    else
        local pos = redis.call('HGET', model_key .. ':LAST', 'position')
        if not pos or math.abs(tonumber(sig) - tonumber(pos)) > 0.00001 then
            if_pass = false
        end
    end
    if not if_pass then
        local strategy_name, stockfullcode, action_time = split_str[2], split_str[3], split_str[4]
        local at_str = string.sub(action_time, 1, 4) .. '-' .. string.sub(action_time, 5, 6) .. '-' ..
            string.sub(action_time, 7, 8) .. ' ' .. string.sub(action_time, 9, 10) .. ':' ..
            string.sub(action_time, 11, 12) .. ':' .. string.sub(action_time, 13, 14)
        redis.call('ZADD', model_key, tonumber(action_time), key)
        local ret1 = redis.call('HMSET', key, 'stockfullcode', stockfullcode, 'position', sig, 'action_time', at_str, 'update_time', update_time, 'price', price, 'ref', ref_str)
        local ret2 = redis.call('HMSET', key:gsub(action_time, 'LAST'), 'stockfullcode', stockfullcode, 'position', sig, 'action_time', at_str, 'update_time', update_time, 'price', price, 'ref', ref_str)
        if ret1.ok and ret2.ok then
            cnt = cnt + 1
            local pubKey = 'PUBSUB:' .. split_str[1] .. ':' .. strategy_name .. ':' .. stockfullcode
            redis.call('PUBLISH', pubKey, key .. ':' .. sig .. ':' .. price .. ':' .. ref_str)
        end
    end
end
return cnt'''
        else:
            raise TypeError(f'不支持的发布类型: {type}')
        return client.register_script(lua_body)


class SignalReceiver(object):
    def __init__(self, strategy_name, recv_type,
                 db_str='redis://:redisxy123456R88@backupstrategynode.redis.rds.aliyuncs.com:6401', db=1,
                 connection_errors='retry', retry_times=10,
                 logger: logging.Logger = None):
        """

        :param strategy_name:
        :param recv_type:
        :param db_str:
        :param db:
        :param connection_errors: {'raise', 'ignore', 'retry'}, default 'retry'
        :param retry_times: retry times, default 10
        :param logger:
        """
        self.strategy_name = strategy_name
        self.recv_type = recv_type
        self.logger = logger
        db_str = _rebuild_db_str(db_str, db)
        thread_safe_pool = redis.BlockingConnectionPool.from_url(db_str, decode_responses=True,
                                                                 socket_keepalive=True,
                                                                 socket_keepalive_options={
                                                                     socket.TCP_KEEPIDLE: 120,
                                                                     socket.TCP_KEEPCNT: 2,
                                                                     socket.TCP_KEEPINTVL: 20})
        self.subscribe_client = redis.StrictRedis(connection_pool=thread_safe_pool)
        self.pubsub = self.subscribe_client.pubsub()
        self.reader_client = redis.from_url(db_str, decode_responses=True)
        retry = 10
        success = True
        if connection_errors == 'raise':
            try:
                self.subscribe_client.ping()
                self.reader_client.ping()
            except Exception as ex:
                success = False
                raise ex
        elif connection_errors == 'ignore':
            try:
                self.subscribe_client.ping()
                self.reader_client.ping()
            except:
                success = False
                if logger:
                    logger.error(traceback.format_exc())
        elif connection_errors == 'retry':
            retry = retry_times if retry_times > 0 else 1e10
            while retry > 0:
                retry -= 1
                success = True
                try:
                    self.subscribe_client.ping()
                except:
                    success = False
                    if logger:
                        logger.error(traceback.format_exc())
                try:
                    self.reader_client.ping()
                except:
                    success = False
                    if logger:
                        logger.error(traceback.format_exc())
                if success:
                    break
                time.sleep(3)
        if logger:
            logger.info(f'subscribe client connection {"successful" if success else "failed"}')
            logger.info(f'reader client connection {"successful" if success else "failed"}')
        self.subscribed_channel = []
        self.channel_prefix = 'BARSTATUS' if recv_type is EnumPublishType.BarStatus else 'POSITION'
        self.callback_dict = {}
        self.is_stop = False
        self.recv_thread = threading.Thread(target=self.start_listen)

    def get_history_signal(self, stockfullcode, start_time, end_time):
        start_score = pd.to_datetime(start_time).strftime('%Y%m%d%H%M%S')
        end_score = pd.to_datetime(end_time).strftime('%Y%m%d%H%M%S')
        model_key = f'{self.channel_prefix}:{self.strategy_name}:{stockfullcode}'
        key_list = self.reader_client.zrangebyscore(model_key, start_score, end_score)
        if len(key_list) > 0:
            with self.reader_client.pipeline() as pipe:
                for key in key_list:
                    pipe.hmget(key, 'position', 'price', 'ref')
                signals = pipe.execute()
                sig_list = []
                for i in range(len(key_list)):
                    strategy, code, dt_str = SignalReceiver._parse_key(key_list[i])
                    pos = signals[i][0]
                    px = signals[i][1]
                    ref = signals[i][2]
                    pos = pos if pos is None else float(pos)
                    px = px if px is None else float(px)
                    tmp_ref = {}
                    try:
                        tmp_ref = json.loads(ref)
                    except:
                        tmp_ref = ref
                    ref = tmp_ref
                    sig_list.append((strategy, code, pd.to_datetime(dt_str), pos, px, ref))
                return sig_list
        else:
            if self.logger:
                self.logger.warn(f'no history signal.{stockfullcode} - {start_time} - {end_time}')
            return []

    def subscribe(self, stockfullcode, on_signal):
        channel = f'PUBSUB:{self.channel_prefix}:{self.strategy_name}:{stockfullcode}'
        self.callback_dict[stockfullcode] = on_signal
        if not self.recv_thread.is_alive():
            self.recv_thread.start()
        self.pubsub.subscribe(**{channel: on_signal})
        if self.logger:
            self.logger.info(f'send subscription command: {channel}')

    def start_listen(self):
        self.is_stop = False
        if self.logger:
            self.logger.info('the receiving thread is started.')
        while not self.is_stop:
            try:
                if self.pubsub.subscribed:
                    msg = self.pubsub.get_message(timeout=1)
                    if msg is None:
                        continue
                    msg_type = msg.get('type', '')
                    if msg_type != 'subscribe':
                        continue
                    chl = msg.get('channel', '')
                    if chl == '':
                        continue
                    split_arr = chl.split(':')
                    if self.logger:
                        self.logger.info(f'subscribe to {split_arr[-2]} - {split_arr[-1]} successfully')
                    else:
                        print(f'subscribe to {split_arr[-2]} - {split_arr[-1]} successfully')
            except:
                if self.logger:
                    self.logger.error(traceback.format_exc())

    def join(self, timeout=None):
        """
        等待接收线程退出
        :return:
        """
        if not timeout:
            self.recv_thread.join()
        else:
            self.recv_thread.join(timeout)

    def stop(self):
        """
        停止接收订阅信号,不支持重启,若要重启,重新初始化SignalReceiver对象
        :return:
        """
        if self.pubsub.subscribed:
            self.pubsub.unsubscribe()
        self.is_stop = True

        if self.logger:
            self.logger.info('the receiving thread is stopped.')

    @staticmethod
    def _parse_key(key):
        str_arr = key.split(':')
        return str_arr[1], str_arr[2], str_arr[3]


class RedisDataFeed(object):
    def __init__(self, db_str='redis://:redisxy123456R88@publicpystrategy.redis.rds.aliyuncs.com:6401', db=0):
        db_str = _rebuild_db_str(db_str, db)
        thread_safe_pool = redis.BlockingConnectionPool.from_url(db_str, decode_responses=True)
        self.datafeed_client = redis.StrictRedis(connection_pool=thread_safe_pool)
        self.datafeed_lua_inst = RedisDataFeed.register_lua(self.datafeed_client)
        self.redis_md_prefix = 'MARKETDATA'
        self.redis_quote_column = ['date_trade', 'tday', 'span', 'open', 'close', 'high', 'low', 'volume', 'amount',
                                   'oi', 'ask1price', 'ask1volume', 'bid1price', 'bid1volume', 'contract']

    def get_history_data_org(self, stockfullcode, freq, start_dt=None, end_dt=None, dataframe=True):
        """
        从指定redis数据源读取实时数据
        :param stockfullcode:  支持传入列表
        :param freq: 1m, 目前只有一分钟
        :param start_dt: yyyy-mm-dd hh:mm:ss
        :param end_dt: yyyy-mm-dd hh:mm:ss
        :param dataframe: 返回Dataframe/Dictionary(key: stockfullcode, value: quote_df)
        :return:
        'date_trade',   行情时间戳
        'span',         频率 单位(秒)
        'open',         开盘价
        'close',        收盘价
        'high',         最高价
        'low',          最低价
        'volume',       成交量
        'amount',       成交金额
        'oi',           持仓量
        'ask1price',    卖1价
        'ask1volume',   卖1量
        'bid1price',    买1价
        'bid1volume',   买1量
        'contract'      主力合约别名，如SFIF9001
        """
        # 设置column列表
        quote_total_dict = dict()
        list_stockfullcode = [stockfullcode, ] if isinstance(stockfullcode, str) else stockfullcode
        start_dt = RedisDataFeed._to_datetime(start_dt)
        end_dt = RedisDataFeed._to_datetime(end_dt)
        # 转化fre为秒
        span_value = RedisDataFeed.convert_freq2num(freq_str=freq)
        # 用字典存储stockfullcode, idx_key
        stock_idx_key_dict = dict()
        for each_stock in list_stockfullcode:
            idx_key = '{}:{}:{}'.format(self.redis_md_prefix, each_stock, span_value)
            stock_idx_key_dict[each_stock] = idx_key
        # 全数字化时间戳
        start_score = start_dt.strftime('%Y%m%d%H%M%S')
        end_score = end_dt.strftime('%Y%m%d%H%M%S')
        for each_stock, each_idx_key in stock_idx_key_dict.items():
            lua_result = self.datafeed_lua_inst(keys=[each_idx_key], args=[start_score, end_score])
            if lua_result is None or len(lua_result) == 0:
                quote_total_dict[each_stock] = pd.DataFrame()
                continue
            # 将lua_list中的item变为JSON格式
            json_obj_list = [json.loads(item) for item in lua_result]
            # 转化为dataframe
            quote_df = pd.DataFrame.from_records(data=json_obj_list)

            # 日期格式转化
            quote_df['DateTrade'] = pd.to_datetime(quote_df['DateTrade'])

            # 构造tday
            if span_value < 86400:
                df_tday = get_tday_df(quote_df.iloc[0]['DateTrade'].date() - pd.Timedelta(days=10),
                                      quote_df.iloc[-1]['DateTrade'].date() + pd.Timedelta(days=10)).copy()
                df_tday.loc[:, 'next_tday'] = df_tday['tday'].shift(-1)
                df_tday = df_tday.astype(np.datetime64)
                quote_df.loc[:, 'date_trade_tmp'] = quote_df['DateTrade']
                quote_df.loc[:, 'date_trade_tmp'] = np.where(quote_df['DateTrade'].dt.hour < 6,
                                                             quote_df['DateTrade'] - pd.Timedelta(hours=6),
                                                             quote_df['DateTrade'])
                quote_df.loc[:, 'date'] = quote_df['date_trade_tmp'].dt.date.astype(np.datetime64)
                quote_df = pd.merge(quote_df, df_tday, how='left', left_on='date', right_on='tday')

                quote_df.sort_index(inplace=True, ascending=False)
                quote_df.loc[:, 'tday'] = np.where((quote_df['date_trade_tmp'].dt.hour > 16), quote_df['next_tday'],
                                                   quote_df['tday'])
                # 格式化tday为int
                quote_df['today'] = quote_df['tday'].dt.strftime('%Y%m%d')
            else:
                quote_df['tday'] = quote_df['DateTrade'].dt.strftime('%Y%m%d')

            # 排序
            quote_df.sort_values(by=['DateTrade'], inplace=True)

            # 字段读出来带有逗号,需要去除
            quote_df['Amount'] = quote_df['Amount'].str.replace(',', '')
            quote_df['Ask1Price'] = quote_df['Ask1Price'].str.replace(',', '')
            quote_df['Bid1Price'] = quote_df['Bid1Price'].str.replace(',', '')
            quote_df['Open'] = quote_df['Open'].str.replace(',', '')
            quote_df['Low'] = quote_df['Low'].str.replace(',', '')
            quote_df['High'] = quote_df['High'].str.replace(',', '')
            quote_df['Close'] = quote_df['Close'].str.replace(',', '')
            quote_df['OpenInterest'] = quote_df['OpenInterest'].str.replace(',', '')

            # 定义类型
            quote_df = quote_df.astype(
                dtype={
                    'Amount': np.float,
                    'Ask1Price': np.float,
                    'Ask1Volume': np.float,
                    'Bid1Price': np.float,
                    'Bid1Volume': np.float,
                    'Close': np.float,
                    'High': np.float,
                    'Low': np.float,
                    'Open': np.float,
                    'OpenInterest': np.int64,
                    'UpdateTime': np.datetime64,
                    'Volume': np.int64
                })

            # 重命名Columns
            quote_df.rename(columns=lambda c: c.lower(), inplace=True)
            quote_df.rename(columns={'openinterest': 'oi', 'datetrade': 'date_trade'}, inplace=True)

            # 构造symbol, contracts
            quote_df['contract'] = each_stock
            quote_df['span'] = span_value

            # 排序
            quote_df = quote_df[self.redis_quote_column]

            # 加入字典
            quote_total_dict[each_stock] = quote_df
        if not dataframe:
            return quote_total_dict
        total_quote_df = pd.concat(quote_total_dict.values()) if len(
            quote_total_dict.values()) > 0 else RedisDataFeed._return_data(
            return_frame=dataframe)
        return total_quote_df

    @staticmethod
    def register_lua(client):
        lua_body = '''
                local isKeyValid = redis.call('EXISTS', KEYS[1])
                local r = {}
                if isKeyValid == 1 then
                    local hashKeyList = redis.call('ZRANGEBYSCORE', KEYS[1], ARGV[1], ARGV[2])
                    for k,v in pairs(hashKeyList) do
                        local hashValue = redis.call('HGETALL', v)
                        local hashData = {}
                        if hashValue then
                            for idx = 1, #hashValue, 2 do
                                hashData[hashValue[idx]] = hashValue[idx + 1]
                            end
                            r[k] = cjson.encode(hashData)
                        end
                    end
                end
                return r
            '''
        return client.register_script(lua_body)

    @staticmethod
    def _to_datetime(tday=None):
        """
        获取tday
        :param tday:
        :return:
        """
        return pd.to_datetime(dt.datetime.now()) if tday is None else pd.to_datetime(tday)

    @staticmethod
    def convert_freq2num(freq_str):
        """
        将1分钟以上，1d以下的分钟线做转化
        :param freq_str:
        :return:
        """
        assert isinstance(freq_str, str)
        if freq_str == 'tick':
            return -1
        # 将freq_str等于单个周期单位的转化为标准周期的转化
        if freq_str in ['m', 'd', 'w', 'M', 'Q']:
            freq_str = '1{}'.format(freq_str)
        freq_regex = re.compile(r'^(?P<num>(\d+))(?P<unit>(s|m|d|w|M|Q))$')
        freq_num_unit = re.match(freq_regex, freq_str)
        if freq_num_unit is not None:
            freq_num = freq_num_unit.group('num')
            freq_unit = freq_num_unit.group('unit')
            # 如果是传入的是日线单位d，直接返回
            if freq_unit == 's':
                return int(freq_num)
            elif freq_unit == 'd':
                return int(freq_num) * 86400
            elif freq_unit == 'm':
                return int(freq_num) * 60
            elif freq_unit == 'w':
                return int(freq_num) * 86400 * 7
            elif freq_unit == 'M':
                return int(freq_num) * 86400 * 30
            elif freq_unit == 'Q':
                return int(freq_num) * 86400 * 90
        else:
            raise TypeError(f'不支持的频率转化类型: {freq_str}')

    @staticmethod
    def _return_data(return_frame=True):
        return pd.DataFrame() if return_frame else dict()


def _cache_tday_list():
    global TDAY_CACHE
    if not TDAY_CACHE:
        TDAY_CACHE = list()
    conn = pyodbc.connect(DB_HIS_STRING)
    cursor = conn.cursor()
    sql = f"""
    SELECT tday FROM lookup_tday ORDER BY tday ASC
    """
    TDAY_CACHE = [pd.to_datetime(item[0]).date() for item in cursor.execute(sql)]


def get_tday_df(start_tday=None, end_tday=None):
    global TDAY_CACHE
    if not TDAY_CACHE:
        _cache_tday_list()
    start_tday = dt.datetime(1900, 1, 1).date() if start_tday is None else pd.to_datetime(start_tday).date()
    end_tday = dt.datetime(2079, 6, 6).date() if end_tday is None else pd.to_datetime(end_tday).date()
    tday_list = [tday for tday in TDAY_CACHE if start_tday <= tday <= end_tday]
    tday_df = pd.DataFrame(data=tday_list, columns=['tday'])
    return tday_df
