# -*- coding:utf-8 -*-
import os
import logging
import pandas as pd
import datetime as dt
import numpy as np
import unittest
import json
import collections
from rio3.base import constant
from rio3.base.lookup_table import FUTURE_LEGAL_PERIOD
from rio3.base.constant import NAS_FUTURE_LEGAL_PERIOD_JSON_PATH
from rio3.datafeed.common import _get_market, _get_future_symbol, convert_freq2num
from rio3.datafeed.dbIO import get_previous_tday, get_tday_df
from rio3.datafeed.redis_feed import get_trade_period_json
from rio3.log.setup import setup_file_logger, setup_screen_logger
from rio3.utils.date import str_to_dtime

TRADE_PERIOD_DF = pd.DataFrame()

logger = logging.getLogger('root')

TRADE_PERIOD_DF_COLUMNS = ['symbol_code', 'date_start', 'date_end', 'time_start', 'time_end']

offset_period_tuple = collections.namedtuple('OffsetPeriod', 'quote_ca_time, quote_open_time')

kline_period_filter_tuple = collections.namedtuple('KlinePeriodFilter',
                                                   'period_start_date, period_end_date, '
                                                   'filter_time_start, filter_time_end')

merged_period_tuple = collections.namedtuple('PeriodMerge', 'trade_time, close_time')

TIME_FORMAT = '%H:%M:%S'

# 初始化集合过滤周期全局变量
FILTER_KLINE_PERIOD_DICT = dict()

# 初始化需要合并的时间段的全局变量
MERGE_PERIOD_DICT = dict()

# 初始化修改时间偏移量
OFFSET_PERIOD_DICT = dict()


def get_quote_offset_periods(market_type):
    offset_period_dict = get_all_quote_offset_periods()
    for key, item in offset_period_dict.items():
        if market_type == key:
            return offset_period_dict[key]


def get_all_quote_offset_periods():
    global OFFSET_PERIOD_DICT
    if not OFFSET_PERIOD_DICT:
        # 商品期货的集合竞价时间和正式开盘时间
        # 2013年之前以日盘交易时间开始
        cf_offset_period_one = offset_period_tuple(quote_ca_time=str_to_dtime('08:59:00'),
                                                   quote_open_time=str_to_dtime('09:00:00'))
        # 2013年之后以前一天的夜盘交易时间开始
        cf_offset_period_two = offset_period_tuple(quote_ca_time=str_to_dtime('20:59:00'),
                                                   quote_open_time=str_to_dtime('21:00:00'))

        # 股指期货的集合竞价时间和正式开盘时间
        # 2016年以前的竞价时间和开盘时间
        if_offset_period_one = offset_period_tuple(quote_ca_time=str_to_dtime('09:14:00'),
                                                   quote_open_time=str_to_dtime('09:15:00'))

        # 2016年以后的竞价时间和开盘时间
        if_offset_period_two = offset_period_tuple(quote_ca_time=str_to_dtime('09:29:00'),
                                                   quote_open_time=str_to_dtime('09:30:00'))

        # 股票的集合竞价时间和正式开盘时间
        stock_offset_period = offset_period_tuple(quote_ca_time=str_to_dtime('09:25:00'),
                                                  quote_open_time=str_to_dtime('09:30:00'))

        OFFSET_PERIOD_DICT['commodity_future'] = [cf_offset_period_one, cf_offset_period_two]
        OFFSET_PERIOD_DICT['index_future'] = [if_offset_period_one, if_offset_period_two]
        OFFSET_PERIOD_DICT['stock'] = [stock_offset_period]

    return OFFSET_PERIOD_DICT


# 通过stock_type或者过滤周期
def get_filter_quote_kline_periods_by_stype(stock_type):
    filter_quote_kline_periods = get_filter_quote_kline_periods()
    return filter_quote_kline_periods[stock_type]


# 记录集合竞价的周期，时间
# [集合竞价]
# 商品期货9点整的1分钟线是集合竞价。早7点-8点59分的1分钟数据需要被过滤
# 股指期货在2016年之前9点15分的1分钟线是集合竞价。早7点-9点14分的1分钟线数据需要被过滤
# 股指期货在2016年之后9点整的1分钟线是集合竞价。早7点-8点59分的1分钟线数据需要被过滤
# 股票9点25分的1分钟是集合竞价，早7点-9点24分之前的1分钟数据需要被过滤
def get_filter_quote_kline_periods():
    global FILTER_KLINE_PERIOD_DICT
    if not FILTER_KLINE_PERIOD_DICT:
        # 夜盘结束误差值大于1分钟的Kline需要被过滤
        common_night_filter_period = kline_period_filter_tuple(period_start_date=constant.DATETIME_MIN,
                                                               period_end_date=constant.DATETIME_MAX,
                                                               filter_time_start='02:32:00', filter_time_end='08:00:00')

        # 日盘结束误差值大于1分钟的Kline需要被过滤
        common_day_filter_period = kline_period_filter_tuple(period_start_date=constant.DATETIME_MIN,
                                                             period_end_date=constant.DATETIME_MAX,
                                                             filter_time_start='15:17:00', filter_time_end='20:00:00')

        # 股票的集合竞价过滤时间 （全部周期：早7点-9点24分之前的1分钟数据需要被过滤）
        stock_call_auction_period = kline_period_filter_tuple(period_start_date=constant.DATETIME_MIN,
                                                              period_end_date=constant.DATETIME_MAX,
                                                              filter_time_start='07:00:00', filter_time_end='9:24:00')

        # 股指期货的集合竞价过滤时间 （2016年1月4日前：早7点-9点14分的1分钟线数据需要被过滤）
        if_call_auction_period_one = kline_period_filter_tuple(period_start_date=constant.DATETIME_MIN,
                                                               period_end_date=dt.datetime(2016, 1, 4),
                                                               filter_time_start=dt.datetime.strptime('07:00:00',
                                                                                                      TIME_FORMAT).time(),
                                                               filter_time_end=dt.datetime.strptime('09:14:00',
                                                                                                    TIME_FORMAT).time())
        # 股指期货的集合竞价过滤时间 （2016年1月4日到至今：早7点-8点59分的1分钟线数据需要被过滤）
        if_call_auction_period_two = kline_period_filter_tuple(period_start_date=dt.datetime(2016, 1, 4),
                                                               period_end_date=constant.DATETIME_MAX,
                                                               filter_time_start=dt.datetime.strptime('07:00:00',
                                                                                                      TIME_FORMAT).time(),
                                                               filter_time_end=dt.datetime.strptime('08:59:00',
                                                                                                    TIME_FORMAT).time())
        # 商品期货的集合竞价过滤时间 （全部周期：早7点-8点59分的1分钟数据需要被过滤）
        cf_call_auction_period = kline_period_filter_tuple(period_start_date=constant.DATETIME_MIN,
                                                           period_end_date=constant.DATETIME_MAX,
                                                           filter_time_start=dt.datetime.strptime('07:00:00',
                                                                                                  TIME_FORMAT).time(),
                                                           filter_time_end=dt.datetime.strptime('08:59:00',
                                                                                                TIME_FORMAT).time())
        FILTER_KLINE_PERIOD_DICT['stock'] = [common_night_filter_period, common_day_filter_period,
                                             stock_call_auction_period]
        FILTER_KLINE_PERIOD_DICT['index_future'] = [common_night_filter_period, common_day_filter_period,
                                                    if_call_auction_period_one, if_call_auction_period_two]
        FILTER_KLINE_PERIOD_DICT['commodity_future'] = [common_night_filter_period, common_day_filter_period,
                                                        cf_call_auction_period]
    return FILTER_KLINE_PERIOD_DICT


# 交易时间段过滤合并原则
# 11:31 -> 11:30
# 12:59 -> 13:00
# 20:59 -> 21:00
# 21:00 < time < 4:00, 晚上最后一笔 31 分 和 30 分的交易数据需要合并
# 21:00 < time < 4:00, 晚上最后一笔 01 分 和 00 分的交易数据需要合并
def get_merged_quote_periods():
    global MERGE_PERIOD_DICT
    if not MERGE_PERIOD_DICT:
        common_merge_period = [merged_period_tuple(trade_time='11:31:00', close_time='11:30:00'),
                               merged_period_tuple(trade_time='12:59:00', close_time='13:00:00'),
                               merged_period_tuple(trade_time='20:59:00', close_time='21:00:00')]
        commodity_merge_period = [merged_period_tuple(trade_time='10:16:00', close_time='10:15:00')]
        commodity_merge_period.extend(common_merge_period)
        MERGE_PERIOD_DICT['index_future'] = common_merge_period
        MERGE_PERIOD_DICT['commodity_future'] = commodity_merge_period
        MERGE_PERIOD_DICT['stock'] = common_merge_period
    return MERGE_PERIOD_DICT


# 通过stock_type获取合并周期
def get_merged_quote_periods_by_stype(stock_type):
    merged_quote_periods = get_merged_quote_periods()
    return merged_quote_periods[stock_type]


def get_stock_period_df(stockfullcode=None):
    """
    返回股票的交易时间定义
    :param stockfullcode:
    :return:
    """
    period_am = [pd.to_datetime(dt.datetime(1990, 1, 1)), pd.to_datetime(constant.DATETIME_MAX),
                 pd.to_datetime('09:30:00').time(), pd.to_datetime('11:30:00').time()]
    period_pm = [pd.to_datetime(dt.datetime(1990, 1, 1)), pd.to_datetime(constant.DATETIME_MAX),
                 pd.to_datetime('13:00:00').time(), pd.to_datetime('15:00:00').time()]
    data_list = [period_am, period_pm]
    data_df = pd.DataFrame(data=data_list, columns=['date_start', 'date_end', 'time_start', 'time_end'])
    if stockfullcode is not None and isinstance(stockfullcode, str):
        data_df['symbol_code'] = stockfullcode
    return data_df


# 传入Stockfullcode获取该品种的正常交易时间表
def get_trade_period_df(stockfullcode=None):
    global TRADE_PERIOD_DF
    if TRADE_PERIOD_DF is None or TRADE_PERIOD_DF.empty:
        if os.path.exists(NAS_FUTURE_LEGAL_PERIOD_JSON_PATH):
            with open(NAS_FUTURE_LEGAL_PERIOD_JSON_PATH) as data_file:
                period_json_dict = json.loads(data_file.read())
        else:
            json_str = get_trade_period_json(cache_path='redis')
            try:
                period_json_dict = json.loads(json_str)
            except:
                period_json_dict = FUTURE_LEGAL_PERIOD
        # 循环period_json，构造stockfullcode=>trade_time的字典组合
        data_row_list = list()
        for symbol_code, period_list in period_json_dict.items():
            for item_dict in period_list:
                data_row = list()
                data_row.append(symbol_code)
                data_row.append(pd.to_datetime(item_dict['date_start']).date())
                data_row.append(pd.to_datetime(item_dict['date_end']).date())
                data_row.append(pd.to_datetime(item_dict['time_start']).time())
                data_row.append(pd.to_datetime(item_dict['time_end']).time())
                data_row_list.append(data_row)
        TRADE_PERIOD_DF = pd.DataFrame(columns=TRADE_PERIOD_DF_COLUMNS, data=data_row_list)
    try:
        if stockfullcode is not None and TRADE_PERIOD_DF is not None:
            symbol_code = '{}{}'.format(_get_market(stockfullcode), _get_future_symbol(stockfullcode))
            matched_trade_period_df = TRADE_PERIOD_DF[TRADE_PERIOD_DF['symbol_code'] == symbol_code]
        else:
            matched_trade_period_df = TRADE_PERIOD_DF
    except Exception as e:
        logging.warning(
            'msg: No future trading period definition for {}, please check lookup_table.py or FutureLegalPeriod.json. '
            'Detail: {}'.format(stockfullcode, e))
        matched_trade_period_df = pd.DataFrame()
    return matched_trade_period_df


def convert_dt_to_tday(dt_quote_df, dt_column='DateTrade', tday_column='tday'):
    """
    构造交易日列
    :param dt_quote_df:
    :param dt_column:
    :param tday_column:
    :return:
    """
    assert isinstance(dt_quote_df, pd.DataFrame)
    if dt_quote_df.empty:
        return dt_quote_df
    if dt_column not in dt_quote_df.columns:
        logger.warning('当前数据中不存在 {} '.format(dt_column))
        return dt_quote_df
    quote_columns = dt_quote_df.columns.tolist()
    quote_columns.extend([tday_column])
    # 获取交易日列表
    df_tday = get_tday_df(dt_quote_df.iloc[0][dt_column].date() - pd.Timedelta(days=10),
                          dt_quote_df.iloc[-1][dt_column].date() + pd.Timedelta(days=10)).copy()
    df_tday.loc[:, 'next_tday'] = df_tday['tday'].shift(-1)
    df_tday = df_tday.astype(np.datetime64)
    dt_quote_df.loc[:, 'date_trade_tmp'] = dt_quote_df[dt_column]
    dt_quote_df.loc[:, 'date_trade_tmp'] = np.where(dt_quote_df[dt_column].dt.hour < 6,
                                                    dt_quote_df[dt_column] - pd.Timedelta(hours=6),
                                                    dt_quote_df[dt_column])
    dt_quote_df.loc[:, 'date'] = dt_quote_df['date_trade_tmp'].dt.date.astype(np.datetime64)
    dt_quote_df = pd.merge(dt_quote_df, df_tday, how='left', left_on='date', right_on='tday')

    dt_quote_df.sort_index(inplace=True, ascending=False)
    dt_quote_df.loc[:, tday_column] = np.where((dt_quote_df['date_trade_tmp'].dt.hour > 16), dt_quote_df['next_tday'],
                                               dt_quote_df['tday'])
    dt_quote_df.sort_values(by=[dt_column], inplace=True)

    return dt_quote_df[quote_columns]


def get_legal_period_data(data_df, stockfullcode, freq=None, is_call_auction=False):
    """
    # 通过查找FutureLegalPeriod.json的期货交易时间，过滤掉非交易时段的数据
    :param data_df: 原始数据
    :param stockfullcode: 期货代码，目前只定义了期货时间。股票的交易时间是固定的，可以自行过滤
    :param freq: 级别，如果非Tick级别的，时间过滤项DateTrade，如果是Tick级别的，时间过滤项是Time
    :param is_call_auction: 是否保留集合竞价的bar数据
    :return:
    """
    assert isinstance(data_df, pd.DataFrame)
    if data_df.empty:
        return data_df
    # 判断行情是何种格式
    data_format = 'N/A'
    if 'date' in data_df.columns and 'ets' in data_df.columns:
        data_format = 'zq'
    elif 'DateTrade' in data_df.columns and 'Tday' in data_df.columns:
        data_format = 'std'
    # 获取当前数据的周期
    if 'span' in data_df.columns:
        span = data_df['span'].unique()
    else:
        if freq is not None:
            span = convert_freq2num(freq_str=freq)
        else:
            logger.warning('无法确定当前数据频率，请指定数据频率')
            return data_df
    # 如果传入周期大于等于1天
    if span >= 86400:
        return data_df
    symbol_period_df = get_trade_period_df(stockfullcode=stockfullcode)
    # 如果没有得到交易时间，就直接返回原来的数据
    if symbol_period_df is None or symbol_period_df.empty:
        logger.info(u'[{}] trading period info not found, Return original data!'.format(stockfullcode))
        return data_df
    # 震球格式类型需要构造DateTrade时间字段
    if data_format == 'zq':
        data_df.loc[:, 'date_trade_temp'] = data_df.date * 1000000 + data_df.ets
        data_df.loc[:, 'date_trade_temp'] = pd.to_datetime(data_df['date_trade_temp'], format='%Y%m%d%H%M%S')
    # 原始格式只需要赋值即可
    elif data_format == 'std':
        data_df.loc[:, 'date_trade_temp'] = data_df['DateTrade']
    else:
        logger.warning('时间构成字段不足，请确认格式，ZQ/STD')
        return data_df
    filter_date_column = 'date_trade_temp'
    if data_df is not None and not data_df.empty and symbol_period_df is not None:
        filtered_data_df = data_df.copy()
        # 按照起始和结束时间段划分
        grouped_period_iterator = symbol_period_df.groupby([symbol_period_df.date_start, symbol_period_df.date_end])
        # 将起始时间和具体时间
        # 将交易时间点内的数据逐步筛选，然后删除掉正常交易点的数据，最后剩下来的就是非交易时间点数据，从整体的data_df中去除
        # 需要被过滤掉的数据集合变量
        excluded_data_df = pd.DataFrame()
        for period, period_time_df in grouped_period_iterator:
            period_time_df_copy = period_time_df.copy()
            # 重新设置过滤时间点的time:
            tday_start = get_previous_tday(period[0]).replace(hour=20, minute=0, second=0, microsecond=0)
            tday_end = pd.to_datetime(period[1]).replace(hour=16, minute=0, second=0, microsecond=0)
            # 筛选出date时间段的数据
            period_datetime_df = filtered_data_df[
                (filtered_data_df[filter_date_column] >= tday_start) & (
                        filtered_data_df[filter_date_column] <= tday_end)]
            # 循环筛选出time时间段的数据
            period_datetime_df = period_datetime_df.set_index(filter_date_column)
            for each_time_period in period_time_df_copy.itertuples():
                # 如果读取半小时线，会因为首尾的时间定义而过滤掉10:30的线
                if span >= 1800:
                    # 特别处理商品期货10:30的行情
                    if each_time_period.time_start == dt.time(10, 30):
                        each_period_datetime_df = period_datetime_df.between_time(each_time_period.time_start,
                                                                                  each_time_period.time_end,
                                                                                  include_start=True)
                    # 特别处理期货15:15的行情，这个时候需要开放结束时间为15:30
                    elif each_time_period.time_end == dt.time(15, 15):
                        each_period_datetime_df = period_datetime_df.between_time(each_time_period.time_start,
                                                                                  dt.time(15, 30),
                                                                                  include_start=True)
                    else:
                        each_period_datetime_df = period_datetime_df.between_time(each_time_period.time_start,
                                                                                  each_time_period.time_end,
                                                                                  include_start=is_call_auction)
                else:
                    # 过滤出每一个时间片内的数据
                    each_period_datetime_df = period_datetime_df.between_time(each_time_period.time_start,
                                                                              each_time_period.time_end,
                                                                              include_start=is_call_auction)
                # 从当前period_time_df中过滤掉
                period_datetime_df = period_datetime_df[~period_datetime_df.index.isin(each_period_datetime_df.index)]
            # 核查当前时间段剩余的行情数目
            if not period_datetime_df.empty:
                excluded_data_df = pd.concat([excluded_data_df, period_datetime_df])
        # 从总的data_df中去除excluded_data_df
        if not excluded_data_df.empty:
            data_df = data_df[~data_df[filter_date_column].isin(excluded_data_df.index)]
    drop_columns = ['Tday', 'date_trade_temp'] if data_format == 'zq' else ['date_trade_temp']
    for drop_column in drop_columns:
        if drop_column in data_df.columns:
            data_df = data_df.drop([drop_column], axis=1)
    return data_df


def get_merged_trading_time_period(stockfullcode):
    """
    构造合约的交易周期表，按照日夜盘区分，只保留开盘时间和收盘时间
    :param stockfullcode:
    :return:
    """
    market_name = _get_market(stockfullcode=stockfullcode)
    symbol_trade_period_list = list()
    # 如果是股票，手动构造交易时间
    if market_name in ['SH', 'SZ']:
        symbol_trade_period_list.append([stockfullcode, 20000101, 20790606, 93000, 150000, 'day'])
    else:
        # 获取期货品种的交易周期定义
        symbol_trade_time_period_df = get_trade_period_df(stockfullcode=stockfullcode).copy()
        symbol_trade_time_period_df['date_start'] = pd.to_datetime(
            symbol_trade_time_period_df['date_start']).dt.strftime('%Y%m%d').astype(np.int64)
        symbol_trade_time_period_df['date_end'] = pd.to_datetime(
            symbol_trade_time_period_df['date_end']).dt.strftime('%Y%m%d').astype(np.int64)
        symbol_trade_time_period_df['time_start'] = pd.to_datetime(symbol_trade_time_period_df['time_start'],
                                                                   format='%H:%M:%S').dt.strftime('%H%M%S').astype(
            np.int64)
        symbol_trade_time_period_df['time_end'] = pd.to_datetime(symbol_trade_time_period_df['time_end'],
                                                                 format='%H:%M:%S').dt.strftime('%H%M%S').astype(
            np.int64)
        # 对周期定义进行合并，只留日盘总周期和夜盘总周期
        for each_period_pair, period_pair_df in symbol_trade_time_period_df.groupby(by=['date_start', 'date_end']):
            # 过滤出日盘时间
            period_pair_df_copy = period_pair_df.copy()
            day_period_pair = period_pair_df_copy[period_pair_df_copy['time_start'] < 200000]
            night_period_pair = period_pair_df_copy[period_pair_df_copy['time_start'] > 200000]
            # 构造当前周期日盘和夜盘的首尾时间
            if not day_period_pair.empty:
                symbol_trade_period_list.append(
                    [stockfullcode, each_period_pair[0], each_period_pair[1], day_period_pair['time_start'].min(),
                     day_period_pair['time_end'].max(), 'day'])
            if not night_period_pair.empty:
                symbol_trade_period_list.append(
                    [stockfullcode, each_period_pair[0], each_period_pair[1], night_period_pair['time_start'].min(),
                     night_period_pair['time_end'].max(), 'night'])
    # 得出过滤掉盘中时间段的交易时间范围
    symbol_trade_period_df = pd.DataFrame(data=symbol_trade_period_list,
                                          columns=['stockfullcode', 'date_start', 'date_end', 'time_start',
                                                   'time_end', 'trading_period'])
    return symbol_trade_period_df


def get_trading_border_time(quote_df):
    """
    获取每一根K线的边界开盘/收盘时间
    :param quote_df:
    :param type:
    :return:
    """
    if quote_df.empty:
        return quote_df
    return_quote_df = quote_df
    return_quote_df_list = list()
    # 按照symbol做group by，获取每一个合约的交易时间
    for each_symbol, symbol_quote_df in quote_df.groupby(by=['symbol']):
        # 获取首尾交易时间周期
        symbol_trade_period_df = get_merged_trading_time_period(stockfullcode=each_symbol)
        # 构造于quote_df的symbol做merge的条件列
        symbol_quote_df_copy = symbol_quote_df.copy()
        # 将quote_df与symbol_trade_period_df做合并，得到quote_df * len(symbol_trade_period_df)的数据
        symbol_quote_df_copy = pd.merge(left=symbol_quote_df_copy, right=symbol_trade_period_df, left_on='symbol',
                                        right_on='stockfullcode', how='left')
        # 再对date做date_start/end之间的过滤，ets做time_start/end之间的过滤
        # 拆分日/盘进行过滤
        day_symbol_quote_df = symbol_quote_df_copy[
            (symbol_quote_df_copy['ets'] > 40000) & (symbol_quote_df_copy['ets'] < 160000)]
        # night_symbol_quote_df = symbol_quote_df_copy[~symbol_quote_df_copy.index.isin(day_symbol_quote_df.index)]
        night_symbol_quote_df = symbol_quote_df_copy[
            (symbol_quote_df_copy['ets'] < 40000) | (symbol_quote_df_copy['ets'] > 200000)]

        day_symbol_quote_df_copy = day_symbol_quote_df[
            (day_symbol_quote_df['date'] >= day_symbol_quote_df['date_start']) & (
                    day_symbol_quote_df['date'] <= day_symbol_quote_df['date_end']) & (
                    day_symbol_quote_df['sts'] >= day_symbol_quote_df['time_start']) & (
                    day_symbol_quote_df['ets'] <= day_symbol_quote_df['time_end'])]
        night_symbol_quote_df_copy = night_symbol_quote_df[
            (night_symbol_quote_df['date'] >= night_symbol_quote_df['date_start']) & (
                    night_symbol_quote_df['date'] <= night_symbol_quote_df['date_end']) & (
                    night_symbol_quote_df['trading_period'] == 'night')]
        symbol_total_quote_df = pd.concat([day_symbol_quote_df_copy, night_symbol_quote_df_copy])
        # 排序
        symbol_total_quote_df = symbol_total_quote_df.sort_values(by=['date', 'ets'])
        # 加入合计列表
        return_quote_df_list.append(symbol_total_quote_df)
    if len(return_quote_df_list) > 0:
        return_quote_df = pd.concat(return_quote_df_list)
    return return_quote_df


def get_trading_time_border_diff(quote_df):
    """
    获取每一根K线距离当前日/盘周期的开盘时间差和收盘时间差
    由于要进行周期开/收盘时间合并，该函数默认过滤掉不在合理交易时间范围内的数据
    :param quote_df:
    :return:
    """
    if quote_df.empty:
        return quote_df
    return_quote_df = quote_df
    return_quote_df_list = list()
    if 'time_start' not in quote_df.columns or 'time_end' not in quote_df.columns:
        quote_df = get_trading_border_time(quote_df=quote_df)
    # 开始循环解析quote_df
    for each_symbol, symbol_quote_df in quote_df.groupby(by=['symbol']):
        symbol_quote_df_copy = symbol_quote_df.copy()
        # 过滤不合理交易时间段数据
        symbol_quote_df_copy = get_legal_period_data(data_df=symbol_quote_df_copy, stockfullcode=each_symbol).copy()
        is_commodity_future = False
        market_name = _get_market(each_symbol)
        # 根据品种，构造下午开盘时间，如果是股指和股票，时间为13:00:00，如果是商品，时间为13:30:00
        if market_name in ['SF', 'SH', 'SZ']:
            afternoon_start_time = pd.Timedelta(hours=13, minutes=0, seconds=0)
            noon_minutes = pd.Timedelta(hours=1, minutes=30, seconds=0)
        else:
            is_commodity_future = True
            afternoon_start_time = pd.Timedelta(hours=13, minutes=30, seconds=0)
            noon_minutes = pd.Timedelta(hours=2, minutes=0, seconds=0)
        # 构造计算时间线格式为datetime
        symbol_quote_df_copy['ets_dt'] = pd.to_datetime(symbol_quote_df_copy['ets'].copy().astype(np.str).str.zfill(6),
                                                        format='%H%M%S')
        symbol_quote_df_copy['start_border_dt'] = pd.to_datetime(
            symbol_quote_df_copy['time_start'].copy().astype(np.str).str.zfill(6),
            format='%H%M%S')
        symbol_quote_df_copy['end_border_dt'] = pd.to_datetime(
            symbol_quote_df_copy['time_end'].copy().astype(np.str).str.zfill(6), format='%H%M%S')
        # 再拆分为timedelta
        symbol_quote_df_copy['ets_time'] = pd.to_timedelta(symbol_quote_df_copy['ets_dt'].dt.hour,
                                                           unit='h') + pd.to_timedelta(
            symbol_quote_df_copy['ets_dt'].dt.minute, unit='m') + pd.to_timedelta(
            symbol_quote_df_copy['ets_dt'].dt.second, unit='s')
        symbol_quote_df_copy['start_border_time'] = pd.to_timedelta(symbol_quote_df_copy['start_border_dt'].dt.hour,
                                                                    unit='h') + pd.to_timedelta(
            symbol_quote_df_copy['start_border_dt'].dt.minute, unit='m') + pd.to_timedelta(
            symbol_quote_df_copy['start_border_dt'].dt.second, unit='s')
        symbol_quote_df_copy['end_border_time'] = pd.to_timedelta(symbol_quote_df_copy['end_border_dt'].dt.hour,
                                                                  unit='h') + pd.to_timedelta(
            symbol_quote_df_copy['end_border_dt'].dt.minute, unit='m') + pd.to_timedelta(
            symbol_quote_df_copy['end_border_dt'].dt.second, unit='s')

        # 计算与收盘边界的时间差，获取两个时间段的差值，单位为分钟，支持跨天相减
        symbol_quote_df_copy['end_border_diff_time'] = np.where(
            symbol_quote_df_copy['end_border_time'] >= symbol_quote_df_copy['ets_time'],
            (symbol_quote_df_copy['end_border_time'] - symbol_quote_df_copy['ets_time']).copy(),
            ((symbol_quote_df_copy['end_border_time'] + pd.Timedelta(days=1)) - symbol_quote_df_copy['ets_time']))
        # 计算与开盘时间的时间差
        symbol_quote_df_copy['start_border_diff_time'] = np.where(
            symbol_quote_df_copy['start_border_time'] <= symbol_quote_df_copy['ets_time'],
            symbol_quote_df_copy['ets_time'] - symbol_quote_df_copy['start_border_time'],
            ((symbol_quote_df_copy['ets_time'] + pd.Timedelta(days=1)) - symbol_quote_df_copy['start_border_time']))
        # 去除休盘噪音时间，需要考虑商品和股指的休息时间
        day_middle_end_time = pd.Timedelta(hours=11, minutes=30, seconds=0)
        if is_commodity_future:
            # 商品的休息为上午10:15 ~ 10:30, 下午为13:30，所以算休盘时间差的时候，需要扣除这些时间
            day_start_time = pd.Timedelta(hours=8, minutes=0, seconds=0)
            day_rest_start_time = pd.Timedelta(hours=10, minutes=15, seconds=0)
            day_rest_end_time = pd.Timedelta(hours=10, minutes=30, seconds=0)
            # 处理计算end_border_diff_time
            # 当交易时间为日盘交易时间，并且小于等于10: 15的时候，需要扣去上午休盘时间和中午休盘时间
            symbol_quote_df_copy['end_border_diff_time'] = np.where(((symbol_quote_df_copy[
                                                                          'ets_time'] <= day_rest_start_time) & (
                                                                             symbol_quote_df_copy[
                                                                                 'ets_time'] > day_start_time)),
                                                                    symbol_quote_df_copy[
                                                                        'end_border_diff_time'] - pd.Timedelta(
                                                                        minutes=15) - noon_minutes,
                                                                    symbol_quote_df_copy['end_border_diff_time'])
            # 当交易时间 >= 10: 30, 并且小于下午开盘时间的时候，只需要扣除中午休盘时间
            symbol_quote_df_copy['end_border_diff_time'] = np.where(((symbol_quote_df_copy[
                                                                          'ets_time'] >= day_rest_end_time) & (
                                                                             symbol_quote_df_copy[
                                                                                 'ets_time'] > day_start_time) & (
                                                                             symbol_quote_df_copy[
                                                                                 'ets_time'] < afternoon_start_time)),
                                                                    symbol_quote_df_copy[
                                                                        'end_border_diff_time'] - noon_minutes,
                                                                    symbol_quote_df_copy['end_border_diff_time'])
            # 处理计算start_border_diff_time, 如果商品的时间大于10:30 并且小于11:30，需要减去15分钟休盘时间
            symbol_quote_df_copy['start_border_diff_time'] = np.where(((symbol_quote_df_copy[
                                                                            'ets_time'] >= day_rest_end_time) & (
                                                                               symbol_quote_df_copy[
                                                                                   'ets_time'] <= day_middle_end_time)),
                                                                      symbol_quote_df_copy[
                                                                          'start_border_diff_time'] - pd.Timedelta(
                                                                          minutes=15),
                                                                      symbol_quote_df_copy['start_border_diff_time'])
            # 如果时间大于并且各自开盘的时间, 商品需要扣除中午休息时间和15分钟休息时间
            symbol_quote_df_copy['start_border_diff_time'] = np.where(((symbol_quote_df_copy[
                                                                            'ets_time'] >= afternoon_start_time) & (
                                                                               symbol_quote_df_copy[
                                                                                   'ets_time'] <= pd.Timedelta(hours=16,
                                                                                                               minutes=0,
                                                                                                               seconds=0))),
                                                                      symbol_quote_df_copy[
                                                                          'start_border_diff_time'] - pd.Timedelta(
                                                                          minutes=15) - noon_minutes,
                                                                      symbol_quote_df_copy['start_border_diff_time'])
        else:
            # 股指和股票的休息为11:30 ~ 13:00，所以算休盘时间差的时候，需要扣除这些时间
            symbol_quote_df_copy['end_border_diff_time'] = np.where(
                symbol_quote_df_copy['ets_time'] < afternoon_start_time, symbol_quote_df_copy[
                    'end_border_diff_time'] - noon_minutes, symbol_quote_df_copy['end_border_diff_time'])
            # 如果时间大于并且各自开盘的时间, 股指需要扣除中午休息时间，
            symbol_quote_df_copy['start_border_diff_time'] = np.where(((symbol_quote_df_copy[
                                                                            'ets_time'] >= afternoon_start_time) & (
                                                                               symbol_quote_df_copy[
                                                                                   'ets_time'] <= pd.Timedelta(hours=16,
                                                                                                               minutes=0,
                                                                                                               seconds=0))),
                                                                      symbol_quote_df_copy[
                                                                          'start_border_diff_time'] - noon_minutes,
                                                                      symbol_quote_df_copy['start_border_diff_time'])
        return_quote_df_list.append(symbol_quote_df_copy)
    if len(return_quote_df_list) > 0:
        return_quote_df = pd.concat(return_quote_df_list)
        return_quote_df['start_border_diff_minutes'] = return_quote_df['start_border_diff_time'].dt.total_seconds() / 60
        return_quote_df['end_border_diff_minutes'] = return_quote_df['end_border_diff_time'].dt.total_seconds() / 60
        return_quote_df.drop(
            ['stockfullcode', 'trading_period', 'ets_dt', 'start_border_dt', 'end_border_dt', 'ets_time',
             'start_border_time',
             'end_border_time', 'start_border_diff_time', 'end_border_diff_time'], axis=1, inplace=True)
        return_quote_df.reset_index(drop=True, inplace=True)
    return return_quote_df


# 通过stock_type或者过滤周期
def get_filter_quote_kline_periods_by_stype(stock_type):
    filter_quote_kline_periods = get_filter_quote_kline_periods()
    return filter_quote_kline_periods[stock_type]


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_get_legal_period_data(self):
        period_data_df = get_trade_period_df()
        pass

    def test_get_stock_period_df(self):
        data1 = get_stock_period_df()
        data2 = get_stock_period_df('SH600000')
        pass


if __name__ == "__main__":
    logger = logging.getLogger('root')
    setup_file_logger()
    setup_screen_logger()
    unittest.main()
