# -*- coding:utf-8 -*-
import logging
import datetime as dt
import unittest
from rio3.base.object import BaseObject
from rio3.base.enumerator import ENUM_ED_TYPE
from rio3.log.setup import setup_file_logger, setup_screen_logger
from rio3.datafeed.common import _get_market, _get_contract_code


class BasicQuoteInfo(BaseObject):
    def __init__(self, name, date, tday, o=0, h=0, l=0, c=0, v=0, amt=0, oi=0, sp1=0, sv1=0, bp1=0, bv1=0, mid=0,
                 midv=0):
        super(BasicQuoteInfo, self).__init__(name=name)
        self.date = date
        self.tday = tday
        self.o = float(o)
        self.h = float(h)
        self.l = float(l)
        self.c = float(c)
        self.v = int(v)
        self.amt = float(amt)
        self.oi = int(oi)
        self.sp1 = float(sp1)
        self.sv1 = int(sv1)
        self.bp1 = float(bp1)
        self.bv1 = int(bv1)
        self.mid = float(mid)
        self.midv = float(midv)

    def __add__(self, other):
        # 将两个同品种的股票相加
        if other.name != self.name:
            raise TypeError(u'两个不同的合约无法相加！{}, {}'.format(self.name, other.name))
        if self.date == max(self.date, other.date):
            info_latter = self
            info_former = other
        else:
            info_latter = other
            info_former = self
        bqinfo = BasicQuoteInfo(
            name=self.name,
            date=max(self.date, other.date),
            tday=info_latter.tday,
            o=info_former.o,
            h=max(self.h, other.h),
            l=min(self.l, other.l),
            c=info_latter.c,
            v=self.v + other.v,
            amt=self.amt + other.amt,
            oi=info_latter.oi,
            sp1=info_latter.sp1,
            sv1=info_latter.sv1,
            bp1=info_latter.bp1,
            bv1=info_latter.bv1,
            mid=info_latter.mid,
            midv=info_latter.midv
        )
        return bqinfo

    @property
    def stockfullcode(self):
        return self.name

    @property
    def date_trade(self):
        return self.date

    @property
    def amt_volumn(self):
        return self.v

    @property
    def amt_amount(self):
        return self.amt

    @property
    def val_open(self):
        return self.o

    @property
    def val_high(self):
        return self.h

    @property
    def val_low(self):
        return self.l

    @property
    def val_close(self):
        return self.c

    @property
    def val_preclose(self):
        return -1

    @property
    def dailyPreclose(self):
        return -1

    @property
    def amt_wp(self):
        return -1

    @property
    def amt_np(self):
        return -1

    @property
    def val_sp1(self):
        return self.sp1

    @property
    def val_sv1(self):
        return self.sv1

    @property
    def val_bp1(self):
        return self.sp1

    @property
    def val_bv1(self):
        return self.sv1

    @property
    def val_mid(self):
        return self.mid

    @property
    def val_midv(self):
        return self.midv

    @property
    def val_OI(self):
        return self.oi

    def calc_ed_info(self, ed_info_list, ed_type=ENUM_ED_TYPE.NoneEd):
        new_bq_info = BasicQuoteInfo(name=self.name, date=self.date, tday=self.tday, o=self.o, h=self.h, l=self.l,
                                     c=self.c, v=self.v, amt=self.amt, oi=self.oi)
        if ed_type == ENUM_ED_TYPE.NoneEd:
            return new_bq_info
        if ed_type == ENUM_ED_TYPE.PreEd:
            ed_info_list.sort(key=lambda e: e.date_ed, reverse=False)
            for ed_info in ed_info_list:
                if (ed_info.is_future and new_bq_info.date < ed_info.date_ed + dt.timedelta(hours=20)) or \
                        (not ed_info.is_future and new_bq_info.date < ed_info.date_ed):
                    new_bq_info.c = self.get_pre_ed_price(new_bq_info.c, ed_info)
                    new_bq_info.o = self.get_pre_ed_price(new_bq_info.o, ed_info)
                    new_bq_info.h = self.get_pre_ed_price(new_bq_info.h, ed_info)
                    new_bq_info.l = self.get_pre_ed_price(new_bq_info.l, ed_info)
                    new_bq_info.amt = self.get_pre_ed_amount(new_bq_info.amt, ed_info, new_bq_info.v)
                    new_bq_info.v = self.get_pre_ed_volume(new_bq_info.v, ed_info)
        elif ed_type == ENUM_ED_TYPE.PostEd:
            ed_info_list.sort(key=lambda e: e.date_ed, reverse=True)
            for ed_info in ed_info_list:
                if (ed_info.is_future and new_bq_info.date > ed_info.date_ed + dt.timedelta(hours=20)) or \
                        (not ed_info.is_future and new_bq_info.date > ed_info.date_ed):
                    new_bq_info.c = self.get_post_ed_price(new_bq_info.c, ed_info)
                    new_bq_info.o = self.get_post_ed_price(new_bq_info.o, ed_info)
                    new_bq_info.h = self.get_post_ed_price(new_bq_info.h, ed_info)
                    new_bq_info.l = self.get_post_ed_price(new_bq_info.l, ed_info)
                    new_bq_info.amt = self.get_post_ed_amount(new_bq_info.amt, ed_info, new_bq_info.v)
                    new_bq_info.v = self.get_post_ed_volume(new_bq_info.v, ed_info)
        else:
            raise NotImplementedError(u'不支持的复权模式{}'.format(ed_type))
        return new_bq_info

    @staticmethod
    def get_pre_ed_price(org_price, ed_info):
        return (org_price + ed_info.pgbl * ed_info.pgjg - ed_info.fhbl * 0.9) / (1 + ed_info.sgbl + ed_info.pgbl)

    @staticmethod
    def get_pre_ed_volume(org_vol, ed_info):
        if ed_info.is_future:
            return org_vol
        return org_vol * (1 + ed_info.pgbl + ed_info.sgbl)

    @staticmethod
    def get_pre_ed_amount(org_amount, ed_info, org_vol):
        if org_vol == 0:
            return 0
        if org_amount < 0:
            return org_amount
        return BasicQuoteInfo.get_pre_ed_volume(org_vol, ed_info) * BasicQuoteInfo.get_pre_ed_price(
            org_amount / org_vol, ed_info)

    @staticmethod
    def get_post_ed_price(org_price, ed_info):
        return org_price * (1 + ed_info.sgbl + ed_info.pgbl) - ed_info.pgbl * ed_info.pgjg + ed_info.fhbl * 0.9

    @staticmethod
    def get_post_ed_volume(org_vol, ed_info):
        if ed_info.is_future:
            return org_vol
        return org_vol / (1 + ed_info.pgbl + ed_info.sgbl)

    @staticmethod
    def get_post_ed_amount(org_amount, ed_info, org_vol):
        if org_vol == 0:
            return 0
        if org_amount < 0:
            return org_amount
        return BasicQuoteInfo.get_post_ed_volume(org_vol, ed_info) * BasicQuoteInfo.get_post_ed_price(
            org_amount / org_vol, ed_info)

    def __str__(self):
        result_string = 'stockfullcode - {}, date - {}, open - {}, high - {}, low - {}, close - {}, volume - {}, amount - {}'.format(
            self.stockfullcode, self.date, self.o, self.h, self.l, self.c, self.v, self.amt
        )
        return result_string

    def __unicode__(self):
        result_string = u'代码 - {}, 日期 - {}, 开盘 - {}, 最高 - {}, 最低 - {}, 收盘 - {}, 成交手数 - {}, 成交量 - {}'.format(
            self.stockfullcode, self.date, self.o, self.h, self.l, self.c, self.v, self.amt
        )
        return result_string


class QuoteInfoList(BaseObject):
    # 传入的list元素为basicquoteinfo
    def __init__(self, name, quote_list):
        super(QuoteInfoList, self).__init__(name=name)
        self.quote_list = quote_list

    def get_max_volume_sum(self):
        # 传入一个quote_list,返回volume最大的
        max = self.quote_list[0]
        for i in range(len(self.quote_list)):
            if self.quote_list[i].v > max.v:
                max = self.quote_list[i]
        return max


class EdInfo(BaseObject):
    def __init__(self, name, date_ed, date_execute=None, stockfullcode_main='', fhbl=0, pgbl=0, pgjg=0, sgbl=0,
                 id_stock=-1, id_stock_main=-1, preclose=0):
        super(EdInfo, self).__init__(name=name)
        self.logger = logging.getLogger('root')
        self.date_ed = date_ed
        if date_execute is None:
            self.date_execute = date_ed
        else:
            self.date_execute = date_execute
        self.stockfullcode_main = stockfullcode_main
        self.fhbl = float(fhbl)
        self.pgbl = float(pgbl)
        self.pgjg = float(pgjg)
        self.sgbl = float(sgbl)
        self.id_stock = id_stock
        self.id_stock_main = id_stock_main
        self.preclose = preclose

    # Overloading function "str()"
    def __str__(self):
        return "stockfullcode: {}, " \
               "date_ed: {}, " \
               "date_execute: {}, " \
               "stockfullcode_main: {}, " \
               "fhbl: {}, " \
               "pgbl: {}, " \
               "pgjg: {}, " \
               "sgbl: {}" \
            .format(
            self.stockfullcode,
            self.date_ed,
            self.date_execute,
            self.stockfullcode_main,
            self.fhbl,
            self.pgbl,
            self.pgjg,
            self.sgbl)

    # Overloading operator "=="
    def __eq__(self, other):
        try:
            if (self.stockfullcode == other.stockfullcode
                    and self.date_ed == other.date_ed
                    and self.date_execute == other.date_execute
                    and self.stockfullcode_main == other.stockfullcode_main
                    and self.fhbl == other.fhbl
                    and self.pgbl == other.pgbl
                    and self.pgjg == other.pgjg
                    and self.sgbl == other.sgbl):
                return True
            else:
                self.logger.error('不一致的条目')
                self.logger.error(str(self))
                self.logger.error(str(other))
                return False
        except AttributeError:
            raise TypeError("These arguments are not comparable!")

    @property
    def stockfullcode(self):
        return self.name

    @property
    def is_future(self):
        return is_future(self.stockfullcode)

    def __str__(self):
        result_string = 'stockfullcode - {}, stockfullcode_main - {}, date_ed - {}, date_execute - {}, fhbl - {}, pgbl - {}, pgjg - {}, sgbl - {}'.format(
            self.stockfullcode, self.stockfullcode_main, self.date_ed, self.date_execute, self.fhbl, self.pgbl,
            self.pgjg, self.sgbl)
        return result_string


def is_future(stockfullcode):
    if stockfullcode.upper().startswith('SF'):
        return True
    if stockfullcode.upper().startswith('DL'):
        return True
    if stockfullcode.upper().startswith('SQ'):
        return True
    if stockfullcode.upper().startswith('ZZ'):
        return True
    if stockfullcode.upper().startswith('SE'):
        return True
    return False


def compare_contract_size(stockfullcode_a, stockfullcode_b):
    """
    比较两个合约的新旧，必须是同一个市场的合约。
    ZZ市场需要特殊处理：比较的首个数字为0和1的时候，1>0, 比较的首个数组为0和9的时候，0>9
    :param stockfullcode_a:
    :param stockfullcode_b:
    :return: True or False
    """
    market_name_a = _get_market(stockfullcode_a)
    market_name_b = _get_market(stockfullcode_b)
    result = False
    if market_name_a != market_name_b:
        raise TypeError("Different stock market name: {} and {}".format(stockfullcode_a, stockfullcode_b))
    stock_code_object_a = _get_contract_code(stockfullcode_a)
    stock_code_object_b = _get_contract_code(stockfullcode_b)
    if stock_code_object_a is not None and stock_code_object_b is not None:
        if market_name_a != 'ZZ':
            result = int(stock_code_object_a) > int(stock_code_object_b)
        else:
            stock_code_a_first_num = int(stock_code_object_a[0])
            stock_code_b_first_num = int(stock_code_object_b[0])
            if stock_code_a_first_num == 0 and stock_code_b_first_num == 9:
                result = True
            elif stock_code_a_first_num == 9 and stock_code_b_first_num == 0:
                result = False
            else:
                result = int(stock_code_object_a) > int(stock_code_object_b)
    return result


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass


if __name__ == "__main__":
    logger = logging.getLogger('root')
    setup_file_logger()
    setup_screen_logger()
    unittest.main()
