# -*- coding: utf-8 -*-
"""日更JQData数据"""
import os
import shutil
import numpy as np
import datetime as dt
import tushare as ts
import traceback
import unittest
import logging
import zipfile
import pandas as pd
import tushare as ts
from multiprocessing import Pool
from sqlalchemy import func, or_, and_, desc, asc, not_
import jqdatasdk as jqsdk
from rio3.base.constant import DEFAULT_INDEX_STOCK
from rio3.log.setup import setup_screen_logger, setup_file_logger
from rio3.utils.mail import send_mail
from jqdatasdk import *
from functools import wraps
from datetime import datetime, timedelta
from rio3.base.constant import H5_JQ_KLINE_DIR, H5_JQ_DAILY_DIR, DATETIME_MAX, DATETIME_MIN
from rio3.datafeed.research.jqdata_sdkIO import DEFAULT_ACCOUNT_NAME, DEFAULT_PASSWORD, BACKUP_ACCOUNT_NAME, \
    BACKUP_PASSWORD, normalize_code_to_jq, normalize_code_to_local, get_jq_query_concat_data
from rio3.datafeed.research.jqdata_httpIO import _http_get_all_securities
from rio3.datafeed.research.tushareIO import DEFAULT_TUSHARE_TOKEN, _get_trading_days
from rio3.datafeed.common import _get_market, _get_future_symbol, convert_freq2num
from rio3.utils.filesys import list_files

logger = logging.getLogger('root')

# 默认需要下载的K线频率
DEFAULT_FREQS = ['1m', '5m', '15m', '30m', '60m', '1d']
DEFAULT_FIELDS = ['open', 'close', 'high', 'low', 'volume', 'money', 'factor', 'avg', 'pre_close', 'high_limit',
                  'low_limit']
DEFAULT_TICK_FIELDS = ['time', 'current', 'high', 'low', 'volume', 'money', 'a1_v', 'a1_p', 'b1_v', 'b1_p']

TICK_UNIQUE_SUBSET_COLUMNS = ['Price', 'Volume', 'Amount', 'OpenInterest', 'SP1', 'SV1', 'BP1', 'BV1']


# 行业概念数据API
# https://www.joinquant.com/help/api/help?name=plateData#%E7%94%B3%E4%B8%87%E4%B8%80%E7%BA%A7%E8%A1%8C%E4%B8%9A

class LackOfDownloadTickError(Exception):
    """
    JQ下载数据不足
    """
    pass


# 标准周期对应Bar的数量
FREQ_BAR_NUM_DICT = {
    '1m': 800,
    '5m': 400,
    '15m': 200,
    '30m': 100,
    '60': 50,
    '1d': 10
}

# 构造交易市场编码对应JQ后缀
EXCHANGE_CODE2SUFFIX_DICT = {
    705001: 'XSHG',
    705002: 'XSHE',
    705003: 'XSHE',
    705004: 'XSHE',
    705005: 'XSHG',
    705006: 'XSHE'
}

# JQData的字段转微盛数据的字段，对应的字典
JQ2WS_COLUMNS_DICT = {
    'time': 'Time',
    'display_name': 'name',
    'name': 'Name',
    'date': 'DateTrade',
    'current': 'Price',
    'volume': 'Volume',
    'money': 'Amount',
    'open': 'Open',
    'high': 'High',
    'low': 'Low',
    'close': 'Close',
    'position': 'OpenInterest',
    'open_interest': 'OpenInterest',
    'contract': 'Contract',
    'a1_p': 'SP1', 'a1_v': 'SV1',
    'a2_p': 'SP2', 'a2_v': 'SV2',
    'a3_p': 'SP3', 'a3_v': 'SV3',
    'a4_p': 'SP4', 'a4_v': 'SV4',
    'a5_p': 'SP5', 'a5_v': 'SV5',
    'b1_p': 'BP1', 'b1_v': 'BV1',
    'b2_p': 'BP2', 'b2_v': 'BV2',
    'b3_p': 'BP3', 'b3_v': 'BV3',
    'b4_p': 'BP4', 'b4_v': 'BV4',
    'b5_p': 'BP5', 'b5_v': 'BV5'
}


def timer(func):
    @wraps(func)
    def wrap_func(*args, **kwargs):
        start = datetime.now()
        func(*args, **kwargs)
        end = datetime.now()
        usage = end - start
        print('用时(s)：', usage.seconds)

    return wrap_func


def show_example(func):
    @wraps(func)
    def export_examples(*args, **kwargs):
        df_diction = func(*args, **kwargs)
        for k, v in df_diction.items():
            print(f'导出{k}.csv')
            v.to_csv(f'D:\\data\\option\\example\\{k}.csv', index=False)

    return export_examples


def get_jq_api(username=None, password=None):
    """
    获取默认jqsdk
    :return: 返回注册过的jqsdk
    """
    username = username if username is not None else DEFAULT_ACCOUNT_NAME
    password = password if password is not None else DEFAULT_PASSWORD
    if not jqsdk.is_auth():
        try:
            jqsdk.auth(username, password)
        except Exception as ex:
            print(ex)
            raise Exception(f'JQ account auth failed, {username}/{password}')
        logger.info(f"get_jq_api接口使用账号:'{username}'")
    return jqsdk


def _exist_fetch_daily_dir(security_type, tday, freq, market=None, symbol=None, cache_root_dir=None):
    """
    判断是否需要缓存日更的数据, 判断是否已经存在日更文件夹
    :param cache_dir:
    :param security_type:
    :param tday:
    :param freq:
    :return:
    """
    cache_dir = cache_root_dir if cache_root_dir is not None else H5_JQ_DAILY_DIR
    if security_type in ['index', 'stock', 'conbond', 'sw']:
        cache_h5_dir = os.path.join(cache_dir, security_type, tday, freq)
    elif security_type == 'futures':
        if market is not None and symbol is not None:
            cache_h5_dir = os.path.join(cache_dir, security_type, tday, market, symbol, freq)
        else:
            cache_h5_dir = os.path.join(cache_dir, security_type, tday, freq)
    return os.path.exists(cache_h5_dir)


def cache_index_jq_data(security_list, tday, start_date, end_date, fields, freq, security_type=None, fq=None,
                        cache_dir=None):
    """
    缓存指数数据
    :param security_list:
    :param tday:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param security_type:
    :param fq:
    :param cache_dir:
    :return:
    """
    tday_freq_num = 0
    # 判断是否需要进行日更操作
    exist_fetch_daily_dir = _exist_fetch_daily_dir(security_type=security_type, tday=tday, freq=freq)
    if exist_fetch_daily_dir:
        logger.info(f"[NO ACTION]已存在'{security_type}'在交易日'{tday}' 频率'{freq}'的日更文件，无需下载JQ数据")
        return tday_freq_num
    # 从DEFAULT_INDEX_STOCK中获取需要缓存的指数列表
    default_index_stock = normalize_code_to_jq(DEFAULT_INDEX_STOCK.copy())
    index_stock_list = list(set(security_list).intersection(set(default_index_stock)))
    # 尝试从JQData获取数据
    jq_kline_data = get_jq_api().get_price(security=index_stock_list, start_date=start_date,
                                           end_date=end_date,
                                           fields=fields, frequency=freq, fq=fq)
    # 如果有数据，则开始缓存
    if not jq_kline_data.empty:
        cache_root_dir = cache_dir if cache_dir is not None else H5_JQ_DAILY_DIR
        cache_h5_dir = os.path.join(cache_root_dir, security_type, tday, freq)
        if not os.path.exists(cache_h5_dir):
            logger.info(f"创建目录 '{cache_h5_dir}' ")
            os.makedirs(cache_h5_dir)
        # 循环fields保存单字段数据
        for each_field in fields:
            h5_cache_field_path = os.path.join(cache_h5_dir, '.'.join([each_field, 'h5']))
            field_df = jq_kline_data[each_field]
            # 构造datetrade
            # field_df['datetrade'] = pd.to_datetime(field_df.index).strftime('%Y%m%d%H%M%S')
            field_df.reset_index(drop=True)
            len_of_data = len(field_df)
            # 缓存成HDF5文件
            try:
                field_df.to_hdf(path_or_buf=h5_cache_field_path, key='data', complevel=6, complib='blosc',
                                format='table')
                logger.info(
                    f"缓存'{security_type}'在交易日'{tday}'的'{freq}'的级别的字段'{each_field}'成功, "
                    f"共'{len_of_data}条' ")
                tday_freq_num += len_of_data
            except Exception as ex:
                logger.error(f"缓存'{security_type}'在交易日'{tday}'的'{freq}'的级别的字段'{each_field}'失败, "
                             f"错误信息：'{ex}' ")
    logger.info(
        f"'{security_type}'在交易日'{tday}'的'{freq}'的级别的字段共缓存'{','.join(fields)}'成功, 共'{tday_freq_num}条' ")
    return tday_freq_num


def cache_stock_jq_data(security_list, tday, start_date, end_date, fields, freq, security_type=None, fq=None,
                        cache_dir=None):
    """
    缓存每个交易日的股票数据，股票品种来源是根据tday获取get_securilty_list(stock)
    :param security_list:
    :param tday:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param security_type:
    :param fq:
    :param cache_dir:
    :return:
    """
    tday_freq_num = 0
    # 判断是否需要进行日更操作
    exist_fetch_daily_dir = _exist_fetch_daily_dir(security_type=security_type, tday=tday, freq=freq)
    if exist_fetch_daily_dir:
        logger.info(f"[NO ACTION]已存在'{security_type}'在交易日'{tday}' 频率'{freq}'的日更文件，无需下载JQ数据")
        return tday_freq_num
    # 尝试从JQData获取数据
    jq_kline_data = get_jq_api().get_price(security=security_list, start_date=start_date,
                                           end_date=end_date,
                                           fields=fields, frequency=freq, fq=fq)
    # 获取后复权的factor，不能与上面不复权的数据一起取
    if 'factor' in fields and security_type == 'stock':
        jq_factor_data = get_jq_api().get_price(security=security_list, start_date=start_date, end_date=end_date,
                                                fields=['factor'], frequency=freq, fq='post')
    else:
        jq_factor_data = pd.DataFrame()
    # 如果有数据，则开始缓存
    if not jq_kline_data.empty:
        cache_root_dir = cache_dir if cache_dir is not None else H5_JQ_DAILY_DIR
        # security_type + tday + market + symbol + freq
        cache_h5_dir = os.path.join(cache_root_dir, security_type, tday, freq)
        if not os.path.exists(cache_h5_dir):
            logger.info(f"创建目录 '{cache_h5_dir}' ")
            os.makedirs(cache_h5_dir)
        # 循环fields保存单字段数据
        for each_field in fields:
            h5_cache_field_path = os.path.join(cache_h5_dir, '.'.join([each_field, 'h5']))
            if each_field == 'factor' and not jq_factor_data.empty:
                field_df = jq_factor_data[each_field]
            else:
                field_df = jq_kline_data[each_field]
            if not field_df.empty:
                # 构造datetrade
                # field_df['datetrade'] = pd.to_datetime(field_df.index).strftime('%Y%m%d%H%M%S')
                field_df.reset_index(drop=True)
                len_of_data = len(field_df)
                # 缓存成HDF5文件
                try:
                    field_df.to_hdf(path_or_buf=h5_cache_field_path, key='data', complevel=6, complib='blosc')
                    logger.info(
                        f"缓存'{security_type}'在交易日'{tday}'的'{freq}'的级别的字段'{each_field}'成功, "
                        f"共'{len_of_data}条' ")
                    tday_freq_num += len_of_data
                except Exception as ex:
                    logger.error(f"缓存'{security_type}'在交易日'{tday}'的'{freq}'的级别的字段'{each_field}'失败, "
                                 f"错误信息：'{ex}' ")
    logger.info(
        f"'{security_type}'在交易日'{tday}'的'{freq}'的级别的字段共缓存'{','.join(fields)}'成功, 共'{tday_freq_num}条' ")
    return tday_freq_num


def cache_future_jq_data(security_df, tday, start_date, end_date, fields, freq, security_type=None, fq=None,
                         cache_dir=None):
    """
    由于期货品种交易时间的差异，KLine的数量不一致，不能一次行传给get_price，需要按照品种获取，所以需要对security_list按照symbol做groupby
    :param security_df:
    :param tday:
    :param start_date:
    :param end_date:
    :param fields:
    :param freq:
    :param security_type:
    :param fq:
    :param cache_dir:
    :return:
    """
    tday_freq_num = 0
    # 先提取市场
    security_df['market'] = security_df.code.str.extract(r'.\.(?P<market>[a-zA-Z]+)')
    security_df['symbol'] = security_df.code.str.extract(r'(?P<symbol>[\D]+).')
    for market, market_security_df in security_df.groupby(security_df.market):
        for symbol, symbol_security_df in market_security_df.groupby(market_security_df.symbol):
            # 判断是否需要进行日更操作
            exist_fetch_daily_dir = _exist_fetch_daily_dir(security_type=security_type, tday=tday, freq=freq,
                                                           market=market, symbol=symbol)
            if exist_fetch_daily_dir:
                logger.info(f"[NO ACTION]已存在'{security_type}'在交易日'{tday}' 频率'{freq}'的日更文件，无需下载JQ数据")
                return tday_freq_num
            # 尝试从JQData获取数据
            jq_kline_data = get_jq_api().get_price(security=symbol_security_df['code'].tolist(), start_date=start_date,
                                                   end_date=end_date,
                                                   fields=fields, frequency=freq, fq=fq)
            # 如果有数据，则开始缓存
            if not jq_kline_data.empty:
                cache_root_dir = cache_dir if cache_dir is not None else H5_JQ_DAILY_DIR
                # security_type + tday + market + symbol + freq
                cache_h5_dir = os.path.join(cache_root_dir, security_type, tday, market, symbol, freq)
                if not os.path.exists(cache_h5_dir):
                    logger.info(f"创建目录 '{cache_h5_dir}' ")
                    os.makedirs(cache_h5_dir)
                # 循环fields保存单字段数据
                for each_field in fields:
                    h5_cache_field_path = os.path.join(cache_h5_dir, '.'.join([each_field, 'h5']))
                    field_df = jq_kline_data[each_field]
                    if not field_df.empty:
                        # 构造datetrade
                        # field_df['datetrade'] = pd.to_datetime(field_df.index).strftime('%Y%m%d%H%M%S')
                        field_df.reset_index(drop=True)
                        len_of_data = len(field_df)
                        # 缓存成HDF5文件
                        try:
                            field_df.to_hdf(path_or_buf=h5_cache_field_path, key='data', complevel=6, complib='blosc',
                                            format='table')
                            logger.info(
                                f"缓存'{security_type}, {market}, {symbol}'在交易日'{tday}'的'{freq}'的级别的字段'{each_field}'成功, "
                                f"共'{len_of_data}条' ")
                            tday_freq_num += len_of_data
                        except Exception as ex:
                            logger.error(
                                f"缓存'{security_type}, {market}, {symbol}'在交易日'{tday}'的'{freq}'的级别的字段'{each_field}'失败, "
                                f"错误信息：'{ex}' ")
    logger.info(
        f"'{security_type}'在交易日'{tday}'的'{freq}'的级别的字段共缓存'{','.join(fields)}'成功, 共'{tday_freq_num}条' ")
    return tday_freq_num


def cache_conbond_jq_data(start_date, end_date, freq='1d', cache_dir=None):
    # 获取给定时间段内的交易日结合，如果不指定，获取当天交易日
    fetch_trading_days = _get_trading_days(start_date=start_date, end_date=end_date)
    tday_total_num = 0
    for fetch_tday in fetch_trading_days:
        # 判断是否需要进行日更操作
        exist_fetch_daily_dir = _exist_fetch_daily_dir(security_type='conbond', tday=fetch_tday, freq=freq)
        if exist_fetch_daily_dir:
            logger.info(f"[NO ACTION]已存在'conbond'在交易日'{fetch_tday}' 频率'1d'的日更文件，无需下载JQ数据")
            return tday_total_num
        fetch_tday_str = dt.datetime.strftime(pd.to_datetime(fetch_tday), '%Y-%m-%d')
        # 获取申万一级的行业列表
        sw_daily_df = get_jq_api().bond.run_query(
            query(bond.CONBOND_DAILY_PRICE).filter(bond.CONBOND_DAILY_PRICE.date == fetch_tday_str))
        if not sw_daily_df.empty:
            cache_root_dir = cache_dir if cache_dir is not None else H5_JQ_DAILY_DIR
            cache_h5_dir = os.path.join(cache_root_dir, 'conbond', fetch_tday, freq)
            if not os.path.exists(cache_h5_dir):
                logger.info(f"创建目录 '{cache_h5_dir}' ")
                os.makedirs(cache_h5_dir)
            h5_cache_field_path = os.path.join(cache_h5_dir, '.'.join(['conbond_' + fetch_tday, 'h5']))
            # 构造date
            sw_daily_df['date'] = pd.to_datetime(sw_daily_df.date).dt.strftime('%Y%m%d')
            sw_daily_df.drop(['id'], inplace=True, axis=1)
            len_of_data = len(sw_daily_df)
            # 缓存成HDF5文件
            try:
                sw_daily_df.to_hdf(path_or_buf=h5_cache_field_path, key='data', complevel=6, complib='blosc',
                                   format='table')
                logger.info(
                    f"缓存conbond在交易日在交易日'{fetch_tday_str}'的'{freq}'的级别的数据成功, "
                    f"共'{len_of_data}条' ")
                tday_total_num += len_of_data
            except Exception as ex:
                logger.error(f"缓存conbond在交易日'{fetch_tday_str}'的'{freq}'的级别的数据失败, "
                             f"错误信息：'{ex}' ")
    return tday_total_num


def cache_sw_jq_data(start_date, end_date, sw_level='sw_l1', freq='1d', cache_dir=None):
    """
    缓存申万1级行业数据，目前只有日线行情, 按平板数据格式存储
    :param start_date:
    :param end_date:
    :param sw_level:
    :param cache_dir:
    :return:
    """
    # 获取给定时间段内的交易日结合，如果不指定，获取当天交易日
    fetch_trading_days = _get_trading_days(start_date=start_date, end_date=end_date)
    tday_total_num = 0
    for fetch_tday in fetch_trading_days:
        # 判断是否需要进行日更操作
        exist_fetch_daily_dir = _exist_fetch_daily_dir(security_type='sw', tday=fetch_tday, freq=freq)
        if exist_fetch_daily_dir:
            logger.info(f"[NO ACTION]已存在'sw'在交易日'{fetch_tday}' 频率'1d'的日更文件，无需下载JQ数据")
            return tday_total_num
        fetch_tday_str = dt.datetime.strftime(pd.to_datetime(fetch_tday), '%Y-%m-%d')
        # 获取申万一级的行业列表
        sw_industries = get_jq_api().get_industries(name=sw_level, date=fetch_tday_str)
        if sw_industries is not None and len(sw_industries) > 0:
            sw_daily_df = get_jq_api().finance.run_query(
                query(finance.SW1_DAILY_PRICE).filter(finance.SW1_DAILY_PRICE.date == fetch_tday_str))
            if not sw_daily_df.empty:
                cache_root_dir = cache_dir if cache_dir is not None else H5_JQ_DAILY_DIR
                cache_h5_dir = os.path.join(cache_root_dir, 'sw', fetch_tday, freq)
                if not os.path.exists(cache_h5_dir):
                    logger.info(f"创建目录 '{cache_h5_dir}' ")
                    os.makedirs(cache_h5_dir)
                h5_cache_field_path = os.path.join(cache_h5_dir, '.'.join([sw_level + '_' + fetch_tday, 'h5']))
                # 构造date
                sw_daily_df['date'] = pd.to_datetime(sw_daily_df.date).dt.strftime('%Y%m%d')
                sw_daily_df.drop(['id'], inplace=True, axis=1)
                len_of_data = len(sw_daily_df)
                # 缓存成HDF5文件
                try:
                    sw_daily_df.to_hdf(path_or_buf=h5_cache_field_path, key='data', complevel=6, complib='blosc',
                                       format='table')
                    logger.info(
                        f"缓存'{sw_level}'在交易日'{fetch_tday_str}'的'{freq}'的级别的数据成功, "
                        f"共'{len_of_data}条' ")
                    tday_total_num += len_of_data
                except Exception as ex:
                    logger.error(f"缓存'{sw_level}'在交易日'{fetch_tday_str}'的'{freq}'的级别的数据失败, "
                                 f"错误信息：'{ex}' ")
    return tday_total_num


def fetch_jq_price(stockfullcode, start_date=None, end_date=None, freqs=None, fields=['']):
    df = get_jq_api().get_price(security=stockfullcode, start_date=start_date, end_date=end_date, fields=fields,
                                frequency=freqs)
    cache_path = os.path.join(r'\\nas\share\HDF5\JQData\Segments', stockfullcode + '.h5')
    df.to_hdf(path_or_buf=cache_path, key='data')
    print(len(df))


def _get_fetch_date_time_range(fetch_tday):
    """
    获取给定交易日的实际行情时间，从上一个交易日的20:00-当个交易日的16:00
    :param fetch_tday:
    :return:
    """
    fetch_tday_str = dt.datetime.strftime(pd.to_datetime(fetch_tday), '%Y-%m-%d')
    # 构造行情的起始时间，永远是该交易日前一天20:00:00的数据
    # 获取前一天交易日
    previous_tday = get_jq_api().get_trade_days(count=2, end_date=fetch_tday_str)[-2]
    fetch_day_start = pd.to_datetime(previous_tday) + pd.to_timedelta('20:00:00')
    # 构造行情的结束时间，永远是该交易日当天到16:00:00为止
    fetch_day_end = pd.to_datetime(fetch_tday) + pd.to_timedelta('16:00:00')
    return fetch_day_start, fetch_day_end


def fetch_source_data_to_csv(freq, security_type, start_date=None, end_date=None, fields=None, cache_extract_dir=None,
                             cache_archive_dir=None, replace=True):
    """
    下载JQData的Tick或者Kline数据
    :param freq:
    :param security_type:
    :param start_date:
    :param end_date:
    :param fields:
    :param cache_extract_dir:
    :param cache_archive_dir:
    :param replace:
    :return:
    """
    tday_error_extract_contract = dict()
    # 获取给定时间段内的交易日结合，如果不指定，获取当天交易日
    fetch_trading_days = _get_trading_days(start_date=start_date, end_date=end_date)
    # 获取需要缓存的品种列表
    for fetch_tday in fetch_trading_days:
        # 获取给定时间段内的交易日结合，如果不指定，获取当天交易日
        fetch_tday_str = dt.datetime.strftime(pd.to_datetime(fetch_tday), '%Y-%m-%d')
        # 获取当前交易日的交易日品种, 需要根据security_type选择列表
        security_info_dict = get_jq_security_code_info(security_type=security_type, fetch_tday=fetch_tday_str)
        # fetch_security_df = fetch_security_df[fetch_security_df['code'].str.contains('CCFX')]
        if len(security_info_dict) > 0:
            # 做实验，只下载前10个
            # security_list = security_list[0:15]
            cache_result_dict = cache_jq_data2csv(security_code=security_info_dict, tday=fetch_tday, freq=freq,
                                                  fields=fields, security_type=security_type,
                                                  cache_extract_dir=cache_extract_dir,
                                                  cache_archive_dir=cache_archive_dir, replace=replace)
            result_dict = dict()
            if len(cache_result_dict) > 0:
                error_contract_list = list()
                # 检查extract的result=0或者-1的品种
                for extract_contract, result in cache_result_dict['extract'].items():
                    if result == -1:
                        error_contract_list.append(extract_contract)
                # 检查archive的文件是否存在
                archive_result = False if not os.path.exists(cache_result_dict['archive']) else True
                # 判断是否有出错的问题
                if len(error_contract_list) > 0 or not archive_result:
                    result_dict['extract'] = error_contract_list
                    result_dict['archive'] = archive_result
                # 加入总结果集
                tday_error_extract_contract[fetch_tday] = result_dict
    return tday_error_extract_contract


def get_bond_security_code_info(bond_type, start_tday='2021-05-01', query_limit=3000):
    """
    根据bond_type获取债券合约列表
    705001	上交所
    705002	深交所主板
    705003	深交所中小板
    705004	深交所创业板
    705005	上交所综合业务平台
    705006	深交所综合协议交易平台
    :param bond_type:
    :param start_tday:
    :param query_limit:
    :return:
    """
    return_dict = dict()
    query_table = None
    if bond_type == 'bond':
        query_table = get_jq_api().bond.BOND_BASIC_INFO
    elif bond_type == 'conbond':
        query_table = get_jq_api().bond.CONBOND_BASIC_INFO
    queries = (
        query_table,
    )
    filters = (
        and_(query_table.list_date <= DATETIME_MAX, query_table.list_date >= start_tday,
             query_table.exchange_code.in_([705001, 705002, 705003, 705004, 705005, 705006]))
    )
    return_data_df = get_jq_query_concat_data(f_query=get_jq_api().bond.run_query, query_args=queries,
                                              filter_args=filters, step=query_limit, verbose=2)
    # 整理格式
    if not return_data_df.empty:
        # 整理EXCHANGE_CODE2SUFFIX_DICT为dataframe
        exchange_code2suffix_df = pd.DataFrame.from_dict(EXCHANGE_CODE2SUFFIX_DICT, orient='index',
                                                         columns=['suffix']).reset_index()
        exchange_code2suffix_df.rename(columns={'index': 'exchange_code'}, inplace=True)
        # 做merge
        return_data_df = pd.merge(left=return_data_df, right=exchange_code2suffix_df, left_on='exchange_code',
                                  right_on='exchange_code', how='left')
        # 定义新column,带后缀
        return_data_df['full_code'] = return_data_df[['code', 'suffix']].agg('.'.join, axis=1)
        return_dict.update(dict(zip(return_data_df['full_code'], return_data_df['short_name'])))
    return dict(sorted(return_dict.items()))


def get_jq_security_code_info(security_type, fetch_tday):
    """
    根据security_type获取指定交易日的
    :param security_type:
    :param fetch_tday:
    :return:
    """
    security_df_dict = dict()
    if security_type == 'stock':
        # 下载股票数据需要同时下载stock/index/etf
        for each_type in ['stock', 'index', 'etf']:
            fetch_security_df = _http_get_all_securities(code=each_type, date=fetch_tday)
            if not fetch_security_df.empty:
                security_df_dict.update(dict(zip(fetch_security_df['code'], fetch_security_df['display_name'])))
    else:
        fetch_security_df = _http_get_all_securities(code=security_type, date=fetch_tday)
        if not fetch_security_df.empty:
            security_df_dict.update(dict(zip(fetch_security_df['code'], fetch_security_df['display_name'])))
    return dict(sorted(security_df_dict.items()))


def cache_jq_data2csv(security_code, tday, freq, security_type, fields=None, cache_extract_dir=None,
                      cache_archive_dir=None, replace=True):
    """
    缓存每个交易日的Tick/Kline数据，股票品种来源是根据tday获取get_securilty_list(stock)
    :param security_code:
    :param tday:
    :param fields: 为None时，默认返回所有字段
    :param security_type:
    :param cache_extract_dir:
    :param cache_archive_dir:
    :return: dict 按交易日储存每一个合约的csv下载结果:
    -1代表下载失败，0代表没有数据，大于0的数据则代表下载成功，空结果集代表无需下载
    """
    tday_contract_quote_result = dict()
    # 获取准确的行情区间,精确到小时
    fetch_start_dt, fetch_end_dt = _get_fetch_date_time_range(fetch_tday=tday)
    # 将tday转为数字排列的日期格式
    tday_str_num = pd.to_datetime(tday).strftime('%Y%m%d')
    security_info_dict = {security_code: ''} if isinstance(security_code, str) else security_code
    # 如果fields等于默认值，则根据security_type进行调整
    # 判断是否需要进行日更操作
    exist_fetch_daily_dir = _exist_fetch_daily_dir(security_type=security_type, tday=tday, freq=freq,
                                                   cache_root_dir=cache_extract_dir)
    # 如果存在当天的文件夹并且不是替换，则无需下载
    if exist_fetch_daily_dir and not replace:
        logger.info(f"[NO ACTION]已存在'{security_type}'在交易日'{tday}' 频率'{freq}'的日更文件，无需下载JQ数据")
        return tday_contract_quote_result
    # 创建当日的tick文件夹
    cache_root_dir = cache_extract_dir if cache_extract_dir is not None else H5_JQ_DAILY_DIR
    # security_type + tday + market + symbol + freq
    cache_csv_dir = os.path.join(cache_root_dir, security_type, tday, freq)
    # 如果是replace, 并存在当天文件夹，则需要删掉当天的文件夹
    if os.path.exists(cache_csv_dir) and replace:
        shutil.rmtree(cache_csv_dir)
    # 保存当日的csv_file路径
    csv_file_path_list = list()
    # 去掉主力合约和次主力
    security_info_dict = {code: display_name for code, display_name in security_info_dict.items() if
                          '8888' not in code and '9999' not in code}
    for each_security_code, display_name in security_info_dict.items():
        # 尝试从JQData获取Tick数据
        try:
            if freq == 'tick':
                quote_df = get_jq_api().get_ticks(security=each_security_code, start_dt=fetch_start_dt,
                                                  end_dt=fetch_end_dt,
                                                  fields=fields, skip=False)
            else:
                fields = ['date', 'open', 'high', 'low', 'close', 'volume', 'money']
                if security_type == 'futures':
                    fields.extend(['open_interest'])
                # 设置get_bars的count
                freq_count = FREQ_BAR_NUM_DICT.get(freq, 1000)
                quote_df = get_jq_api().get_bars(security=each_security_code, count=freq_count, end_dt=fetch_end_dt,
                                                 unit=freq,
                                                 fields=fields, include_now=True, fq_ref_date=None)
            # 如果有数据，则开始缓存
            if not quote_df.empty:
                if freq != 'tick':
                    quote_df['date'] = pd.to_datetime(quote_df['date'])
                    quote_df = quote_df[(quote_df['date'] >= fetch_start_dt) & (quote_df['date'] <= fetch_end_dt)]
                len_of_quote = len(quote_df)
                quote_df['name'] = display_name
                quote_df['contract'] = normalize_code_to_local(each_security_code)
                # 缓存成csv文件
                try:
                    # 重命名columns
                    format_jq_quote_fields(quote_df=quote_df)
                    # 重新计算单位时间内的Amount和Volume
                    _re_calculate_amount_volume(quote_df=quote_df)
                    # 过滤数据
                    # _filter_quote_data(quote_df=quote_df, freq=freq)
                    # 重设index
                    quote_df.set_index(keys='Time', inplace=True) if freq == 'tick' else quote_df.set_index(
                        keys='DateTrade', inplace=True)
                    # 将每一个品种保存成csv文件
                    quote_csv_path = get_quote_csv_export_path(security_type=security_type,
                                                               stockfullcode=each_security_code,
                                                               cache_csv_dir=cache_csv_dir, tday_str_num=tday_str_num)

                    quote_df.to_csv(path_or_buf=quote_csv_path, encoding='utf_8_sig')
                    # 加入集合，方便打包处理
                    csv_file_path_list.append(quote_csv_path)
                    logger.info(
                        f"[缓存CSV成功]: 代码:'{each_security_code}'在交易日'{tday}'的'{freq}'数据成功, "
                        f"共'{len_of_quote}条' ")
                    data_result_num = len_of_quote
                except Exception as ex:
                    data_result_num = -1
                    logger.error(f"[缓存CSV失败]: 代码:'{each_security_code}'在交易日'{tday}'的'{freq}'数据失败, "
                                 f"错误信息：'{ex}' ")
            # 如果为空
            else:
                data_result_num = 0
                logger.warning(
                    f"[获取Tick数据为空]: 代码:'{each_security_code}'在交易日'{tday}'的'{freq}'数据为空, 数量：{len(quote_df)} ")
        except Exception as ex:
            data_result_num = -1
            logger.error(f"[获取Tick数据出错]: 代码:'{each_security_code}'在交易日'{tday}'的'{freq}'数据出错, "
                         f"错误信息：'{ex}' ")
        # 写入结果集
        # 如果结果集中已经有当日记录，则继续加入当前合约的结果信息
        if 'extract' in tday_contract_quote_result.keys():
            tday_contract_quote_result['extract'][each_security_code] = data_result_num
            # 如果结果集中没有当日记录，创建第一条全新的记录
        else:
            current_contract_result_dict = {each_security_code: data_result_num}
            tday_contract_quote_result['extract'] = current_contract_result_dict
    # 判断保存成csv的文件数量是否和需要获取的合约数量保持一致
    msg = '[缓存数据统计]： 合约数量/实际下载: {}/{}'.format(len(security_info_dict.keys()), len(csv_file_path_list))
    logger.info(msg)
    if len(csv_file_path_list) != len(security_info_dict.keys()):
        logger.warning('[无数据合约]: 交易日:{}, 频率:{}, 品类:{}, 缺失合约: {}'.format(tday, freq, security_type, list(
            set(csv_file_path_list).difference(set(security_info_dict.keys())))))

    if len(csv_file_path_list) / len(security_info_dict.keys()) <= 0.95:
        exception_msg = '下载数据不足: 交易日:{}, 频率:{}, 品类:{}, 详细: {}'.format(tday, freq, security_type, msg)
        logger.warning('[无数据合约]: 交易日:{}, 频率:{}, 品类:{}, 缺失合约: {}'.format(tday, freq, security_type, list(
            set(csv_file_path_list).difference(set(security_info_dict.keys())))))
        # 出错邮件通知
        subject = f'[Error][期货合约Tick数据检查][交易日: {tday}]'
        content_dict = {'content': exception_msg}
        send_task_result_email(to_address=['huahuabill@qq.com'], subject=subject, content=content_dict)
        raise LackOfDownloadTickError(exception_msg)
    # 打成压缩包，命名为tick_ymd日期
    try:
        tday_contract_quote_result['archive'] = None
        tick_zip_file_name = '{}_{}_{}.zip'.format(freq, security_type, tday_str_num)
        zip_file_path = os.path.join(os.path.join(cache_root_dir, security_type, tday), tick_zip_file_name)
        # 打包csv文件,注意需要根据security_type保留解压出来的源文件路径
        make_zip(csv_file_path_list=csv_file_path_list, output_filename=zip_file_path, security_type=security_type)
        logger.info(f"[打包Tick成功]': 类型:{security_type}, 交易日:{tday}, 打包位置:{zip_file_path}' ")
        # 不删除csv源文件，等待后续的task做读取
        # shutil.rmtree(cache_h5_dir)
        # 将CSV的打包文件剪切到备份目录
        archive_zip_file_store_dir = os.path.join(cache_archive_dir, security_type, str(pd.to_datetime(tday).year),
                                                  str(pd.to_datetime(tday).strftime('%Y%m')), freq)
        # 判断是否需要创建
        if make_dir(archive_zip_file_store_dir):
            archive_zip_file_store_path = os.path.join(archive_zip_file_store_dir, tick_zip_file_name)
            if os.path.exists(archive_zip_file_store_path):
                os.remove(archive_zip_file_store_path)
            shutil.move(src=zip_file_path, dst=archive_zip_file_store_dir)
            archive_zip_file_path = os.path.join(archive_zip_file_store_dir, tick_zip_file_name)
            logger.info(f"[归档Tick成功]': 类型:{security_type}, 交易日:{tday}, 归档位置:{archive_zip_file_path}' ")
            tday_contract_quote_result['archive'] = archive_zip_file_path
        else:
            logger.error(f"[打包/复制CSV压缩文件失败]: 错误:{ex}")
    except Exception as ex:
        logger.error(f"[打包/复制CSV压缩文件失败]: 错误:{ex}")
    return tday_contract_quote_result


def send_task_result_email(to_address, subject, content='', server_info=None, from_address='huahuabill@qq.com'):
    """
    发送任务状态邮件，一般失败的时候调用
    :param to_address:
    :param subject:
    :param content:
    :param server_info:
    :param from_address:
    :return:
    """
    send_mail(
        str_from_address=from_address,
        list_to_address=to_address,
        str_subject=subject,
        dict_content=content,
        use_ssl=True)


def _filter_quote_data(quote_df, freq):
    """
    对K线数据进行过滤
    :param quote_df:
    :param freq:
    :return:
    """
    if freq == 'tick':
        # 去掉重复的行情，即使时间戳是不一样的
        quote_df.drop_duplicates(subset=TICK_UNIQUE_SUBSET_COLUMNS, inplace=True)


def _re_calculate_amount_volume(quote_df):
    """
    JQData的源数据，Amount和Volume都是累积记录，需要自己计算出差值
    :param quote_df:
    :return:
    """
    quote_df['UnitAmount'] = quote_df['Amount'] - quote_df['Amount'].shift(1)
    quote_df['Amount'] = np.where(quote_df['UnitAmount'].isnull(), quote_df['Amount'], quote_df['UnitAmount'])
    quote_df['UnitVolume'] = quote_df['Volume'] - quote_df['Volume'].shift(1)
    quote_df['Volume'] = np.where(quote_df['UnitVolume'].isnull(), quote_df['Volume'], quote_df['UnitVolume'])
    quote_df.drop(['UnitAmount', 'UnitVolume'], inplace=True, axis=1)


def make_dir(dir_path):
    oper_result = False
    # 判断dir_path是否存在，不存在就创建
    try:
        if not os.path.exists(dir_path):
            logger.info(f"[创建目录成功]:'{dir_path}'")
            os.makedirs(dir_path)
        oper_result = True
    except Exception as ex:
        oper_result = False
        logger.error(f"[创建目录失败]:'{dir_path}'")
    return oper_result


def get_quote_csv_export_path(security_type, stockfullcode, cache_csv_dir, tday_str_num):
    """
    对csv分类，如果是股票，market_name直接跟股票代码，如果是期货，market_name + symbol
    :param security_type:
    :param stockfullcode:
    :param cache_csv_dir:
    :param tday_str_num:
    :return:
    """
    local_contract = normalize_code_to_local(jq_stockfullcode=stockfullcode)
    market_name = _get_market(stockfullcode=local_contract)
    contract_root_dir = os.path.join(cache_csv_dir, market_name)
    if security_type == 'futures':
        symbol = _get_future_symbol(stockfullcode=local_contract)
        contract_root_dir = os.path.join(contract_root_dir, symbol)
    if not os.path.exists(contract_root_dir):
        logger.info(f"创建目录 '{contract_root_dir}' ")
        os.makedirs(contract_root_dir)
    csv_export_path = os.path.join(contract_root_dir, '{}_{}.csv'.format(local_contract, tday_str_num))
    return csv_export_path


def format_jq_quote_fields(quote_df):
    """
    将JQData的数据列命名为微盛格式
    :param quote_df:
    :param freq:
    :return:
    """
    if quote_df.empty:
        return quote_df
    rename_dict = dict()
    for jq_column in quote_df.columns:
        ws_column = JQ2WS_COLUMNS_DICT.get(jq_column, jq_column)
        rename_dict[jq_column] = ws_column
    # 重命名
    quote_df.rename(columns=rename_dict, inplace=True)
    # 去掉不需要的columns
    for drop_column in ['high', 'low']:
        if drop_column in quote_df.columns:
            quote_df.drop([drop_column], inplace=True, axis=1)


def make_zip(csv_file_path_list, output_filename, security_type):
    """
    打包Quote CSV文件
    :param csv_file_path_list:
    :param output_filename:
    :return:
    """
    zip_file = zipfile.ZipFile(output_filename, 'w')
    for csv_file_path in csv_file_path_list:
        csv_file_name = os.path.basename(csv_file_path)
        # 带上market_name, 如果是futures, 则需要分离出symbol
        contract_tday_src = os.path.splitext(csv_file_name)[0]
        contract_tday_arr = contract_tday_src.split('_')
        market_name = _get_market(contract_tday_arr[0])
        # 默认路径, market + contract, 适用股票
        path_csv_file_name = os.path.join(market_name, csv_file_name)
        # 调整路径, market + symbol + contract, 适用期货
        if security_type == 'futures':
            symbol = _get_future_symbol(contract_tday_arr[0])
            path_csv_file_name = os.path.join(market_name, symbol, csv_file_name)
        zip_file.write(filename=csv_file_path, arcname=path_csv_file_name)
    zip_file.close()


def fetch_jqdata(start_date=None, end_date=None, freqs=DEFAULT_FREQS, fields=DEFAULT_FIELDS, security_type='stock',
                 cache_dir=None, fq=None):
    """
    指定stockfullcode或者type,在给定的时间区间内缓存JQData的数据
    :param start_date: 开始交易日
    :param end_date: 结束交易日
    :param freqs: 缓存频率
    :param fields: 缓存域
    :param type: 缓存类型 'stock', 'index', 'futures', 'options'
    :param cache_dir: 缓存目录
    :param fq: 复权
    :param stockfullcode: 指定缓存代码
    :return:
    """
    # 获取给定时间段内的交易日结合，如果不指定，获取当天交易日
    fetch_trading_days = _get_trading_days(start_date=start_date, end_date=end_date)
    # 获取需要缓存的品种列表
    for fetch_tday in fetch_trading_days:
        tday_total_num = 0
        # 获取给定时间段内的交易日结合，如果不指定，获取当天交易日
        fetch_tday_str = dt.datetime.strftime(pd.to_datetime(fetch_tday), '%Y-%m-%d')
        fetch_day_start, fetch_day_end = _get_fetch_date_time_range(fetch_tday=fetch_tday_str)
        # 获取当前交易日的交易日品种
        fetch_security_df = _http_get_all_securities(code=security_type, date=fetch_tday_str)
        if not fetch_security_df.empty:
            security_list = fetch_security_df['code'].tolist()
            tday_freq_num = 0
            for freq in freqs:
                try:
                    # 如果freq是1d, 需要把start/end设置为同一个tday
                    if freq == '1d':
                        fetch_day_start = fetch_tday_str
                        fetch_day_end = fetch_tday_str
                    # 缓存股票数据
                    if security_type == 'stock':
                        tday_freq_num += cache_stock_jq_data(security_list=security_list, tday=fetch_tday,
                                                             start_date=fetch_day_start, end_date=fetch_day_end,
                                                             fields=fields, freq=freq, security_type=security_type,
                                                             fq=fq, cache_dir=cache_dir)
                    # 缓存期货数据
                    elif security_type == 'futures':
                        tday_freq_num += cache_future_jq_data(security_df=fetch_security_df, tday=fetch_tday,
                                                              start_date=fetch_day_start, end_date=fetch_day_end,
                                                              fields=fields, freq=freq, security_type=security_type,
                                                              fq=fq, cache_dir=cache_dir)
                    # 缓存指数数据
                    elif security_type == 'index':
                        tday_freq_num += cache_index_jq_data(security_list=security_list, tday=fetch_tday,
                                                             start_date=fetch_day_start, end_date=fetch_day_end,
                                                             fields=fields, freq=freq, security_type=security_type,
                                                             fq=fq, cache_dir=cache_dir)

                    tday_total_num += tday_freq_num
                except Exception as ex:
                    logger.error(f"获取'{security_type}'在交易日'{fetch_tday}'的'{freq}'的级别失败, 错误信息：'{ex}' ")
        logger.info(
            f"'{security_type}'在交易日'{fetch_tday}'缓存成功, 共'{tday_total_num}条' ")
    # 获取数据结束后注销退出
    if jqsdk.is_auth():
        logger.info(u"正在退出JQData")
        jqsdk.logout()


def get_extracted_source_dir(archived_file_path, cache_extract_dir=r'D:\TempCSV\jq_data\source'):
    """
    根据archived_file_path获取解压地址，先检查给定的cache_extract_dir中是否有，如果没有再对archived_file_path进行解压
    :param archived_file_path:
    :param cache_extract_dir:
    :return:
    """
    extracted_source_dir = None
    archive_file_name = os.path.basename(archived_file_path)
    archive_file_name_arr = archive_file_name.split('_')
    if len(archive_file_name_arr) > 0:
        archive_tday = os.path.splitext(archive_file_name_arr[2])[0]
        security_type = archive_file_name_arr[1]
        freq = archive_file_name_arr[0]
        # 构造默认解压地址
        excepted_extract_dir = os.path.join(cache_extract_dir, security_type, str(archive_tday), freq)
        # 判断解压路径是否存在
        if os.path.exists(excepted_extract_dir) and len(list_files(excepted_extract_dir, '.csv')) > 0:
            extracted_source_dir = excepted_extract_dir
        # 如果不存在，则对归档zip文件进行解压
        else:
            with zipfile.ZipFile(archived_file_path) as zip_file:
                zip_file_list = zip_file.namelist()
                if make_dir(excepted_extract_dir):
                    for each_file in zip_file_list:
                        zip_file.extract(each_file, excepted_extract_dir)
            zip_file.close()
    return extracted_source_dir


def get_quote_from_csv(extracted_dir):
    """
    从csv文件读取行情
    :param freq:
    :return:
    """
    total_csv_quote_df = pd.DataFrame()
    csv_quote_files = list_files(dir_path=extracted_dir, suffix='.csv')
    total_csv_quote_list = list()
    for csv_quote_file_path in csv_quote_files:
        contract_quote_df = pd.read_csv(filepath_or_buffer=csv_quote_file_path)
        if not contract_quote_df.empty:
            total_csv_quote_list.append(contract_quote_df)
    if len(total_csv_quote_list) > 0:
        total_csv_quote_df = pd.concat(total_csv_quote_list)
    return total_csv_quote_df


class TestFetchJqDataIO(unittest.TestCase):
    def setUp(self):
        setup_screen_logger()
        setup_file_logger()
        self.logger = logging.getLogger('root')

    def tearDown(self):
        pass

    def test_get_extracted_source_dir(self):
        archived_file_path = r'D:\TempCSV\jq_data\archive\futures\2020\202012\1d\1d_futures_20201224.zip'
        extract_dir = get_extracted_source_dir(archived_file_path=archived_file_path)
        pass

    def test_get_trading_days(self):
        start = '2019-01-01'
        end = '2020-12-01'
        trade_days = _get_trading_days(start_date=start, end_date=end)
        self.assertEqual(465, len(trade_days))

    def test_get_security_list(self):
        start = '2019-01-01'
        end = '2020-12-01'
        trade_days = _get_trading_days(start_date=start, end_date=end)
        for tday in trade_days:
            tday = dt.datetime.strftime(pd.to_datetime(tday), '%Y-%m-%d')
            security_list = _http_get_all_securities(code='stock', date=tday)
            pass

    def test_fetch_stock_jqdata(self):
        """
        Task:Stock Step1: 从JQData下载股票全市场的1分钟/1天的数据
        :return:
        """
        start = '2021-10-08'
        end = '2021-11-06'
        freq_list = ['1m', '1d']
        fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'avg', 'factor']
        fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'avg', 'factor']
        # fetch_jqdata(start_date=start, end_date=end, security_type='stock', freqs=freq_list, fields=fields)
        fetch_jqdata(start_date=start, end_date=end, security_type='stock', freqs=freq_list)
        pass

    def test_fetch_index_jqdata(self):
        """
        Task:Index Step1: 从JQData下载股票指数的1分钟/1天的数据
        :return:
        """
        start = '2021-07-15'
        end = '2021-10-12'
        freq_list = ['1m', '1d']
        # fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'avg']
        fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'factor', 'avg', 'pre_close', 'high_limit',
                  'low_limit']
        fetch_jqdata(start_date=start, end_date=end, security_type='index', freqs=freq_list, fields=fields)

    def test_fetch_future_jqdata(self):
        start = '2018-01-01'
        end = '2020-03-24'
        freq_list = ['1m', '1d']
        fields = ['open', 'close', 'high', 'low', 'volume', 'money', 'high_limit', 'low_limit', 'pre_close',
                  'open_interest', 'avg', 'factor']
        fetch_jqdata(start_date=start, end_date=end, security_type='futures', freqs=freq_list, fields=fields)

    def test_fetch_sw_jqdata(self):
        start = '2004-01-01'
        end = '2020-02-13'
        sw_level = 'sw_l1'
        cache_sw_jq_data(start_date=start, end_date=end, sw_level=sw_level)
        pass

    def test_fetch_conbond_jqdata(self):
        start = '2018-09-01'
        end = '2021-12-31'
        cache_conbond_jq_data(start_date=start, end_date=end)
        pass

    def test_fetch_stockfullcode_price(self):
        start = '2016-01-01 00:00:00'
        end = '2020-2-12 15:00:00'
        freq = '1d'
        fields = ['open', 'close', 'low', 'high', 'volume', 'money', 'high_limit', 'low_limit', 'avg', 'pre_close']
        stockfullcode = ['399905.XSHE', '000300.XSHG']
        for each_stock in stockfullcode:
            fetch_jq_price(stockfullcode=each_stock, start_date=start, end_date=end, freqs=freq, fields=fields)

    def test_fetch_stock_tick_jqdata(self):
        """
        尝试获取股票的Tick数据
        :return:
        """
        security_type = 'stock'
        start_dt = '2020-12-03 09:00:00'
        end_dt = '2020-12-04 16:00:00'
        cache_extract_dir = r'D:\TempCSV\stock_data\jq_tick'
        cache_archive_dir = r'D:\TempCSV\future_data\archive'
        fetch_result = fetch_source_data_to_csv(start_date=start_dt, end_date=end_dt, security_type=security_type,
                                                fields=None,
                                                cache_extract_dir=cache_extract_dir,
                                                cache_archive_dir=cache_archive_dir)
        pass

    def test_fetch_future_tick_jqdata(self):
        start = '2021-03-31'
        end = '2021-03-31'
        security_type = 'futures'
        cache_extract_dir = r'D:\TempCSV\jq_data\source'
        cache_archive_dir = r'D:\TempCSV\jq_data\archive'
        replace = True
        freq = 'tick'
        fetch_result = fetch_source_data_to_csv(start_date=start, end_date=end, security_type=security_type,
                                                cache_extract_dir=cache_extract_dir,
                                                cache_archive_dir=cache_archive_dir,
                                                replace=replace, freq=freq)
        pass

    def test_get_quote_from_csv(self):
        extracted_dir = r'D:\TempCSV\jq_data\source\futures\20201224\1d'
        csv_quote_df = get_quote_from_csv(extracted_dir=extracted_dir)
        print(csv_quote_df)
        pass

    def test_cache_jq_data2csv(self):
        cache_jq_data2csv(security_code=normalize_code_to_jq('SFT2203'), tday='2022-01-12', freq='tick',
                          security_type='futures')
        pass

    def test_get_bond_security_code_info(self):
        data1 = get_bond_security_code_info(bond_type='bond')
        data2 = get_bond_security_code_info(bond_type='conbond')
        pass

    def test_send_task_result_email(self):
        tday = '2022-02-23'
        freq = 'tick'
        security_type = 'futures'
        msg = 'test_msg'
        exception_msg = '下载数据不足: 交易日:{}, 频率:{}, 品类:{}, 详细: {}'.format(tday, freq, security_type, msg)
        subject = f'[Error][期货合约Tick数据检查][交易日: {tday}]'
        content_dict = {'content': exception_msg}
        send_task_result_email(to_address=['huahuabill@qq.com'], subject=subject, content=content_dict)
        pass


if __name__ == "__main__":
    unittest.main()
