# -*- coding:utf-8 -*-
import logging
import os
import datetime as dt
import unittest
import time as t
import types
import collections
import numpy as np
import pandas as pd
from dateutil import rrule, parser
from collections import defaultdict
import matplotlib.pyplot as plt
import matplotlib.cbook as cbook
import matplotlib.ticker as ticker

OA_DATE_START = dt.datetime(1899, 12, 30)
TOTAL_SECONDS_A_DAY = 86400.0
TOTAL_MICROSECONDS_A_DAY = 86400000000.0
TIME_FORMAT = '%H:%M:%S'


# 将Time字符串转化成对象
def str_to_dtime(time_str, format=TIME_FORMAT):
    return dt.datetime.strptime(time_str, format)


def to_strftime(obj, date_format='%Y-%m-%d'):
    if isinstance(obj, str) \
            or isinstance(obj, dt.datetime) \
            or isinstance(obj, pd.Timestamp) \
            or isinstance(obj, int) \
            or isinstance(obj, dt.date):
        t = pd.to_datetime(obj).to_pydatetime()
        return t.strftime(date_format)
    raise TypeError(u'to_datetime：不支持的转换类型! {}'.format(type(obj)))


def to_datetime(obj):
    if isinstance(obj, pd.Timestamp):
        return dt.datetime(year=obj.year, month=obj.month, day=obj.day, hour=obj.hour, minute=obj.minute,
                           second=obj.second, microsecond=obj.microsecond)
    if isinstance(obj, dt.datetime):
        return obj
    if isinstance(obj, dt.date):
        return dt.datetime(obj.year, obj.month, obj.day)
    if isinstance(obj, int) or type(obj) == np.int64 or isinstance(obj, np.long):
        if obj < 30000000:
            y = obj / 10000
            m = (obj / 100) % 100
            d = obj % 100
            H = M = S = 0
        else:
            y = obj / 10000000000
            m = (obj / 100000000) % 100
            d = (obj / 1000000) % 100
            H = (obj / 10000) % 100
            M = (obj / 100) % 100
            S = obj % 100
            if H == 24:
                return dt.datetime(y, m, d, 0, M, S) + dt.timedelta(days=1)
        return dt.datetime(y, m, d, H, M, S)
    if isinstance(obj, str):
        return parser.parse(obj)
    raise TypeError(u'to_datetime：不支持的转换类型! {}'.format(type(obj)))


def to_timedelta(obj):
    if isinstance(obj, str) or type(obj) == dt.datetime:
        t = to_datetime(obj)
        return dt.timedelta(
            hours=t.hour,
            minutes=t.minute,
            seconds=t.second,
            microseconds=t.microsecond
        )
    raise TypeError(u'to_timedelta：不支持的转换类型! {}'.format(type(obj)))


def to_unix_microseconds(obj):
    if isinstance(obj, dt.datetime):
        return t.mktime(obj.timetuple()) * 1000
    raise TypeError(u'to_unixtime：不支持的转换类型! {}'.format(type(obj)))


def to_oadate(obj):
    if isinstance(obj, pd.Timestamp):
        obj = pd.to_datetime(obj)
        delta = obj - OA_DATE_START
    elif isinstance(obj, dt.datetime):
        delta = obj - OA_DATE_START
    elif isinstance(obj, str):
        delta = pd.to_datetime(obj).to_pydatetime() - OA_DATE_START
    else:
        raise TypeError(u'to_oadate：不支持的转换类型! {}'.format(type(obj)))
    return delta.days + delta.seconds / TOTAL_SECONDS_A_DAY + delta.microseconds / TOTAL_MICROSECONDS_A_DAY


def from_oadate(oa_date):
    if isinstance(oa_date, np.float) or isinstance(oa_date, np.int):
        return dt.timedelta(days=oa_date) + OA_DATE_START
    else:
        raise TypeError(u'to_oadate：不支持的转换类型! {}'.format(type(oa_date)))


# pd.to_datetime('20180906124502.500', format='%Y%m%d%H%M%S.%f')
def convert_tick_date(tick_date):
    return int(to_strftime(tick_date, "%Y%m%d"))


def convert_tick_time(tick_date):
    return float(to_strftime(tick_date, "%H%M%S.%f")[:-3])


def convert_tick_datetime(tick_date):
    return int(to_strftime(tick_date, ("%Y%m%d%H%M%S%f"))[:-3])


def convert_trade_date(trade_date):
    return int(to_strftime(trade_date, "%Y%m%d%H%M%S"))


def convert_tday(tday):
    return int(to_strftime(tday, '%Y%m%d'))


def get_tday(tday):
    """
    获取tday，去除时分秒
    :param tday:
    :return:
    """
    if tday is None:
        tday = dt.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
    else:
        tday = pd.to_datetime(tday).to_pydatetime()
    return tday


def get_date_scope(start_date, end_date):
    date_scope_tuple_list = list()
    date_scope_tuple = collections.namedtuple('DateScope', 'start_date, end_date')
    dict_date_list = defaultdict(list)
    for each_date in rrule.rrule(rrule.DAILY, dtstart=to_datetime(start_date), until=to_datetime(end_date)):
        dict_date_list[each_date.year].append(each_date)
    for each_year, date_list in sorted(dict_date_list.items()):
        if len(date_list) > 0:
            date_tuple = date_scope_tuple(start_date=date_list[0], end_date=date_list[-1])
            date_scope_tuple_list.append(date_tuple)
    return date_scope_tuple_list


def json_serial(obj):
    """
    JSON serializer for objects not serializable by default json code
    :param obj:
    :return:
    """
    if isinstance(obj, (dt.datetime, dt.date, pd.Timestamp)):
        return obj.isoformat('%Y-%m-%d %H:%M:%S')
    raise TypeError("Type %s not serializable" % type(obj))


def get_date_range(start, end, freq='1T'):
    index = pd.date_range(start=start, end=end, freq=freq)
    return index.tolist()


def draw_chart():
    data_path = r'D:\TempCSV\50ETF_1m.h5'
    data = pd.read_hdf(data_path)
    data = data.iloc[-100:]

    # 将时间序列转化成datetime，范例数据里时间是index
    date = pd.to_datetime(data.index.values)

    # first we'll do it the default way, with gaps on weekends
    fig, axes = plt.subplots(ncols=2, figsize=(8, 4))
    ax = axes[0]
    ax.plot(date, data.close.values)
    ax.set_title("Default")
    fig.autofmt_xdate()

    # next we'll write a custom formatter
    data_length = len(data)
    ind = np.arange(data_length)  # the evenly spaced plot indices

    def format_date(x, pos=None):
        this_ind = np.clip(int(x + 0.5), 0, data_length - 1)
        return date[this_ind].strftime('%Y-%m-%d %H:%M:%S')

    ax = axes[1]
    ax.plot(ind, data.close.values)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(format_date))
    ax.set_title("Custom tick formatter")
    fig.autofmt_xdate()
    plt.show()


class mytest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_draw_chart(self):
        draw_chart()
        pass

    def test_convert_tick_time(self):
        tick_date = '2018-08-20 20:45:03.50000'
        float_tick_time = convert_tick_time(tick_date)
        pass

    def test_convert_tick_date(self):
        tick_date = '2018-08-20 20:45:03.50000'
        float_tick_date = convert_tick_date(tick_date)
        pass

    def test_get_date_range(self):
        start = '2018-09-06 09:00:00'
        end = '2018-09-06 15:00:00'
        freq = '1T'
        date_range = get_date_range(start=start, end=end, freq=freq)
        self.assertEqual(361, len(date_range))

    def test_24_to_0(self):
        date_str = 20160921240000
        t = to_datetime(date_str)
        print(t)

    def test_to_unixtime(self):
        now = dt.datetime(1990, 3, 29, 0, 0)
        unix_now = to_unix_microseconds(now)
        print(unix_now)

    def test_to_oadate(self):
        now = dt.datetime.now()
        now = now.replace(microsecond=0)
        oa_date = to_oadate(now)
        now2 = from_oadate(oa_date)
        self.assertEqual(now, now2)

    def test_to_strftime(self):
        date_str = 20160908153000
        t = to_datetime(date_str)
        print(t)

    def test_str_to_datetime(self):
        date_time = dt.datetime.now()
        t = to_strftime(date_time, "%Y%m%d%H%M%S")
        print(t)

    def test_to_oadate(self):
        date_str = '2016-09-08 12:09:45'
        oa_date = to_oadate(date_str)
        pass

    def test_from_ondate(self):
        date_str = from_oadate(42151.63538194444)
        pass

    def test_get_days(self):
        start_date = dt.datetime(year=2010, month=2, day=2)
        end_date = dt.datetime(year=2017, month=12, day=20)
        date_scope_list = get_date_scope(start_date, end_date)
        print(date_scope_list)


def setup_logger(logger_dir='.', name="root", ch_level=logging.INFO, file_level=logging.DEBUG):
    if not os.path.exists(logger_dir):
        os.mkdir(logger_dir)
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    ch = logging.StreamHandler()
    ch.setLevel(ch_level)
    ch.setFormatter(formatter)

    log_file = logger_dir + os.sep + os.path.basename(os.path.abspath(__file__)) + ".log"

    fh = logging.FileHandler(log_file, "w")
    fh.setLevel(file_level)
    fh.setFormatter(formatter)

    logger.addHandler(fh)
    logger.addHandler(ch)

    return logger, fh


if __name__ == "__main__":
    logger_dir = os.path.dirname(os.path.abspath(__file__)) + os.sep + "log" + dt.datetime.today().strftime("%m%d")
    logger, file_handler = setup_logger(logger_dir=logger_dir, name="root", ch_level=logging.INFO,
                                        file_level=logging.DEBUG)
    unittest.main()
