# -*- coding:utf-8 -*-
import os
import re
import shutil
import unittest
import logging
import binascii
from hashlib import md5
import zipfile
import rarfile
import sys
import numpy as np
import pandas as pd
import datetime as dt
import h5py as h5

try:
    import ConfigParser
except:
    import configparser as ConfigParser
rarfile.UNRAR_TOOL = "C:\\Program Files\\WinRAR\\unrar.exe"
# Set to '/' to be more compatible with zipfile
rarfile.PATH_SEP = '\\'


def get_main_file_path():
    main_pyc = sys.modules['__main__'].__file__
    main_pyc = os.path.normpath(main_pyc)
    main_py = main_pyc.replace('.pyc', '.py')
    return main_py


def get_main_file_name():
    main_file = get_main_file_path()
    main_file_name = os.path.basename(main_file)
    return main_file_name


def get_main_file_dir():
    main_file = get_main_file_path()
    main_file_dir = os.path.dir(main_file)
    return main_file_dir


def get_file_crc(filepath):
    try:
        blocksize = 1024 * 64
        f = open(filepath, "rb")
        str = f.read(blocksize)
        crc = 0
        while (len(str) != 0):
            crc = binascii.crc32(str, crc)
            str = f.read(blocksize)
        f.close()
    except:
        print('get file crc error!')
        return 0
    return crc


def check_ball(ballpath):
    filetype = os.path.splitext(ballpath)[1]
    if filetype == ".rar":
        try:
            rf = rarfile.RarFile(ballpath)
            rf.testrar()
            rf.close()
            return True
        except:
            return False
    elif filetype == ".zip":
        try:
            zf = zipfile.ZipFile(ballpath)
            zf.testzip()
            zf.close()
            return True
        except:
            return False


class BadFileError(Exception): pass


class BadBallFileType(Exception): pass


def extract_ball(ball_path, target_path="."):
    if not os.path.exists(target_path):
        os.makedirs(target_path)
    file_type = os.path.splitext(ball_path)[1].lower()
    if not check_ball(ball_path):
        raise BadFileError("Bad ball file %s" % (ball_path))
    if file_type == ".rar":
        ball_file_obj = rarfile.RarFile(ball_path)
    elif file_type == ".zip":
        ball_file_obj = zipfile.ZipFile(ball_path)
    else:
        raise BadBallFileType(u'无法识别的压缩包文件{}'.format(ball_path))
    ball_file_obj.extractall(target_path)
    ball_file_obj.close()
    return target_path


def get_extract_package_bf(ball_path):
    file_type = os.path.splitext(ball_path)[1]
    if file_type == ".rar":
        bf = rarfile.RarFile(ball_path, "r")
    elif file_type == '.zip':
        bf = zipfile.ZipFile(ball_path, "r")
    else:
        raise BadBallFileType(u'无法识别的压缩包文件{}'.format(ball_path))
    return bf


def extract_file_from_bf(bf, bf_file_path, extract_file_path):
    if os.path.exists(extract_file_path):
        return extract_file_path
    else:
        # 如果待解压文件存在
        if bf_file_path in bf.namelist():
            # 从bf.namelist中过滤掉目录以及size为0的文件，其统一特征为file_size = 0
            if bf.getinfo(bf_file_path).file_size != 0:
                # 创建待解压文件的目录
                extract_file_dir = os.path.dirname(extract_file_path)
                if not os.path.exists(extract_file_dir):
                    os.makedirs(extract_file_dir)
                fw = open(extract_file_path, 'wb')
                fw.write(bf.read(bf_file_path))
                fw.close()
                return extract_file_path
    return None


def extract_file(ball_path, filename, destpath="."):
    file_type = os.path.splitext(ball_path)[1]
    if not check_ball(ball_path):
        raise BadFileError("Bad ball file %s" % (ball_path))
    if file_type == ".rar":
        bf = rarfile.RarFile(ball_path, "r")
    elif file_type == '.zip':
        bf = zipfile.ZipFile(ball_path, "r")
    if not os.path.exists(destpath):
        os.makedirs(destpath)
    for ball_file in bf.namelist():
        if ball_file.split(os.sep)[-1] == filename:
            if not os.path.exists(destpath):
                os.makedirs(destpath)
            extracted_file_path = destpath + os.sep + filename
            fw = open(extracted_file_path, "w")
            fw.write(bf.read(ball_file))
            fw.close()
            return extracted_file_path
    return None


# 获取指定后缀名的文件集合
def list_files(dir_path, suffix=None):
    file_list = []
    for root, dirs, files in os.walk(dir_path):
        for each_file in files:
            file_path = os.path.join(root, each_file)
            if suffix is not None:
                if each_file.endswith(suffix):
                    file_list.append(file_path)
                    continue
            else:
                file_list.append(file_path)
    file_list = sorted(file_list)
    return file_list


def filter_suffix(s, *end_string):
    array = map(s.endswith, end_string)
    if True in array:
        return True
    else:
        return False


def cal_md5(str):
    m = md5()
    m.update(str)
    return m.hexdigest()


def cal_md5_for_file(file):
    m = md5()
    a_file = open(file, 'rb')
    m.update(a_file.read())
    a_file.close()
    return m.hexdigest()


def zip_dir(dir_path, zip_file=None):
    zipped_num = 0
    if os.path.exists(dir_path):
        if os.path.isdir(dir_path):
            if not zip_file:
                zip_file = os.path.dirname(os.path.abspath(dir_path)) + os.sep + os.path.basename(dir_path) + ".zip"
            zf = zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED)
            for root, dirnames, filenames in os.walk(dir_path):
                for filename in filenames:
                    filepath = os.path.join(root, filename)
                    arcpath = filepath.replace(os.path.dirname(dir_path), "", 1)
                    zf.write(filepath, arcpath)
                    zipped_num += 1
            zf.close()
    else:
        raise IOError("Dir %s doesn't exists!" % (dir_path))
    return zipped_num


def copy_specific_dir(org_path, dest_path, search_prefix, filter_dir):
    pat_prefix = re.compile('^' + search_prefix)

    # Get the original directory
    org_dirs = os.listdir(org_path)

    if not os.path.exists(dest_path):
        os.makedirs(dest_path)

    for dir_name in org_dirs:
        if os.path.isdir(org_path + os.sep + dir_name):
            search_result = pat_prefix.search(dir_name)
            if search_result:
                filter_dir_path = org_path + os.sep + dir_name + os.sep + filter_dir
                if os.path.isdir(filter_dir_path):
                    dest_filter_dir_path = dest_path + os.sep + dir_name + os.sep + filter_dir
                    if os.path.exists(dest_filter_dir_path):
                        # delete directory
                        shutil.rmtree(dest_filter_dir_path)
                        # copy directory
                    shutil.copytree(filter_dir_path, dest_filter_dir_path)


def write_to_hdf(df, h5_file_path, group_name):
    assert isinstance(df, pd.DataFrame)
    if not os.path.exists(os.path.dirname(h5_file_path)):
        os.makedirs(os.path.dirname(h5_file_path))
    destpath = h5_file_path
    np_array = df.to_records()
    f = h5.File(destpath, 'a')
    g_cls = f.get('/oneminute', getclass=True)
    if g_cls is None or g_cls == 'default':
        g = f.create_group('/oneminute')
    else:
        g = f['/oneminute']
    set_name = group_name
    ds_cls = g.get(set_name, getclass=True)
    if ds_cls is None or ds_cls == 'default':
        ds = g.create_dataset(set_name, data=np_array, chunks=True, compression='gzip', compression_opts=9)
    else:
        init = g[set_name].value
        np_array = np.append(init, np_array)
        del g[set_name]
        ds = g.create_dataset(set_name, data=np_array, chunks=True, compression='gzip', compression_opts=9)
    f.close()


def read_ini(ini_file_path, section_name=None, option_name=None):
    if not os.path.exists(ini_file_path):
        return None
    config = ConfigParser.ConfigParser()
    config.readfp(open(ini_file_path, 'r'))
    # 当section name和option name都为None的时候，读取整个ini文件的内容
    if section_name is None and option_name is None:
        section_dict = dict()
        for section in config.sections():
            option_dict = dict()
            for option in config.options(section):
                option_dict[option] = config.get(section, option)
            section_dict[section] = option_dict
        return section_dict
    # 当section name不为空，则选择性读取
    elif section_name is not None:
        # 只有当section_name存在的时候，才进行读取
        if section_name in config.sections():
            # 当option_name不指定的时候，读取给定section中所有的option，放入字典返回
            if option_name is None:
                option_dict = dict()
                for option in config.options(section_name):
                    option_dict[option] = config.get(section_name, option)
                return option_dict
            # 当option_name指定的时候，直接进到section中获取该值
            elif option_name is not None and option_name in config.options(section_name):
                return config.get(section_name, option_name)
    return None


def write_ini(ini_file_path, section_name, option_dict=dict(), mode='wb'):
    config = ConfigParser.ConfigParser()
    config.add_section(section_name)
    for opt_key, opt_value in option_dict.items():
        config.set(section=section_name, option=opt_key, value=opt_value)
    config.write(open(ini_file_path, mode))


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_copy_ini_files(self):
        # sourcefile_path, destination_path, filted_filepath, to be copied files name
        result = copy_specific_dir(r"D:\TempCSV\INI_Test\source", r"D:\TempCSV\INI_Test\target", "trade", "INI")

        # Judge whether the copied file exists
        # self.assertEqual(True, os.path.exists("D:\\test_dir\\trade_dest\\trade_testfile\\INI"))

    # def test_print_data_table(self):
    #     print print_data_table(["host", 'name', 'status', 'date_execute', 'date_completed'], [])

    def test_extract_future_files(self):
        # 期货数据路径参数
        # rar file name: wsfu0401fb.rar
        # rar file internal folder name: 20160401
        # rar file exchange name: DL/SQ/ZZ
        # rar file csv file path: \\DESKTOP-NQ7U5OD\StockData\commodity_future\2016\201604\\wsfu0401fb.rar\\20160401\\DL\\DLxxxxx.csv
        source_dir = r"\\DESKTOP-NQ7U5OD\StockData\commodity_future\2013\201303"
        exchange_type = "DL"
        target_csv = "DLc01.csv"
        target_dir = r"D:\TempCSV\pydata\future"
        for each_file in os.listdir(source_dir):
            file_type = os.path.splitext(each_file)[1]
            each_file_path = source_dir + os.sep + each_file
            if file_type == ".rar":
                bf = rarfile.RarFile(each_file_path, "r")
            elif file_type == ".zip":
                bf = zipfile.ZipFile(each_file_path, "r")
            else:
                continue
            csv_path = "2013" + each_file.replace("wsfu", "").replace("fb.zip", "").replace("fb.rar", "")
            csv_path += os.sep + exchange_type + os.sep + target_csv
            extracted_file_path = target_dir + os.sep + each_file + os.sep + exchange_type + os.sep + target_csv
            if not os.path.exists(os.path.dirname(extracted_file_path)):
                os.makedirs(os.path.dirname(extracted_file_path))
            try:
                csv_content = bf.read(csv_path)
                fw = open(extracted_file_path, "w")
                fw.write(csv_content)
                fw.close()
            except:
                pass

    def test_extract_file(self):
        bf = rarfile.RarFile(r'\\DESKTOP-NQ7U5OD\StockData\stock\2007\200705\ws0518fb5.rar', "r")
        list = []
        for ball_name in bf.infolist():
            bf_info = bf.getinfo(ball_name)
            if bf_info.file_size != 0:
                list.append(ball_name)
        print(list)

    def test_list_files(self):
        files = list_files(r'\\DESKTOP-NQ7U5OD\StockData\stock\2007\200701', '.rar')
        print(files)

    def test_write_ini(self):
        ini_file_path = r'\\NAS\share\HDF5\TempCSV\stock\update.ini'
        section_name = 'stock'
        start_date_str = '2017-01-20'
        end_date_str = '2017-01-29'
        option_dict = dict()
        option_dict['StartDate'] = start_date_str
        option_dict['EndDate'] = end_date_str
        write_ini(ini_file_path, section_name, option_dict)

    def test_read_ini(self):
        ini_file_path = r'\\NAS\share\HDF5\TempCSV\stock\update.ini'
        all_ini_content = read_ini(ini_file_path)
        section_ini_content = read_ini(ini_file_path, 'if')
        option_ini_content = read_ini(ini_file_path, 'stock', 'startdate')
        pass


if __name__ == "__main__":
    logger = logging.getLogger('root')
    unittest.main()
