# -*- coding: UTF-8 -*-
import logging
import re
import pandas as pd
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.mime.application import MIMEApplication
from email.header import Header
import datetime as dt

import pytz
from rio3.base import constant
from rio3.log.setup import setup_screen_logger

# python 2.3.*: email.Utils email.Encoders
from email.utils import COMMASPACE, formatdate
import email

import os
import smtplib
import base64
import unittest
import imaplib

DICT_MAIL_FORMAT = {'TEXT': 'plain', 'HTML': 'html'}


def savefile(filename, data, path):
    try:
        if not os.path.exists(path):
            os.makedirs(path)
        filepath = path + os.sep + filename
        f = open(filepath, 'wb')
    except Exception as ex:
        print(ex)
        f.close()
    f.write(data)
    f.close()


# 获得字符编码方法
def get_charset(message, default="ascii"):
    # Get the message charset
    return message.get_charset()


def parseEmail(msg, attach_dir):
    mailContent = None
    fname = None
    data = None
    for part in msg.walk():
        if not part.is_multipart():
            filename = part.get_filename()
            charset = get_charset(part)
            # 是否有附件
            if filename:
                h = email.Header.Header(filename)
                dh = email.Header.decode_header(h)
                fname = dh[0][0]
                encodeStr = dh[0][1]
                if encodeStr is not None:
                    if charset is None:
                        fname = fname.decode(encodeStr, 'gbk')
                    else:
                        fname = fname.decode(encodeStr, charset)
                data = part.get_payload(decode=True)
                # 保存附件
                if fname is not None or fname != '':
                    savefile(fname, data, attach_dir)
            else:
                if charset is None:
                    mailContent = part.get_payload(decode=True)
                else:
                    mailContent = part.get_payload(decode=True).decode(charset)
    attach_path = attach_dir + os.sep + fname if fname is not None else None
    return (mailContent, attach_path)


def get_mail(dict_server, attachment_dir=None, start_date=None, subject=None, sentby=None):
    mail_content_list = []
    if start_date is None:
        start_date = dt.datetime(1900, 1, 1)
    else:
        start_date = pd.to_datetime(start_date)
    start_date = start_date.replace(tzinfo=pytz.timezone('Asia/Shanghai'))
    sub_pat = None
    if subject is not None:
        sub_pat = re.compile(subject)
    if dict_server['ssl']:
        imapServer = imaplib.IMAP4_SSL(dict_server['name'], dict_server['port'])
    else:
        imapServer = imaplib.IMAP4(dict_server['name'], dict_server['port'])
    imapServer.login(dict_server['user'], dict_server['passwd'])
    imapServer.select()
    # 邮件状态设置，新邮件为Unseen
    # Message statues = 'All,Unseen,Seen,Recent,Answered, Flagged'
    resp, items = imapServer.uid('search', None, 'ALL')
    number = 1
    for i in items[0].split():
        # get information of email
        resp, mailData = imapServer.fetch(i, "(RFC822)")
        if mailData is None:
            continue
        if mailData[0] is None:
            continue
        if mailData[0][1] is None:
            continue
        mailText = mailData[0][1]
        msg = email.message_from_string(mailText)
        ls = msg["From"].split(' ')
        strfrom = ''
        if (len(ls) == 2):
            strfrom = ls[1]
        else:
            strfrom = msg["From"]
        if sentby is not None:
            if strfrom != sentby:
                continue
        send_date = msg["Date"]
        send_date = pd.to_datetime(send_date)
        if send_date < start_date:
            continue
        sub = email.Header.decode_header(msg["Subject"])
        if sub[0][1] is not None:
            mail_subject = sub[0][0], sub[0][1]
        else:
            mail_subject = sub[0][0]
        if sub_pat is not None:
            search_result = sub_pat.search(mail_subject)
            if not search_result:
                continue
        mailContent, attachment = parseEmail(msg, attachment_dir)
        mail_content_list.append((mailContent, attachment))

    imapServer.close()
    imapServer.logout()
    return mail_content_list


def send_mail(str_from_address='huahuabill@qq.com', list_to_address=None, str_subject='', dict_content=None,
              dict_server=None, use_ssl=False, ssl_port=465):
    if dict_server is None:
        dict_server = {}
    if dict_content is None:
        dict_content = {}
    if not list_to_address:
        list_to_address = constant.DEFAULT_EMAIL_ADDRS
    assert type(list_to_address) == list
    smtp_server = dict_server.get('name', 'smtp.qq.com')
    smtp_user = dict_server.get('user', 'huahuabill@qq.com')
    smtp_passwd = dict_server.get('passwd', 'ftnmizdjxavwcacc')
    content_format = dict_content.get('format', DICT_MAIL_FORMAT['TEXT'])
    content_files = dict_content.get('files', [])
    content_txt = dict_content.get('content', '')

    msg = MIMEMultipart()
    msg['From'] = str_from_address
    msg['to'] = COMMASPACE.join(list_to_address)  # COMMASPACE==','
    msg['subject'] = Header(str_subject, 'utf-8')
    msg['Date'] = formatdate(localtime=True)
    msg.attach(MIMEText(content_txt, content_format, 'utf-8'))

    for att_file in content_files:
        part = MIMEBase('application', 'octet-stream')  # 'octet-stream': binary data
        part.set_payload(open(att_file, 'rb').read())
        email.encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(att_file))
        msg.attach(part)

    if use_ssl:
        smtp = smtplib.SMTP_SSL(smtp_server, port=ssl_port)
    else:
        smtp = smtplib.SMTP(smtp_server)
    smtp.login(smtp_user, smtp_passwd)
    smtp.sendmail(str_from_address, list_to_address, msg.as_string())
    smtp.close()


def send_html_mail(content, receiver, sender='info@xuanyeinvest.com', password='SuM8k3djDqQqUjs9',
                   smtpserver='smtp.exmail.qq.com', subject='宣夜投资-产品净值报告', list_attach_file_path=None):
    """
    发送(带附件的)html格式邮件
    :param content: str, html文本
    :param receiver: str, 收件人邮箱地址
    :param sender: str, 发件人邮箱地址
    :param password: str, 发件人邮箱密码/授权码(某些邮箱需要授权码,而不是登录密码)
    :param smtpserver: str, 发件人邮箱smtp服务器地址
    :param subject: str, 邮件标题
    :param list_attach_file_path: str or list, 附件文件完整路径
    :return:
    """
    msgRoot = MIMEMultipart('related')
    msgRoot['Subject'] = subject
    msgRoot['From'] = '宣夜投资 <{}>'.format(sender)
    msgRoot['To'] = receiver

    msgText = MIMEText(content, 'html', 'utf-8')
    msgRoot.attach(msgText)

    if list_attach_file_path is not None:
        if isinstance(list_attach_file_path, str):
            list_attach_file_path = [list_attach_file_path, ]
        from collections import Iterable
        assert isinstance(list_attach_file_path, Iterable)
        for file_path in list_attach_file_path:
            if os.path.exists(file_path):
                attach = MIMEApplication(open(file_path, 'rb').read())
                attach.add_header('Content-Disposition', 'attachment',
                                  filename=os.path.split(file_path)[1].encode('gb2312'))
                msgRoot.attach(attach)

    smtp = smtplib.SMTP_SSL()
    smtp.connect(smtpserver)
    smtp.login(sender, password)
    smtp.sendmail(sender, receiver, msgRoot.as_string())
    smtp.quit()


class MyTest(unittest.TestCase):
    def setUp(self):
        setup_screen_logger()
        self.logger = logging.getLogger('root')

    def tearDown(self):
        pass

    def test_send_mail(self):
        to_mail_list = ['huahuabill@163.com']
        send_mail(list_to_address=to_mail_list, str_subject='send_mail_test1', dict_content={'content': 'aaaa'},
                  use_ssl=True)

    def test_get_mail(self):
        mail_content_list = get_mail(constant.SETTLEMENT_EMAIL_ACCONT, 'D:\\temp_email\\', subject=u'华鑫期货',
                                     start_date='2016-11-28')
        self.assertEqual(len(mail_content_list) > 0, True)
        for content, attachment in mail_content_list:
            self.logger.info(attachment)
