# -*- coding:utf-8 -*-
import datetime as dt
import os
import re
import unittest
import csv
import io
import numpy as np
import pandas as pd
from rio3.utils.date import to_strftime
from rio3.base.constant import SECURITY_DIR_DICT

tick_columns = ['Time', 'Price', 'Volume', 'Amount', 'Marketposition',
                'SP1', 'SP2', 'SP3', 'SP4', 'SP5', 'SV1', 'SV2', 'SV3', 'SV4', 'SV5',
                'BP1', 'BP2', 'BP3', 'BP4', 'BP5', 'BV1', 'BV2', 'BV3', 'BV4', 'BV5',
                'isBuy']
idx_future_market_list = SECURITY_DIR_DICT['index_future']
cmmd_future_market_list = SECURITY_DIR_DICT['commodity_future']
double_max_value = 1.79769313486232E+308

# 过滤重复数据的组合条件
TICK_UNIQUE_SUBSET_COLUMNS = ['Time', 'Price', 'Volume', 'Amount', 'SP1', 'SV1', 'BP1', 'BV1']


def check_tick_csv(validator_name=None):
    def refine_decorator(func):
        def func_wrapper(*args, **kwargs):
            pass

        return func_wrapper

    return refine_decorator


# 用于修复原油数据BP1中存在数据异常的问题
def check_double_max(df_record, check_column='BP1', replace_column='Price', replace_column2='SP1'):
    # 先找到check_columns中存在1.79769313486232E+308的数据
    double_max_df = df_record[df_record[check_column].astype(np.float64) == double_max_value].copy()
    # 如果存在double_max_df的情况
    if not double_max_df.empty:
        # 用Price的价格代替BP1的价格，如果没有，就用SP1
        double_max_df[check_column] = np.where(double_max_df[check_column].astype(np.float64) == double_max_value,
                                               double_max_df[replace_column], double_max_df[replace_column2])
        # 通过index配对删掉double_max_df的数据
        df_record = df_record[~df_record.Time.isin(double_max_df.Time)]
        # 合并double_max_df和剩下的df_record
        df_record = pd.concat([df_record, double_max_df])
        # 排序
        df_record = df_record.sort_values(by=['Time'])
    return df_record


def unique_time_rows(func_refine_csv):
    def func_wrapper(csv_file_path, *args, **kwargs):
        refined_csv_file_path = csv_file_path
        if os.path.exists(csv_file_path):
            with open(csv_file_path, 'rb') as csv_file:
                spam_reader = csv.reader(csv_file)
                refine_csv_file = io.BytesIO()
                spam_writer = csv.writer(refine_csv_file)
                # 记录原时间
                last_row_time = None
                for i, row in enumerate(spam_reader):
                    # 过滤第一条Header
                    if i != 0:
                        # 将当前时间转化为datetime类型
                        dt_current_row_time = pd.to_datetime(row[0]).to_pydatetime()
                        if last_row_time is not None:
                            dt_last_row_time = pd.to_datetime(last_row_time).to_pydatetime()
                            # 判断是否有重复时间
                            if dt_current_row_time <= dt_last_row_time:
                                # 检测到有重复时间戳，取上一次的时间累加100毫秒
                                refine_row_time = to_strftime(
                                    dt_last_row_time + dt.timedelta(microseconds=100), "%Y-%m-%d %H:%M:%S.%f")
                                row[0] = refine_row_time
                        last_row_time = row[0]
                    spam_writer.writerow(row)
                refine_csv_file.seek(0)
                refined_csv_file_path = refine_csv_file
        func_return = func_refine_csv(refined_csv_file_path, *args, **kwargs)
        return func_return

    return func_wrapper


def filter_invalid_df_rows(func_refine_df):
    def func_wrapper(df_records, *args, **kwargs):
        df_records = df_records[df_records['Time'].apply(is_valid_date)]
        func_return = func_refine_df(df_records, *args, **kwargs)
        return func_return

    return func_wrapper


# GroupBy所有的记录，对重复的时间戳增加时间偏移量
def allocate_time_offset(func_refine_df):
    def func_wrapper(df_records, *args, **kwargs):
        # df_records.Time = df_records.Time.apply(pd.to_datetime)
        org_records_len = len(df_records)
        unique_records_len = len(df_records.Time.unique())
        if org_records_len != unique_records_len:
            # 检测是否有重复时间的记录
            time_column_grouped = df_records.groupby(df_records['Time'])
            time_offset_list = list()
            for key, grp_df in time_column_grouped:
                dup_num = len(grp_df)
                if dup_num > 1:
                    for i in range(dup_num):
                        time_offset_list.append(dt.timedelta(seconds=i / float(dup_num)))
                else:
                    time_offset_list.append(dt.timedelta(seconds=0))
            df_records.Time = df_records.Time + time_offset_list
        func_return = func_refine_df(df_records, *args, **kwargs)
        return func_return

    return func_wrapper


# GroupBy重复的记录，增加时间戳
def allocate_time_offset_adv(func_refine_df):
    def func_wrapper(df_records, *args, **kwargs):
        # 先从df_records中根据Time筛选所有的重复选项，如果没有重复，直接返回
        duplicated_df = df_records[df_records.duplicated(['Time'], keep=False)]
        # 如果有重复的选项
        if not duplicated_df.empty and len(duplicated_df) > 0:
            # 对重复的记录按Time进行分组
            dup_time_column_grouped = duplicated_df.groupby(duplicated_df['Time'])
            time_offset_list = list()
            for key, grp_df in dup_time_column_grouped:
                each_dup_nums = len(grp_df)
                for i in range(each_dup_nums):
                    time_offset_list.append(dt.timedelta(seconds=i / float(each_dup_nums)))
            # 如果长度相等，则进行时间偏移增量
            if len(time_offset_list) == len(duplicated_df):
                # 先删掉所有重复的记录，过滤出不重复的记录
                none_duplicated_df = df_records[~df_records.Time.isin(duplicated_df.Time)]
                # 为重复项增加时间偏移量
                # pandas 0.19.x版本以后不可以直接用series + list，需要都转化成Series再相加
                # 重置index
                duplicated_df = duplicated_df.reset_index()
                # 默认根据index序号配对相加时间偏移量
                # duplicated_df.Time = duplicated_df.Time + pd.Series(time_offset_list)
                duplicated_df.Time = duplicated_df.Time.values + pd.Series(time_offset_list).values
                # 将没有重复的数据和增加过时间偏移量的数据组合
                df_records = pd.concat([none_duplicated_df, duplicated_df], sort=True)
                # 按照Time进行排序保证时间顺序
                df_records = df_records.sort_values('Time')
        func_return = func_refine_df(df_records, *args, **kwargs)
        return func_return

    return func_wrapper


def filter_invalid_rows(func_refine_csv):
    def func_wrapper(csv_file_path, *args, **kwargs):
        refined_csv_file_path = csv_file_path
        if os.path.exists(csv_file_path):
            with open(csv_file_path, 'rb') as csv_file:
                spam_reader = csv.reader(csv_file)
                invalid_num_list = []
                # 循环原csv文件的内容
                for i, row in enumerate(spam_reader):
                    # 第一行是标题，排除
                    if i != 0 and len(row) > 0:
                        time_column = row[0]
                        # 如果除标题以外存在时间的行内容不是合法的格式
                        if not is_valid_date(time_column):
                            invalid_num_list.append(i)
                # 如果error数组不为空，说明csv中有错误
                if len(invalid_num_list) > 0:
                    # 创建一个spam writer
                    refine_csv_file = io.BytesIO()
                    spam_writer = csv.writer(refine_csv_file)
                    # 再次读取原csv文件
                    with open(csv_file_path, 'rb') as org_csv_file:
                        spam_reader = csv.reader(org_csv_file)
                        for i, row in enumerate(spam_reader):
                            # 过滤掉检查出来的结果行
                            if i not in invalid_num_list:
                                spam_writer.writerow(row)
                    refine_csv_file.seek(0)
                    refined_csv_file_path = refine_csv_file
        func_return = func_refine_csv(refined_csv_file_path, *args, **kwargs)
        return func_return

    return func_wrapper


'''
商品期货的Tick文件中：
    2003 至 now 源文件都命名为Open_Int
股指期货的Tick文件分两种:
    SF  (1档数据)：2010-01-01 至 2012-05-31 其中Open_Int这一列在CSV文件中取名为OI
    SF5 (5档数据)：2012-06-01 至 now        其中Open_Int这一列在CSV文中中取名为OpenInt
以上所有对OpenInt的命名，在转化过程中统一修改为Marketposition

股指期货的SF和SF5数据统一转化为'Time', 'Price', 'Volume', 'Amount', 'Marketposition',
                             'SP1', 'SP2', 'SP3', 'SP4', 'SP5', 'SV1', 'SV2', 'SV3', 'SV4', 'SV5',
                             'BP1', 'BP2', 'BP3', 'BP4', 'BP5', 'BV1', 'BV2', 'BV3', 'BV4', 'BV5',
                             'isBuy'
1. 程序需要对SF中没有的其余四档数组进行补全，为每一列填充-1作为补全值
2. 程序需要对SF5中不需要的Columns进行删选
'''


def format_columns(func_refine_df):
    def func_wrapper(df_records, exchange_type, *args, **kwargs):
        # 先判断市场名，如果商品期货文件，需要重命名Open_Int成Marketposition，保持其余列的完整
        if exchange_type in cmmd_future_market_list:
            if 'Open_Int' in df_records.columns:
                df_records = df_records.rename(columns={'Open_Int': 'Marketposition'})
            if 'OpenInterest' in df_records.columns:
                df_records = df_records.rename(columns={'OpenInterest': 'Marketposition'})
        # 如果是股指期货文件，需要重命名OI或OpenInt成Marketposition，同时过滤不需要的列或者补全缺失的列
        elif exchange_type in idx_future_market_list:
            if 'OI' in df_records.columns:
                df_records = df_records.rename(columns={'OI': 'Marketposition'})
            if 'OpenInterest' in df_records.columns:
                df_records = df_records.rename(columns={'OpenInterest': 'Marketposition'})
            # 过滤不需要存储的列
            for each_column in df_records.columns:
                if each_column not in tick_columns:
                    del each_column
        # 补全列
        current_df_columns_set = set(df_records.columns)
        supply_columns_set = set(tick_columns) - current_df_columns_set
        for column in supply_columns_set:
            df_records[column] = -1
        # 重新按照column定义df_records的顺序
        df_records = df_records[tick_columns].copy()
        # 过滤掉重复的数据
        df_records.drop_duplicates(subset=TICK_UNIQUE_SUBSET_COLUMNS, inplace=True)
        func_return = func_refine_df(df_records, exchange_type, *args, **kwargs)
        return func_return

    return func_wrapper


# 判断是否是一个有效的日期字符串
def is_valid_date(str_val):
    try:
        # time.strptime(str_val, "%Y-%m-%d %H:%M:%S.%f")
        pd.to_datetime(str_val).to_pydatetime()
        return True
    except Exception as e:
        return False


@filter_invalid_rows
def get_valid_csv_resource(csv_file_resource):
    return csv_file_resource


@allocate_time_offset_adv
def get_valid_csv_df(df_records):
    return df_records


@format_columns
def get_format_csv_df(df_records, exchange_type):
    return df_records


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_refine_csv(self):
        csv_file_path = r"D:\TempCSV\SH000001_20090122.csv"
        refine_csv = get_valid_csv_resource(csv_file_path)
        pass

    def test_csv_file_format(self):
        csv_stock_code_date_format = re.compile(r'(\D{2})(\D*)(\d+)_(\d{8}).csv')
        stock_code = 'DLa000001_20160103.csv'
        match_res = re.match(csv_stock_code_date_format, stock_code)
        if match_res:
            exchange_type = match_res.group(1)
            symbol = match_res.group(2)
            code = match_res.group(3)
            date = match_res.group(4)
            # print match_res.group(1)


if __name__ == "__main__":
    unittest.main()
